;; @(#) tinycache.el -- Maintain a cache of visited files [compile,dired]

;; @(#) $Id: tinycache.el,v 1.33 1997/05/31 11:41:48 jaalto NotReleased $
;; @(#) $Keywords: extensions $
;; $KnownCompatibility: 19.28 $

;;{{{ Id

;; Copyright (C) 1996,97 Jari Aalto
;; Author:       Jari Aalto	<jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto	<jari.aalto@poboxes.com>
;; Created:      May 20 1996
;;
;; To get information on this program use ident(1) or do M-x tica-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el

;; LCD Archive Entry:
;; tinycache|Jari Aalto|jari.aalto@poboxes.com|
;; Maintain a cache of visited files: compile, grep, dired. Flush-cache|
;; 31-May-1997|$Revision: 1.33 $|~/misc/tinycache.el.Z|

;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.


;;}}}
;;{{{ Install

;; ....................................................... &t-install ...
;;   Put this file on your Emacs-Lisp load path, add following into your
;;   ~/.emacs startup file. Rip code with with tinylib.el/ti::m-pkg-rip-magic
;;
;;* _
;;*      (require 'tinycache)
;;
;;   or prefer this; your ~/.emacs loads up quicker
;;
;;* _
;;*      (eval-after-load "compile" '(require 'tinycache))
;;*      (eval-after-load "dired"   '(require 'tinycache))
;;
;;   To disable this package:
;;
;;	C-u M-x tica-advice-control
;;
;;   If you have any questions, use this function
;;
;;      M-x tica-submit-bug-report
;;
;;   If you use gnuserv.el, be sure that you load th packages in this
;;   order: GNUSERV, TINYCACHE.



;;}}}
;;{{{ Documentation

;; ..................................................... &t-commentary ...
;;; Commentary:
;;
;;  Preface, overview of features
;;
;;	This package is meant to be used with dired and compilation
;;	buffers. When you load file from either one, the file is
;;	cached. This way you can view files easily and when you
;;	have finished you can flush the cache and get rid of all
;;	vieved files.
;;
;;  Dired description
;;
;;	When you load file from dired for viewing (dired-view-file), the
;;	file is cached. You can load several files for viewing and when you
;;	have finished, you can call tica-flush (Defaults to C-c k in dired)
;;	to remove all the cached files from emacs.
;;
;;	This way you don't end up having files that you're not interested
;;	in any more. Using the cache makes browsing buch of files very
;;	easy. Each dired buffer has it's own cache. The cache is also
;;	flushed if you kill the dired buffer.
;;
;;  Compilation cache description
;;
;;	Maintain also a cache of buffers visiting files via the
;;	`next-error' and `compile-goto-error' commands; each compile/grep
;;	buffer has its own cache.  To kill the cached buffers manually, use
;;	`C-c C-d' (compile-flush-cache) in the compile/grep buffer;
;;	deleting the compile/grep buffer automatically kills the cached
;;	buffers.  To disable the cache, set `compilation-find-file-cache'
;;	to a non-list value (e.g. 'disable).
;;
;;	After loading this file, every file that is loaded by calling some
;;	compile function, ie. compile-goto-error, is cached if it is not in
;;	emacs already.
;;
;;	Ie. when you fix some small errors in other files, you may not want
;;	to keep those files in emacs after you've done; remember, those got
;;	loaded during the calls to compile-goto-error. The easiest way to
;;	get rid of these extra files, that were not originally in emacs, is
;;	to:
;;
;;	    A. kill compilation buffer, C-x k *compilation*
;;	    B. Call M-x tica-flush directly
;;
;;	See igrep.el also how you can browse(grep) files easily and when
;;	you've done, you can call this package top get rid of those browsed
;;	files.
;;
;;  Cache minor mode indication --  Controlling the cache flag
;;
;;	Modeline indication shows for loaded buffer
;;
;;	    "+C"    if file is loaded as cached.
;;	    "+c"    If you have manually turned off the cache
;;
;;	And for root buffer where the file were loaded, normally
;;	compilation or dired buffer, the modeline shows
;;
;;	    "+CN"  where N is number of files currently in the cache
;;
;;	Sometimes you want to keep some file that belongs to the cache
;;	and you don't want to loose it when you execute M-x tica-flush
;;	or when you kill the root buffer.
;;
;;	For that purpose there function tica-mode to turn off the cache
;;	for current buffer. When the cache mark says "+c" in the
;;	modeline, it tells you that the file will not be killed when you
;;	tica-flush is called.
;;
;;	Note: the root buffer's xx count is not updated when you kill
;;	buffer that was cached. So if the count says 10, and you kill 3
;;	files that, the count will still still say 10. The count is
;;	updated only when you load some *new* file from the root buffer.
;;	At his tim all the buffers cached are checked and the ones that
;;	do not exist any more are removed.
;;
;;  Buffer list commands [C-x C-b]
;;
;;	There are some additional commands added to buffer list which
;;	helps you to keep track of the cached files better. The "c"
;;	prefix is chosen for (c)ache related commands.
;;
;;	    C-c c m	mark all cached files
;;	    C-c c d	mark as deleted
;;	    C-c c u	unmark cached files.
;;
;;  Dired mode commands
;;
;;	Similar to buffer list, there is some dired commands too
;;
;;	    C-c c k	tica-flush, remove all cached files from this dired
;;	    C-c c m	tica-dired-mark
;;	    C-c c u	tica-dired-unmark
;;
;;  Thanks
;;
;;	Kevin Rodgers, his igrep.el gave me an idea for this. The
;;	original cache code from where this pacakge evolved was
;;	written by Kevin under name compile-cache.el

;; ....................................................... &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; May	31	1996	[jari]		19.28	v1.28-1.33	NotReleased
;; - Internal changes. Library function chanegs.
;; - Changed all non standard boolean tests to use when and unless
;; - Checkdoc 1.29 clear. defcustom support. Now the view-mode function
;;   is determined better.
;;
;; Mar	11	1996	[jari]		19.28	v1.27		NotReleased
;; - Corrected tica-mode function. it now affects the flag always
;;   and prints warning messages if the tica-:info does not exist for
;;   the buffer. (That's weird situation, it should exist).
;; - Corrected turning the view mode on after find-file
;; - Rearranged keysetting with new library fucntion define-key-if-free,
;;   so that user's keys are not carelessly overwritten. The buffer
;;   menu also uninstalls itself automatically when installation
;;   has been done.
;; - The dired find-file no longer turns on cache flag if the buffer
;;   was already loaded in emacs (bug)
;;
;; Jan	15	1996	[jari]		19.28	v1.25-1.26	NotReleased
;; - Added the cache count code. Reformatted documentation.
;;
;; Dec	17	1996	[jari]		19.28	v1.20-1.24	NotReleased
;; - Now works in dired buffers too: when you view a file it will
;;   be cached. Makes browsing easy. New Version number reflects that.
;; - Now the cache count is shown in modeline in a root buffer:
;;   compile, grep or dired.
;; - Substantial code changes and rearrangements.
;;   tica-dired-mark			:+ new user command
;;   tica-dired-unmark			:+ new user command
;;
;; Sep	27	1996	[jari]		19.28	v1.15		NotReleased
;; - Added msb.el menu installation for cached files as Dan Nicolaescu
;;   suggested. Cleared byteComp errors.
;;   tica-install-msb			:+
;;
;; Sep	26	1996	[jari]		19.28	v1.14		NotReleased
;; - Found several serious promlems from the code. Now the killed buffer
;;   case is more carefully tested.
;; - Adding to the tica-:cache was errorneus in compilation-find-buffer
;; - Now uses "+c " for mode off and " +C" for mode on indicator. It
;;   was not good if there were no indicator if the tica-mode for
;;   buffer was off.
;; - Corrected missing tinylib autoloads.
;;   tica-flush-all				:+ new user function
;;   tica-buffer-list-mark-cache-off            :+ new user function
;;   tica-buffer-list-map-over-off-buffers      :+
;;
;; Sep	25	1996	[jari]		19.28	v1.13		NotReleased
;; - Rearranged loading. now uses more autoloads.
;; - Changed "-bl-" to more decriptive "-buffer-list-".
;; - Made tica-:cache permanent local, so that it doesn't get wiped away
;;   when user start new compilation.
;; - Bug: Dan Nicolaescu reported that killing the compilation
;;   buffer results an error, maybe we try to kill non-existent buffer.
;;   [That should not happen, because buffer existence is tested
;;   in code...no idea right now why that error occurs]
;;
;; Sep	12	1996	[jari]		19.28	v1.12		NotReleased
;; - <done@nexus.sorostm.ro> (Dan Nicolaescu) suggested that there should
;;   be a way to remove a file from cache, so that it is not killed when
;;   you do M-x tica-flush.
;; - Now this package shows the cache status in the modeline, which
;;   you can toggle like in any other minor mode. When the mode is on
;;   file is part of the cache.
;; - Wrote lot of new code, eg. added support for buffer-list C-x C-b
;;
;; Sep	11	1996	[jari]		19.28	v1.9-1.11	NotReleased
;; - Added copy-sequence call to compile-internal advice to make sure
;;   the hook is local.
;; - Added autolaod installation as suggested by
;;   done@nexus.sorostm.ro (Dan Nicolaescu)
;;
;; Sep	5	1996	[jari]		19.28	v1.7-1.8	test
;; - done@nexus.sorostm.ro (Dan Nicolaescu) reported infinite loop
;;   (ie. Lisp nesting exceeds max-lisp-eval-depth) when using
;;   the sequence:
;;
;;    M-x load-file .../tinychache.el
;;    M-x load-file .../gnuserv.el
;;    M-x gnuserv-start
;;
;;    By looking at the backtrace, it seems that gnuserv.el defines it's
;;    own kill-buffer function with server-real-kill-buffer. Can't tell
;;    that for sure without looking at the source.
;;
;;  - I changed the kill-buffer advice for now. Currently only for
;;    test purpose.
;;  - Copied piece of code from gnuserv.el to cope wiith kill-buffer
;;    redefinition in that package.
;;
;; Aug	15	1996	[jari]		19.28	v1.5-1.6	NotReleased
;; - If there was no compilation buffer, it generated fatal error
;;   in kill-buffer-hook. Corrected.
;;
;; Aug	12	1996	[jari]		19.28	v1.4		NotReleased
;; - The kill-buffer-hook doesn't exist in 19.28. Added new advice
;;   to overcome it.
;; - done@nexus.sorostm.ro (Dan Nicolaescu) reported infinite loop in
;;   19.32. This was due to wron function name in remove-hook call
;;   in tica-flush

;;}}}
;;{{{ setup: require, variables

;;; Code:
;;; ......................................................... &require ...

(eval-when-compile (require 'advice))
(require 'tinylibm)

(eval-and-compile
  (autoload 'compilation-find-buffer	"compile"   t t)
  (autoload 'dired-view-file		"dired"	    t t)
  (autoload 'dired-get-filename		"dired")
  (autoload 'dired-mark			"dired"	    t t)
  (autoload 'dired-unmark		"dired"	    t t)
  )

(defgroup-tiny TinyCache tica-: extensions
    "Maintain a cache of visited files [compile,dired].
    overview of features

	This package is meant to be used with dired and compilation
	buffers. When you load file from either one, the file is
	cached. This way you can view files easily and when you
	have finished you can flush the cache and get rid of all
	vieved files."
    )



;;; ............................................. &emacs-compatibility ...
;;;  We must install this for older emacs versions ( 19.30< )
;;;

(unless (boundp 'kill-buffer-hook)
  ;; gnuserv.el
  ;; - We have to do this now, because if user loads files in this order:
  ;;
  ;;	    tinycache
  ;;	    gnuserv
  ;;
  ;;   The gnuserv won't see the real function, because tinycache
  ;;   put advice around the function. And using the adviced function
  ;;   as "real" causes infinite loop.
  ;;
  ;;   If user doesn't use gnuserv, this just defines one exta function,
  ;;   which does no harm.
  ;;
  (or (fboundp 'server-real-kill-buffer)
      (fset 'server-real-kill-buffer (symbol-function 'kill-buffer)))

  (defvar kill-buffer-hook nil
    "Defined in tinycache.el package. Hook run just before buffer
     is killed.")

  (defadvice kill-buffer (around tica act)
    "Run kill-buffer-hook before buffer iss killed."
    (run-hooks 'kill-buffer-hook)
    (let ((kill-buffer-hook nil)       ;; prevent accidents
	  )
      ad-do-it
      ))
  )

;;; ........................................................ &v-public ...

(defcustom tica-:mode-on-string  " +C"
  "*Cache property on indicator. File can be flushed."
  :type 'string :group 'TinyCache
  )

(defcustom tica-:mode-off-string " +c"
  "*Cache property off indicator. File will not be flushed."
  :type 'string :group 'TinyCache
  )


(defcustom tica-:load-hook '(tica-install-msb)
  "*Hook run when file has been loaded."
  :type 'hook :group 'TinyCache
  )


(defcustom tica-:find-file-buffer-hook 'tica-maybe-view-mode
  "*Hook run inside buffer which is loaded from compile output."
  :type 'hook :group 'TinyCache
  )

;;; ....................................................... &v-private ...


(defvar tica-:mode-name  tica-:mode-on-string
  "String in the modeline to mark if that file is part the cache or not.
This is canged by program and not a user variable.")


(defvar tica-mode  nil
  "Mode on/off variable.")
(make-variable-buffer-local 'tica-mode)
(put 'tica-mode             'permanent-local t)


(defvar tica-:info  nil
  "Variable to keep information about the cache.
Contains the compile buffer pointer from where the file was loaded.")
(make-variable-buffer-local 'tica-:info)
(put 'tica-:info             'permanent-local t)


(defvar tica-:mode-user-flag  nil
  "Non-nil means that user has touched the `tica-mode' flag in this buffer.")
(make-variable-buffer-local 'tica-:mode-user-flag)
(put 'tica-:mode-user-flag   'permanent-local t)


(defvar tica-:cache nil
  "List of buffers created by `compilation-find-file'. Local to each buffer.
Format: (BUFFER-POINTER BP ..)")

;; If user starts new compilation in *grep* buffer, it wipes the results
;; but the cache must remain there, since he may have loaded files from
;; previouos compilation.
;;
(put 'tica-:cache            'permanent-local t)


;;; ....................................................... &v-version ...

(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinycache.el"
 "tica"
 tica-:version-id
 "$Id: tinycache.el,v 1.33 1997/05/31 11:41:48 jaalto NotReleased $"
 '(tica-:version-id
   )))


;;}}}

;;{{{ code: misc functions

;;; ........................................................ functions ...

;;; ----------------------------------------------------------------------
;;;
(defun tica-advice-control (&optional disable)
  "Turn advices on. Optionally DISABLE."
  (ti::m-ad-control
   '(compile-internal compilation-find-file) "^tica"  disable 'verb))


;;; ----------------------------------------------------------------------
;;;
(defun tica-install-mode  ()
  "Install tica-mode."
  ;;  Make sure we can display string in modeline
  ;;
  (if (null (assq 'tica-mode minor-mode-alist))
      (ti::k-add-minor-mode 'tica-mode
			    'tica-:mode-name
			    (make-sparse-keymap);; No keymap
			    )))


;;; ----------------------------------------------------------------------
;;;
(defun tica-define-default-keys  ()
  "Define keys to `compilation-minor-mode-map'"
  (interactive)
  (let* (map
	 )
    ;;  Avoid byte compilation whinings this way....
    ;;
    (when (and (boundp  'compilation-minor-mode-map)
	       (setq map (symbol-value 'compilation-minor-mode-map))
	       (keymapp map)
	       )
      (use-prefix-key map "\C-cc")
      (define-key-if-free map "\C-cck" 'tica-flush)
      )

    (when (and (boundp  'compilation-minor-mode-map)
	       (setq map (symbol-value 'compilation-minor-mode-map))
	       (keymapp  map)
	       )
      (use-prefix-key map "\C-cc")
      (define-key-if-free map "\C-cck" 'tica-flush)
      )

    (when (and (boundp  'dired-mode-map)
	       (setq map (symbol-value 'dired-mode-map))
	       (keymapp  map))
     (use-prefix-key map "\C-cc")
     (define-key-if-free map "\C-cck" 'tica-flush)
     (define-key-if-free map "\C-ccm" 'tica-dired-mark)
     (define-key-if-free map "\C-ccu" 'tica-dired-unmark)
     )

    ))


;;; ----------------------------------------------------------------------
;;;
(defun tica-buffer-list-bindings-on  ()
  "Add default bindings to buffer list, \\[list-buffers]."

  ;;  Choose "c" for cache commands
  (use-prefix-key Buffer-menu-mode-map "\C-cc")

  ;;  With these you can see the cachec files
  ;;
  (define-key-if-free Buffer-menu-mode-map "\C-ccd"
    'tica-buffer-list-mark-deleted)

  (define-key-if-free Buffer-menu-mode-map "\C-ccm"
    'tica-buffer-list-mark)

  (define-key-if-free Buffer-menu-mode-map "\C-ccu"
    'tica-buffer-list-unmark)

  (define-key-if-free Buffer-menu-mode-map "\C-ccf"
    'tica-buffer-list-mark-cache-off)

  ;;  Bye, bye. No need to call us again. Installation already done.
  ;;
  (remove-hook 'buffer-menu-mode-hook 'tica-define-default-keys)
  )

;;; ----------------------------------------------------------------------
;;;
(defun tica-install-msb ()
  "Install new cache menu to msb.el if it is loaded."
  (let* ((elt
	  '((and (boundp 'tica-mode)
		 tica-mode)
	    1005
	    "Cached compile files (%d)"))
	 (sym  'msb-menu-cond)
	 menu
	 )
    (when (and (featurep 'msb)
	       ;;  Install only once.
	       ;;  symbol-value just silences byteComp
	       (setq menu (symbol-value sym))
	       (not (member menu elt)))
      (push elt menu)
      (set sym menu)
      )))


;;; ----------------------------------------------------------------------
;;;  - The following says that "Load view-(minor)mode in Emacs 19.30, but
;;;    for other (X)Emacs load mview-mode if it exists."
;;;  - The view-mode is minor mode in later emacs releases, others
;;;    are encouraged to get mview.el which is minor mode for any emacs.
;;;
;;;            ~/modes/mview.el.Z
;;;            Mike Williams           <mikew@gopher.dosli.govt.nz>
;;;
(defun tica-maybe-view-mode ()
  "Turen on view (minor) mode if needed."
  (interactive)
  (let* ((zip-re   ;; arc-mode's zip file prompt
	  ;; M Filemode      Length  Date         Time      File
	  ;; - ----------  --------  -----------  --------  ---------
	  ;; -rw-r--r--      9695  10-Sep-1993  17:53:46  Makefile
	  ;; -rw-r--r--      7441   8-Sep-1993  14:21:20  README
	  ;;
	  ".*Filemode[ \t]+Length[ \t]+Date[ \t]+Time[ \t]+File[ \t]*$")


	 func
         )
    ;;  Now; if you load from dired a .zip or .tar file; you don't
    ;;  want view mode on, check buffer file name.
    ;;
    (cond
       ((and (emacs-p)
	     (< emacs-minor-version 30)
	     (feature-load 'mview)
	     ;;
	     ;; Not loaded from dired-view-file, but from compile buffer?
	     ;;
	     (not (eq major-mode 'view-mode))
	     )
	(when (fboundp 'view-mode)	;Turn off this mode.
	  (sfuncall 'view-mode 0)	;in 19.28 won't work, but try anyway
	  )
	(setq func 'mview-mode)			;Use minor mode instead
	)
       (t
	(unless (fboundp 'view-mode) (autoload 'view-mode "view" t t))
	(setq func 'view-mode)
	))

    (save-excursion
      (PMIN)
      (cond
       ((and (looking-at zip-re)
	     (fboundp 'archive-mode))
	(sfuncall 'archive-mode))
       (t
	(funcall func)
	)))
    ))

;;}}}
;;{{{ code: Buffer list control

;;; ..................................................... &buffer-list ...


;;; ----------------------------------------------------------------------
;;;
(defmacro tica-:mode-on-string-p ()
  "Check if `tica-mode' is on"
  (` (memq tica-mode '(t on))))

;;; ----------------------------------------------------------------------
;;;
(put 'tica-map-over-buffers 'lisp-indent-hook 1)
(defmacro tica-map-over-buffers (off &rest body)
  "Map over all cached buffers.
If OFF is non-nil, maps over buffers whose `tica-:cache' is off
and do BODY.

In macro you can refer to variables:

    'BuffeR'	as current buffer
    'NamE'	buffers's name
    'LisT'	current list of buffers to loop over.

Setting 'list' to nil terminates this macro."
  (`
   (let* ((LisT (tica-:cached-files (, off)))
	  BuffeR
	  NamE
	  )
     (list-loop LisT BuffeR
       (setq NamE (buffer-name BuffeR))
       (if (null NamE)  (setq NamE nil)) ;NoOp, Silence ButeComp

       (,@ body)
       ))
   ))


;;; ----------------------------------------------------------------------
;;;
(put 'tica-buffer-list-map-over-buffers 'lisp-indent-hook 0)
(defmacro tica-buffer-list-map-over-buffers (&rest body)
  "Special Buffer list macro to execute BODY at found buffer line."
  (`
   (tica-map-over-buffers nil
     (setq NamE (regexp-quote NamE))
     (PMIN)
     (when (re-search-forward NamE nil t)
       (beginning-of-line)
       (,@ body)
       ))
   ))

;;; ----------------------------------------------------------------------
;;;
(put 'tica-buffer-list-map-over-off-buffers 'lisp-indent-hook 0)
(defmacro tica-buffer-list-map-over-off-buffers (&rest body)
  "Special Buffer list macro to execute BODY at found buffer line."
  (`
   (tica-map-over-buffers 'off
     (setq NamE (regexp-quote NamE))
     (PMIN)
     (when (re-search-forward NamE nil t)
       (beginning-of-line)
       (,@ body)
       ))
   ))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tica-buffer-list-mark  ()
  "Mark Cached files in buffer list."
  (interactive)
  (tica-buffer-list-map-over-buffers (Buffer-menu-mark)))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tica-buffer-list-unmark  ()
  "Mark Cached files in buffer list."
  (interactive)
  (tica-buffer-list-map-over-buffers (Buffer-menu-unmark)))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tica-buffer-list-mark-deleted  ()
  "Mark Cached files in buffer list."
  (interactive)
  (tica-buffer-list-map-over-buffers (Buffer-menu-delete)))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tica-buffer-list-mark-cache-off  ()
  "Mark files whose cache property has been turned off."
  (interactive)
  (tica-buffer-list-map-over-off-buffers (Buffer-menu-mark)))


;;}}}
;;{{{ Dired buffer

;;; ........................................................... &dired ...

;;; ----------------------------------------------------------------------
;;;
(defun tica-dired-mark  (&optional unmark verb)
  "Mark cached files. Optionally UNMARK. VERB."
  (interactive)
  (let* ((list  tica-:cache)
	 (i     0)
	 elt
	 file
         )
    (VERB)
    (save-line-col nil nil
      (list-loop list elt
	(setq file (buffer-name elt))
	(PMIN)
	(when (re-search-forward (format "%s$" file))
	  (cond
	   (unmark
	    (beginning-of-line)
	    (when (eq (following-char)
		      (symbol-value 'dired-marker-char))
	      (inc i)
	      (dired-unmark 1)
	      ))
	   (t
	    (inc i)
	    (dired-mark 1)
	    )))
	))
    (if verb (message "%d cached files %smarked" i  (if unmark "un" "") ))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tica-dired-unmark  (&optional verb)
  "Unmark cached files. VERB."
  (interactive)
  (VERB)
  (tica-dired-mark 'unmark verb))

;;}}}

;;{{{ code: cache control

;;; ........................................................... &cache ...


;;; ----------------------------------------------------------------------
;;;
(defun tica-add-local-hook  ()
  "Make `kill-buffer-hook' local to this buffer. And add `tica-flush' to it.
When you kill the dired buffer, cached buffers loaded from this
buffer are also killed."

  ;; Make sure this hook doesn not contain tica-flush outside
  ;; the compilation buffer

  (remove-hook 'kill-buffer-hook 'tica-flush)

  (save-excursion

    ;;   Select right buffer where to localize the hook
    (cond
     ((eq major-mode 'dired-mode))

     ((boundp 'compilation-last-buffer)
      (set-buffer (symbol-value 'compilation-last-buffer)))

     (t
      (error "inproper use of tica-add-local-hook: no dired/compilation")
      ))

    ;;  force the hook local to buffer
    ;;
    (add-hook-local 'kill-buffer-hook nil)

    ;;  Make sure there are no constants in the hook.
    ;;
    (setq  kill-buffer-hook (delq t kill-buffer-hook))
    (setq  kill-buffer-hook (delq nil kill-buffer-hook))
    (add-hook 'kill-buffer-hook 'tica-flush)
    ))



;;; ----------------------------------------------------------------------
;;;
(defun tica-activate  (this-buffer buffer)
  "THIS-BUFFER is the root buffer (dired,compile) and put BUFFER to cache."
  (interactive)

  (let (list
	new-list
	elt
	)

    ;;  update cache list
    ;;
    (job-buffer this-buffer
      (make-local-variable 'tica-:cache)
      (setq list tica-:cache)
      (list-loop list elt		;Remove buffers that do not exist
	(if (not (killed-buffer-p elt))
	    (push elt new-list)))

      (setq tica-:cache new-list)

      (if (not (memq buffer tica-:cache))
	  (push buffer tica-:cache))

      (tica-mode-root-buffer)
      )

    ;;  Turn on cache mode
    ;;
    (job-buffer buffer
      (setq tica-:info this-buffer)
      ;;  - We don't touch the tica-mode if user has taken control
      ;;  - If the flag shows untouched buffer, mark the buffer
      ;;    to cache.
      ;;

      (unless tica-:mode-user-flag
	(setq tica-mode 'on)
	(setq tica-:mode-name tica-:mode-on-string)
	(ti::xe-modeline-update)
	)

      (run-hooks 'tica-:find-file-buffer-hook)
      )))


;;; ----------------------------------------------------------------------
;;;
(defun tica-:cached-files (&optional off)
  "Return all cached files by looking at every compilation buffer.
This excludes files that user has manually instructed not to be in
cache (tica-mode is off for buffer)

if optional OFF is non-nil, then return buffers whose `tica-:cache'
has been turned manually off."
  (let* ((blist (buffer-list))
	 (modes '(compilation-mode compilation-minor-mode dired-mode))
	 blist2
	 elt elt2
	 buffers
	 )
    (list-loop blist elt
      (job-buffer elt
	(when (and (memq major-mode modes)
		   tica-:cache)
	  (setq blist2 tica-:cache)

	  ;;  We must check the tica-mode for every cached file.
	  ;;
	  (list-loop blist2 elt2
	    (job-buffer elt2
	      (cond
	       ((null off)
		(if (tica-:mode-on-string-p)
		    (push elt2 buffers)))
	       (t
		(if (null (tica-:mode-on-string-p))
		    (push elt2 buffers))
		))))
	  )))
    buffers
    ))



;;; ----------------------------------------------------------------------
;;;
(defun tica-mode-root-buffer (&optional remove)
  "Update root buffer's `tica-mode' flag. Optionally REMOVE from cache.
This is the buffer where `tica-:cache' resides."
  (make-variable-buffer-local 'tica-:mode-name)
  (cond
   (remove
    (setq tica-mode nil))
   (t
    (setq tica-mode 'on)
    (setq tica-:mode-name (format "%s%d" tica-:mode-on-string (length tica-:cache)))
    )))


;;; ----------------------------------------------------------------------
;;; - This function must be called by user only!
;;;
;;;###autoload
(defun tica-mode (&optional arg)
  "Toggle cache flag for this buffer with ARG.
If the file does not belong to compile cache, calling this function
does nothing. If the file is in the cache, the modeline displays mode name.

Removing the file from cache means that the file is not killed when
the cache is flushed with \\[tica-flush]."
  (interactive)
  (unless tica-:info
    (message-quick "Can't find tica-:info, Buffer is not part of the cache?"))

  (if (null tica-mode)
      (message "This buffer is not in cache controlled.")
    (setq tica-:mode-user-flag t)
    (cond
     ((memq arg '(0 -1))
      (setq tica-mode 'off))

     ((eq arg nil)
      (if (tica-:mode-on-string-p)
	  (setq tica-mode 'off)
	(setq tica-mode 'on)))
     (t
      (setq tica-mode 'on)
      ))

    (message
     (format "Buffer is %s in cache control"
	     (if (tica-:mode-on-string-p)
		 (progn
		   (setq tica-:mode-name tica-:mode-on-string)
		   "now"
		   )
	       (setq tica-:mode-name tica-:mode-off-string)
	       "no more")))
    (ti::xe-modeline-update)
    ))



;;}}}

;;{{{ code: flush

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tica-flush-all-compilation (&optional verb)
  "Kill all cached files by stepping through all compilation buffers. VERB."
  (interactive)
  (let* ((count  0)
	 (verb   (or verb (interactive-p)))
	 )

    (tica-map-over-buffers nil
      (when (not (killed-buffer-p BuffeR))
	(kill-buffer BuffeR)
	(inc count))
	)

    (if verb
	(message (format "Flushed %d buffers." count)))

    ))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tica-flush (&optional verb)
  "Kill buffers listed in `tica-:cache'. VERB.
You must be in the Compilation/Dired buffer to execute this command.

If you're not in dired buffer, function tries to find compilation
buffer and kill compilation cached files."
  (interactive)
  (let* ((cache-buffer  (current-buffer))
	 count
	 do-it
	 buffers
	 buffer
	 )
    (setq verb (or (interactive-p) verb))

    (cond
     ((not (eq major-mode 'dired-mode))
      ;;  This calls error if no compile buffer found...
      ;;
      (setq cache-buffer (safe-do (compilation-find-buffer)))
      ))


    (when cache-buffer			;if there is buffer for us
      (job-buffer cache-buffer
	(tica-mode-root-buffer 'remove)
	;;  This variable is local to buffer
	;;
	(setq buffers tica-:cache
	      count   (length tica-:cache))

	(list-loop buffers buffer
	  (setq do-it nil)
	  (if (killed-buffer-p buffer)
	      (setq do-it t)
	    (job-buffer buffer
	      ;;  kill-buffer-hook is local to each buffer.
	      ;;  prevent alias loop
	      (let ((kill-buffer-hook kill-buffer-hook)
		    )
		(remove-hook 'kill-buffer-hook 'tica-flush)
	       (when (tica-:mode-on-string-p)	;it's allowed to kill
		 (kill-buffer buffer)
		 (setq do-it t)
		 )
	       )))
	  (if do-it
	      (setq tica-:cache
		    (delq (car buffers) tica-:cache)))
	  )))

    (if verb
	(message (format "Cache flushed [%s] files. " count)))

    ))

;;}}}
;;{{{ code: advice

;;; .......................................................... &advice ...

;;; ----------------------------------------------------------------------
;;;
(defadvice compilation-find-file (around tica activate)
  "Cache newly visited files in `tica-:cache';
use `\\[tica-flush]' in compilation buffer,
to kill these loaded files."
  (let ((this-buffer   	(current-buffer))
	(file		filename)	;ADVICE variable
	buffer
	)
    (tica-install-mode)

    ;;  Already in emacs ?
    (setq buffer (get-file-buffer file))
    ;; ... ... ... ... ... ... ... ... ... ... ... ... ...  do-advice  ..
    (prog1
	ad-do-it
      (unless buffer			;Loaded new file
	(if (setq buffer ad-return-value)
	    (tica-activate this-buffer buffer))
	))
    ))




;;; ----------------------------------------------------------------------
;;;
(defadvice compile-internal (after tica activate)
  "Automatically kill the buffers listed in `tica-:cache'
kill-buffer-hook when the `compile' buffer is killed."
  (tica-add-local-hook))


;;; ........................................................... &dired ...


;;; ----------------------------------------------------------------------
;;;
(defadvice dired-view-file (around tica act)
  "Cache newly visited files in `tica-:cache';
Kill the dired buffer to kill cached files."
  (let ((this-buffer   	(current-buffer))
	(file		(dired-get-filename))
	buffer
	)
    (tica-install-mode)

    ;; ... ... ... ... ... ... ... ... ... ... ... ... ...  do-advice  ..


    (cond
     ((file-directory-p file)
      (setq buffer t))			;skip this one
     ((get-file-buffer file)		;Already in emacs ?
      (setq buffer t))
     (t
      (setq buffer (get-file-buffer file))
      ))

    ;; ... ... ... ... ... ... ... ... ... ... ... ... ...  do-advice  ..
    (prog1
	ad-do-it
      (when (and (null buffer)			;Loaded new file
		 (setq buffer (get-file-buffer file)))
	(tica-activate this-buffer buffer)
	))
    ))



;;}}}

;;   delayed install
;;
(add-hook 'buffer-menu-mode-hook 'tica-buffer-list-bindings-on)
(add-hook 'compilation-mode-hook 'tica-define-default-keys)
(add-hook 'dired-mode-hook       'tica-add-local-hook)
(add-hook 'dired-mode-hook       'tica-define-default-keys)


(provide   'tinycache)
(run-hooks 'tica-:load-hook)

;; tinycache.el ends here
