;; @(#) tinycerr.el -- General parser for compile errors.

;; @(#) $Id: tinycerr.el,v 1.31 1997/04/21 09:27:32 jaalto NotReleased $
;; @(#) $Keywords: lib, language, error parsing, compile $
;;
;; $KnownCompatibility: 19.28,30 $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1995,96,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      May 02 1995
;;
;; To get information on this program use ident(1) or do M-x tice-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el


;; LCD Archive Entry:
;; tinycerr|Jari Aalto|jari.aalto@poboxes.com|
;; General parser for compile errors.|
;; 21-Apr-1997|$Revision: 1.31 $|~/misc/tinycerr.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ........................................................ &t-install ...
;; - Put this file on your Emacs-Lisp load path, add following into your
;;   to some of your own custom language parser file
;;
;;	(require 'tinycerr)
;;
;;   or more prefered
;;
;;	(autoload 'tice-compile "tinycerr")
;;
;;; If you have any questions, use this function
;;
;;      M-x tice-submit-bug-report


;;}}}
;;{{{ Documentation

;;; Commentary:
;; ..................................................... &t-commentary ...


;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "tice-" in front of
;;   every function & variable. It stands for '(ti)ny (c)ompile
;;   (e)rror handler
;; - variable names contain letters 'tice-:'.

;; ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;; Repeat: This is no good for you on ITS OWN, you have to have
;;         parser build on top of it!
;;
;; ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;;
;;  Preface, may 1995
;;
;;	I got tired of goto-line, when I had a flint output in the
;;	compile buffer, I thought, that there has to be some way to write
;;	one's own error parser. I looked around a bit, but I could see
;;	anything *good* in standard 19.28 distribution.
;;
;;	The compile.el was too limited, because it relied on regexps and
;;	lint output was spread over multiple lines --> I started to cde
;;	this to replace compile.el
;;
;;	First the tinygflint.el and tinycerr.el were one and the same
;;	package, but I realized, that I should separate them, so that
;;	general error parser could provide a base whe other's build.
;;
;;	Good luck for writing your own language parser with the aid of
;;	this package; hope it helps you to get going!
;;
;;  When you would use this package
;;
;;
;;	o   This is LIBRARY file, so it can't stand on it's own, you have to
;;	    write a parser around it. See example of usage in ohio archive file
;;	    tinygflint.el , which parses C/C++ error checker's, The Gimpel
;;	    Flint, output.
;;	o   Provide functions to parse Compilation buffer easily
;;
;;	If your compiler produces simple error, such that fit into ONE
;;	line, then you propable be better off with std. Emacs 19.xx
;;	package compile.el. Just go and add some more regexps and it will
;;	parse errors for you.
;;
;;	If your compiler generates errors that can spread into multiple lines,
;;	then I'd suggest that you give this package a whirl.
;;
;;  How to use
;;
;;	The best thing is that you immediately load tinygflint.el ,which
;;	is constructed over the top of this .el. Don't worry if you don't
;;	know what the heck is flint and what is it good for. Just get
;;	the package, print it and watch how the parser uses this file.
;;
;;	When you're creating parser for some other language, you should
;;	change the appropriate variables. The tice-compile always resets
;;	itself to these values before continuing and you can change them
;;	by putting your own function into variable tice-:local-var-func.
;;	An example function is supplied, which you could use:
;;
;;	    tice-set-defaults
;;
;;	When you want to use normal compilation after tinycerr, use following
;;	command tore store all bindings in compilation buffer
;;
;;	    M-x tice-restore-compile-map
;;
;; Compilation preparations
;;
;;	You should configure your compile command so that the it prints
;;	it version number and vendor name, just before it starts compiling.
;;	If there is no such switch available [very usual, when using native
;;	system compiler CC and cc], you should make a compile script, that
;;	provides the necessary information. Like the following, put the script
;;	under name ~/bin/CC and you don't notice using script at all, that
;;	absolute path makes sure the _real_ compiler is called. And of course
;;	The path has to have you ~/bin before the /usr/bin/ directory for
;;	this trick to succeed.
;;
;;	Try command 'what' or 'strings' over your compiler, it will possibly
;;	reveal the necessary comnpiler version information too.
;;
;;	    !#/bin/sh
;;	    # - You could use echo `what /usr/bin/CC`
;;	    # - here is the same text string. This is lot faster when we don't
;;	    #   need to run 'what' command.
;;	    echo "CompilerVersion: HP C++ B2402/B2404  A.03.20"
;;	    /usr/bin/CC $*
;;
;;   Don't forgot to check that it works:
;;
;;	    % chmod +x ...
;;	    % rehash                  # rebuild the tables
;;	    % which CC                # is my path prior the system paths?
;;	    ~/bin/CC
;;
;;	    % echo $PATH              # you need to modify this if NOK ...
;;
;;  Highlighting used
;;
;;	This package doesn't use any of the existing highlighting packages
;;	such as hilit19, face-lock or font-lock. The coloring functions
;;	come from general lisp library tinylibo.el. Take a look at it and
;;	you can easily use those function for your own purposes.
;;	All overlays used in this package have property
;;
;;	    'owner
;;
;;	And the properties values is set to value returned by function
;;
;;	    (tice-v 'package)
;;
;;	This makes it possible to use other highlighting packages too with
;;	this module, provided that they can identify their own overlays.
;;	You don't like colors? Want to get rid of them fast? Okay, there
;;	is function which dehilits all by removing all overlays owned by
;;	TICE.
;;
;;	    tice-unparse
;;	    tice-parse-again
;;
;;	The 'again' function can be used only after initial compilation,
;;	because it requires that all the needed variables are already set.
;;	If some variables get lost you may not get hilit back and the
;;	jumps won't work --> run compile again.
;;
;;  Moving to errors
;;
;;	All the basic commands are in
;;
;;	    tice-compile-map
;;
;;	Error movements NEXT/PREV moves only to lines that have
;;	_file_ and _line_ information immediately available. They jump
;;	over other lines that have mouse-face property, but the file and
;;	line information is not fully available. You can click over these
;;	'guess' line to see if there is some special error parser
;;	provided by the TOP parser that looks around the point.
;;
;;  Note: keyboard vs. mouse
;;
;;	Remember that using the mouse commands aren't the same than using
;;	the keyboard.  Mouse commands moves immediately the point, and
;;	then executes the command whereas keyboard command activates at
;;	the current point the cursor is on.
;;
;;  Note: using folding.el or outline
;;
;;	The goto-error does not work properly if source buffer is folded.
;;	Please execute fold-open-buffer before you use the compile output's
;;	goto-error.

;;}}}
;;{{{ history

;; ........................................................ &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; Apr	21	1996	[jari]		19.28	v1.31		NotReleased
;; - internal changes.
;;
;; Sep	25	1996	[jari]		19.28	v1.28-1.30	NotReleased
;; - Rearranged loading. now uses autoloads.
;; - Changed all '(lambda forms to use (function (lambda
;;
;; Aug  20	1996    [jari]          19.28   v1.26-1.27	NotReleased
;; - All Errors reported by XE 19.14 ByteCompiler corrected. Don't
;;   know if this module still works in XE...
;;
;; Jun  28	1996    [jari]          19.28   v1.25           NotReleased
;; - Removed extra require commands: 'faces 'mouse. XE doesn't have
;;   feature 'faces. Tip by <samuels@noise.bsd.uchicago.edu>.
;;
;; May  29	1996    [jari]          19.28   v1.24           NotReleased
;; - If you run TICE, and immediately compile, then the tiec-finish
;;   function was still beeing run. Ans the TICE keymap left to
;;   compilation buffer. Corrected.
;;
;; May  25	1996    [jari]          19.28   v1.23           NotReleased
;; - Changed function in tinylibo.el and it needed change here.
;;   tice-ov-remove-area		:! parameter order to lib 'o'
;;
;; May  16	1996    [jari]          19.28   v1.22           NotReleased
;; - Corrected some variable names. Seems working fine in 19.30 too.
;;
;; Apr  19      1996    [jari]          19.28   v1.21           NotReleased
;; - Reverted to previous version 1.19, because I found better way
;;   to cope with XE. There is now new lib tinylibxe.el
;;
;; Apr  18      1996    [jari]          19.28   v1.20           NotReleased
;; - Removed all Emacs specific overlay commands and used portable
;;   lib 'O' commands instaed. This should be XE compatible.
;;
;; Feb	11	1996	[jari]		19.28	v1.19		NotReleased
;; - there were some typos that prevented using this module. Now corrected.
;;
;; Feb	10	1996	[jari]		19.28	v1.18		NotReleased
;; - lisp-mnt.el's lm-verify checked. Rewrote version and feedback funcs.
;;   Corrected some byte compiler errors. Added ### autoloads.
;;
;; Sep	4	1995	[jari]		19.28	v1.17		NotReleased
;; - Checked that byte compiles without errors. [many fixes]
;; - Added more documentation, changed bookmarks to standard convention
;; - some substle changes [code rearranged a lot, so don't trunst on diff]
;;
;; Aug	28	1995    [jaalto]        19.28	v1.16		NotReleased
;; - tice-restore-compile-map, tice-set-compile-map, just minor changes.
;;
;; May	24	1995    [jari]		19.28	v1.15		NotReleased
;; - It's been long time I last see this file..Hmmm? Seems that I haven't
;;   touched the basic code much. Some minor fine tuning.
;; + tice-compile-prepare , tice-compile-restore
;;
;; May	10	1995    [jari]		19.28	v1.14		NotReleased
;; - Started seriously developing libraries for lisp programs. The
;;   overlay handling is done by library tinylibo.el now.
;; - moved/modified: tice-ov-mouse-on-p , tice-ov-get-prop
;;   tice-ov-get-prop-v , tice-ov-get-mouse , tice-ov-get-mouse
;;
;; May	6	1995    [jari]		19.28	v1.13		NotReleased
;; - Removed all 'make-buffer-local...' commands, I decided that there
;;   was no point of using them. User must set the variables _always_
;;   before he compiles, so default settings do no good.
;;   Instead we set the default TICE package defaults through function
;;   tice-set-defaults
;; ! Rearranged huge amount of code due to compatibility problems. I
;;   started writing totally different package tiny-hpc++.el, which
;;   uses this engine to do C++ compiling. It turned out that degining
;;   package for flint only had led to various false assumptions :(
;;   Now it is much more robust.
;; ! tice-get-compiler-version , tice-is-supported , Compile section
;;   carefully rewritten.
;;
;; May	6	1995    [jari]		19.28	v1.12		NotReleased
;; - Hmm, the nexp/prev error movement was badly formed. Rewrote whole
;;   procedures again. Now the next-file&move is obsolete function and
;;   paragraph marks are not necessary at the moment...
;; - While changes are few, they affect greatly to the use. Gosh!
;;   saved previous ves as _binary_
;; - tice-goto-error-next , tice-goto-error-prev rewrote
;; + macros: BOLP, EOLP , function tice-be deleted
;; + tice-goto-error-blink , tice-goto-error-scan
;;
;; May	5	1995    [jari]		19.28	v1.11		NotReleased
;; - Again major upgrade.
;; - code arranges so much that diff doen't show well changes...
;; ! error in tice-ov-mouse-on-p corrected.
;; ! tice-is-mouse-line totally rewritten, now uses low level funcs,
;;   tice-read-prop too.
;; ! tice-goto-error rewritten
;; + Mouse binding + [np] for next and previous error
;; + VAR: tice-:paragraph-separate, tice-:paragraph-start
;;        tice-:blink-delay
;; + FUN: tice-get-next-file&line , tice-prop-mouse-line-on
;;        tice-prop-mouse-line-off , tice-prop-mouse-line
;;        tice-goto-error-prev , tice-goto-error-next
;;        tice-goto-error-1
;;
;;
;; May	4	1995    [jari]		19.28	v1.8-1.10	NotReleased
;; - This is amajor upgrade.
;; - Compile function changed so that the tice-dummy removed,
;;   it always jumps to 'tice-compile-finish.
;; + unparse, parse-again: so that hilit can be turned on/off
;; + lot's of new/changed functions: tice-ov-remove-area,
;;   tice-ov-mouse-on-p, tice-ov-get-prop, tice-ov-get-prop-v
;;   tice-ov-get-eval , tice-get-single-prop , tice-get-file
;;   tice-get-line , tice-prop-mouse-line , tice-ov-get-prop
;;   tice-prop-mouse-line-on , tice-prop-mouse-line-off
;; + Now every instanse adds property 'owner to overlays and
;;   text properties.
;; + inc: macro added
;;
;; May	4	1995    [jari]		19.28	v1.7		NotReleased
;; - Accidentally saved file in pure ascii ^M, Not folds...corrected.
;;
;; May	4	1995    [jari]		19.28	v1.7		NotReleased
;; - I have run version 1.6 succesfuly for some time. Seems to be
;;   stable for a little t.cc test. Full blown errors are not tested
;;   yet. It may choke to them.
;; + tice-ov-get-prop , tice-ov-get-eval
;;   The overlay LIST functions created, so that handling is much
;;   easier, maybe slower too.
;; + tice-compile-finish-beg-hook, tice-compile-finish-end-hook
;;   The begin hook added, beacuse it is usefull for filtering
;;   purposes. The end hook renamed accordingly.
;;
;; May	3	1995    [jari]		19.28	v1.2-1.6	NotReleased
;; - These are intermediate versions and do not necessarily work
;;   properly. Lot's of problems and corrections concerning overlays
;;   and mouse hiliting.
;;
;; May	2	1995    [jari]		19.28	v1.1		Created


;; To do list:
;; ========================================
;; - Some of these functions are currently transferred to general
;;   tinylibXXX.el lisp libraries and when the libraries are commonly
;;   available, the overlapping functions are removed from this file.


;; Thank you list:
;; Simon.Marshall@mail.esrin.esa.it (Simon Marshall)
;;   [ author of best selling lazy-lock.el, If you don't already have	]
;;   [ it, go and get it fast! 						]
;; - Simon kindly suggested that the overlay should have 'owner
;;   property, then the properties are not copied along with paste.
;;   Now it's possible to have clean cut & copy out of compile buffer.




;;}}}



;;; Code:

;;{{{ setup: require

;;; ......................................................... &require ...

(require 'compile)
(require 'tinylibm)
(require 'tinylibo)

(eval-and-compile
  (if (xemacs-p) (require 'tinylibxe))
  )

(eval-when-compile (use-dynamic-compilation))


;;}}}
;;{{{ setup: -- hooks

;;; ......................................................... &v-hooks ...

;;  You could use this hook for message filtering purposes
;;
(defvar tice-:compile-finish-beg-hook nil
  "*Hook that is run, when the compile command itself has finished.
In this point buffer is not yet parsed, only the compiler version
string is stored [see tice-v function].")


(defvar tice-:compile-finish-end-hook nil
  "*Hook that is run after the compile and the parsing is done.
It is called after the *-finish-[ok|err]* functions are called.*

If either of finish-[ok|err] returns error, the hook is not called.
")

(defvar tice-:define-keys-hook '(tice-keys-defaults)
  "*Hook that defines keybindings.")


(defvar tice-:load-hook nil
  "*Hook run when file has been loaded.")


;;}}}
;;{{{ setup: -- bind

;;; .......................................................... &v-bind ...


;;   We use our own mouse-2 to search the errors, not compile buffer's
;;   own search
;;   Keybindings are defined by running variable tice-:define-keys-hook
;;
(defvar tice-compile-map (make-sparse-keymap)
  "local keymap.")


;;}}}
;;{{{ setup: -- private


;;; ....................................................... &v-private ...
;;; - Please use tice-v function to access variables, because the variable
;;;   names here may be changed.
;;;


(defconst tice-:package-sym 'tice
  "value to use for property 'owner")

(defvar tice-:compile-finish-orig nil
  "Storage of original compile function.")

;;  This is hardcoded in compile.el::compile-internal
(defvar tice-:compile-buffer "*Compilation*"
  "*Compile output buffer")

(defconst tice-:compiler-str nil
  "The Compiler version string will be stored here.")

(defconst tice-:compile-finish-status nil
  "Status of compilation is stored here.")


(defvar tice-:load-allow-re nil
  "See tice-:load-allow-re-func. This variable is always set to nil after
compilation. NIL means to match all files thus efectively allowing
loading all files.")

(defvar tice-:overlay-mark nil
  "For Marker")

(defvar tice-:overlay-mouse nil
  "For Mouse")

(defconst tice-:ret nil
  "Global return value from certain functions.")


(defconst tice-:re-face-table-default
  '(
    ((".*\\(Fatal\\)"   . 1)   .  tice-:f-fatal)
    ((".*\\(Error\\)"   . 1)   .  tice-:f-err)
    ((".*\\(Warning\\)" . 1)   .  tice-:f-warn)
    ((".*\\(Info\\)"    . 1)   .  tice-:f-info)
    ((".*\\(Note\\)"    . 1)   .  tice-:f-note)
    )
  "List in a form \(\(REGEXP . SUB-LEVEL\) . FACE_VARIABLE \) which are
used for marking text. User should not touch this.")



;;}}}
;;{{{ setup: user conf

;;; ........................................................ &v-public ...
;;; - These variables will be set in function tice-set-defaults
;;;   YOU CAN'T DEFINE all of them THEM with setq, because they are
;;;   always overwritten before compile happend.

(defvar tice-:verbose t
  "*Allow information messages.")

(defvar tice-:msg-sleep nil
  "*sit-for time for message display. Nil for no waits.")

(defconst tice-:blink-delay nil
  "*sit-for time for when cursor point is momemtarily shown.
Nil to no delays.")


;; your best bet is to keep PARA-START and PARA-SEP equal.
(defvar tice-:paragraph-separate "^\\s *$\\|^\f"
  "*See paragraph-separate.")

(defvar tice-:paragraph-start tice-:paragraph-separate
  "*See paragraph-start.")

(defvar tice-:max-lines nil
  "*How many lines at maximum is marked. If you have very long
compilation output, it may take quite a while to parse whole buffer.
Normallly you only need information of the first 10-30, correct the error
and run compile again.")

(defvar tice-:goto-unmarked nil
  "*If you unmark the line with 'u' you can choose if the goto
command on that line takes effect. If this variable is t, you can use goto
if it is nil, jumps are inhibited.")


(defvar tice-:compile-finish-ok-re nil
  "*Regext that matches status of the compilation when exited,
    this represents ok condition. Anything else in considered error and
tice-:compile-finish-err-func is called")

(defvar tice-:local-var-func nil
  "Function to ser verride variables for this compilation. Called in
various places.")

(defvar tice-:goto-err-func nil
  "*Function that is called if the line does not contain
enough information to jump to file & line. This function is
called only if the line is 'candidate', meaning that it had matched
the first parameter[regexp] of tice-parse.

The function must return non-nil if it takes care of the jumping to the
right point. If it returns nil, nothing further is done.
")

(defvar tice-:compile-finish-ok-func nil
  "*Function to be run after tice compile ends succesfully")

(defvar tice-:compile-finish-err-func nil
  "*Function to be run after tice compile was unsuccesfull")

(defvar tice-:compile-cmd nil
  "*Command to run in compile buffer.")

(defvar tice-:compile-cmd-func nil
  "*Function to compose the compile command. It is called with 3
parameters: full file-name, add-opt and cmd-set
If this function exist it overrides tice-:compile-cmd  variable.")

(defvar tice-:mark-area-func 'tice-mark-area
  "*Function that return list of begin and end of the mark.")

(defvar tice-:compiler-ver-re "CompilerVersion:"
  "*Regexp that is tried against tice-:compiler-str. If re matches
it meabs that the version is supported.

Even if the compiler has no option, which makes it output the version
string, you should generally output one manually as a first line of
compilation output. Use separate shell script that outputs appropriate
information before starting the compiler.

This can be added to standard UNIX makefile:

#==== Building rule for the program for HP-UX /bin/CC
$(PROGRAM):     $(OBJS)
                @ echo \"\n\nlinking whole program, CompilerVersion: X.xx \"
                $(LNK) -o $@ $(LNKFLAGS) $(OBJS) $(PROJLIBS) $(LNKLIBS)
                @ echo \"\nlinking finished.\n\"

Makefile note: remember that there are TABS at the front of this rule.
")



(defvar tice-:compiler-ver-func 'tice-get-compiler-version
  "*Function which returns the string which represents compiler version.
The return value is stored to tice-:compiler-str")


(defvar tice-:compiler-support-re "."
  "*Regexp that is tried to match against Compiler's version string.
Dot mathes all versions.")


(defvar tice-:compiler-support-func 'tice-is-supported
  "*Function that returns t/nil if this compile version is supported.
It will be called with a string parameter that is returned by
function tice-:compiler-ver-func. ")


;;  Remeber that the filename tested here is buffer-name which
;;  return "t.cc<2>"  forms too.
;;
(defvar tice-:compile-file-re "[.]\\(cc\\|c\\).*[>]?$"
  "*Files that can be compiled.")


(defvar tice-:compile-file-sorry-msg "Sorry, can't compile this file."
  "*Message to be displayed if tice-:compile-file-re not matched.")

(defvar tice-:load-no-re "/usr/include"
  "*files that are not loaded for errror inspection.")

(defvar tice-:load-ask t
  "*If file is not in the buffer, ask if it is loaded for you.")

(defvar tice-:load-allow-re-func nil
  "*Function that return RE to allow loading files, for example if you're
developing files starting with ABCxxxx in a project, you may want to visit
only those file's errors. Returning nill means that _all_ files are
valid, the tice-:load-no-re is applied after this.

The function is called immediately when the compilation has finished.
The variable tice-:load-allow-re is initially set to nil, but
if this function returns non-nil, the value returned will
be assigned to it.

")

(defvar tice-:re-face-table tice-:re-face-table-default
  "*The face table that is used during marking the error texts.")

;;}}}
;;{{{ setup: user faces

;;;  ......................................................... &v-faces ...

(defvar  tice-:f-mouse 'highlight
  "*Mouse select face.")

(defvar  tice-:f-mark
  (progn
    (make-face 'tice-:f-mark)
    (set-face-foreground 'tice-:f-mark "black")
    (set-face-background 'tice-:f-mark "yellow")
    'tice-:f-mark)
  "*face for mark.")

(defvar tice-:f-fatal
  (progn
    (make-face 'tice-f-fatal)
    (set-face-foreground 'tice-f-fatal "red")
    'tice-f-fatal)
  "*face for Fatals.")


(defvar tice-:f-err
  (progn
    (make-face 'tice-f-err)
    (set-face-foreground 'tice-f-err "blue")
    'tice-f-err)
  "*face for 'Errors'")


(defvar tice-:f-warn
  (progn
    (make-face 'tice-f-warn)
    (set-face-foreground 'tice-f-warn "yellow")
    'tice-f-warn)
  "*face for Warnings.")

(defvar tice-:f-info
  (progn
    (make-face 'tice-f-info)
    ;;  It happens that "X server cannot allocate color"
    ;;  for exotic colors
    (condition-case nil
	(set-face-foreground 'tice-f-info "turquoise")
      (error				;let's be conservative then
       (set-face-foreground 'tice-f-info "grey")))
    'tice-f-info)
  "*face for Informational Warnings.")

(defvar tice-:f-note
  (progn
    (make-face 'tice-f-note)
    (set-face-foreground 'tice-f-note "green")
    'tice-f-note)
  "*face for Notes")

(defvar tice-:p-file 'file-name
  "*Property to put for the _file_name_.")

(defvar tice-:p-line 'file-line
  "*Property to put to the _line_number_ of file name.")

;;}}}

;;{{{ setup: -- version

;;; ...................................................... &v-version' ...

(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinycerr.el"
 "tice"
 tice-:version-id
 "$Id: tinycerr.el,v 1.31 1997/04/21 09:27:32 jaalto NotReleased $"
 '(tice-:version-id

   tice-:compile-finish-beg-hook
   tice-:compile-finish-end-hook
   tice-:define-keys-hook
   tice-:load-hook
   tice-compile-map
   tice-:compile-finish-orig
   tice-:compile-buffer
   tice-:load-allow-re
   tice-:overlay-mark
   tice-:overlay-mouse
   tice-:verbose
   tice-:msg-sleep
   tice-:paragraph-separate
   tice-:paragraph-start
   tice-:max-lines
   tice-:goto-unmarked
   tice-:compile-finish-ok-re
   tice-:local-var-func
   tice-:goto-err-func
   tice-:compile-finish-ok-func
   tice-:compile-finish-err-func
   tice-:compile-cmd
   tice-:compile-cmd-func
   tice-:mark-area-func
   tice-:compiler-ver-re
   tice-:compiler-ver-func
   tice-:compiler-support-re
   tice-:compiler-support-func
   tice-:compile-file-re
   tice-:compile-file-sorry-msg
   tice-:load-no-re
   tice-:load-ask
   tice-:load-allow-re-func
   tice-:re-face-table
   tice-:f-mouse
   tice-:f-mark
   tice-:f-fatal
   tice-:f-err
   tice-:f-warn
   tice-:f-info
   tice-:f-note
   tice-:p-file
   tice-:p-line

   tice-:package-sym
   tice-:compiler-str
   tice-:compile-finish-status
   tice-:ret
   tice-:re-face-table-default
   tice-:blink-delay
   )))



;;}}}
;;; ########################################################### &funcs ###

;;{{{ macros

;;;  .......................................................... &macros ...


;;; ----------------------------------------------------------------------
;;;
(defsubst tice-dummy (&rest args) "No op." nil)


;;}}}

;;{{{ misc

;;; ................................................ variable-handling ...

;;; ----------------------------------------------------------------------
;;;
(defun tice-v (sym)
  "External package should use this function to read the
TICE's variables, because they are buffer local."
  (cond
   ((eq sym 'para-beg)        tice-:paragraph-start)
   ((eq sym 'face-table)      tice-:re-face-table)
   ((eq sym 'para-s)          tice-:paragraph-separate)
   ((eq sym 'compile-buf)     tice-:compile-buffer)
   ((eq sym 'compile-cmd)     tice-:compile-cmd)
   ((eq sym 'compiler-str)    tice-:compiler-str)
   ((eq sym 'package)         tice-:package-sym)
   ((eq sym 'fin-status)      tice-:compile-finish-status)
   ((eq sym 'load-allow-re)   tice-:load-allow-re)
   ((eq sym 'ask)             tice-:load-ask)
   (t
    (error "No such symbol %s" sym))
   ))


;;; ----------------------------------------------------------------------
;;;  - You should copy this and make the necessary modifications
;;;    when you're making your own parser.
;;;
(defun tice-set-defaults ()
  "set some TICE defaults."
  (let* ()
    (setq-default

     ;;  too bad this is hard-coded in compile.el...
     tice-:compile-buffer	"*Compilation*"

     tice-:load-allow-re	nil	            ;accept all
     tice-:paragraph-separate   "^\\s *$\\|^\f"
     tice-:paragraph-start 	tice-:paragraph-separate
     tice-:re-face-table        tice-:re-face-table-default

     tice-:verbose		t
     tice-:msg-sleep  		nil
     tice-:blink-delay		0.5
     tice-:max-lines		100
     tice-:goto-unmarked	t
     tice-:compile-finish-ok-re "finish"
     tice-:compile-finish-ok-func    nil
     tice-:compile-finish-err-func   nil
     tice-:compile-cmd		nil
     tice-:compile-cmd-func	nil
     tice-:mark-area-func	'tice-mark-area
     tice-:compiler-ver-re	"CompilerVersion:"
     tice-:compiler-ver-func	'tice-get-compiler-version
     tice-:compiler-support-re	"."
     tice-:compiler-support-func 'tice-is-supported
     tice-:load-no-re		 "/usr/include"
     tice-:load-ask		 t
     tice-:load-allow-re-func	 nil
     )
    ))

;;; ............................................................. misc ...

;;; ----------------------------------------------------------------------
;;;
(defun tice-keys-defaults ()
  "Insert default bindings. "
  (interactive)

  (ti::k-put-abc-map tice-compile-map 'ignore) ;disable writing to buffer

  (define-key tice-compile-map [mouse-2] 'tice-goto-error)
  (define-key tice-compile-map " "
    (function
     (lambda ()
       (interactive) (tice-goto-error nil))))


  (define-key tice-compile-map [mouse-3] 'tice-mouse-line-toggle)
  (define-key tice-compile-map "u"
    (function
     (lambda ()
       (interactive) (tice-mouse-line-toggle nil))))


  ;;  The DOWN event usually brings up the Font menu, disable it.
  (define-key tice-compile-map [C-mouse-3] 'tice-mouse-mark)
  (define-key tice-compile-map [C-down-mouse-3] 'ignore)
  (define-key tice-compile-map "m"
    (function
     (lambda ()
       (interactive) (tice-mouse-mark nil))))

  ;; Alt - ctrl
  (define-key tice-compile-map [C-M-mouse-3] 'tice-mouse-mark-jump)
  (define-key tice-compile-map "j"
    (function
     (lambda ()
       (interactive) (tice-mouse-mark-jump nil))))

  (define-key tice-compile-map [S-mouse-1] 'tice-goto-error-prev)
  (define-key tice-compile-map "n"
    (function
     (lambda ()
       (interactive) (tice-goto-error-prev nil))))

  (define-key tice-compile-map [S-mouse-2] 'tice-goto-error-next)
  (define-key tice-compile-map "p"
    (function
     (lambda ()
       (interactive) (tice-goto-error-next nil))))

  t					;just some return value
  )


;;; ----------------------------------------------------------------------
;;;
(defun tice-restore-compile-map ()
  "Restores compile map."
  (interactive)
  (use-local-map compilation-minor-mode-map)
  )

;;; ----------------------------------------------------------------------
;;;
;;; Hmm, ** DO NOT USE THIS **
;;;
(defun tice-set-compile-map ()
  "Redefines compilation-minor-mode-map."
  (let* ((map (current-local-map))
	 )

    (error "do not use this func")
;;; (d! "compile map~")
    (define-key map [mouse-2] 'tice-goto-error)

    ;;  clear compile.el keys
    (define-key map "\C-c\C-k" 'ignore)
    (define-key map "\M-n" 'ignore)
    (define-key map "\M-p" 'ignore)
    (define-key map "\M-{" 'ignore)
    (define-key map "\M-}" 'ignore)

    ))



;;; ----------------------------------------------------------------------
;;;
(defun tice-msg (msg &optional eflag)
  "Print package's error message. Signal error if eflag raised"
  (let* ((verb  tice-:verbose)
	 (sleep tice-:msg-sleep)
	 )
    (if (null verb) nil
      (message
       (concat (if eflag "*err "  "")
	       "tice: " msg))
      ;;  There may be successive messages...
      (and sleep (sit-for sleep)))
    ))


;;; ----------------------------------------------------------------------
;;; - This is default checking funtion, which top parser can use.
;;;   see tinygflint.el
;;;
(defun tice-load-allowed-p (file)
  "Checks if file is allowed to load."
  (let* ((allow-re  (or tice-:load-allow-re "."))
	 (no-re     (or tice-:load-no-re "blurb#x"))
	 )
    (if (and (stringp file)
	     (string-match allow-re file)
	     (not (string-match no-re file)))
	t nil)
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tice-read-match (re &optional level)
  "Return string matched at LEVEL. default is level 0 = full match"
  (let* ((lev (or level 0))
	 )
    (if (not (and re (looking-at re) (match-end lev)))
	nil
      (buffer-substring (match-beginning lev) (match-end lev))
      )))



;;; ----------------------------------------------------------------------
;;;
(defun tice-get-c++-include-re  (re &optional level)
  "Returns include RE by looking with RE and picking the file in LEVEL.
Default is level 0."
  (let* ((ref (or (tice-read-match re 1) ""))
	 (extra
	  (cond
	   ;;  There are include <sys/time.h> and <time.h>
	   ((string-match "/sys/" ref) "sys/")
	   (t "")))
	 (inc (concat extra (file-name-nondirectory ref)))
	 ret
	 )
    (if (not (and re (not (equal ref ""))))
	nil
      (setq ret (concat "^#[ \t]*include[ \t]+[<\"][ \t]*" inc))
      )
;;;    (d! "~inc" inc ret)
    ret))




;;; ----------------------------------------------------------------------
;;;
(defun tice-mark-area  ()
  "Returns mark area"
  (let* ((mb (BOLP))
	 (mee (EOLP))
	 (me (- mee 7))		;7 char mark to the right
	 )
    (if (< me mb)			;not that long line.
	(list mb mee)
      (list me mee)
      )
    ))



;;; ----------------------------------------------------------------------
;;; - Acts like pop-to-buffer
;;;
(defun tice-pop-file  (file line)
  "Goes to FILE at line LINE, loads file is it's not in emacs."
  (let* (;;  Buffer name has no directories,
	 ;;  -- In case of t.cc<2> we can't know which is the right one
	 (f     (file-name-nondirectory file))
	 (buf   (get-buffer f))
	 (exist buf)			;exist ?
	 )

;;;    (d! "~pop" f exist)
    (if buf nil
      (if (null (setq exist (file-exists-p file)))
	  nil
	(setq buf (find-file-noselect file))))
    (if (null exist)
	nil			;not exist
      (pop-to-buffer buf)
      (goto-line line))
    ))



;;}}}

;;{{{ mouse
;;; ########################################################### &mouse ###
;;; Mouse command handling



;;; ----------------------------------------------------------------------
;;;
(defun tice-mouse-line-toggle (event)
  "Toggless mouse face on the line. "
  (interactive "e")
  (let* (ov
	 ret
	 )
    (if event (mouse-set-point event))	;move point there

    (setq ov (ti::o-get-mouse))
    (setq ret (ti::o-mouse-on-p ov))

    (cond
     ((null ov)
      nil)				;not a mouse line
     ((eq ret nil)
      (tice-prop-mouse-line-on))	;set
     (t
      (tice-prop-mouse-line-off)))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tice-mouse-mark-jump (event)
  "Jumps to mark."
  (interactive "e")
  (let* ((ov tice-:overlay-mark)
	 p
	 )
;;;    (if event (mouse-set-point event))	;move point there
    (if (not (and ov
		  (setq p (overlay-start ov))))
	(tice-msg "No mark.")
      (goto-char p)
      (beginning-of-line))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tice-mouse-mark (event)
  "Marks current position."
  (interactive "e")
  (let* (
	 (prop tice-:f-mark)
	 (func tice-:mark-area-func)
	 mb me
	 el
	 )
    (if event (mouse-set-point event))	;move point there
    (or tice-:overlay-mark			;overlay to use ?
	(progn
	  ;;	  make-local-variable 'tice-:overlay-mark
	  (setq tice-:overlay-mark (make-overlay (point) (point)))
	  (overlay-put tice-:overlay-mark  'owner (tice-v 'package))
	  (overlay-put  tice-:overlay-mark 'face prop)
	  (overlay-put  tice-:overlay-mark 'mark t)
	  ))


    ;;   get the area region to be marked
    (setq el (funcall func))

    (if (not (eq 2 (length el)))
	(error "tice-:mark-area-func invalid return value")
      (setq mb (nth 0 el) me  (nth 1 el))
;;;      (d! "~mark" mb me)
      (move-overlay tice-:overlay-mark mb me)

      )))

;;}}}

;;{{{ lowlevel overlays

;;; ###################################################### &properties ###

;;; ......................................................... &overlay ...



;;; %%not used yet
;;; ----------------------------------------------------------------------
;;;
(defun tice-ov-get-prop-v (ovl prop &optional val)
  "Return first _overlay_ that has property PROP set to value VAL.
Defaul is t for VAL."
  (let* (ov)
    (setq ov (ti::o-get-prop-val ovl (list prop (or val t))))
;;;    (d! "~1" prop val (prin1-to-string ov))
;;;    (setq ov (tice-ov-get-prop-v))
;;;    (d! "~2" (prin1-to-string ov))
    ov
    ))


;;; #not used
;;; ----------------------------------------------------------------------
;;;
(defun tice-ov-get-eval (ovl eval)
  "Return _overlay_, if the item in the overlay list OVL stisfies
the EVAL form.

You should use variable 'ovx' to refer to everlay item in the EVAL.
"
  (let (ov ovx)
    (while (and (null ov) ovl)
      (setq ovx (car ovl))
      (if (eval ov)
	  (setq ov ovx)		;ok, reuse it
	(setq ovl (cdr ovl))))
    ov
    ))




;;; ----------------------------------------------------------------------
;;;
(defun tice-ov-remove-area (owner &optional beg end)
  "Removes all overlays owned by OWNER. Default area is whole buffer."
  (let* (
	 (p   (or beg (point-min)))
	 (max (or end (point-max)))
	 )
    (ti::o-remove-area p max (list 'owner owner) t)
    ))



;;}}}
;;{{{ properties



;;; ----------------------------------------------------------------------
;;;
(defun tice-ov-get-mouse (&optional point)
  "Checks if the line has 'mouse-face overlay. Default is to check at
current POINT.

Returns:
  nil          ,no mouse face
  ov           ,overlay
"
  (let* (ov )
    (save-excursion
      (goto-char (or point (point)))
      (beginning-of-line)
      (setq ov (ti::o-get-mouse))
      ov
      )))



;; ----------------------------------------------------------------------
;;;
(defun tice-prop-line (prop val)
  "handles overlay property PROP VAL for the whole line."
  (let* (
	 (mb (BOLP)) (me (EOLP))	;line  BEG,END
	 ov				;overlay
	 ovl				;overlay list
	 )

    (save-excursion
      (if (null (setq ovl (overlays-at (point))))
	  nil				;none exists
	;;  reuse the overlay if it is mouse overlay
	(setq ov (ti::o-get-prop ovl (list prop)))
;;;	(d! "~set-prop" val prop "reuse" (if ov "t" "nil")
;;;	    (buffer-substring mb me))
	)
      (if ov nil			;reused
	(setq ov (make-overlay mb me))
	(overlay-put ov 'owner (tice-v 'package))
	)

      (overlay-put ov prop val)
      )))



;;; ----------------------------------------------------------------------
;;; - This is not for hiliting...
;;;
(defun tice-prop-look-at (list prop &optional prop-val)
  "Set's properties of text to PROP-VAL according to LIST table.

Input values:
  LIST      \( \(re level\) ..\)    list pairs
  PROP      property name
  PROP-VAL  value [default is 't]
"
  (interactive)
  (let* ((prop-val  (or prop-val t))
	 (PACKAGE   (tice-v 'package))
	 el
	 lev				;subexpression level
	 rex
	 mb me
	 )
    (while list
      (setq el  (car list)
	    lev (nth 0 el)
	    rex  (nth 1 el))
;;;      (d! "_" prop (point) (looking-at rex) (match-beginning lev) rex)
      (if (not (and (looking-at rex)
		    (setq mb (match-beginning lev))
		    (setq me (match-end lev))))
	  nil				;not found
	;;  mark that this text area
	(put-text-property mb me prop prop-val)
	(put-text-property mb me 'owner PACKAGE)
;;;	(d! "p-marked" prop (buffer-substring mb me))
	(setq list nil)		;stop right there !
	)
      (and list (setq list (cdr list))))
;;;    (d! "prop done")
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tice-look-at (list)
  "Hilights the text according to LIST table."
  (interactive)
  (let* ((loop t)
	 (PACKAGE (tice-v 'package))
	 re
	 lev
	 face
	 el
	 ov
	 )
    ;;   Suppose point is in the right place
    ;;   Try all possible matches
    (while (and loop list)
      (setq el (car list)
	    re (car (car el))
	    lev (cdr (car el))
	    face (eval (cdr el)))

;;;      (d! re (looking-at re) (point) (tir-read-line))

      (if (not (and (looking-at re)
		    (match-end lev)))
	  nil
	(setq loop nil)
;;;	(d! re face)
	(setq ov (make-overlay (match-beginning lev) (match-end lev)))
	(overlay-put ov 'face face)
	(overlay-put ov 'owner PACKAGE)
	(overlay-put ov 'err-type t)
	)
      (setq list (cdr list))
      )
    ))



;;; ----------------------------------------------------------------------
;;;
(defun tice-read-prop (prop &optional beg end val)
  "Returns continuous text block which is marked with PROP,
whose value is VAL [default is t] from optional BEG END.

Sets global variable tice-:ret to list of \(beg end\) read positions.
Nil if not found.
"
  (let* ((val (or val t))
	 (end (or end (point-max)))
	 pb pe				;point beg, end
	 data
	)

    ;;   It's appropriate to return only one value, the most
    ;;	 important one. and the rest through ENV
    ;;
    (setq tice-:ret nil)

    (save-excursion
      (and beg (goto-char beg))

      (and (setq pb (text-property-any (point) end prop val))
	   (setq pe (next-property-change pb))
	   (< pe end)
	   (setq data (buffer-substring pb pe))
	   (setq tice-:ret (list pb pe))
	   )

;;;      (d! "~read-prop" prop val "_" pb pe data)
      ;;  hmmm, it returns the property lists also, I haven't
      ;;  yet decided how to get plain ASCII ...
      (if (sequencep data)
	  data
	data)
      )))




;;; ----------------------------------------------------------------------
;;;
(defun tice-get-single-prop (prop &optional point)
  "Returns string with PROP starting from current point .

Returns:
  nil
  \(str beg end \)
"
  (let* (data
	 )
    (save-excursion
      (if point (goto-char (point)))
      (setq data (tice-read-prop prop (point) (point-max)))
      (if data
	  (list data (nth 0 tice-:ret) (nth 1 tice-:ret))
	nil)
    )))


;;; ................................................... &prop-hi-level ...
;;; These are TOPMOST level functions and they are prefered.
;;;

;;; ----------------------------------------------------------------------
;;;
(defun tice-get-file (&optional point)
  "Returns file string in a list form \(str beg end \)"
  (tice-get-single-prop tice-:p-file (or point (point))))

;;; ----------------------------------------------------------------------
;;;
(defun tice-get-line (&optional point)
  "Returns line string in a list form \(str beg end \)"
  (tice-get-single-prop tice-:p-line (or point (point))))


;;; ----------------------------------------------------------------------
;;;
(defun tice-get-file&line (&optional beg end)
  "Returns file & line info around area. Default is from current line."
  (let* (
	 (beg		(or beg (BOLP)))
	 (end		(or end (EOLP)))
	 (prop-file	tice-:p-file)
	 (prop-line	tice-:p-line)
	 line file
	 ret
	 )
    (setq file (tice-read-prop prop-file beg end))
    (setq line (tice-read-prop prop-line beg end))
    (and  line (setq line (string-to-int line)))
    (if  (and file line)
	(setq ret (list file line)))
;;;    (d! "~f&l" beg end file line)
    ret
    ))

;;;#please do not use this
;;; ----------------------------------------------------------------------
;;;
(defun tice-get-next-file&line (&optional point)
  "Returns file & line from point forward which is around current paragraph.
 You have to have set paragraph-start and paragraph-separator in order
this to work reliably. The variables with this, point is *:

	1:
	2:  txt txt txt txt txt txt txt
	3:  txt txt txt [*] txt txt txt
	4:
	5:  txt txt ..

	Now you should press backward-paragraph and end up to
	beginning of line 1 or 2. This is para begin. With
	forward-paragraph you should get to line 4, para end.

Return
  nil
  \(file lineNbr file-begin-point line-begin-point this-paragraph-end\)
"
  (let* (
	 beg end
	 fl fn fb 			;file list, beg, end pos
	 ll ln lb 			;line list
	 line nbr
;;;	 file line
	 ret
	 )
    (save-excursion
      (setq fl (tice-get-file point))
      (if (null fl) nil
	(progn
	  ;;  No more
	  nil
	  )

	;; - We guess thge direction by looking how close the name is
	;;   to the left of right of BEG and END
	(setq fn (nth 0 fl)
	      fb (nth 1 fl)
;;;	      fe (nth 2 fl)
	      )
	(goto-char fb)
;;;	(d! "file beg" fb (read-full-line))

	;; - The file is forward, but where is the line number ?
	;;   line can be after/before the file itself:
	;;         t.cc  619  Note 919:
	;;         (line 88, file /users/jaalto/t.cc ...)
	;;
	;; - The search area is delimited by paragraphs
	(setq beg (progn (backward-paragraph)
;;			 (forward-line 1)
			 (point))
	      end (progn (forward-paragraph) (point))
	      )
	(goto-char beg)

	(setq  ll (tice-get-line point))
	(and ll
	     (setq
	      ln (nth 0 ll)
	      lb (nth 1 ll)
;;;	      le (nth 2 ll)
	      ;;  Convert to nbr , 0 is not good enuff'
	      line (if (not (eq (setq nbr (string-to-int ln)) 0)) nbr)
	      ))
;;;	(d! "para beg" beg end "L" line ll)
	;;  Is the found LINE txt out of paragrh region. If it is
	;;  we have no way to know what FILE and LINE are we supposed
	;;  to get
	;;
	(if (and line
		 (< lb end))
	    (setq ret (list fn line fb lb end)))
	)
;;;      (d! "~next" ret)
      ret
      )))


;;; ----------------------------------------------------------------------
;;;
(defun tice-prop-mouse-line-on (&optional face)
  "Set mouse face ON for the whole line.";
  (tice-prop-line 'mouse-face tice-:f-mouse))

;;; ----------------------------------------------------------------------
;;;
(defun tice-prop-mouse-line-off (&optional face)
  "Set mouse face for the whole line. Defalut is to clear."
  (tice-prop-line 'mouse-face 'default))


;;; ----------------------------------------------------------------------
;;;
(defun tice-prop-mouse-line (&optional face)
  "Set mouse face for the whole line. Defalut is to clear."
  (tice-prop-line 'mouse-face face))


;;}}}

;;{{{ goto error



;;; ----------------------------------------------------------------------
;;;
(defun tice-goto-error-blink (l)
  "Goes to file and line. Blinks. Input is special LIST, cursor stays
in compile buffer.

Returns
  nil or t if moved.
"
  (let* ((delay  tice-:blink-delay)
	 fn ln
	 )
    (if (null l)
	nil
      (setq fn (nth 0 l) ln (nth 1 l))
      (tice-goto-error-1 fn ln)
      (if delay (sit-for delay))
      (pop-to-buffer (tice-v 'compile-buf))
      t
      )
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tice-goto-error-scan (&optional back)
  "Searches occurrance of error and Blinks. Optionally BACK"
  (let* ((arg	 (if back  -1 1))
	 fn ln  l hit
	 )
    (forward-line arg)			;move away immediately
    (while (and (null l) (null (bobp))(null (eolp))
		)
      (setq hit (ti::o-mouse-on-p (tice-ov-get-mouse)))
      (if (null hit) nil
	(setq l  (tice-get-file&line (point) (EOLP)))
	(if (null l) nil		;no returned list
	  (setq fn (nth 0 l) ln (nth 1 l))
	  (if (null (and fn ln)) nil	;both must exist
	    (tice-goto-error-blink
	     (list fn ln  1 1 1) ))
	  ))
      (forward-line arg)
      )
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tice-goto-error-next (event)
  "Goes to the error at mouse click/or point."
  (interactive "e")
  (tice-goto-error-scan)
  )


;;; ----------------------------------------------------------------------
;;;
(defun tice-goto-error-prev (event)
  "Goes to the error at mouse click/or point."
  (interactive "e")
  (tice-goto-error-scan t)
  )


;;; ----------------------------------------------------------------------
;;; - This is an umbrella filter
;;;
(defun tice-goto-error-1 (file line)
  "Goes to the error only if allow."
  (let* ((ff	     (file-name-nondirectory file))
	 (no-re      (or tice-:load-no-re "blurb#x"))
	 (ask        tice-:load-ask)
	 (load	     t)
	 )
    (cond
     ((get-buffer ff)			;if it's on the buffer already
      (pop-to-buffer ff)
      (goto-line line))

     ((string-match no-re file)
      (tice-msg "Load not allowed."))

     (t
      (if (string-match no-re file)
	  (tice-msg "Not allowed to enter, see tice-:load-no-re")
;;;	  (d! "asking" file ask (get-buffer file))
	(if ask (setq load (y-or-n-p (concat "Load " file "? "))))
	(if (null load) nil
	  (tice-pop-file file line))
	))
     )
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tice-goto-error (event)
  "Goes to the error at mouse click/or point."
  (interactive "e")
  (let* ((err-func   tice-:goto-err-func)
	 candidate
	 file line el
	 status ping
	 )
;;; (d! "~!tice goto")
    (save-excursion
      (if (null event)
	  nil			;keypress then
	(set-buffer (window-buffer (posn-window (event-start event))))
	(goto-char (posn-point (event-start event))))

      (setq ping (point))		;cursor position

      ;;  Is this line with selectable mouse property ?
      (setq candidate (tice-ov-get-mouse ping))

;;;      (d! "ping " ping candidate)

      (if (null candidate)
	  nil
	(setq el   (tice-get-file&line))	;get the information
	(and el (setq file (nth 0 el)  line (nth 1 el)))
	))


;;;    (d! "~goto-error" line file)
    ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. ..
    ;; - Okay , save excursion done, parameters have been gotten
    ;;
    (if (not (ti::o-mouse-on-p candidate))
	;;  The line has no mouse cursor or it's disabled
	(tice-msg "disabled line.")
      ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. ..
      ;;  Okay, the line is selectable
      (if (not (and line file))
        (progn
	  ;;  Try calling some special function that can hadle difficult
	  ;;  lines, that were not marked with regexp pairs.
	  (if (null  candidate) nil
	    (and err-func (setq status (funcall err-func ping)))
	    ;;  Print error only is the function returned nil
	    (if status nil
	      (tice-msg "Not enough error info on this line")))
	  )
	;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. ..
	;;  - Rest of the if
	;;    Now ready to pick the action
	(tice-goto-error-1 file line)
	))
    ))

;;}}}

;;{{{ parse

;;; ########################################################### &parse ###

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tice-unparse ()
  "Removes all overlays owned by TICE. "
  (interactive)
  (let* (
	 (PACKAGE (tice-v 'package))
	 )
    (tice-ov-remove-area PACKAGE)

    ;;  This kills all, if there is some hilit package on, they are
    ;;  removed too.
    ;;  -- Commented out since I don't use text-properties's any more.
    ;;  (set-text-properties (point-min) (point-max) nil)
    ;;  (set-buffer-modified-p nil)
    ))





;;; ----------------------------------------------------------------------
;;; - This should hilit the buffer again and do other things.
;;;
;;;###autoload
(defun tice-parse-again ()
  "Parses the *compile* buffer again. This function can be used only
after the compilation is executed at least once. It does it jobs in the
compile buffer, but does not make it visible.

Returns nil         ,if there is an error
        anything    ,whatever the original parser returns
"
  (interactive)
  (save-excursion
    (if (set-buffer  (tice-v 'compile-buf)) nil
      (error "Compile buffer does not exist."))
    (let* (
;;;	   (do-func       tice-:compile-finish-ok-func)
	   (supp-func     tice-:compiler-support-func)
	   (compiler-str  (tice-v 'compiler-str))
	   (E		  0)		;error condition
	   (e-table
	    '((1  . "Compiling seems to be uncomplete...")
	      (2  . "Compilation check func missing.")
	      (3  . "Parsing not supported for this compiler version.")
	      ))
	   )

      ;;  if the results fails inc error counter + return nil so that
      ;;  it jumps out of AND
      ;;

      (if (and (if (tice-compile-finish-ok-p) t (inc E) nil)
	       (if supp-func t (inc E) nil)
	       (if (funcall supp-func compiler-str) t (inc E) nil))
	  (funcall tice-:compile-finish-ok-func)
	(tice-msg (concat "tice-parse: " (cdr (assoc E e-table))))
	nil
	)
      )))




;;; ----------------------------------------------------------------------
;;;
(defun tice-parse-entry (re file-re line-re)
  "This fuction should be the _entry_ point to parsing  from custom packages.
See tice-parse for parameter explanation.
"
  (let* ((end				;only maximum lines
	  (save-excursion
	    (goto-char (point-min))
	    (forward-line tice-:max-lines)
	    (point)))
	 )
    (tice-parse re file-re line-re (point-min) end)
    ))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tice-parse (re file-re line-re &optional start max)
  "Prepares buffer with error messages, normally, compile result,
to have mouse binding over errors.

RE              regexp for possible line candidate. These lines are
                marked as mouse selectable.
The [fl]-[re]   list of Regexps and sub-levels:
                '\( \(level regexp) .. \)
                Matched portion acts as valid LINE or valid FILE
"
  (let* (
	 (allow-re    tice-:load-allow-re)
	 (hilit-table (tice-v 'face-table))
	 (prop-file   tice-:p-file)
	 (prop-line   tice-:p-line)
	 el				;element
	 file
	 )
;;;   (d! "~parse")
    (save-excursion
      (and start (goto-char start))

      ;;  Find possible line candidate
      ;;  do not test position if max is nil
      (while (and (re-search-forward re nil t)
		  (if max (< (point) max) t)
		  )
	(end-of-line)
	(beginning-of-line)

;;;	(d! "candidate" (read-full-line))

	(tice-look-at hilit-table)	;Hilit text

	;;  Now set properties for text.
	(tice-prop-look-at file-re prop-file)
;;;	(d! "file ok")

	(tice-prop-look-at line-re prop-line)

	;;  Mark for mouse hilit also, turn line ON
	(tice-prop-mouse-line-on)

	;;  Do we want to see this file ?
	;;
	(setq el   (tice-get-file&line)	;get the information
	      file (nth 0 el)
	      )
	(if (not (and file allow-re))
	    nil
	  (if (string-match allow-re file)
	      nil			;this file is ok
	    ;;   file didn't mach, so dehilit it for mouse
;;;	    (d! "dehilit" (read-full-line))
	    (tice-mouse-line-toggle nil)
	    ))
	;; pick next
	(end-of-line)
	)

      (set-buffer-modified-p nil)
      )
    ))

;;}}}
;;{{{ compile

;;; ######################################################### &compile ###

;;; ----------------------------------------------------------------------
;;;
(defun tice-get-compiler-version ()
  "Return compiler version."
  (let* ((re tice-:compiler-ver-re)
	 ret
	 )
    (if (stringp re)
	(save-excursion
	  (goto-char (point-min))
	  (if (re-search-forward re nil t)
	      (setq ret (read-full-line)))))
    ret
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tice-is-supported (str)
  "Compares compilation version STRING against support regexp"
  (let* ((re  tice-:compiler-ver-re)
	 )
    (if (and (stringp str) (stringp re))
	(string-match re str)
      nil)
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tice-compile-finish-ok-p ()
  "Checks if compile was finished ok. Expects that the status is saved."
  (let* ((stat-ok-re    tice-:compile-finish-ok-re)
	 (status	(tice-v 'fin-status))
	 )
    ;;  prevent errors if variables are nil
    (string-match (or stat-ok-re "x") (or status "y"))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tice-compile-finish (buf status)
  "Function called by COMPILE when finished.

References:
	tice-compile-finish-beg-hook
          tice-:compile-finish-ok-re
	  tice-:compile-finish-ok-func
	  tice-:compile-finish-err-func
	tice-compile-finish-end-hook
"
  (let* (re
	 (err-func    tice-:compile-finish-err-func)
	 (ok-func     tice-:compile-finish-ok-func)
	 (ver-func    tice-:compiler-ver-func)
	 (supp-func   tice-:compiler-support-func)
	 (allow-func  tice-:load-allow-re-func)
;;;	 (max         tice-:max-lines)
	 (stat-ok-re  tice-:compile-finish-ok-re) ;; finish staus ok ?
	 (tice-map    tice-compile-map)
	 (call t)
	 )
    (tice-compile-restore)		;restore orig compile cmd

    (save-excursion
      (set-buffer buf)
;;;      (tice-set-compile-map)		;redefine compile.el map
      (use-local-map tice-map)

      ;;  >> save some globl variables
      (setq tice-:compile-finish-status status)
      (setq tice-:compiler-str (or (and ver-func
					(funcall ver-func))
				   ""))
;;;      (d! ver-func supp-func tice-:compiler-str)


      (run-hooks 'tice-:compile-finish-beg-hook)

;;;      (d! "~Support" supp-func stat-ok-re status)
      ;;  "Exited abnormally with status NBR" ??
      (if (not (and (string-match stat-ok-re status)
		    (if supp-func
			(funcall supp-func tice-:compiler-str)
		      nil)
		    ))
	  (and err-func (setq call (funcall err-func)))

	;;  << Set global
	(and allow-func
	     (setq re (funcall allow-func))
	     (setq tice-:load-allow-re re))

;;;	(d! "~hilit" ok-func)
	;;  Now do hiliting and marking
	(and ok-func (setq call (funcall ok-func)))
	))
    (if call (run-hooks 'tice-:compile-finish-end-hook))
    ))



;;; ----------------------------------------------------------------------
;;;
(defun tice-emacs-var-put ()
  "Puts some of the emacs variables into 'local' mode."
  (make-local-variable 'paragraph-separate)
  (make-local-variable 'paragraph-start)
  )

;;; ----------------------------------------------------------------------
;;;
(defun tice-emacs-var-set ()
  "Sets some of  emacs variables into. Make sure you're in Compile buffer
 when you call this."
  (setq paragraph-separate  tice-:paragraph-separate
	paragraph-start     tice-:paragraph-start
	))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tice-compile-prepare ()
  "Prepare compilation to call TICE compilation finish function."

  ;;  First save the previous value
  (setq-default
   tice-:compile-finish-orig     compilation-finish-function
   compilation-finish-function  'tice-compile-finish
   ))

;;; ----------------------------------------------------------------------
;;;
(defun tice-compile-restore ()
  "This function restores original function which is called
when compilation finishes. Be sure to call tice-compile-prepare first."
  ;;  Restore
  (if (string-match "tice" (symbol-name tice-:compile-finish-orig))
      ;; Hmm, our funbction should not be left there?
      ;;
      (setq-default compilation-finish-function nil)
    (setq-default
     compilation-finish-function  tice-:compile-finish-orig))
  )



;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tice-compile (&optional add-opt cmd-set)
  "Kills all local variables before running Compile.
The ADD-OPT usually has additional options, but CMD-SET is arg that is
passed to possible command compose function only.

References:
	tice-:local-var-func
	tice-compile-finish  ,use this for actions after compile
"
  (interactive)
  (let* (
	 ;; ........... tice variables
	 cmd				;compile command
	 cmd-func			;function to generate command
	 compile-buf
	 allow-re
	 def-msg			;default error message

	 ;; ........... other vars
	 (fid   "tice-compile: ")	;function id
	 (cb    (current-buffer))	;abbreviation
	 (bn	(buffer-name cb))	;buffer name
	 (bf	(buffer-file-name cb))	;buffer file
	 (obuf	(current-buffer))	;orig buffer
	 )
    ;;  ... ... ... ... ... ... ... ... ... ... ... ... ... ... ... ... ...
    (tice-set-defaults)
    (and (fboundp tice-:local-var-func) (funcall tice-:local-var-func))

    (setq
     compile-buf  (tice-v 'compile-buf)
     cmd          (tice-v 'compile-cmd)
     cmd-func     tice-:compile-cmd-func
     allow-re     (or tice-:compile-file-re ".") ;allow to compile only these
     def-msg			;default message
     (concat fid tice-:compile-file-sorry-msg " [" bn "]"))

    ;;  Make sure we have fresh buffer before compiling
    ;;
    (set-buffer  (get-buffer-create compile-buf))


    (if (or (null bn)			;no file name: *scratch* ?
	    (null (string-match allow-re bn))) ;not Compilable file
	(progn
	  ;;  wait a little, beacuse mouse move clears the message
	  (message def-msg)
	  (sleep-for 1)
	  )
      ;;  ... ... ... ... ... ... ... ... ... ... ... ... ... ... ... ... .
      ;;
      (if (null cmd-func)
	  (setq cmd (concat cmd " " (or (stringp add-opt) "") bf))
	(setq cmd (funcall cmd-func bf add-opt cmd-set))
	)
;;;      (d! "~tice-c final cmd: " add-opt cmd-set "_" cmd )


      ;;  ... ... ... ... ... ... ... ... ... ... ... ... ... ... ... ... .
      ;;  Now run compile
      ;;  - Unfortunately this kills all local variables!
      ;;  - We have to call that local-var func again.
      (tice-compile-prepare)
      (and cmd (compile cmd))

      (run-hooks 'tice-:define-keys-hook)
      (tice-emacs-var-put)
      (tice-emacs-var-set)

      (and (fboundp 'tice-:local-var-func) (funcall tice-:local-var-func))
      )
    (set-buffer obuf)			;return
    ))

;;}}}


(provide   'tinycerr)
(run-hooks 'tice-:load-hook)

;;; tinycerr.el ends here
