;; @(#) tinydesk.el -- Saves/restores files of emacs session

;; @(#) $Id: tinydesk.el,v 1.44 1997/06/19 09:31:46 jaalto NotReleased $
;; @(#) $Keywords: extensions $
;;
;; $KnownCompatibility: 19.28 XEmacs 19.14 $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1995,96,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Feb 13 1995
;;
;; To get information on this program use ident(1) or do M-x tid-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el


;; LCD Archive Entry:
;; tinydesk|Jari Aalto|jari.aalto@poboxes.com|
;; Saves/restores files of emacs session|
;; 19-Jun-1997|$Revision: 1.44 $|~/misc/tinydesk.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ........................................................ &t-install ...
;;   Put this file on your Emacs-Lisp load path, add following into your
;;   ~/.emacs startup file
;;   Code can be easily extracted with tinylib.el / ti::m-pkg-rip-magic
;;
;;* _
;;*       (require 'tinydesk)
;;
;;   or use the autoload feature. Notice that the automatic "file
;;   state backup feature" gets enables only when this file is loaded.
;;   If you want that feature, then use require.
;;
;;* _
;;* 	(autoload 'tid-mode		    "tinydesk" t t)
;;* 	(autoload 'tid-save-state	    "tinydesk" t t)
;;* 	(autoload 'tid-unload		    "tinydesk" t t)
;;* 	(autoload 'tid-recover-state	    "tinydesk" t t)
;;* 	(autoload 'tid-edit-state-file	    "tinydesk" t t)
;;
;;   Suggested keybinding. You need these original x-4 bindings rarely.
;;
;;* _
;;* 	(define-key ctl-x-4-map "s" 'tid-save-state)     ;; free in 19.28
;;
;;* _
;;* 	;;  This is find-file-read-only-other-window, do you use it ever?
;;* 	;;
;;* 	(define-key ctl-x-4-map "r" 'tid-recover-state)
;;* 	(define-key ctl-x-4-map "e" 'tid-edit-state-file) ;; free in 19.28
;;* 	(define-key ctl-x-4-map "u" 'tid-unload)          ;; free in 19.28
;;
;;
;;   If you have any questions, always use function
;;
;;      M-x tid-submit-bug-report       ,send question, feedback,bug report

;;}}}
;;{{{ Documentation

;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "tid-" in front of
;;   every function & variable. It stands for '(ti)ny (d)esktop.
;; - variable names contain letters 'tid-:'.

;;  Preface Feb 1995
;;
;;	I have had these function for ages and I just decided to remove
;;	them from my own lisp library of various useful functions and
;;	put them into separate .el
;;
;;	I'm usually working with X-tube at my work, which I left open day to
;;	day. In fact I seldom even logout, so my emacs files just sits nicely,
;;	so I haven't had need for sophisticated session saver yet.
;;
;;	What I do need, is, that sometimes I have to go to the lab next floor
;;	to see what's troubling my C++ program. I must have some way
;;	to transfer the files that I was just editing and bring them into lab
;;	where I need to run the debugger for a while.
;;	--> These functions save my configuration into file, that
;;      can be opened again in emacs somewhere else.
;;
;;	Hope someone finds use for this also, although there exist much more
;;	better desktop savers, which saves points, marks and modes...
;;
;;  Overview of features
;;
;;	o   Simple desktop: only filenames and directories are read/saved.
;;
;;	    Unlike the other desktop savers, this one can also UNLOAD files
;;	    from emacs. You just tell it to remove 'these files listed in
;;	    state file state.XXX', and those files will be removed from
;;	    your emacs buffers. You can collect 'projects' and switch
;;	    between them easily: after project1, It can can
;;	    be unload and load project3 instead.
;;
;;	o   Parse any file that includes filenames and comments
;;
;;	o   If there were any invalid entries in the state file,
;;	    the state file contents is shown to user and the entries which
;;	    had problems are marked.
;;
;;	o   State file editing (tid-mode):
;;
;;	    --  load single file on the line
;;	    --  clear face properties from buffer, so that they don't
;;		disturb your view.
;;	    --  parse files for loading.
;;	    --  Show files that cannot be loaded.
;;
;;	o   In regular intervals save the state of emacs (files loaded)
;;	    If emacs crashes you can recover the previous session.
;;	    See function `tid-auto-save' for more. Sililar functionality
;;	    is in new emacs releases, but this package was designed in
;;	    Emacs 19.28.
;;
;;  Quick start
;;
;;	If you're just eager to use the package, here are the basics.
;;	I suppose you have copied the installation setup as is.
;;
;;	o   You have emacs session open with bunch of files. Now you
;;	    believe that it's time to save this session. You do
;;	    C-x 4 s and give some name "state.c" if you edited c project.
;;
;;	Now, it all depends what you want to do after that. If you find more
;;	files to emacs; or kill some unwanted buffers, you can re-execute
;;	C-x 4 s whenever you like. You can even edit the state file with
;;	C-x 4 e to remove some files that you don't want to include to
;;	that "project".
;;
;;	o   Next time you open emacs you can load any state file with
;;	    C-x 4 r "state.c"
;;
;;	If you want to switch between projects; unload first the current
;;	project with C-x 4 u "state.c" and reload some other project
;;	with C-x 4 r "state.cc"
;;
;;  Automatic one time session saver
;;
;;	Some people just want to save the session on exit and reopen it
;;	when emacs starts again. I must say that this is not necessarily
;;	the best, because when you start emacs for some quick job, you
;;	don't necessarily want it to load the saved session (loading all
;;	files take time considerably). Loading emacs with -q is not the
;;	choice, if you still like to have your other emacs goodies active.
;;
;;	Here is semi-automatic save and restore, put all these lines
;;	near the end of your ~/.emacs. It saves the file info when
;;	emacs exists and asks if you want to return to saved session
;;	on emacs startup.
;;
;;* _
;;* 	(setq tid-:dir-default "~/projects/")
;;* 	(setq my-tid-session  (concat tid-:dir-default "state.last-session"))
;;* _
;;* 	(add-hook 'kill-emacs-hook 'my-save-session)
;;* 	(defun my-save-session ()
;;* 	  "Save loaded file information."
;;* 	  ;;  Or (tid-save-state my-tid-session '(4))
;;* 	  ;;  if you want to save dired buffers too.
;;* 	  ;;
;;* 	  (tid-save-state my-tid-session) nil)
;;* _
;;* 	(if (y-or-n-p "Recover session ")
;;* 	    (tid-recover-state my-tid-session))
;;
;;
;;  Face setup
;;
;;	This program uses some faces to catch your attention when
;;	you're working with the state files. I you restore state from
;;	a file and some file reference cannot be loaded, the state
;;	file will be shown to you and the problematic lines are
;;	highlighted.
;;
;;	If you open the state file for editing, you can selectively
;;	load files. The mouse pointer will change and the text is
;;	again highlighted.  To make the highlight work for you, you
;;	must set some colors like this
;;
;;         (set-face-foreground 'italic "LightBlue")
;;
;;  About saving the files
;;
;;	While you may save your session files with any name I suggest here
;;	one convention that you could use. Name every filename so,
;;	that they have common prefix:
;;
;;	    M-x tid-save-state   ;; or any hotkey you have bound this
;;	    state.XXX
;;
;;	where the XXX describes the name of the state file you just saved.
;;
;;	Later on it's easier to use emacs file name cmpletion capability
;;	to load the file you want. If you don't exactly remember what files
;;	you saved, or which sessions you have in dir, you just type
;;
;;	    state.[TAB]
;;
;;	when `tid-recover-state' ask for filename.
;;	Prefix arg to tid-save-state saves also directories loaded
;;
;;  Automatic state file saving
;;
;;	Has it ever happened to you that emacs crashed mystically when
;;	you  were in the middle of your daily routines. You had several
;;	C++ files open, perl code, text files, RMAIL, ...
;;
;;	19.29+ has new features that makes it recover your session, but
;;	if you want that feature now, then this file helps you.
;;
;;	It installs tid-auto-save function to write-file-hooks. In
;;	regular intervals all your emacs session files are stored into
;;	the state file. After a crash you can easily recover you
;;	session by reading that state file.
;;
;;  Development note
;;
;;	No, I won't duplicate desktop.el functionality to save points
;;	and modes and so on. This is for simple state restoring only.
;;
;;  Advertise
;;
;;	Emacs 19.27 is known to ship with desktop.el, but someone mentioned
;;	that there existed much more powerful desktop saver.
;;	Please note, that version may differ now.
;;
;;      ftp://ftp.ae.keio.ac.jp/pub/emacs-lisp/util/ windows-1.8.tar.gz


;;}}}
;;{{{ history

;; ........................................................ &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; Jun	19	1997	[jari]		19.28	v1.41-1.44	NotReleased
;; - dired-directory is now hidden from ByteCompiler, thanks to bug
;;   report from Rick Flower <flower@ms31.sp.trw.com>
;; - Rick Flower <flower@ms31.sp.trw.com> reported tid-:save-default-title
;;   bug. Corrected.
;; - Checkdoc.el 1.42: docrings corrected. Defcustom support.
;; - Wrote sections 'Quick start' 'Automatic one time session saver'
;;
;; May	14	1997	[jari]		19.28	v1.38-1.40	NotReleased
;; - Keybinding converter update. Internal changes.
;; - Changed all non standard boolean tests to use when and unless
;;
;; Feb	21	1997	[jari]		19.28	v1.37		NotReleased
;; - tid-:save-no-file-re variable is now case sensitive. Added
;;   deleting /tmp/ dir files by defult.
;;
;; Jan	24	1997	[jari]		19.28	v1.35-36	NotReleased
;; - Added new commands to StateFileEdit mode. 'u unloads files from
;;   emacs. 'x expunges 'unloaded' lines and 'r removes all file marks.
;; - Corrected auto-save bug.
;;
;; Jan	21	1997	[jari]		19.28	v1.32-1.34	NotReleased
;; - Alain Picard <alainp@goodfellow.in.ot.com.au> pointed out that symlinked
;;   files should be found with function find-buffer-visiting and not with
;;   get-file-buffer when unloading the files.
;;
;; Dec	3	1996	[jari]		19.28	v1.30-31	NotReleased
;; - Small fixes
;;
;; Nov	23	1996	[jari]		19.28	v1.28-1.29	NotReleased
;; - Now all autoloads are in the 'm' lib.
;; - Sergei B. Pokrovsky <pok@el.nips.nsk.su> suggested that when loading
;;   new state file the previous files from previous session
;;   should be unloaded. I added prefix arg to command tid-recover-state
;;   to do just that. He didn't "merge" projects like I, so he prefered
;;   one emacs/per project
;; - He also recommended that the first file should be displayed when
;;   the state file was loaded. This is now supported.
;; - And he wanted to be able to save directories too: It is now
;;   possible by giving prefix arg to tid-save-state.
;;
;; Nov	1	1996	[jari]		19.28	v1.25-1.27	NotReleased
;; - Chnaged StateFileEdit prefix from C-c to "'", which is more comfortable.
;; - Added lib m require due to macros.
;;
;; Sep	25	1996	[jari]		19.28	v1.25-1.26	NotReleased
;; - Rearranged loading. Now uses autolaod as much as possible.
;;
;; Aug	21	1996	[jari]		19.28	v1.24		NotReleased
;; - Corrected all XEmacs 19.14 ByteComp errors. Made package
;;   XEmacs compatible. Not sure about the advice in XEmacs though...
;;
;; Jun	19	1996	[jari]		19.28	v1.23		NotReleased
;; - There was wrong keymap mentioned in the installation. Corrected.
;;   Thanks to Martin Thurn <mthurn@loon.irnet.rest.tasc.com>.
;;
;; May	16	1996	[jari]		19.28	v1.22		NotReleased
;; - Added missing ARG to tid-version function.
;;
;; Apr	26	1996	[jari]		19.28	v1.21		NotReleased
;; - Renamed keymap tid-map --> tid-mode-map, -func --> --> function.
;;   key-func --> key-hook. Added ###autoload tags.
;; - When there is no filename at front, default-directory is assumed.
;;   tid-get-file			:+
;;   tid-use-map			:+
;;
;; Apr	15	1996	[jari]		19.28	v1.20		NotReleased
;; - Corrected naming mismatch, tdd --> tid.
;;
;; Mar	8	1996	[jari]		19.28	v1.19		NotReleased
;; - Changed version variables not to have ":"
;;
;; Feb	19	1996	[jari]		19.28	v1.18		NotReleased
;; - I noticed that It would be nice if the current state were
;;   saved periodically, so that I could recover from emacs crash.
;; - I think emacs 19.29+ introduced the autosave files, but it's
;;   not quite the same...
;;   tid-:auto-save-counter		:+
;;   tid-:auto-save-interval		:+ user var
;;   tid-:auto-save-name-function	:+
;;   tid-auto-save			:+
;;   save-buffer			:+ adviced
;;
;; Feb	17	1996	[jari]		19.28	v1.17		NotReleased
;; - I happened to save state which help many ange ftp@site buffer, and
;;   when I loaded the state back it started ange ftp immediately for
;;   these ftp sites...Uhum, that was not my intention. Now there is
;;   variable that controls which files are allowed to put into state
;;   file.
;;   tid-:save-no-file-re		:+ do not save these files
;;   tid-submit-bug-report		:+
;;   tid-comment			:+
;;   tid-get-save-dir			:! bug, didn't remove "loaded" flag
;;   tid-mark-region			:! bug, didn't clear flags
;;   tid-get-save-files			:! non save files handling
;;
;; Feb	12	1996	[jari]		19.28	v1.16		NotReleased
;; - Added suggested keybindings and 'About saving the files'.
;; - Added keybindings for non-mouse too. User "marks" for users that do
;;   not see colors. Eg. when parsing the buffer there is "loaded"
;;   or "invalid" tags in the EOL comment of state file.
;; - Rewrote some functions and clarified the comment variable.
;; - I suspect this version has errors...
;;
;; Nov	17	1995	[jari]		19.28	v1.15		NotReleased
;; - Removed all library functions, because they are in tinylib. Rewrote
;;   the feedback functions. Should make this work in -nw emacs too,
;;   eg the unloadable file info is only seen in X. I should put some kind
;;   of comment marker there and rewrite the keybinding for non-mouse..
;; - now lisp-mnt.el lm-verify checked module.
;;
;; Nov	17	1995	[jari]		19.28	v1.14		Release_2
;; - Time to release this new version. This is exactly the same as 1.13,
;;   but may have some cosmetic text changes.
;;
;; Nov	13	1995	[jari]		19.28	v1.13		NotReleased
;; - Run byte compile test and noticed that inc macro was missing.
;;   inc				:+ from my libs
;;
;; Nov	2	1995	[jari]		19.28	v1.12		NotReleased
;; - some things corrected and rethought.
;;   use-tmp-buffer			:! optional clear parameter
;;   tid-save-state			:! rewritten
;;
;; Nov	2	1995	[jari]		19.28	v1.11		NotReleased
;; - The unload function had some troubles that are now corrected.
;;   use-tmp-buffer			:+ from my libs
;;   tid-unload				:! rewritten
;;
;; Oct	27	1995	[jari]		19.28	v1.10		NotReleased
;; - Rearranged code a bit. Now there is actual tid-mode funtion
;;   that can be turned on.
;; - Now it pops up the buffer and marks the files it couldn't load
;;   and put user in edit state mode automatically.
;;   tid-set-face-non-files-buffer	:! now uses narrow
;;   tid-edit-state-file		:! rewritten
;;   tid-mode				:+ new
;;
;;
;; Oct	27	1995	[jari]		19.28	v1.9		NotReleased
;; - Started replacing functions with library code, so that they could be
;;   easily removed later. Now keys are uer configurable through funcall.
;; - Separate major mode created for handling the State files.
;; - Reprogrammed some functions
;;   tid-load-hook		:+ when loding this file
;;   tid-:loaded-file-list		:+ for hook users
;;   tid-:rejected-file-list	:+ for hook users
;;   tid-:mode-name		:+
;;   tid-read-word		:+ deleted and replaces (lib func now)
;;   tid-get-save-dir		:! changed
;;   tid-only-files-buffer	:! changed
;;   tid-edit-state-file	:! now uses major mode
;;   tid-recover-state		:! reprogrammed
;;   tid-ff-buffer		:! reprogrammed
;;   tid-end-slash		:- moved to lib
;;   tid-kill-line		:- moved to lib
;;   tid-user-home		:- moved to lib
;;   tid-emacs-files		:- moved to lib
;;
;; Aug	23	1995	[jari]		19.28	v1.8		NotReleased
;; - Many corrections. bugs when saving to non-dir and when retrieving
;;   file that already was in emacs, allthouhg had different directory.
;;   tid-:trash-tmp-buffer    	:+ trashes temp buf automatically.
;;   tid-trash			:+
;;   tid-unload			:! rewritten
;;   tid-save-state		:! bug, saving to non-existent dir
;;   tid-recover-state		:! bug, if same filename, but diff dir.
;;
;; May	10	1995	[jari]		19.28	v1.7		NotReleased
;; - Hmmm, When I had many state files, I wanted to load files
;;   1 and 2 .. but later I wanted to remove all 2 files, so I wrote
;;   funtion to do just this.
;; - tid-unload kills all buffers listed in file
;; - Added tid-:trash-tmp-buffer to automatically kill tmp file used.
;;
;; Apr	7	1995	[jari]		19.28	v1.7		Release_1
;; - If default save directory included "/" it was discarded due
;;   to bug in tid-end-slash.
;; - Changed mouse-2 --> mouse-2 , because mouse-2 is used for paste
;; - Corrected bug in tid-ff-buffer that always returned: '0 files loaded'
;; - Added parameter FILE to tid-edit-state-file, so that it can be
;;   be called direcly.
;;
;; Mar	31	1995	[jari]		19.28	v1.6		NotReleased
;; - Fine tuned tid-edit-state-file: Few more checks if the file
;;   is already loaded. Now it doesn't create duplicate windows,
;;   also clears previous faces before processing.
;;
;; Mar	31	1995	[jari]		19.28	v1.5		NotReleased
;; - Corrected the tid-:comment-re variable, there must not me '.*\\('
;;   at the front, becasue .* aways matches at least one, so the sub-level
;;   starts from position 2, which is wrong. Changed to have '[ \t]*'
;;   instead, which is more accurate. Also added + for #;// marks
;; - Corrected bug in tid-set-face-non-files-region, which forgot
;;   to take care of comments and ignore them.
;; - Relocated the redraw-screen, so that the sceen isn't so jumpy when
;;   faces are removed, added.
;;
;; Mar	31	1995	[jari]		19.28	v1.4		NotReleased
;; - Added more mouse commands and more user functions, Now
;;   C-S-Mouse 2 shows unloadable lines. Great for verifying the lines.
;; - function that rips off all comments from the file.
;; - added some more hooks + title string insertion.
;; - added tid-kill-line, tid-only-files-buffer , tid-only-files-region
;;   tid-set-face-non-files-buffer , tid-set-face-non-files-region
;; - defconst made defvar in several places
;;
;; Mar	30	1995	[jari]		19.28	v1.3		NotReleased
;; - Well, this is really some package. Sent to g.e.sources
;;   to have people testing this. If I don't here complaints or
;;   bugs, then I'll send this to ohio.
;;
;; Mar	30	1995	[jari]		19.28	v1.2		NotReleased
;; - I took a look at this again, I've been bust lately.
;; - Oh man, I really started to add functions to it, now has configurable
;;   save directory, properties.... just the kind of .el I want it to be.
;;
;; Feb	13	1995	[jari]		19.28	v1.1		Created


;;}}}


;;; Code:


(require 'advice)
(require 'tinylibm)

(eval-when-compile (use-dynamic-compilation))

(defgroup-tiny TinyDesk tid-: extensions
  "Simple desktop: only filenames and directories are read/saved.

	    Unlike the other desktop savers, this one can also UNLOAD files
	    from emacs. You just tell it to remove 'these files listed in
	    state file state.XXX', and those files will be removed from
	    your emacs buffers. You can collect 'projects' and switch
	    between them easily: after project1, It can can
	    be unload and load project3 instead.
    ")


;;{{{ setup: hooks


;;; ......................................................... &v-hooks ...

(defcustom tid-:mode-define-keys-hook '(tid-default-bindings)
  "*List of functions to run which define keys to `tid-mode-map'."
  :type 'hook :group 'TinyDesk
  )


(defcustom tid-:load-hook nil
  "*Hook run when file has been loaded."
  :type 'hook :group 'TinyDesk
  )

(defcustom tid-:save-before-hook nil
  "*Hook run just before _putting_ files to STATE file.
begins. This is your chance to do something to the buffers."
  :type 'hook :group 'TinyDesk
  )

(defcustom tid-:save-after-hook nil
  "*Hook just before _saving_ of STATE file.
The files are there, possibly in sorted order, and the title is there."
  :type 'hook :group 'TinyDesk
  )

(defcustom tid-:mode-hook nil
  "*Hook run after the edit mode is turned on."
  :type 'hook :group 'TinyDesk
  )

(defcustom tid-:recover-before-hook nil
  "*Hook run after recover file is loaded, just before processing start."
  :type 'hook :group 'TinyDesk
  )

(defcustom tid-:recover-after-hook nil
  "*Hook run after recover file is _parsed_ AND no errors during load."
  :type 'hook :group 'TinyDesk
  )

;;}}}
;;{{{ setup: -- private

;;; ....................................................... &v-private ...


(defvar tid-mode-map nil
  "Local keymap for STATE files loaded by edit."
  )

(defconst tid-:comment-chars ";"
  "Allowed comment characters.")


(defvar tid-:dir-last nil
  "Directory that were used for last save.")


(defvar tid-:tmp-buffer  "*tmp*"
  "The workbuffer used, created and killed when needed.")


(defvar tid-:trash-tmp-buffer  t
  "If non-nil, the workbuffer is always deleted.")

(defvar tid-:message-column 60
  "Column where to put possible messages regarding file.")


(defvar tid-:auto-save-counter 0
  "Counter incremented every every time `write-file' event happens.
See. `tid-auto-save'.")


(defconst tid-:loaded-file-list nil
  "Overwritten. List. Contain files that were loaded by `tid-ff-buffer'.
Hooks may check the contents of this.")

(defconst tid-:rejected-file-list nil
  "Overwritten. List. Contain files that were *not* loaded by `tid-ff-buffer'.
Reason may be anything: incorrect filename, path, garbage at line...
Hooks may check the contents of this.")


(defconst tid-:comment-start-level 1
  "Which sub expression is the comment start.")

(defvar tid-:last-state-file  nil
  "Last state file loaded is stored here.")


;;}}}
;;{{{ setup: user config

;;; ................................................... &v-user-congig ...

(defcustom tid-:mode-name "StateFileEdit"
  "*Editing STATE files mode name."
  :type 'string :group 'TinyDesk
  )


(defcustom tid-:dir-default "~/projects"
  "*Default directory where to save and restore files."
  :type 'directory :group 'TinyDesk
  )


(defcustom tid-:dir-suggesion 'default
  "*What is the suggested save directory:
'last        the latest that user used
'defauilt    always offer the default directory"
  :type  '(choice
	   (const last)
	   (const default)
	   )
  :group 'TinyDesk
  )


(defcustom tid-:auto-save-interval 5
  "*Interval between doing auto-save for current emacs state.
If you set it to 5, after every 5th `write-file' the state is saved.

The interval cannot be smaller than 5. It is reseted to 10 if
it's smaller that 5.

See. `tid-auto-save'"
  :type '(integer :tag "counter") :group 'TinyDesk
  )


(defcustom tid-:auto-save-name-function  nil
  "*Function to return a full path name for auto-save file.
If this variable is nil, then default name is derived from frame's
first element and it will be stored to directory `tid-:dir-default'

See. source code of `tid-auto-save'"
  :type 'function :group 'TinyDesk
  )


(defcustom tid-:save-no-file-re
  (concat

   ;;  Do save mail buffers; because you want to call M-x rmail
   ;;  instead.
   ;;
   "RMAIL\\|VM\\|MH"

   ;;  No ange ftp buffers
   ;;
   "\\(ftp\\|anonymous\\)@.*/"

   ;;  No files from these directories
   "\\|^/tmp/\\|/junk/\\|/trash/\\|/[aA]utosaved?/"
   )
  "*Regexp of files that are not saved to state file.
match is case sensitive. If you do want not case sensitive match, you
have to do set this variable to nil and use your own line delete:

   (setq tid-:save-after-hook	'my-tid-:save-after-hook)
   (defun my-tid-:save-after-hook ()
      (flush-lines REGEXP))"

   :type '(string :tag "Regexp") :group 'TinyDesk
   )


(defcustom tid-:save-title
  ;; The \$\Docid: was used by tinystamp.el, nowadays tinymy.el
  (concat
   ";;\n"
   ";;  description:  emacs desktop session file\n"
   ";;\n"
   ";;       \$\Docid: $\n\n\n"
   )
  "*Title to attach the beginning of file before saving."
  :type 'string :group 'TinyDesk
  )


(defcustom tid-:sort-when-saved t
  "*Non-nil for sorting the file list. Nil preserve `buffer-list' order."
  :type 'boolean :group 'TinyDesk
  )


;;  Set to nil if you don't want title.
;;
(defcustom tid-:get-save-file-function 'tid-get-save-files
  "*Function returning list of filenames that are stored to STATE file.
This function isn't run if `tid-save-state' is run with parameter FILES

Arguments:
  mode		flag passed by `tid-save-state'"

  :type 'function :group 'TinyDesk
  )


(defcustom tid-:face-table
  '(
    (file-pick .  highlight)
    (error     .  italic)
    )
  "*Faces used for marking text."
  :type '(list
	  (cons
	   (const  file-pick)
	   (symbol :tag "Face")
	   )
	  (cons
	   (const error)
	   (symbol :tag "Face")
	   )
	  )
  :group 'TinyDesk
  )

;;}}}
;;{{{ setup: -- version

;;; ....................................................... &v-version ...

(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinyindent.el"
 "tii"
 tii-:version-id
 "$Id: tinydesk.el,v 1.44 1997/06/19 09:31:46 jaalto NotReleased $"
 '(tid-version-id
   tid-:mode-define-keys-hook
   tid-mode-map
   tid-:load-hook
   tid-:save-before-hook
   tid-:save-after-hook
   tid-:mode-hook
   tid-:recover-before-hook
   tid-:recover-after-hook
   tid-:dir-last
   tid-:tmp-buffer
   tid-:trash-tmp-buffer
   tid-:message-column
   tid-:loaded-file-list
   tid-:rejected-file-list
   tid-:comment-start-level
   tid-:mode-name
   tid-:dir-suggesion
   tid-:save-no-file-re
   tid-:comment-chars
   tid-:save-title
   tid-:sort-when-saved
   tid-:face-table
   )))


;;}}}

;;; ########################################################### &Funcs ###



;;{{{ Install: bindings

;;; ........................................................ &bindings ...

(defun tid-default-bindings ()
  "Define default keybindings to `tid-mode-map'."

  (tinyck-maybe-activate 'emacs)

  ;;  - Don't want to use mouse-2 because it's for PASTE.
  ;;  - The others are put to mouse-2 because there is not
  ;;    not always 3 button mouse available.
  (define-key tid-mode-map [mouse-3] 'tid-mouse-load-file)

  ;;  - When editing a file, those colors might be too annoyinng,
  ;;    so you can remove properties with this. Loading is disabled too
  ;;  - Remeber, emacs is slow with this... wait some time.
  (define-key tid-mode-map [S-mouse-2] 'tid-clear-buffer-properties)

  ;;  To make buffer loadable by mouse again, run this
  (define-key tid-mode-map [C-mouse-2] 'tid-mark-buffer-loadable)

  ;;  To mark files that are not loadable, check for possibly typo in filename
  (define-key tid-mode-map [C-M-mouse-2] 'tid-set-face-non-files-buffer)

  ;; ............................................. users with no mouse ...

  ;; Select fast prefix "'" for keys, avoid carpar tunnel caused by
  ;; C-c syndrome
  ;;
  ;; The C-m is RET key.
  ;;
  (define-key tid-mode-map "'\C-m" 'tid-load-file)

  (define-key tid-mode-map "'c" 'tid-clear-buffer-properties)
  (define-key tid-mode-map "'l" 'tid-mark-buffer-loadable)
  (define-key tid-mode-map "'n" 'tid-set-face-non-files-buffer)
  (define-key tid-mode-map "'u" 'tid-unload-current-file)
  (define-key tid-mode-map "'x" 'tid-expunge-unloaded)
  (define-key tid-mode-map "'r" 'tid-remove-file-coments)

  (tinyck-maybe-activate 'emacs 'dis)
  )


(unless tid-mode-map
  (setq tid-mode-map (make-sparse-keymap))
  (run-hooks 'tid-:mode-define-keys-hook)
  )

;;}}}


;;{{{ code: misc


;;; ----------------------------------------------------------------------
;;;
(defsubst tid-comment ()
  "Return comment."
  (make-string 2 (string-to-char (substring tid-:comment-chars 0 1 ))))


;;; ----------------------------------------------------------------------
;;;
(defsubst tid-comment-re ()
  "Return comment regexp."
  (concat "[^\n"  tid-:comment-chars "]*\\([" tid-:comment-chars "].*\\)"))


;;; ----------------------------------------------------------------------
;;;
(defsubst tid-read-word ()
  "Read filename word."
  ;;   Hmmm, we allow quite a filename here !
  ;;   It's not my problem if user has such filenames ...
  (ti::b-read-word "-a-zA-Z0-9_/.!@#%&{}[]+:;~`<>")
  )


;;; ----------------------------------------------------------------------
;;;
(defsubst tid-tmp-buffer (&optional clear)
  "Return tmp buffer. optionally CLEAR it."
  (use-tmp-buffer tid-:tmp-buffer clear))


;;; ----------------------------------------------------------------------
;;;
(defsubst tid-get-file  (file)
  "Add default directory to FILE if it has no directory."
  (if file
      (if (not (string-match "/" (or file "")))
	  (setq file (concat default-directory file))))
  file
  )


;;; ----------------------------------------------------------------------
;;;
(defsubst tid-use-map ()
  "Use local \\{tid-mode-map} on this buffer."
  (use-local-map tid-mode-map))


;;; ----------------------------------------------------------------------
;;;
(defun tid-get-save-dir ()
  "Return suggested save directory."
  (let* ((type	    tid-:dir-suggesion)
	 (last	    tid-:dir-last)
	 (dir	    tid-:dir-default)
	 (ret	    dir)		;set default return value
	 )

    (if (and (eq type 'last)
	     (stringp last)
	     (file-writable-p last))
	(setq ret last))

    (if (null ret)			;make sure it's not nil
	(setq ret dir))			;use default then
    (ti::s-verify-ends ret "/")
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tid-only-files-buffer ()
  "Remove all comments and empty lines from buffer and leave 1st word."
  (interactive)
  (tid-only-files-region (point-max) (point-min)))


;;; ----------------------------------------------------------------------
;;;
(defun tid-only-files-region (beg end)
  "Remove comments BEG END and empty lines from region and leave 1st word.
This way you can rip off all comments and leave filenames."
  (interactive "r")
  (let* (
	 (sub-level	tid-:comment-start-level)
	 (comment-re	(tid-comment-re))
	 (empty-re      "^[ \t]*$\\|$")
	 mark-end
	 p maxp word tmp
	 )

    (if (> beg end)
	(setq tmp beg  beg end  end tmp))

    (save-excursion
      (goto-char end)
      ;;  have to user markers , beacuse we delete lines and point moves
      ;;  along
      (setq mark-end (point-marker))
      (goto-char beg)
      (while (< (point) (marker-position mark-end))
	(setq p (point)  maxp nil)
	(catch 'next

	  (if (null (looking-at empty-re))
	      nil
	    (ti::b-kill-line)
	    (throw 'next t))


	  (if (null (looking-at comment-re))
	      nil
	    (if (match-beginning sub-level)
		(setq maxp (match-beginning sub-level))))

	  (if (and maxp (eq maxp p))	;BEG of line comment
	      (progn
		(ti::b-kill-line) (throw 'next t)))

	  (setq word (tid-read-word))
;;;%	  (setq word (tid-read-word p maxp))
	  (ti::b-kill-line)

	  ;; The \n make cursor forward
	  (if word (insert word "\n"))
	  )
	)
      )))


;;; ----------------------------------------------------------------------
;;;
(defun tid-trash ()
  "Kill temporary buffer if user has requeted it.

References:
  `tid-:tmp-buffer'
  `tid-:trash-tmp-buffer'"
  (and tid-:trash-tmp-buffer
       (get-buffer tid-:tmp-buffer)
       (kill-buffer  (get-buffer tid-:tmp-buffer))))




;;; ----------------------------------------------------------------------
;;;
(defun tid-dired-table  ()
  "Return dired buffer table '((directory  dired-buffer) ...)."
  (interactive)
  (let ((blist   (buffer-list))
	;;  ByteCompiler doesn't know that I do
	;;  (eq major-mode 'dired-mode) test before I use this varaible,
	;;  so hide it from it.
	;;
	;;  The varaible is defined if it passed the eq test
	;;
	(sym     'dired-directory)
	elt
	list
	)
    (list-loop blist elt
      (job-buffer elt
	(if (eq major-mode 'dired-mode)
	    (nconcc
	     list
	     (list (expand-file-name (symbol-value sym))
		   (current-buffer))))
	))
    list
    ))

;;}}}
;;{{{ code: auto save

;;; ....................................................... &auto-save ...

;;; ----------------------------------------------------------------------
;;; (ad-unadvise 'save-buffer)
;;;
(defadvice save-buffer (after tid-auto-save act)
  "Run `tid-auto-save' after `save-buffer' has completed."
  (tid-auto-save))




;;; ----------------------------------------------------------------------
;;;
(defun tid-auto-save  ()
  "This function is installed in `write-file-hooks'. Periodic autosave.
It increments counter every time file is saved.
Every Nth time the state of the emacs (which files were loaded into emacs)
is saved, so that you can recover the same session if emacs crashes.

The default state name is derived in the following manner
o  use directory `tid-:dir-default'
o  add string \"state.\"
o  get frame's first word, usually the one that gets set when
   you use -name XXX switch in emacs command line. If Emacs is beeing
   run with -nw option, the frame name returns \"terminal\"
o  if there is no frame name, then use \"saved\"

   Possible yielding: ~/projects/state.saved

References:
  `tid-:auto-save-interval'       ,every Nth write
  `tid-:auto-save-name-function'  ,function to generate name of state file."
  (let* ((func       tid-:auto-save-name-function)
	 (dir        (ti::s-verify-ends
		      (or tid-:dir-default "~" )
		      "/"))
	 (frame-name (ti::xe-get-frame-parameter 'name (selected-frame)))
	 (name       (strmatget "^\\([^*\t ]+\\)" 1 frame-name))
	 (fn         (concat dir "state." (or name "saved")));; default
	 (save-to    fn)
	 )
    (if func
	(setq save-to (funcall func)))

    (when (stringp save-to)

      ;;  - Be extra carefull, because we're in write file hook
      ;;  - Make sure we succeed always

      (if (not (integerp tid-:auto-save-counter));; init if not int
	  (setq tid-:auto-save-counter 0))

      (if (not (integerp tid-:auto-save-interval));; user didn't set this?
	  (setq tid-:auto-save-interval 10))

      (if (< tid-:auto-save-interval 5);; Must be more than 5
	  (setq tid-:auto-save-interval 10))

      (inc tid-:auto-save-counter)

      ;;  - time up ? Select name FN if it's string.
      ;;
      (cond
       ((eq 0 (% tid-:auto-save-counter tid-:auto-save-interval))
	;;   Actually tid-save-state generates new call to this
	;;   function but, it won't come in this COND, because the counter
	;;   value is different.

	(setq save-to (if (stringp save-to) save-to fn))

	;; Try chmod, if it fails, then signal error
	;;
	(if (not (file-writable-p save-to))
	    (set-file-modes
	     save-to
	     (file-mode-make-writable (file-modes save-to))
	     ))


	(if (not (file-writable-p save-to))
	    (error "Can't do state autosave: [%s] not writable " save-to))

	(save-window-excursion
	  (save-excursion
	    (tid-save-state save-to)))
	))

      nil
      )))






;;}}}

;;{{{ Code: faces

;;; ............................................................ &face ...

;;; ----------------------------------------------------------------------
;;;
(defun tid-face (face)
  "Return FACE."
  ;;  This way the global variable does not float around the file
  (cdr (assoc face tid-:face-table)))


;;; ----------------------------------------------------------------------
;;;
(defun tid-clear-line-properties ()
  "Remove properties from the line."
  (let* ()
    (set-text-properties (BOLPP) (EOLPP) nil)
    (set-buffer-modified-p nil)
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tid-clear-buffer-properties ()
  "Remove properties and EOL comments from buffer."
  (interactive)
  (let* ((c-chars	tid-:comment-chars)
	 (c-lev		tid-:comment-start-level)
  	 (c-re		(tid-comment-re))
	 )
    (tid-clear-region-properties (point-min) (point-max))
    (save-excursion
      (PMIN)
      (while (not (eobp))

	;;  Skip over BOL comments
	(if (and (not (looking-at (concat "^[ \t]*[" c-chars "]+\\|$")))
		 (looking-at c-re)
		 (match-beginning c-lev)
		 )
	    (delete-region (match-beginning c-lev) (EOLPP)))
	(forward-line 1)
	))
    (set-buffer-modified-p nil)

    (message "*properties cleared from buffer")

    ;;  Little nasty, but Emacs does not update display always...
    (redraw-display)

    ))


;;; ----------------------------------------------------------------------
;;;
(defun tid-clear-region-properties (beg end)
  "Remove properties from BEG END."
  (let* ()
    (set-text-properties beg end nil)
    (set-buffer-modified-p nil)
    ))



;;; ----------------------------------------------------------------------
;;;
(defun tid-line-property-set-error ()
  "Set line face to signify error."
  (let* (beg
	 end
	 )
    (save-excursion
      (beginning-of-line)		(setq beg (point))
      (skip-chars-forward "^ \t\n")	(setq end (point))
      )

    ;; clear first full line
    (put-text-property beg (EOLP) 'face 'default)

    (put-text-property beg end 'face (tid-face 'error))
    (set-buffer-modified-p nil)
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tid-handle-text-property (p text)
  "Look property P and run command under TEXT."
  (let* ((file	     (file-name-nondirectory text))
	 (loaded     (get-buffer file))	;in emacs already ?
	 (comment    (tid-comment))
	 (err-col    tid-:message-column)
	 )
    ;;  - We need the sleep-for, because moving the mouse
    ;;    clears the message and user may not notice it.
    (cond
     ((eq p (tid-face 'file-pick))
      (cond
      ((null (file-exists-p text))
       (message (concat "File not exist: " text)) (sleep-for 0)
       (tid-clear-line-properties)
       )
      (loaded
       (message "File already loaded.") (sleep-for 0)
       (tid-clear-line-properties)
       )
      (t
       (find-file-noselect text)
       (message "Loaded.") (sleep-for 0)
       (tid-clear-line-properties)
       (move-to-column err-col 'force)
       (if (not (looking-at "$\\|[ \t]*$"))
	   (end-of-line))
       (insert comment " loaded")
       (beginning-of-line)
       ))
      )
     )))

;;; ----------------------------------------------------------------------
;;;
(defun tid-mark-buffer-loadable (&optional verb)
  "Parse whole buffer and make first _word_ loadable with mouse. VERB.
Marking is only done if word is valid filename."
  (interactive)
  (save-excursion
    (tid-mark-region (point-min) (point-max)
		     (tid-comment-re)
		     tid-:comment-start-level
		     (or (interactive-p) verb)
		     )
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tid-set-face-non-files-buffer  ()
  "Change face to 'error in those lines whose first word is not valid file."
  (interactive)
  (tid-set-face-non-files-region (point-min) (point-max))
  (if (interactive-p)
      (message "marked non-lodable files"))
  )


;;; ----------------------------------------------------------------------
;;;
(defun tid-set-face-non-files-region (beg end)
  "Change face to 'error in those lines whose first word is not valid file.
Also add comment marker for user that do not have highlight capabities.
Region is BEG END."
  (interactive "r")
  (let* ((empty-re	"^[ \t]*$")
	 (sub-level	tid-:comment-start-level)
	 (c-chars	tid-:comment-chars)
	 (comment       (tid-comment))
	 (comment-re	(tid-comment-re))
	 (err-col	tid-:message-column)
	 word

	 )
    (save-excursion
      (save-restriction
	(narrow-to-region beg end)
	(PMIN)
	(while (not (eobp))

	  ;;  - ignore empty lines and BEG of line comments.
	  (if (or (looking-at empty-re)
		  (and (looking-at comment-re)
		       (eq (match-beginning sub-level) (point))))
	      nil
	    (setq word (tid-read-word))
	    (if (and word
		     (file-exists-p word))
		nil
;;;	    (d! word)
	      (tid-line-property-set-error)	;put color on line

	      ;;  Show to user that does not see the color
	      (move-to-column err-col 'force)

	      ;; If the filename that long, that it goes past err-col ?
	      (cond
	       ((eolp))			;do nothing
	       ((looking-at (concat "[ \t" c-chars "]"))
		(kill-line))		;delete other marks
	       (t
		(end-of-line)))		;no other choices. place is cccupied

	      (insert (concat comment " invalid"))
	      ))
	  (forward-line 1)
	  )))
    ))



;;; ----------------------------------------------------------------------
;;; - This function is not general.
;;; #todo: rewrite it for this module only
;;;
(defun tid-mark-region (beg end &optional com-re sub-level verb)
  "Make all filenames in the buffer loadable by mouse.
Supposes that the first _word_ on the line is filename.
If the first word isn't loadable file, its face isn't changed.
If there is no directory part, then `default-directory' is supposed.

Input:
  BEG END   region
  COM-RE    the file can have comments, but comments can be only
	    _single span type_, that is, only shell like #, or C++
	    like //. Everything after and included  COM-RE is discarded
	    from SUB-LEVEL.
  SUB-LEVEL subexpression match; default is 0.
  VERB	    verbose messages.

Example:

  Myfile.sh   #comment

  com-re     = '.*\\\\(#\\\\)'
  sub-level  = 1 , because there is paren"
  (let* (
	 (err-col	tid-:message-column)
	 (file-face	(tid-face 'file-pick))
	 (sub-level	(or sub-level 0))
	 (c-chars	tid-:comment-chars)
	 (comment	(tid-comment))
	 bp ep				;beg, end points
	 elp				;end line point
	 maxlp				;max line point
	 file
	 )
    (and verb				;this make take a while...
	 (message "Marking files..."))

    (save-restriction
      (narrow-to-region beg end)
      (goto-char (point-min))

      (while (not (eobp))
	(if (looking-at  "^[ \t]*$\\|$")
	    nil				;ignore empty lines

	  (setq elp (EOLP))
	  (setq maxlp elp)

	  ;;  Does there exist comment on the line ?
	  (save-excursion
	    (cond
	     ((and (stringp com-re)
		   (looking-at com-re))
	      (setq maxlp (1- (match-beginning sub-level)))
	      )))

	  (if (< maxlp (1+ (point)))	;beg of line COMMENT exist
	      (progn
;;;		(d! "skipped" (read-full-line))
		nil
		)
	    (skip-syntax-forward " " maxlp) ;ignore whitespace
	    (setq bp (point))

	    (skip-chars-forward "^ \t"  maxlp) ;first space

	    (if (eq bp ep) 		;not moved, maybe "one-word$" ?
		(goto-char maxlp))

	    (setq ep (point))
	    (if (eq bp ep)

		nil			;still not moved ?
	      ;;  Mark the word only if the WORD is valid file
	      ;;  - If the filename has ange-ftp @ char, then mark
	      ;;    automatically. Calling file-exists-p for ange
	      ;;    file would start ange-ftp... and we don't
	      ;;    want that here.
	      ;;
	      (setq file (buffer-substring bp ep))
	      (setq file (tid-get-file file))

;;;	      (d! bp ep (point) file )

	      (goto-char ep)
	      (if (looking-at "[ \t;]")
		  (delete-region (point) elp))	;delete other marks

	      (cond
	       ((get-file-buffer file)	;already in emacs ?

		(move-to-column err-col 'force)
		(if (not (looking-at (concat "$\\|[ \t" c-chars "]")))
		    (end-of-line))	;no other choices, place is cccupied

		(insert (concat comment " loaded"))
		)
	       (t
                ;; ............................... not loaded in emacs ...
		(if (or (string-match "@" file)
			(file-exists-p file))
		    (put-text-property bp ep 'mouse-face file-face))
		)))
	    ))
	(forward-line 1)
	)
      (set-buffer-modified-p nil)
      (and verb				;this make take a while...
	   (message "Marking files...ok"))

      )))

;;}}}
;;{{{ code: mouse

;;; ........................................................... &mouse ...

;;; ----------------------------------------------------------------------
;;;
(defun tid-mouse-load-file (event)
  "Load file under mouse. Use mouse EVENT."
  (interactive "e")
  (mouse-set-point event)		;move point there
  (tid-load-file)
  )

;;; ----------------------------------------------------------------------
;;;
(defun tid-remove-file-coments  ()
  "Remove all comment at `tid-:message-column'."
  (interactive)
  (save-line-col nil nil
    (PMIN)
    (while (re-search-forward (tid-comment) nil t)
      (if (eq (current-column) (+ 2 tid-:message-column))
	  (delete-region (- (point) 2) (EOLP))))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tid-expunge-unloaded  ()
  "Remove lines that have 'unloaded' flag."
  (interactive)
  (save-line-col nil nil
    (PMIN)
    (while (re-search-forward (format "%s unloaded$" (tid-comment)) nil t)
      (ti::b-kill-line))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tid-unload-current-file  ()
  "Remove file on this line from emacs."
  (interactive)
  (let* ((file (tid-get-file (tid-read-word)))
	 buffer
         )
    (when file
      (if (setq buffer (find-buffer-visiting file))
	  (kill-buffer buffer)
	(message "No such buffer."))

      (beginning-of-line)
      (or (and (re-search-forward (tid-comment) nil t)
	       (progn (delete-region (point) (EOLP)) t))
	  (and (move-to-column tid-:message-column 'force)
	       (insert ";;")))
      (insert " unloaded")
      (forward-line 1)
      )))


;;; ----------------------------------------------------------------------
;;;
(defun tid-load-file ()
  "Load file under point."
  (interactive)
  (let* (prop
	 word
	 )
    (setq prop (get-text-property (point) 'mouse-face))

    (cond
     (prop				;property found?
      (setq word  (tid-get-file (tid-read-word))) ;read word under cursor
      (cond
       (word				;grabbed
	(tid-handle-text-property prop word)
	)))
     ((interactive-p)
      (message
       (substitute-command-keys
	(concat
	 "Can't find mouse face...   Mark buffer first with "
	 "\\[tid-mark-buffer-loadable]")))
      ))
    ))

;;}}}

;;{{{ code: edit, unload

;;; ............................................................ &edit ...

;;; ----------------------------------------------------------------------
;;;
;;;###autolaod
(defun tid-unload (file &optional verb)
  "Unload all files from emacs that are in state file FILE.

If VERB is non-nil offer verbose messages [for code calls]; interactive
call always turns on verbose."
  (interactive
   (list
    (let* ((save-dir	(tid-get-save-dir))
	   (save-dir	(if save-dir save-dir "./"))
	   (msg	        (concat "Unload from state file: "))
	   )
      (read-file-name msg  save-dir))
    ))
  (let* ((b      (tid-tmp-buffer))
	 (dlist  (tid-dired-table))
	 (count  0)
	 buffer
	 elt
	 fn
	 )
    (VERB)

    (save-excursion
      (set-buffer b)
      (erase-buffer)
      (insert-file-contents file)

      ;;  - Now process every line. We don't care if we read commented
      ;;    line as "buffer" because the test inside loop return nil
      ;;    for such lines...
      (PMIN)
      (if (eobp)
	  (if verb (message "Empty state file."))
	(while (not (eobp))		;check if exist
	  (setq fn (read-full-line))

	  ;;  File name must start with absolute path
	  ;;  find-buffer-visiting function can find files that
	  ;;  may be symlinks.
	  ;;
	  (when
	      (and (string-match "^/" fn)
		   (or (and (file-directory-p fn)
			    (setq elt (assoc (expand-file-name fn) dlist))
			    (setq buffer (nth 1 elt)))
		       (setq buffer (find-buffer-visiting fn))
		       ))
	    (kill-buffer buffer)
	    (inc count)
	    )
	  (forward-line)
	  )))

    (when verb
      (if (> count 0)
	  (message (format "Removed %d files" count))
	(message "No files removed.")))

    (tid-trash)
    ))


;;; ----------------------------------------------------------------------
;;;
;;;###autolaod
(defun tid-mode (&optional no-face verb)
  "Mark and parse buffer's 1st words as loaded files.
If NO-FACE is non-nil, the default mouse marking isn't performed. VERB.

Comments in the right tell what is the files status:
loaded      = file inside emacs already
invalid     = the path is invalid, no such file exists

Mode description:
\\{tid-mode-map}"
  (interactive "P")
  (VERB)

  ;; - If the file is already in buffer, remove extra marks, like
  ;;   non-loadable files.
  (tid-clear-region-properties (point-min) (point-max))
  (tid-remove-file-coments)

  (if (null no-face)
      (tid-mark-buffer-loadable verb))

  (tid-use-map)				;turn on the map
  (setq  mode-name   tid-:mode-name)
  (setq  major-mode 'tid-mode);; for C-h m

  (when verb
    (message
     (substitute-command-keys
      (concat
       "load \\[tid-load-file], "
       "clear \\[tid-clear-buffer-properties], "
       "error \\[tid-set-face-non-files-buffer], "
       "mark \\[tid-mark-buffer-loadable]"
       )))
    (sleep-for 1)
    )
  (run-hooks 'tid-:mode-hook)
  )


;;; ----------------------------------------------------------------------
;;;
;;;###autolaod
(defun tid-edit-state-file (file)
  "Load state FILE into buffer for editing.
You can add comments and remove/add files. Turns on `tid-mode'.

Following commands are available in `tid-mode'.
\\{tid-mode-map}"
  ;;  I can't use interactive "f" , beacuse I want the completions
  ;;  to come from the save-directory. The "f" uses by default the
  ;;  variable default-directory...
  (interactive
   (list
    (let* ((save-dir	(tid-get-save-dir))
	   (save-dir	(if save-dir save-dir "./"))
	   (msg	(concat "Edit state file: "))
	   )
      (read-file-name msg  save-dir))
    ))
  ;; - If file is already loaded, avoid creating duplicate window
  (pop-to-buffer (find-file-noselect file))
  (tid-mode nil t)
  )




;;}}}
;;{{{ code: save

;;; ............................................................ &save ...

;;; ----------------------------------------------------------------------
;;;
(defun tid-get-save-files (&optional dirs)
  "Return list of files to save, optionally DIRS too."
  (let ((blist (buffer-list))
	;;  See function tid-dired-table for explanation
	;;
	(sym  'dired-directory)
	elt
	list
	)
    (list-loop blist elt
      (job-buffer elt
	(cond
	 ((and dirs
	       (eq major-mode 'dired-mode))
	  (nconcc list (expand-file-name (symbol-value sym))))
	 ((buffer-file-name)
	  (nconcc list (buffer-file-name))
	  ))
	))
    list
    ))


;;; ----------------------------------------------------------------------
;;;
;;;###autolaod
(defun tid-save-state (file &optional mode files verb)
  "Output all files in emacs into FILE.
Notice, that this concerns only buffers with filenames.

Input:
  FILE		,the STATE file beeing saved
  MODE          ,Controls what is saved
		 nil	= only filenames
		 '(4)	= C - u, filenames and directories.
  FILES		,filenames , absolute ones
  VERB		,verbose flag"
  (interactive
   (list
    (read-file-name
     "Save state to: "
     (if (tid-get-save-dir) (tid-get-save-dir) "./"))
    current-prefix-arg
    ))


  (let* ((tmp-buffer	(tid-tmp-buffer 'clear))
	 (save-func	tid-:get-save-file-function)
	 (sort		tid-:sort-when-saved)
	 (title		tid-:save-title)
	 (re-no		tid-:save-no-file-re)
	 buffer
	 elt
	 )
    (VERB)

    (cond
     ((setq buffer (get-file-buffer file))
      (pop-to-buffer buffer)
      (error "Please save to new file or kill buffer: %s" file )
      ))


    (run-hooks 'tid-:save-before-hook)
    (setq files (or files
		    (and (fboundp save-func)
			 (funcall save-func mode))))

    (if (null files)
	(if verb			;no files
	    (message "no files to save"))
      ;; ... ... ... ... ... ... ... ... ... ... ... ... ... . do save . .
      (if (or  (null file)
	       (and (file-exists-p file)
		    (null (file-name-directory file))
		    ))
	  (error (format  "access problem with: '%s'" file)))

      ;;  We kill this buffer later, so we don't need save-excursion
      (set-buffer tmp-buffer)
;;; (pop-to-buffer tmp-buffer)


      ;; ... ... ... ... ... ... ... ... ... ... ... ...  insert files . .
      (list-loop files elt
	;;  Remove some files...
	(if (or (not (stringp re-no))
		(and (stringp re-no)
		     (not
		      (string-match-case re-no elt))))
	    (insert (car files) "\n"))
	)


      ;; ... ... ... ... ... ... ... ... ... ... ... ... ... ...  sort . .
      (if sort
	  (sort-lines nil (point-min) (point-max)))

      ;; ... ... ... ... ... ... ... ... ... ... ... ... ... ... title . .
      (if (null (stringp title))
	  nil				;no title
	(PMIN)
	(insert title))

      (run-hooks 'tid-:save-after-hook)
      (write-region (point-min) (point-max) file)

      (not-modified) (message "")

      (kill-buffer tmp-buffer)
      (if (interactive-p)
	  (message (concat "State saved to file " file)))
      ;; ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ catch ^^^
      )

    (tid-trash)
  ))

;;}}}
;;{{{ code: recover

;;; ......................................................... &recover ...

;;; ----------------------------------------------------------------------
;;;
;;;###autolaod
(defun tid-recover-state (file &optional ulp pop verb)
  "Load all files listed in FILE into emacs.
FILE can have empty lines or comments. No spaces allowed at the
beginning of filename. The state FILE itself is not left inside
emacs if everything loads well. When all files are already
in emacs, you may see '0 files loaded' message.

In case there were problems, the FILE will be shown and the
problematic lines are highlighted.

Prefix arg sets flag ULP, unload previous.


Input:

  FILE		state file to load

  ULP		'unload previous' if non-nil then unload previously
		loaded files according to `tid-:last-state-file'

  POP           if non-nil, then pop to first buffer in the saved
                state file. This flag is turned automatically in
                interactive calls

  VERB		non-nil enables verbose messages. If called interactively
		this is enabled automatically.

References:
  `tid-:last-state-file'	    ,hold the state file that was loaded."
  (interactive
   (list
    (read-file-name "load state file: "
		    (if (tid-get-save-dir)
			(tid-get-save-dir)
		      "./"))
    current-prefix-arg
    t
    ))

  (let* ((count		0)
	 (state-file    (expand-file-name file))
	 (last-state	tid-:last-state-file)
	 buffer
	 kill-buffer
	 err
	 not-loaded
	 l
	 first-entry
	 )
    (VERB)

    ;; o  read the config file
    ;; o  raise the kill flag if the file ISN'T already loaded, user
    ;;    may be editing it.
    ;; o  There may be buffers with the same name, but different dirs..
    ;;
    (if (setq buffer (get-file-buffer state-file))
	nil				;yup, it's loaded
      (setq kill-buffer t)		;different directory
      (setq buffer (find-file-noselect state-file))
      )

    ;; ... ... ... ... ... ... ... ... ... ... ... ... unload previous ...
    (if (and ulp (stringp last-state))
	(if (not (file-exists-p last-state))
	    (message-quick
	     (format "Cannot unload, file does not exist '%s' " last-state))
	  (tid-unload last-state)
	  ))


    (save-excursion
      (set-buffer (use-tmp-buffer buffer))
      (setq l (tid-ff-buffer))
      (setq  count	    (nth 0 l)
	     err	    (nth 1 l)
	     not-loaded	    (nth 2 l)
	     first-entry    (nth 3 l)
	     )
      (when (and pop first-entry)
	(pop-to-buffer (get-file-buffer first-entry)))

      (cond
       ((null err)
	(if verb (message (format "%d files loaded" count)))
	(run-hooks 'tid-:recover-after-hook)

	;;  kill the buffer only if it was loaded by us
	(and kill-buffer (kill-buffer buffer))
	)
       (t
	;;  Show user the failed files
	(message (concat "Not loaded> " not-loaded))
	(sleep-for 0)
	(pop-to-buffer buffer)
	(tid-mode 'no-face 'verbosee)
	(tid-set-face-non-files-buffer)
	(PMIN)
	))

      (setq tid-:last-state-file file)
      )))



;;; ----------------------------------------------------------------------
;;;
(defun tid-ff-buffer ()
  "Abbreviation ff, Find-file-buffer,  Load all files listed in buffer.
Point is not preserved.

References:
  `tid-:loaded-file-list'
  `tid-:rejected-file-list'

Return:
   (count err not-loaded-string first-entry)
   count		,how many files actually loaded
   err			,error while load
   not-loaded-string	,files that had problems.
   first-entry		,first entry"
  (interactive)
  (let* ((count		0)
	 (sub-level	tid-:comment-start-level)
	 (ignore-re     (tid-comment-re))
	 (empty-re      "^[ \t]*$")
	 first-entry			;
	 bp
	 not-loaded
	 load				;file ont the line to be processed
	 maxp				;max point
	 word
	 err				;per file basis
	 ERR				;return status
	 )


    (setq   tid-:loaded-file-list   nil	;<< reset GLOBALS
	    tid-:rejected-file-list	nil)

    (run-hooks 'tid-:recover-before-hook)
    (PMIN)				;there is *no* save excursion

    (while (not (eobp))

      (setq bp (point)  err nil)	;BEG of line
      (setq maxp (save-excursion (end-of-line) (point)))
      (beginning-of-line)

      (catch 'next
        ;; ... ... ... ... ... ... ... ... ... ... ... ... .. comments ...
	(if (looking-at empty-re)	;emty lines
	    (throw 'next t))

	(if (null (looking-at ignore-re)) ;no comment in this line
	    nil
	  (setq maxp  (match-beginning sub-level))
;;	  (d! "ignore" maxp (read-full-line))
	  )

	(if (= maxp bp)			;full comment line ?
	    (throw 'next t))

        ;; ... ... ... ... ... ... ... ... ... ... ...  read file name ...
	;;  Now load the file, raise err if not loaded
	;;  - remove whitespace at front first

	(if (looking-at "[ \t]+")
	    (skip-chars-forward " \t" maxp))
	(setq word (tid-read-word))
;;;	(d! "W" (point) maxp  word)

	;; --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``--
	(when word
	  (setq load (expand-file-name word))

;;;	  (d! "buffer?" (get-file-buffer load) (file-loadable-p load) load)

	  (if (or (null load)
		  (get-file-buffer load))
	      nil			;already in emacs

	    (if (not (file-loadable-p load))
		(setq err t)
	      (if (condition-case nil
		      (progn
			(find-file-noselect load)
			t)
		    (error
		     (setq err t)
		     nil))
		  (progn
		    (setq count (1+ count))
		    (if (null first-entry)
			(setq first-entry word))
		    (nconcc tid-:loaded-file-list word))
		))

;;;	    (d! "load" err load )

	    ))

	;; --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``--
	(if (null err)
	    nil				;no errors to report
	  (setq ERR t)
	  (push word tid-:rejected-file-list)
	  (and (interactive-p)
	       (tid-line-property-set-error))
	  (setq  not-loaded
		 (concat
		  (or not-loaded "")	;start value
		  (or
		   (file-name-nondirectory load)
		   "[nil-word]")
		  " ")))

	  )				;catch line
      (forward-line 1)
      )					;while loop
;;;    (d! "load-end" count ERR not-loaded)
    (list count ERR not-loaded first-entry)
    ))

;;}}}


(provide   'tinydesk)
(run-hooks 'tid-:load-hook)

;;; tinydesk.el ends here
