;; @(#) tinyeat.el -- Eating blocks of text forward, backward

;; @(#) $Id: tinyeat.el,v 1.37 1997/05/31 12:43:25 jaalto NotReleased $
;; @(#) $Keywords: extensions $
;; $KnownCompatibility: 19.28 , XE19.14 $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Documentation

;; Copyright (C) 1995,96,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Apr 2 1995
;;
;; To get information on this program use ident(1) or do M-x tie-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el


;; LCD Archive Entry:
;; tinyeat|Jari Aalto|jari.aalto@poboxes.com|
;; Eating blocks of text forward, backward. No more ESC-d or or DEL/BACKSPACE|
;; 31-May-1997|$Revision: 1.37 $|~/misc/tinyeat.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ....................................................... &t-install ...
;; Put this file on your Emacs-Lisp load path, add following into your
;; emacs startup file. Rip code with tinylib.el / ti::m-pkg-rip-magic
;;
;;* _
;;*  	(require 'tinyeat)
;;
;; Or use autoload, your emacs starts up a bit faster
;;
;;*     (autoload 'tie-forward-preserve         "tinyeat" t t)
;;*     (autoload 'tie-backward-preserve        "tinyeat" t t)
;;*     (autoload 'tie-delete-paragraph         "tinyeat" t t)
;;*     (autoload 'tie-kill-line                "tinyeat" t t)
;;*     (autoload 'tie-zap-line                 "tinyeat" t t)
;;*     (autoload 'tie-kill-line-back           "tinyeat" t t)
;;*     (autoload 'tie-kill-buffer-lines        "tinyeat" t t)
;;*     (autoload 'tie-kill-buffer-lines-min    "tinyeat" t t)
;;* _
;;
;; If you don't want the default binding put this to your emacs
;; prior the above statements and define your own keys.
;;
;;*     (setq tie-:load-hook nil)
;;* _
;;
;; If you have any questions, use this function
;;
;;      M-x tie-submit-bug-report



;;}}}
;;{{{ Documentation

;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "tie-" in front of
;;   every function & variable. It stands for '(ti)ny (e)at
;; - variable names contain letters 'tie-:'

;;  Preface, overview of features
;;
;;	o   Determines how much text should be eaten around current cursor
;;	    position. Normally eats extra spaces, extra newlines, next word
;;	    next statement , next comment ... whatever is appropriate
;;	o   When you grow accustomed to this, it probably almost replaces your
;;	    single BACKSPACE and DEL keys.
;;	o   Can also eat only 'inside' words: WordsThatAreLikeThis
;;
;;  Story behind this package
;;
;;	I got sick of moving cursor around the point and using DEL/BACKSPACE
;;	for example when writing C++ and LISP symbols. Say I have:
;;
;;	    (defun lisp-symbol-name-myname          ()
;;                                  *
;;
;;	And I decide I want to change that 'myname' to something
;;	else. Normally I'd have to reach out for ESC-d for kill-word to
;;	delete 'myname'.  Then I'd type a new name:
;;
;;	    (defun lisp-symbol-name-mynew           ()
;;		                         *
;;
;;	Now I notice that there are extra spaces involved, I don't want
;;	those before the '()' marks for function name, I have to call
;;	fixup-whitespace for that ... damn it's not bound to any keys by
;;	default, so I have to type it the long way round: 'M-x
;;	fixup-whitespace'. now I'm beginning to think: "Yeah, Why I
;;	didn't bind it to keys...".  next thing I notice that the
;;	'mynew' wasn't good name after all, So I decide to delete 3
;;	words backward. Now, how do I do that?
;;
;;	    (defun lisp-symbol-name-mynew ()
;;		       		         *
;;
;;	Where is the command to delete backward ... (I start remembering)
;;	After spending valuable minutes to finding the delete-backward-word
;;	command: "Damn again, there is no such command". I end up tapping
;;	backspace until I reach the correction point:
;;
;;	    (defun lisp- ()
;;	                *
;;
;;	And start typing new name...
;;
;;	    (defun lisp-my-func ()
;;
;;	Then I suddenly notice that there is too many newlines above
;;	my newly created function: "I really should delete those 5
;;	extra ones".  "Now, how do I kill backward ? The kill-line
;;	in C-k kills only forward" ...
;;
;;  Lesson learned
;;
;;	As you can notice, we often spend most of our time to
;;	position the cursor in the right places and deleting text
;;	over there.. over here ..  typing more .. changing our mind
;;	... etc.
;;
;;	I thought it was time to do something creative, so that I
;;	wouldn't have to worry about the deleting of text so much.
;;	This .el provides _smart_ deleting of text, whether you want
;;	to delete forward of backward. Naturally it isn't capable of
;;	miracles, it just does few guesses, and a guess may be
;;	wrong.
;;
;;	If you end up in a situation where you suddenly notice that
;;	vast area of text have just vanished from you buffer,
;;	remember, there is no need to panic. Just send a bug report
;;	to me, and hit UNDO and use normal DEL/BACKSPACE. Emacs is
;;	great, because it lets programs, like this, to do mistakes
;;	... while allowing you to restore what it did.
;;
;;	I'd be surprised if you ever want to discard this .el when you have
;;	tried it. :-/

;;}}}
;;{{{ history

;; ....................................................... &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; May	31	1997	[jari]		19.28	v1.37		NotReleased
;; - lib 'm' changes. checdoc 1.29 clear. defcustom support.
;;
;; May	12	1997	[jari]		19.28	v1.33-1.36	NotReleased
;; - internal changes.
;; - Changed all non standard boolean tests to use when and unless
;;
;; Mar	19	1997	[jari]		19.28	v1.29-1.32	NotReleased
;; - Made tie-delete-whole-word more smarter and user configurable.
;;   See variable tie-:eat-full-word-charset
;; - XE20 corrections; CHAR handling.
;;
;; Jan	15	1997	[jari]		19.28	v1.27-1.28	NotReleased
;; - Now all autoloads are in the 'm' lib.
;;
;; Nov	8	1996	[jari]		19.28	v1.26		NotReleased
;; - Wayne Adams <wadams@galaxy.sps.mot.com> suggested that there should
;;   be function which deleted whole word at point.
;;   tie-delete-whole-word		:+ put to S-backspace
;;
;; Nov	6	1996	[jari]		19.28	v1.25		NotReleased
;; - Corrected (alt k) problem in 19.28. Not it activates tinyck
;;   to handle the definition right and making the key active.
;;
;; Oct	10	1996	[jari]		19.28	v1.24		NotReleased
;; - Corrected Alt-k binding in older emacs versions (required loading
;;   tinyck).
;;
;; Sep	25	1996    [jari]		19.28	v1.21-1.23	NotReleased
;; - ptype@dra.hmg.gb (Proto Type) suggested that tinyck should not be
;;   loaded if emacs is used. Rearranged loading of all functions.
;;   Now this package should be slick to load :-)
;;   tie-yank-overwrite			:+ new user function
;;
;; Aug	20	1996    [jari]		19.28	v1.20		NotReleased
;; - Corrected all XE 19.14 byteCompiler errors.
;;
;; Aug	16	1996    [jari]		19.28	v1.19		NotReleased
;; - Rewrote keybinding installation. Now passes XE 19.14.
;; - Correctes syntax commands to work in XE19.14
;;
;; Jun	13	1996    [jari]		19.28	v1.18		NotReleased
;; - Small error in keybinding installation corrected.
;;
;; Jun	9	1996    [jari]		19.28	v1.17		NotReleased
;; - Rewrote the installation. Added Minibuffer kill functions too.
;;
;; Apr	14	1996    [jari]		19.28	v1.16		NotReleased
;; - The backward deletion hanged too much. Changed algorithm to make it
;;   faster.
;;   tie-word-move-point		:! corrected.
;;
;; Apr	11	1996    [jari]		19.28	v1.15		NotReleased
;; - There was lot to improve in word selection function. Now it
;;   can delete BIGletters and no more stucks in some places, like
;;   when looking at period(.) in text-mode. (a bug)
;;   tie-word-move-point		:! rewrote
;;
;; Apr	9	1996    [jari]		19.28	v1.14		NotReleased
;; - Removed duplicate functions that were already in tinylib.el. Added
;;   require.
;; - Rewrote the work killing, so that it's possible to kill C++ styled
;;   mixed case words more comfortable.
;;   tie-word-move-point		:+ handles word positioning
;;   tie-eat				:! new work killing
;;
;; Feb	12	1996    [jari]		19.28	v1.13		Release_4
;; - Corrected last byteCompiler warnings.
;;
;; Feb	10	1996    [jari]		19.28	v1.12		NotReleased
;; - Rewrote the keybinding suggestions. Added new variable
;;   tie-:non-word-chars which controls eating better.
;; - Rewrote the version functions and used lisp-mnt.el lm-verify
;;   to check layout.
;;
;; Sep	28	1995    [jari]		19.28	v1.11		Release_3
;; - errors corrected: the macros EOLP,BOLP .. missed
;; - added tie-kill-line, which preserves mouse cut buffer.
;; - added interactive to tie-kill-buffer-lines-min
;;
;; Sep	28	1995    [jari]		19.28	v1.10		Release_2
;; - Small error in parameter kill-lines "arg" --> "back"
;;
;; Sep	26	1995    [jari]		19.28	v1.8-9		NotReleased
;; - Added more keybind/autoload statements to the installation
;; - Added missing "##autoload" for new funcs.
;;
;; Sep	26	1995    [jari]		19.28	v1.7		NotReleased
;; - Added trivial buffer text kill functions tie-kill-buffer-lines,
;;   tie-kill-buffer-lines-min
;;
;; Sep	26	1995    [jari]		19.28	v1.6		NotReleased
;; - Corrected little byte compile error. Also one missing ')' in func.
;;
;; Sep	26	1995    [jari]		19.28	v1.5		NotReleased
;; - Added tie-kill-line-back, which reverses the kill-line function.
;;   very handy!
;;
;; Sep	1	1995    [jari]		19.28	v1.4		NotReleased
;; - Little more fine tuning. Deleting multiple newlines is now more
;;   substle: N lines --> 1 sperating line --> Delete that 1 line
;;   The old version didn't keep the cursor at the same place all the time
;;   when killing newlines forward.
;; - Added "Line Cleared" message. Deleted 3 x \n testing from the main
;;   section, because 2 x \n section could do the same.
;; - Added ###autolad directives. Some functions didn't have 'tie-' prefix.
;;
;; Jun	5	1995    [jari]		19.28	v1.3		Release_2
;; - No changes, since this has been working like charm for me.
;; + Added one usefull eat function: current paragraph which is suggested to
;;   be put on C-S-backspace.
;;
;; Apr	2	1995    [jari]		19.28	v1.2		Release_1
;; - Finished all funcs.
;;
;; Apr	2	1995    [jari]		19.28	v1.1		NotReleased
;; - Moved my not-so-good-kill functions here and rewrote everything.

;; Thank you list:
;; nickson@cs.uq.oz.au (Ray Nickson)
;; - His little function tie-looking-back-at gave me an inspiration
;;   to develop this package.
;;
;; Advertise
;; - See tinymouse.el for optimal use of your mouse. Eg with mouse you can
;;   scroll the buffer lot easier than you thought... and there is function
;;   that really overwrites when you paste text with mouse [emacs can't
;;   do this by default, even when ovwrt-mode is on]
;; - Other "tiny tools" in the ~/misc directory of OHIO.




;;}}}

;;; Code:

;;{{{ setup: variables

;;; ......................................................... &require ...

(require 'tinylibm)
(eval-when-compile (use-dynamic-compilation))

(defgroup-tiny TinyEat tie-: extension
    "Eating blocks of text forward, backward.
Overview of features

	o   Determines how much text should be eaten around current cursor
	    position. Normally eats extra spaces, extra newlines, next word
	    next statement , next comment ... whatever is appropriate
	o   When you grow accustomed to this, it probably almost replaces your
	    single BACKSPACE and DEL keys.
	o   Can also eat only 'inside' words: WordsThatAreLikeThis"
    )

;;; ......................................................... &v-hooks ...

(defcustom tie-:load-hook '(tie-install-default-bindings)
  "*Hook that is run when package is loaded."
  :type 'hook :group 'TinyEat
  )

;;; ........................................................ &v-public ...
;;; User configurable

(defcustom tie-:verbose-flag t
  "*Non-nil means : Allow or disallow informational messages to be printed."
  :type 'boolean :group 'TinyEat
  )

(defcustom tie-:bind-install-type '(hp-alt hp-meta)
  "*How to bind the keys.
See source code of function `tie-install-default-bindings.'"
  :type '(choice
	  (const hp-alt)
	  (const hp-meta)
	  )
  :group 'TinyEat
  )


(defcustom tie-:non-word-chars "][-=_~+!@#$%&*:;\n \t'\"`,.<>(){}$<>?/|\\"
  "*Characters that _stop_ eating word.
Certain characters must be in this order, because this string is
converted into regexp later.

012345465..  <-- char positions in this variable.
][-          <-- chars that must be set to this order"
  :type '(string :tag "Charset")  :group 'TinyEat
  )

(defcustom tie-:eat-full-word-charset  "^][ \t\n(){};'\","
  "*Character set to use when determining word boundary.
Normally word is terminated by whitespace or newlines."
  :type '(string :tag "Charset")  :group 'TinyEat
  )


;;}}}
;;{{{ version

;;; ....................................................... &v-version ...

(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinyeat.el"
 "tie"
 tie-:version-id
 "$Id: tinyeat.el,v 1.37 1997/05/31 12:43:25 jaalto NotReleased $"
 '(tie-:version-id
   tie-:load-hook
   tie-:verbose-flag
   tie-:bind-install-type
   tie-:non-word-chars
   tie-:eat-full-word-charset
   )))

;;}}}

;;; ########################################################### &Funcs ###

;;{{{ install

;;; ......................................................... &install ...

(defun tie-install-default-bindings ()
  "Add default bindings."
  (interactive)
  (let* ((type tie-:bind-install-type)
         )
    (if (xemacs-p)
	(tinyck-advice-control))	;turn it on

    (global-set-key "\ek"	  'tie-kill-line-back)

    ;;  Hmm, howabout these...
    ;;
    (global-set-key [?\e backspace]	'erase-buffer)
    (global-set-key [?\e delete]	'erase-buffer)
    (global-set-key "\e\C-?"		'erase-buffer)


    ;; This is ESC + "delete"
    (global-set-key [?\e deletechar]	'tie-kill-line)


    ;; Control or Alt backspace: the "delete" is usually recognized
    ;; in PC as backspace.
    ;;
    ;; Usually M- is same as ALT key in HP-UX.
    ;;
    (global-set-key [C-backspace]	'tie-forward-preserve)
    (global-set-key [C-delete]		'tie-forward-preserve)

    (global-set-key [M-delete]		'tie-backward-preserve)
    (global-set-key [A-delete]		'tie-backward-preserve)
    (global-set-key [M-backspace]	'tie-backward-preserve)
    (global-set-key [A-backspace]	'tie-backward-preserve)

    (global-set-key [S-backspace]	'tie-delete-whole-word)
    (global-set-key [C-S-backspace]	'tie-delete-paragraph)


    ;; Clearing the input in minibuffer prompt
    ;;
    (define-key minibuffer-local-completion-map	[f1] 'tie-zap-line)
    (define-key minibuffer-local-map		[f1] 'tie-zap-line)
    (define-key minibuffer-local-completion-map	[f2] 'tie-kill-line-back)
    (define-key minibuffer-local-map		[f2] 'tie-kill-line-back)


    (when (memq 'hp-meta type)
      ;;  When the M- is same as ALT key in HP-UX.

      ;; Alt + Control + [Shift]
      ;;
      (global-set-key [C-M-backspace]   'tie-kill-buffer-lines)
      (global-set-key [C-M-S-backspace] 'tie-kill-buffer-lines-min)

      ;; Alt + Shift, the opposite of this is plain C-k :=)
      ;;
      (global-set-key [M-S-backspace]   'tie-kill-line-back)

      (global-set-key [M-mouse-2]	'tie-yank-overwrite)
      )

    (when (memq 'hp-alt type)
      ;;  Using this Alt-k, instead of C-k is much better, because
      ;;  it doesn't wipe out the cut buffer.
      ;;

      (tinyck-maybe-activate 'xemacs)
      (global-set-key [(alt k)]		'tie-kill-line)
      (global-set-key [(alt control k)]	'tie-zap-line)
      (tinyck-maybe-activate 'xemacs 'disable)

      ;; Alt + Control + [Shift]
      ;;
      (global-set-key [A-C-backspace]	'tie-kill-buffer-lines)
      (global-set-key [A-C-S-backspace]	'tie-kill-buffer-lines-min)


      ;; Alt + Shift, the opposite of this is plain C-k
      ;;
      (global-set-key [A-S-backspace]   'tie-kill-line-back)
      (global-set-key [A-mouse-2]	'tie-yank-overwrite)

      )

    (if (xemacs-p)
	(tinyck-advice-control 'disable)) ;don't leave conversion ON

    ))

;;}}}
;;{{{ misc

;;; ----------------------------------------------------------------------
;;; - Well, this is just a wrapper, I felt that all kill functions must
;;;   be named "tie"
;;;
;;;###autoload
(defun tie-zap-line ()
  "Kill full file"
  (interactive)
  (beginning-of-line)
  (ti::b-kill-line))			;works with any emacs


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tie-backward ()
  "Eats backward. See `tie-eat' function."
  (interactive)
  (tie-eat t))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tie-backward-preserve ()
  "Eats forward, but handles spaces differently. See `tie-eat' function."
  (interactive)
  (tie-eat t t))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tie-forward ()
  "Eats forward. See `tie-eat' function."
  (interactive)
  (tie-eat))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tie-forward-preserve ()
  "Eats forward, but hadles spaces differently. See EAT function."
  (interactive)
  (tie-eat nil t))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tie-join-lines()
  "Join this and next line with possible space, and go to the joint"
  (interactive)
  (end-of-line
  (if (eobp) nil
    (kill-line)
    (fixup-whitespace))))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tie-delete-whole-word  ()
  "Delete word at point. If sitting on whitespace, call `fixup-whitespace'.

References:
  `tie-:eat-full-word-charset'"
  (interactive)
  (let* ((set  tie-:eat-full-word-charset)
	 beg
	 end
         )
    (if (looking-at "[ \t\n]")
	(fixup-whitespace)
      (skip-chars-backward set) (setq beg (point))
      (skip-chars-forward  set) (setq end (point))
      (delete-region beg end)
      )))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tie-kill-line ()
  "Same as `kill-line', except the killed text isn't put into cut buffer.
This way you can retain mouse selection in cut buffer.
This only interests people who can use mouse."
  (interactive)
  (if (eobp)
      nil				;skip
    (cond
     ((eolp)
      (delete-char 1))
     (t
      (delete-region (point) (EOLP))
      ))
    ))



;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tie-kill-line-back ()
  "`kill-line' backward."
  (interactive)
  (cond
   ((and (not (bobp)) (bolp))
    (backward-delete-char 1)
    )
  (t
   (delete-region (point) (BOLP))
   ))
  )


;;; ----------------------------------------------------------------------
;;; Just to ease calling it from key
;;;
;;;###autoload
(defun tie-kill-buffer-lines-min (&optional back)
  "Kill untill `point-min'. optionally BACK."
  (interactive)
  (tie-kill-buffer-lines 'back))


;;; ----------------------------------------------------------------------
;;; very trivial, but I need this all the time
;;;
;;;###autoload
(defun tie-kill-buffer-lines (&optional back)
  "Kill to the `point-max' or BACK to the `point-min' with ARG."
  (interactive)
  (cond
   (back
    (delete-region (point) (point-min)))
  (t
   (delete-region (point) (point-max)))
  ))



;;}}}
;;{{{ misc2


;;; ----------------------------------------------------------------------
;;; This extremely handy deletion function, I kill C++ code with this
;;; all the time
;;;
;;;###autoload
(defun  tie-delete-paragraph ()
  "Deletes current paragraph, separated by empty lines."
  (interactive "*")
  (let* (beg
	 end
	 (re "^[ \t]*$")
	 )
    (save-excursion
      (if (null (re-search-backward re nil t)) nil
	(beginning-of-line)
	(forward-line 1)		;exlude space
	(setq beg (point))
	))
    (save-excursion
      (if (null (re-search-forward re nil t)) nil
	(beginning-of-line)
	(setq end (point))
	))

    (if (not (and beg end))
	(message "Can't find paragraph bounds.")
      (kill-region beg end))
    ))




;;; ----------------------------------------------------------------------
;;;
(defun tie-space-del-at-point (&optional back preserve)
  "Delete whitespace at point. Optionally BACK.
If optional PRESERVE is given, then deletes towards the BACK only.
if BACK is non-nil the deletion is headed backward."
  (interactive)
  (let* (
	 (charf	  (if back 'skip-chars-backward 'skip-chars-forward))
	 (p	  (point))
	 (verb	  tie-:verbose-flag)
	 (ch	  (ti::b-read-char back 0)) ;sitting on it if looking fwd
	 (ch-p	  (ti::b-read-char back -1))
	 (ch-n	  (ti::b-read-char back 1))
	 )
    (cond
     ((and back (space-p ch-p) (char= ch ?\n))
      (delete-horizontal-space)
      (if verb (message "*line cleared"))
      t)
     ((char= ch ?\n)		;no spaces before, do nothing
      nil)
     ((or (and (space-p ch)   (space-p ch-n)) ;at least two spaces
	  (and (space-p ch-p) (space-p ch)))
      (if (null preserve)
	  (fixup-whitespace)
	(funcall charf " \t")
	(delete-region p (point)))
      t)
     (t
      (delete-horizontal-space)
      t))
  ))



;;; ----------------------------------------------------------------------
;;;
(defun tie-word-move-point (&optional back)
  "Moves to suitable word kill point. Mixed case words are special.
Optionally BACK. See variable `tie-:non-word-chars' how to delimit word parts.

* = cursor position

ThisIsMixedWord --> ThisIsMixedWord
*		        *
THISmixedWord   --> THISmixedWord
*                       *"
  (interactive)
    (let* ((charf	(if back 'skip-chars-backward 'skip-chars-forward))
	   (non-word    tie-:non-word-chars)
	   (nonw-re     (concat "[" non-word "]+"))
	   (ch		(ti::b-read-char back))
	   p str
	   mb me			;match beg end
	   mixed
	   )
      ;;    Check if this is special mixedCase before vaporizing word...
      ;;
      (save-excursion
	(setq p (point))
	(if back
	    (backward-word 1)
	  (forward-word 1))
	(setq str (buffer-substring p (point)))
	(setq mixed (string-match-case "[A-Z][a-z]" str))
	)

      (cond
       (mixed
	(if (eq ch (downcase ch))
	    (funcall charf "a-z")
	  (setq p (point))

	  ;;  Skip all big letters
	  (funcall charf "A-Z")

	  ;;  If this was only one letter, continue deleting. Otw stay put.
	  (if (eq 1 (abs (- p (point))))
	      (funcall charf "a-z"))
	  )

	  ;;  The previous statements only moved 2 first statements
	  ;;	    ThisIsWord	    start,
	  ;;		     *
	  ;;	    ThisIsWord	    after,
	  ;;		   *
	  ;;	    ThisIsWord	    correction. This is needed
	  ;;		  *
	  ;;
	  (if (and back
		   (not (bobp)))
	      (backward-char 1))
	)

       (t

	;; if there is non-word we must remove it.
	;; - There is some problems in backward deltion, eg deleting "...."
	;;   backward in text-mode does not delete all dots. Don't
	;;   know why not.
	;;
	(cond
	 ((if back			;select FWD of BCK looking
	      (cond
	       ((string-match nonw-re (char-to-string ch))

		(re-search-backward nonw-re nil t)
		))
	    (looking-at nonw-re)
	    )

	  (setq mb (match-beginning 0)
		me (match-end 0))
;;;	  (setq N nonw-re MB mb ME me) (d! 1)

	  ;;  1. if there is multiple items like "....", delete only
	  ;;     those
	  ;;  2. if there is only one member like ".member", delete
	  ;;     dot and the word that follows it.
	  ;;

	  (if back (setq p mb) (setq p me)) ;selet direction

	  (if (not (eq 1 (abs (- me mb))))
	      (goto-char p)
	    (goto-char p)
	    (funcall charf (concat "^" non-word))
	    ))
	 (t
	  ;;  The skip-chars-forward _requires_ that the "-"
	  ;;  character is the first item. That's why we have
	  ;;  to add extra "-" to the front of string if user
	  ;;  has defined "-" to be word stopper.
	  ;;
	  (if (string-match-case "-" non-word)
	      (setq non-word (concat  "^-" non-word))
	    (setq non-word (concat "^" non-word)))

	  (funcall charf non-word)

	 ))

	))
      ))



;;}}}
;;{{{ Yanking


;;; ----------------------------------------------------------------------
;;; Don't you hate that the overwrite mode doeesn't support this !!!
;;;
(defun tie-yank-overwrite ()
  "Yanks text by overwriting previous contents."
  (interactive)
  (let* ((bp (use-tmp-buffer "*tmp*" 'clear)) ;we need work buffer first
	 (p  (point))			;insertion point
	 len
	 end
	 )
    (save-excursion
      (set-buffer bp)
      (yank)
      (setq len (1- (point-max)))	;how many chars in there ?
      )
    (cond
     ((= len 0)
      (message "Nothing to yank"))
     (t
      ;;   we must untabify  the line, otw we get unpleasant results
      ;;
      (untabify p (EOLP))
      (setq end (+ p len))
      (if (> end (point-max))
	  (setq end (point-max)))
      (delete-region p end)
      (yank)
      ))
    ))


;;}}}

;;{{{ engine

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tie-eat (&optional back space-preserve verb)
  "Eat *appropriate* text forward, if BACK then backward.

The optional SPACE-PRESERVE changes the space eating (VERB).

A.  when it is NIL and BACK is anything.   * marks the cursor.
         text1 text1        *     text2  text2
    -->  text1 text1 text2  text2                   ;one space left

B.  when it is NON-NIL and BACK nil
         text1 text1        *     text2  text2
    -->  text1 text1        *text2  text2            ;dels right spaces

C.  when it is NON-NIL and BACK t
         text1 text1        *     text2  text2
         text1 text1*     text2  text2               ;dels left spaces


References:
  `tie-:non-word-chars'"
  (interactive)
  (let (
	(p	    (point))
	(syntaxf    (if back 'skip-syntax-backward 'skip-syntax-forward))
	(charf	    (if back 'skip-chars-backward 'skip-chars-forward))
	(verb	    (or verb tie-:verbose-flag))
	ch ch-n
;;;	ch-nn ch-p
	D				; for debug only
	)
    ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. ..
    (setq ch    (ti::b-read-char back 0))       ;; sitting on it if looking fwd
;;;    (setq ch-p  (ti::b-read-char back -1))	;; previous
    (setq ch-n  (ti::b-read-char back 1))	;; next
;;;    (setq ch-nn (ti::b-read-char back 2))	;; following next

;;d     (d!
;;d      (char-to-string ch-p )
;;d      (char-to-string ch )
;;d      (char-to-string ch-n )
;;d      (char-to-string ch-nn))


    (cond
     ;; --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``--
     ;; BEG of buffer or END of buffer
     ((eq nil ch)
      (if verb
	  (message
	   (concat (if (bobp) "Beginning" "End") " of buffer")))
      )

     ;; --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``--
     ((space-p ch)			;one whitespace
      (strcat D  "SPACE")
      (tie-space-del-at-point back space-preserve)
      (if (and verb (looking-at "$"))		;it handled this
	  (message "Line cleared"))
      )

     ;; --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``--
     ;; - Multiple  newlines, squeeze to one only
     ((and (char= ch ?\n) (char= ch-n ?\n))
      (funcall charf "\n")
      (if (null back)
	  (backward-char 1)		;do not join, leave 1 EMPTY newline
	(forward-char 1))
      (strcat D  (prin1cat " " " multi n " p (point)))
      (delete-region p (point))
      )

     ;; --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``--
     ;; - at the end of line I suppose add previous line to it.
     ((char= ch ?\n)
      (strcat D  (prin1cat " " " 1 n " space-preserve))
      (if (tie-space-del-at-point back space-preserve)
	  nil
	(if (null back)			;which direction
	    (delete-char 1)
	  (if (not (eq 0 (funcall syntaxf  "_")))	;try to move
	      (delete-region p (point))	;moved!
	    (backward-char 1)
	    (delete-region p (point))))
	))

     ;; --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``-- --``--
     ;; WORD handling (blocks)
     (t					;eat next word
      (funcall syntaxf " ")		;ignore spaces
      (strcat
       D (prin1cat " " "t syntax " (char-to-string ch)
		   " " (char-to-string (char-syntax ch))))
      ;;   - What is next char after whitespace ??
      ;;   - With these following conditionals we set the point
      ;;     to appropriate position and after COND we run the kill command
      (cond
       ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. ..
       ((and (not  (char-in-list-case ch  '(?- ?_ ?:)))
	     (equal ?w (char-syntax ch)))
	(strcat D "-- 1")
	(tie-word-move-point back)
	)
       ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. ..
       ((and (char-in-list-case ch   '(?- ?_ ?:))
	     (memq (char-syntax ch-n)  '(?w ?\ ))
	     )
	(strcat D "-- 2")
	;;  This is really hard to understand... execpt for the author
	;;  1) Is CH variable start, a delimiter ?
	;;  2) AND is the NEXT-CH word or whitespace

	(funcall charf "-_:")

	;; (funcall syntaxf  "_w")
	;; (funcall syntaxf  " w")
;;;	(setq CH ch CH-N ch-n)  (d! "A sym")
	)
       ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. ..
       (t
	;; punctuation, comment, the rest ... skip non important stuff
	;;
       	(strcat D "-- other")
	(funcall charf "^ \t\na-zA-Z0-9")
	))
      (delete-region p (point))
      ))
;;;    (d! D)
    ))

;;}}}


(provide   'tinyeat)
(run-hooks 'tie-:load-hook)

;;; tinyeat.el ends here
