;; @(#) tinyezip.el -- 'load' support for compressed elisp files.

;; @(#) $Id: tinyezip.el,v 1.43 1997/07/04 18:00:32 jaalto NotReleased $
;; @(#) $Keywords: compression, file, load $
;;
;; $KnownCompatibility: 19.28, XEmacs19.14 $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1995,96,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Sep 7 1995
;;
;; To get information on this program use ident(1) or do M-x tinyezip-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el

;; LCD Archive Entry:
;; tinyezip|Jari Aalto|jari.aalto@poboxes.com|
;; Support for compressed elisp files: require, load, load-library,load-file..|
;; 04-Jul-1997|$Revision: 1.43 $|~/misc/tinyezip.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ....................................................... &t-install ...
;; Put this file on your Emacs-Lisp load path, add following into your
;; ~/.emacs startup file. *THIS MUST BE THE FIRST ENTRY* before any other
;; 'load' or 'require' command. Otherwise compressed elisp
;; files are not recognized.
;;
;;	(load "~/elisp/tinyezip.el")
;;
;; or
;;	(require 'tinyezip)
;;
;; and
;;	(global-set-key "\C-cZ" 'tinyezip-pre-command-install)
;;
;; It is *highly* recommended that you bind this command to some key,
;; because the handler drops off from pre-command-hook quite easily. Eg
;; C-g clears these handlers. This ensures that Keyboard bound commands,
;; that are autoloaded, will succeed from compressed files.
;;
;; To disable/enable totally this package, call
;;
;;	tinyezip-advice-control
;;
;; If you have any questions, use this function
;;
;;      M-x tinyezip-submit-feedback
;;
;; Autoload note: functions can't be set to autoload state in lisp code
;; Warning	: See limitations section.
;; Note		: See example setup at the end of file
;; Tip          : If you want to have this file compressed too, see
;;                example setting at the end.

;;}}}

;;{{{ Documentation

;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - File's name means (tiny) (e)lisp (zip) handling, where "zipping"
;;   is commonly used synonym for compression.


;;  Preface, sep 1995
;;
;;	I got tired of watching my University's quota every time I found
;;	some useful elisp code I could stuff into my ~/elisp directory. I
;;	thought that "I must have this and that .el, because it's so
;;	useful..". Sooner or later I have to raise my hands and give
;;	up. My disk quota didn't allow any more lisp code, besides every
;;	time I received mail I got announcement:
;;
;;	    "Quota exceeded, you have 7 days to remove excess files."
;;
;;	Annoying You bet. After a while I came up with the idea of
;;	configuring basic emacs commands so that they could also load
;;	lisp packages that were in compressed form. Basicly the
;;	designing seemed simple, the commands should check for
;;	compressed files and load it, instead of the raw .el or .elc
;;	file. But there were a catch, jka-compr knew only files that
;;	were loaded into emacs for editing, not about lisp files that
;;	should be just 'evaled'. So, there I was, hacking up the basic
;;	lisp commands with defadvice...
;;
;;  Overview of features
;;
;;	o   Allows commands LOAD, LOAD-LIBRARY, LOAD-FILE, REQUIRE and AUTOLOAD
;;	    to accept jka-compr supported lisp '.el' files that are
;;	    compressed. Note: you can't compress .elc files and have
;;	    them loaded, only .el files.
;;
;;	o   As a special feature: you can change the actual "filename"
;;	    (with hook) before it is loaded by emacs. Eg. to have same
;;	    load-commands, but different files for different
;;	    users/goups/sites
;;
;;	o   Correctly handles aliased commands that turn out to be
;;	    in 'autoload' state.
;;
;;	o   Now you can compress all your lisp library files in your ~/elisp
;;	    dir and save space eg. for quota limit reasons, while still being
;;	    able to load them as usual.
;;
;;	o   You don't have to change anything in your .emacs startup file, all
;;	    will work as usual. (Note: see Limitations. You cannot
;;	    use autoloaded functions inside lisp code without preparations.)
;;
;; Important
;;
;;	You must have newer advice.el, than what 19.28 version ships with,
;;	because it has some minor problem that was discovered during
;;	development of this package. The 19.28 advice.el's version is 2.14
;;	and 2.15 version cures any problems. Mail Hans Chalupsky
;;	<hans@cs.buffalo.edu> directly to get the new advice.el
;;
;;  Jka-compr and this package
;;
;;	jka-compr has native support to un/compress any file that has
;;	specific extension. The handling is done via
;;	file-name-handler-alist and commands like these will load
;;	properly, including any autoloads.
;;
;;	    (load "my-el.gz")
;;
;;	The obvious problem is that you have to manually go and change
;;	all you load commands so that they end in .gz so that
;;	jka-compr takes care of loading. What if you later
;;	uncompress the file? You have to go and update all the load
;;	commands in you .emacs.XXX files. This isn't very nice,
;;	since you should be able to un/compress elisp files
;;	whenever you wish and still have permanent statement
;;
;;	    (load "my-el")
;;
;;	Basicly this is what this package is all about; you don't have to
;;	worry if the file is compressed or not when you use these
;;	advised functions. Besides any 'require' command in any
;;	lisp module work too for compressed files.
;;
;;  Using pre-command-hook
;;
;;	Every time programmer puts hook into [post|pre]-command-hook,
;;	emacs slows down a bit, because it increases overhead. For
;;	me, the pre-command-hook seememd to be last resort to
;;	intercept keyboard mapped autoload command, before it was
;;	handled by emacs. I tried to use smart check, which would
;;	not spend excess time in the hook and while I have loaded
;;	55 lisp packages, I can't tell if emacs slowed down or not.
;;
;;  How this package works
;;
;;	o   When user request 'load FILE', try to find some compressed file
;;	    that JKA knows about by adding extensions ".gz" and ".Z" and
;;	    whatever user has configured JKA to handle.  ** LIMITATION:
;;	    only .gz and .Z and the like that compress one file at a time
;;	    is currently supported. Don't try using .zip or similar.
;;	o   If the FILE contain absolute directory, then look from that
;;	    directory only
;;	o   If no directory is given, find the file along the path.
;;	o   If there was somewhere a compressed file, just load it (because JKA
;;	    will transparently uncompress it), eval it, and kill the buffer
;;	    where file was some fract of second for evaling.
;;	o   and if NO COMPRESSED file was found, just follow normal
;;	    emacs rules.
;;
;;  Load failed -- is there a bug in this package?
;;
;;	If you have had habit of keeping debug-on-error t, you'll see
;;	a little different error messages now, when the basic lisp
;;	load functions are advised. If the last messagege includes
;;	prefix "ad-Orig", say the 'load' command failed:
;;
;;	    ad-Orig-load("myfile") Failed....
;;
;;	It means that the original lisp 'load' function couldn't
;;	locate the file you tried to visit. Messages like these are
;;	not *bugs* of this package. Bugs related to this package
;;	come in picture normally only if you have lisp files in
;;	_compressed_ formats and when it fails to find them. Make
;;	sure that:
;;
;;	o   Your load-path contains the directory the file is loacated,
;;	    if you gave load command without the directory part.
;;	o   Directory name is spelled right in the command
;;	o   pre-command-hook handler must be installed !!
;;	    It may drop off due to various reasons, like pressig C-g too hard.
;;	    This is the most common reason for failing to load, a
;;	    ,typical error shown is:
;;
;;	    Signalling: (file-error "Cannot open load file"myfile.el")
;;
;;	o   You have configured jka-compr-compression-info-list properly
;;	    so that this program can decode the info right. Try
;;	    running (ti::a-jka-compress-ext ) by hand and it should return
;;	    valid identifiers like (".gz" ".Z"). If you see anything like
;;	    REGEXP marks in the returned list, then alarms should ring and
;;	    you should contact me immediately.
;;
;;  In the end
;;
;;	If you notice that your fat ~/elisp has loosed some weight
;;	lately, drop me mail and tell me this package made it happen.
;;	Of course your emacs now starts ever _slower_ that it used to,
;;	but you also gained much more space. Try to optimize everything
;;	so that they use 'autoload' as much as possible and you're much
;;	happier with the startup time.
;;
;;
;;  Note: Is this file slow to load or what?
;;
;;	Yes, I know. It's that advice thing that does it. Never mind
;;	about the load time. You want to save space, not time, okay?
;;
;;  Note: Why can't I have .elc files in compressed form?
;;
;;	This package' purpose was clearly to reduce the sizes of
;;	lisp files, so that you can fit more of them into your account's
;;	directory. If you dwonload a file, it is in .el form and
;;      compressing such text file saves quite a much disk space.
;;
;;	Wanting to have .elc files shows, that you don't have quota problems,
;;	because you can have those bulky .elc files. And you certainly
;;	don't need this package at all, so that's why .elc files are not
;;	supported.


;;}}}
;;{{{ Bugs & limitations and other notes

;; .............................................. &t-bugs-limitations ...
;;
;;  Overloaded emacs commands
;;
;;	If any of the STD emacs functions handled here get overloaded,
;;	ie. replaced with smeone else's definition of that function,
;;	it'll break this package. Currently the only function handled
;;	specially is the following which can be safely redefined, as
;;	is done in execcmd.el (19.28+ keyboard keys command reminder)
;;
;;	    execute-extended-command
;;
;;	The problem may not become visible, if the redefined function
;;	doesn't use following function in it's body:
;;
;;	    (interactive-p)
;;
;;	The reason why you can't redefine those emacs function that are
;;	adviced in this packe is, thet advice's ad-do-it can't pass the
;;	interactive flag information to the original functions.
;;
;;  Compressed files called from lisp functions.
;;
;;	It's not possible handle 'eval' command. This means that
;;	you cannot do
;;
;;          (autoload 'function-call "functions" t t)
;;	    (function-call)C-xC-e
;;
;;	or execute behind any other arbitrary lisp expression that might
;;	include function call to autoloadable function that is stored in
;;	compressed package.. So, in general, if you use command that you KNOW
;;	is autoloaded, use appropriate 'require package' before you're going
;;	to use that command.
;;
;;	Ie. *this is bad thing*
;;
;;	    (autoload  'some-minor-rmode "some-minor-mode"  t t)
;;	    (add-hook  'mail-setup-hook  'my-mail-setup-hook)
;;
;;	    (defun my-mail-setup-hook ()
;;	      "My mail preparations."
;;	      ;;  I want to use this handy minor mode while I edit mail
;;	      (some-minor-rmode t)
;;
;;	What happens is, that lisp tries to execute expression, where
;;	some-minor-mode is used and which is declared as autoload and then
;;	it goes and tries to find associated file to load. But because
;;	this all is handled by lisp itself, there is no tinyezip handler to
;;	take care of detecting possible compressed some-minor-rmode.el.gz
;;	--> Your code breaks, and lisp barfs.
;;
;;	The solution is to use appropriate (require 'some-minor-mode)
;;	before using the command in lisp. This is the right thing to
;;	do if you use this package
;;
;;	    (defun my-mail-setup-hook ()
;;	      "My mail preparations."
;;	      ;;  I want to use this handy minor mode while I edit mail
;;	      (require 'some-minor-mode)    ;; << this is needed !
;;	      (some-minor-rmode t)
;;	      )
;;
;;  Adviced hook functions
;;
;;	Currently only run-hooks command is adviced and it can
;;	detect functions that come from zipped files. Other hook run
;;	commands are not recognized.
;;
;;  XEmacs note
;;
;;      The performance of the adviced function is greatly reduced compared
;;	to the Emacs installation, because XEmacs has different source
;;	tree structure. Packages are spread over multiple dirs in
;;	lisp source tree:
;;
;;	% ls /opt/local/lib/xemacs-19.14/lisp
;;	README          emulators       ilisp           pcl-cvs    version.el
;;	bytecomp        energize        iso             prim       viper
;;	calendar        eos             mailcrypt       psgml      vm
;;	cl              epoch           mh-e            rmail      vms
;;	comint          eterm           modes           sunpro     w3
;;	dired           games           ns              term       x11
;;	edebug          gnus            oobr            tooltalk
;;	ediff           hm--html-menus  packages        url
;;	electric        hyperbole       paths.el        utils
;;
;;
;;	In order to find some package, eg. "paren", the program has to look
;;	into _every_ directory and see if there is "paren.el.gz" or
;;	"paren.el" or "paren.el.Z"
;;
;;	--> It takes 5-10X more times than in Emacs where all the lisp code
;;	are placed in one directory. It would be better if you'd
;;	make a directory which would hold symlinks to all other files
;;	and then include only that to your load-path. This would make
;;	XEmacs's package loading lot faster.
;;
;;	I have optimized the code so that the most common extension ".el"
;;	is searched first and then plain ".gz" and possible other extensions
;;	after that.
;;
;;	If someone has some other good suggestion to improve speed of loading
;;	please mail me, I have run out of ideas to help XEmacs people. I
;;	could cache the compressed files, but what if new package is
;;	intalled...no, I don't believe that cache would be suitable.


;;}}}
;;{{{ history

;; ........................................................ &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; Jul	4	1997	[jari]		19.28	v1.43     	NotReleased
;; - Adviced `load' command didn't retutn t on success. Corrected.
;; - checkdoc 1.42 cleaar.
;;
;; Jun	19	1997	[jari]		19.28	v1.41-1.42     	NotReleased
;; - Internal changes due to library functions.
;;
;; Feb	19	1997	[jari]		19.28	v1.39-1.40     	NotReleased
;; - Added more documentation and slightly changed the zip-elisp-p
;;  function. It puts first tries to find
;;
;;	FILE.el.gz and FILE.el.Z
;;
;;  and onle secondly FILE.gz FILE.Z.
;; - Updated new function to the end offile: example section(compress
;;   household changed)
;;
;; Jan	15	1997	[jari]		19.28	v1.34-1.38     	NotReleased
;; - Noticed that M-x didn't trigger autoloading. That was because
;;   I didn't redefine execute-extended-command, but used execcmd.el's
;;   definition. Now corrected.
;; - Corrected error in finding file: now the suffixes tried are
;;   '(".el.gz" ".el.Z" ".gz" ".Z"). The last two were missing.
;;
;; Nov	21	1996	[jari]		19.28	v1.32-1.33     	NotReleased
;; - The tinyezip-file-eval illed the buffer inside save-excursion,
;;   which was wrong. This caused error in 19.34.
;;
;; Oct	1	1996	[jari]		19.28	v1.30-1.31     	NotReleased
;; - Removed extra 'require tinylib' commands.
;; - Changed all '(lambda forms to use (function (lambda
;;
;; Sep	6	1996	[jari]		19.28	v1.29     	NotReleased
;; - Corrected function tinyezip-file-eval, which didn't generate
;;   the eval buffer right. And it forgot to clear it before using it.
;;
;; Sep	4	1996	[jari]		19.28	v1.28     	NotReleased
;; - Small error in tinyezip-zip-elisp-p corrected. Affected the
;;   finding file along the path.
;;
;; Aug	23	1996	[jari]		19.28	v1.27     	NotReleased
;; - Cleared all XEmacs 19.14 Bytecompiler errors. Optimized code especially
;;   XEmacs tree structure in mind. Corrected bug in jka-compr extension
;;   handling (was XEmacs issue only).
;;
;; Jun	28	1996	[jari]		19.28	v1.25-1.26     	NotReleased
;; - ByteComp errors corrected (XEmacs)
;; - Disabled run-hooks advice in XEmacs.
;;
;; Jun	28	1996	[jari]		19.28	v1.24		NotReleased
;; - Rerwrote tinyezip-zip-elisp-p, now it's much mopre cleaner and
;;   robust. Serious error found from tinyezip-ti::f-newer-exist.
;;
;; Jun	28	1996	[jari]		19.28	v1.23		NotReleased
;; - I just started using my university account again for serious lisp and
;;   found immediately some errors to be corrected.
;; - There were several bugs concerning the filename detection. Eg
;;   the .el extension was always tried, and this caused failing
;;   on files like "~/.emacs.init" (Because it tried to
;;   find "~/.emacs.init.el.gz")
;; - Added advice for run-hooks, because there may be autoload function
;;   that are loaded from compressed packages (I have them in Unbiversity)
;; - Renamed some library function left here to have tinyezip- prefix.
;; - Added renewed my-file-compress example.
;;
;; May	10	1996	[jari]		19.28	v1.22		NotReleased
;; - Renamed lot of functions to have "tinyezip" prefix.
;;
;; Feb	10	1996	[jari]		19.28	v1.21		NotReleased
;; - Used lisp-mnt's lm-verify to check layout. Rewrote feedback funcs.
;; - converted many functionames to use tinyezip prefix. This is important
;;   because this is special modue and I can't use tinylibXX funcs
;;   via 'require
;;
;; Feb	1	1996	[jari]		19.28	v1.20		NotReleased
;; - All library code renamed to have prefix of this modeule. This
;;   were done to avoid possible overloading problems when libraries
;;   are released.
;;
;; Oct	31	1995	[jari]		19.28	v1.19		NotReleased
;; - Added example1, which shows how to keep this file compressed too,
;;   and how to load it in startup phase. Saves space even more...
;; - When keeping different configuration files for emacs in different
;;   sites, I needed some more intelligency. Now the filename beeing
;;   loaded can be changed through hooks. This makes it possible to reflect
;;   the name according to site/user while still using exactly same "load"
;;   commands.
;;   tinyezip-filename-hook	:+ hook
;;   tinyezip-:require-file	:+ var
;;   tinyezip-:require-flag	:+ var
;;   tinyezip-filename-hook-init:+ func
;;   tinyezip-filename-get	:+ func
;;   ...xxx-load		:! changed every load command to call hook
;;
;; Sep	14	1995	[jari]		19.28	v1.18		Release_1&2
;; - Serious error in ti::a-zip-elisp-p corrected. When I added the
;;   "newer" file test, this function changed a lot and many variables were
;;   pointing to wrong files --> almost every time returned nil.
;; - Okay, this is first official release, because seems quite stable now.
;;
;; Sep	14	1995	[jari]		19.28	v1.17		NotReleased
;; - Mistakenly left out some Lib functions I needed in this new versions.
;; - + ti::f-newer-exist , ti::s-index
;;
;; Sep	14	1995	[jari]		19.28	v1.16		NotReleased
;; - Commented out previous code and started from 1.14 base.
;; - Now it can detect which file is newer [.el .el.gz] and load it instead of
;;   always loading ".gz" if one existed.
;; - Added example section to the end of file: hoiw can you accoplish
;;   automatic compression houshold everytime you start emacs.
;;
;; Sep	13	1995	[jari]		19.28	v1.15		NotReleased
;; - Experimental. Not working. Tried to use Hans suggestion for
;;   tinyezip-ad-do-it.
;;
;; Sep	11	1995	[jari]		19.28	v1.14		NotReleased
;; - Added documentation for " Error messages" doc section. Added
;;   keybinging install notice for keeping pre-hook installed.
;; - Note "Copessed takes precedence" --> put to "to do" list
;; - Added ro limitations "LISP SOURCE CODE" section.
;; - The load-library's interactive form was errorneous. Corrected.
;;
;;
;; Sep	10	1995	[jari]		19.28	v1.13		NotReleased
;; - Added more text to the limitations section, this time it involves
;;   writing lisp code.
;;
;; Sep	10	1995	[jari]		19.28	v1.12		NotReleased
;; - put 'compile' flag to execute-extended-command advice, because it
;;   included special ad-do-it macro, it woun't work without it.
;; - 'require' command used  mistakenly tinyezip-ad-do-it when it should
;;   have used regular ad-do-it.
;;
;; Sep	10	1995	[jari]		19.28	v1.12		NotReleased
;; - The tinyezip-file-eval wasn't smart enough: it let my folding.el to fold the
;;   file before emacs could eval it --> emacs didn't see a thing about
;;   the functions in the file! Now it will reset some emacs variables
;;   so that no-one can't take control while it loads the file into buffer.
;;
;; Sep	10	1995	[jari]		19.28	v1.11		NotReleased
;; - After tickling with the 'around' advises I found a way how
;;   to pass the interactive definition to original function. Now
;;   when I need special call to access the interactive forms of
;;   the original function, I use tinyezip-ad-do-it instead of
;;   ad-do-it. It will pass the interactive flag property to the
;;   original function too.
;; - Added limitations documentation.
;;
;; Sep	10	1995	[jari]		19.28	v1.10		NotReleased
;; - Removed advise definition for call-interactively due to problems.
;;   At the same time, found out that it was not needed, because
;;   PRE hook would take over it and detect the command prior it is
;;   passed to call-interactively.
;; - Error in function ti::a-zip-elisp-p corrected, when passed argument
;;   with no dir info --> couldn't find file.
;;
;; Sep	9	1995	[jari]		19.28	v1.9		NotReleased
;; - Function ti::a-jka-compress-ext returned "\\.gz" when it
;;   should have returned only ".gz" ie. without any regexp marks.
;; - Structure of ti::a-zip-elisp-p changed. Now it tries FILE.el and
;;   FILE.el.gz universally, no matter what the initial input was
;; - Due to these errors the package choke, when (require 'my "~/my/my.el")
;;   was tried.
;; - Hm, it seems that advice.el can't handle advising function
;;   call-interactively, because the ad-do-it does not pass the
;;   (interactive-p) to the function, thus efectively breaking the
;;   purpose of call-interactively.
;;
;;   **  You should not advice functions **
;;       o  call-interactively
;;       o  interactive-p
;;   because it'll break all functions that make interactive tests.
;;
;; - Corrected several functions, because there was no proper ad-do-it
;;   adter the special handling: call-interactively,
;;   execute-extended-command, command-execute
;;
;; Sep	9	1995	[jari]		19.28	v1.8		NotReleased
;; - Now when I put the tinyezip-defalias-p and autoload-p tests inside pre-func
;;   I made a substle mistake. I took (and (sequencep sym).. when
;;   I should have taken  (and (sequencep (symbol-function sym)) ..
;;   This prevented any autoload checkings and my handler wasn't called.
;; - typo in func name: tinyezip-post-.. --> tinyezip-pre.. in the install
;;   function.
;; - Now I'm going to run few test with this before I make it next aplha.
;;
;; Sep	9	1995	[jari]		19.28	v1.7		NotReleased
;; - I noticed several flaws. The autoload-p tests should return name of the
;;   real function symbol if it was aliases, not just boolean t -->
;;   corrects execute-extended-command, call-interactively, command-execute
;; - Inserted tinyezip-defalias-p and autoload-p tests directly into
;;   pre-command-hook, so that it would be even faster.
;; - I didn't notice that 'require' command might be passed optional filename
;;   parameter. Now it's been taken cared of too. Does anyone use the
;;   optional parameter? At least I haven't seen one around.
;; - tinyezip-pre-command-install now has optional argument that removes
;;   the handler
;;
;; Sep	8	1995	[jari]		19.28	v1.6		Release_1
;; - Used fastest test I could imagine in tinyezip-pre-command, so that
;;   opverhead is reduced to minimum. If anybody have more faster test, I'd
;;   be happy to use it instead.
;; - Added 'pre-command-hook' documentation. Added back
;;   'execute-extended-command' advise.
;; - This seems to run as advertised, but wait when the
;;   bugs crawl out...
;;
;; Sep	8	1995	[jari]		19.28	v1.5		NotReleased
;; - The last problem is solved. The pre-command-hook can intercept
;;   any command prior it's execution, so now hitting a key that calls
;;   autoloadable function succeeds by loading definition from zipped file
;;   too. Boy, I'll go and pour some Cognac for that..!
;;
;; Sep	7-8	1995	[jari]		19.28	v1.4		NotReleased
;; - Sep 7. The autoload works only if user calls function via M-x
;;   FUNC, But I don't understand what happens when commands executes
;;   directly eg. through a key ? Autoload fails then... *hmpf* I'm going
;;   bed now.
;; - Next day: Okay, now it failed to detect autoload form when the function
;;   was aliased. Corrected the autoload-p test to handle aliased names.
;; - I added advice for the call-interactively, but haven't still found
;;   the way to handle direct calls to autoloaded files (direct kbd cmd)
;;
;; Sep	7	1995	[jari]		19.28	v1.3		NotReleased
;; - It's 01:00am and night is still darkening, should go to bed soon, since
;;   I have to get to work at 06:45am .. Now I noticed that the damn
;;   autoload is tricky guy. Spent tracing the triger point long time and
;;   finally found it from execute-extended-command. Now it should
;;   work.
;;
;; Sep	7	1995	[jari]		19.28	v1.2		NotReleased
;; - Indeed it was beta. The 'require' advice missing altogether.
;;   Also changed code so that when absolute path name is
;;   given, the .el is tried there too.
;;
;; Sep	7	1995	[jari]		19.28	v1.1		NotReleased
;; - This is it! Running and working fine.
;; - Send this as "Beta" to gnu.emacs.sources group.

;; To do list:
;; ========================================
;; - If there is file.el and file.el.gz, it will load "file.el"
;;   with its own tinyezip-file-eval. For speed it maybe should use std
;;   lisp version... marginal effect, so maybe I don't fix this one.

;; Sincere thanks to:
;; ========================================
;; Hans Chalupsky <hans@cs.buffalo.edu> -- author of Advice.el
;; - He patiently responded to my questions about proper advising and
;;   solving some of my problems. Thanks!

;;}}}


;;; Code:

;;{{{ require

;;; ...................................................... &v-tinyezip ...

(require 'cl)
(require 'advice)
(require 'jka-compr)
(require 'backquote)


(eval-and-compile
  (autoload 'ti::m-pkg-version-info    "tinylib")
  (autoload 'ti::m-pkg-submit-feedback "tinylib")
  )


;;}}}
;;{{{ variables

;;; ......................................................... &v-hooks ...

(defvar tinyezip-load-hook '(tinyezip-advice-control)
  "*Run when file has been loaded.")


(defvar tinyezip-filename-hook nil
  "*List of functions or single function.
Functions are called with one parameter: the filename or
required feature name given in the actual load command.

For example you have identical .emacs.keyup files in _two_
different sites. But to manage your projects better, you have
configured some lisp modules so that they have site
extension at the end

	site1			site2
	.emacs.key		.emacs.key
				.emacs.key.site2

Now suppose you're managing the files so that you have
common .emacs which has generic load commands:

	(load \".emacs.key\")

And this .emacs is common to both files. You normally just
copy all the .emacs configuration files from site 1 to site
2. For special purposes you rename some of the config files
to suit site2 better, so you actually want to load file


	.emacs.key.site2

When the .emacs calls command

	(load \".emacs.key\")

in site2. You could make separate load files and separate
config files to all sites, so that your load commands
reflect right files, but that wouldn't be very smart.

Instead you want to have _mother_ files, and with a little
system checking you change the required .emacs.xxx files a
bit to suit other site's needs (suppose you're running HP
and SUN in different machines, or different emacs version.)

Now, with this hook you can do quick check to look around:
which system, or which user; whatever, and *change* the
filename to reflect right target.

If the hook is called from 'require' command, then the
parameter is the first parameter of the command, typically a
FEATURE name. Some variables are also set in this case.


Action:
  If any hook returns string, it will be used as if it file were beeing
  loaded.

References:
  tinyezip-:require-file
  tinyezip-:require-flag

Notes:
  If load command is called interactively, wehn the `interactive-p'
  is satisfied, the hooks are *not* called. User must have the
  ability to totally control the load file name.

  Carefull! Remember that jka-compr, ange-ftp etc. uses load
  commands. Do regexp match against the file name before you go and change
  the name in this hook.")







;;; ... user configurable .................................. &v-config ...


;;; ... private variables ................................. &v-private ...


(defvar tinyezip-:require-file nil
  "FILENAME parameter of 'require' command. See tinyezip-filename-hook.")


(defvar tinyezip-:require-flag nil
  "Non-nil means that 'require' command was used.
See tinyezip-filename-hook")


(defvar tinyezip-:jka-compress-ext  nil
  "File name extensions known.")


;;}}}
;;{{{ version

;;; ... version info ...................................... &v-version ...

(defconst tinyezip-version-id
  "$Id: tinyezip.el,v 1.43 1997/07/04 18:00:32 jaalto NotReleased $"
  "Latest modification time and version number.")


;;; ----------------------------------------------------------------------
;;;
(defun tinyezip-version (&optional arg)
  "Version information. With ARG; show brief info."
  (interactive)
  (ti::m-pkg-version-info "tinyezip.el" arg))

;;; ----------------------------------------------------------------------
;;;
(defun tinyezip-submit-feedback ()
  "Submit suggestions, error corrections, impressions, anything..."
  (interactive)
  (ti::m-pkg-submit-feedback "tinyezip.el"))





;;}}}

;;; ########################################################### &Funcs ###

;;{{{ Install/uninstall






;;; ----------------------------------------------------------------------
;;;
(defun tinyezip-pre-command-install (&optional remove)
  "Install packages `pre-command-hook' again.
If it happens that packages pre command hook drops away,
due to setting the `pre-command-hook' to nil, you can enable
it with this function. It'll ensure you can load definition from
zipped elisp files.

In emergencies, you can supply optional prefix arg REMOVE to remove this
handler."
  (interactive "P")
  (if remove
      (remove-hook 'pre-command-hook 'tinyezip-pre-command)
    (add-hook 'pre-command-hook 'tinyezip-pre-command)))

;;; ----------------------------------------------------------------------
;;;
(defun tinyezip-advice-control (&optional disable verb)
  "Advice or unadvice with prefix arg DISABLE all package's functions. VERB."
  (interactive "P")
  (cond
   (disable
    (ad-disable-advice 'load		'around 'tinyezip-zip-load)
    (ad-disable-advice 'require		'around 'tinyezip-zip-load)
    (ad-disable-advice 'load-file	'around 'tinyezip-zip-load)
    (ad-disable-advice 'run-hooks	'around 'tinyezip-zip-load)
    (ad-disable-advice 'load-library	'around 'tinyezip-zip-load)

    ;;  The "autoload" handler
    (ad-disable-advice 'command-execute 'around 'tinyezip-zip-load)
    (ad-disable-advice 'execute-extended-command  'around 'tinyezip-zip-load)
    (ad-disable-advice 'command-execute 'around 'tinyezip-zip-load)

    (tinyezip-pre-command-install 'remove)
    )
   (t
    (ad-enable-advice 'load		'around 'tinyezip-zip-load)
    (ad-enable-advice 'require		'around 'tinyezip-zip-load)
    (ad-enable-advice 'load-file	'around 'tinyezip-zip-load)
    (ad-enable-advice 'run-hooks	'around 'tinyezip-zip-load)
    (ad-enable-advice 'load-library	'around 'tinyezip-zip-load)
    (ad-enable-advice 'command-execute  'around 'tinyezip-zip-load)
    (ad-enable-advice 'execute-extended-command  'around 'tinyezip-zip-load)

    (tinyezip-pre-command-install)
    ))

  (ad-activate 'load)
  (ad-activate 'require)
  (ad-activate 'load-file)
  (ad-activate 'load-library)


  ;;  Do not enable this, problems in XEmacs...

  (if (not (string-match "XEmacs" (emacs-version)))
    (ad-disable-advice 'run-hooks	'around 'tinyezip-zip-load))
  (ad-activate 'run-hooks)

  (ad-activate 'command-execute)
  (ad-activate 'execute-extended-command)
  )

;;}}}

;;{{{ lib

;;; ...................................................... &lib-macros ...
;;; ** These functions are 'preloaded' in this section, so that
;;;    we don't have to 'require' anything, thus making this file
;;;    self standing .
;;; ** Do not copy these, they are in tinylibXX.el distribution



;;; ----------------------------------------------------------------------
;;;
(defmacro tinyezip-sfuncall (func-sym &rest args)
  "Like funcall, but use `symbol-function' to get the function definition.
Equivalent to lisp call (FUNC-SYM ARGS). This shuts up some whinings of
byte compiler."
  (`
   (apply (symbol-function (, func-sym)) (,@ args) nil)
   ))



;;; ----------------------------------------------------------------------
;;; - The problem with "loading into emacs" is that all kinds of hooks
;;;   are run, eg. folding and outline might get activated when the file is
;;;   loaded --> 'eval' can't see the functions if they are
;;;   behind selective display.
;;;
(defun tinyezip-file-eval (file)
  "This works like `load-file', but we physically load the FILE.
Load into emacs and only use `eval-current-buffer'.

The advantage over `load-file' is that physical loading also uncompresses
the file if there is proper elisp package to handle it, thus your elisp
can be in any file *form* that packages allow for loading."
  (let* (
	 ;;   This makes sure we truly load the file.
	 ;;   If there were that file in emacs, emacs won't load it.
	 (bp  (generate-new-buffer "*tinyezip-tmp*"))

	 ;;   Phrohibit emacs from doing anything fancy while
	 ;;   we load it to buffer
	 (enable-local-eval  nil)
	 (find-file-hooks    nil)        ;; jka doesn't use this
	 )
    (unwind-protect
	(save-excursion
	  (set-buffer bp)
	  (setq buffer-read-only nil)
	  (erase-buffer)
	  (insert-file-contents file)

	  (cond
	   ((fboundp 'eval-current-buffer)
	    (tinyezip-sfuncall 'eval-current-buffer))

	   ((fboundp 'eval-buffer)
	    (tinyezip-sfuncall 'eval-buffer) ;XEmacs 19.14 compatible.
	    )))
      (if (get-buffer bp)		;Make sure buffer is deleted.
	  (kill-buffer bp))
      )))

;;; ----------------------------------------------------------------------
;;;
(defmacro tinyezip-func-car-test (symbol test-val)
  "Test first value with of `symbol-function' SYMBOL with TEST-VAL.
Function must be symbol, not in lambda form."
  (`(if (and (not (sequencep (, symbol))) ;; list ?
	     (symbolp (, symbol))	  ;; chokes if sequencep
	     (fboundp (, symbol))
	     ;;  Eg. symbol-function 'car  doesn't return list.
	     (listp (symbol-function (, symbol)))
	     (eq (, test-val) (car (symbol-function (, symbol)))))
	(, symbol)			;return the function
      nil
      )))


;;; ----------------------------------------------------------------------
;;;
(defun tinyezip-defalias-p (symbol)
  "If function SYMBOL is alias, return it's truename. Otw Return nil.
Input can be anything."
  (let* (sym
	 prev
	 ret
	 )
    (if (or (sequencep symbol)		;lambda form ?
	    (not (symbolp symbol))
	    (not (fboundp symbol)))
	nil
      (setq sym (symbol-function symbol))
      (if (not (symbolp sym))
	  nil
	(while (symbolp sym)		;was alias, go into nesting levels
	  (setq prev sym)
	  (setq sym (symbol-function sym))
	  )
	(setq ret prev)
	))
    ret
    ))



;;; ----------------------------------------------------------------------
;;;
(defmacro tinyezip-autoload-p (symbol)
  "Test if function is in its autoload form. Works with aliased symbols too.

Returns:
  SYMBOL     ,this can be truename of the function if it was aliased
  nil"
  (`
   ;;  Get the REAL name if it is alias or use the func's SYMBOL name
   (let* ((func (or (tinyezip-defalias-p (, symbol)) (, symbol)))
	  )
     (tinyezip-func-car-test func 'autoload))))





(defmacro tinyezip-NEXTP (list)
  "Advance LIST pointer with cdr."
  (` (setq (, list) (cdr (, list)))))




;;; ....................................................... &lib-funcs ...
;;; These functions are copied from tinylib.el or tinylibm.el so that
;;; this package is self standing, because this package will be the very
;;; first one to load in your ~/.emacs
;;;
;;; ** THIS IS ONLY EXEPTION, SINCE THIS PACKAGE IS VERY SPECIAL **
;;; ** NEVER COPY FUNCTIONS LIKE THIS, use the libraries **



;;; ----------------------------------------------------------------------
;;;
(defun tinyezip-read-match (level &optional string)
  "Read match from buffer at LEVEL or from optional STRING at LEVEL.
Return nil, if match at LEVEL doesn't exist.

You have to call `looking-at' etc. before using this function.

Please use
   bufmatget, strmatget instead of calling this if you can

Call:
  (level &optional string)"
  (if (match-end level)
      (if (stringp string)
          (substring
           string
           (match-beginning level) (match-end level))
        (buffer-substring
         (match-beginning level) (match-end level)))
    ))

;;; ----------------------------------------------------------------------
;;;
(defsubst tinyezip-strmatget (re level str)
  "Return RE match at LEVEL from STR. Nil if no match at level.
Call:
  strmatget (re level str)"
  (if (string-match re str)
      (tinyezip-read-match level str)))


;;; ----------------------------------------------------------------------
;;;
(defun tinyezip-ti::f-newer-exist (f1 f2)
  "Return file F1 or F2, which on is newer OR the one that exists.

Return:
  str	,file
  nil   ,none of them exist"
  (cond
   ((and (file-exists-p f1)
	 (file-exists-p f2))
    (if (file-newer-than-file-p f1 f2)
	f1 f2))
   ((file-exists-p f1)
    f1)
   ((file-exists-p f2)
    f2)
   (t
    nil)
  ))



;;; ----------------------------------------------------------------------
;;;
(defun tinyezip-ti::s-verify-ends (str re &optional add-str beg)
  "Match  STR with RE and add ADD-STR string to it.
If ADD-STR is not given, adds RE to the string.

Default is to check end of string, Optionally BEG of string.
The RE may not include anchors.

Returns:
  str    ,modified or not."
  (let* ((RE  (if beg
                  (concat "^" (regexp-quote re))
                (concat (regexp-quote re) "$")))
         (add (or add-str re))          ;which one to add.
         )
    (if (string-match RE str)
        str
      (if beg
          (concat add str)
        (concat str add)
        ))
      ))


;;; ----------------------------------------------------------------------
;;;
(defun tinyezip-ti::f-get-load-path (lib paths)
  "Return full path name for library LIB accross the PATHS."
  (let ((dir paths)
	(loop t)
	file found
	)
    (if (not (listp paths))
	 (error "Need path list" lib paths))

    (while (and loop dir)
      (setq file (concat (tinyezip-ti::s-verify-ends (car dir) "/") lib))
      (if (file-exists-p file)
	  (setq found file  loop nil))
      (setq dir (cdr dir)))
    found
    ))



;;; ----------------------------------------------------------------------
;;; I need both all the time, so this macro is handy
;;;
(defmacro tinyezip-file-components (str dir-var file-var &optional dir-flag)
  "Shortcut for extracting directory part and file part at once.
STR; DIR-VAR; FILE-VAR; DIR-FLAG.

If optional DIR-FLAG is non-nil, then empty directory
portion is returned as \"./\"

STR --> DIR-VAR and FILE-VAR parts"
  (`
   (let* ((dIrI (, str))
          )
     (setq (, dir-var)   (or (file-name-directory dIrI)
                             (and (, dir-flag) "./"))
           (, file-var)  (file-name-nondirectory dIrI)
           ))))



;;; ----------------------------------------------------------------------
;;;
(defun tinyezip-ti::s-index (str char &optional reverse)
  "Read STR and return first CHAR. optionally REVERSE. Postion is 0..nbr.
If REVERSE is non-nil, start searching at the end of string."
  (let ((len (length str))
        (i   -1)
        )
    (cond
     (reverse
      (while (and (>= (setq len (1- len)) 0)
                  (/= (aref str len) char))) ;check character in string
      (if (>= len 0)
          len
        nil))
     (t
      (while (and   (< (setq i (1+ i)) len)
                    (/= (aref str i) char)))
      (if (< i len)
          i
        nil))
     )))



;;; ----------------------------------------------------------------------
;;; This is very usefull for printing any variables.
;;; (mesasge (concat "DEBUG:" (print1cat nil buffer frame overlay..)))
;;;
(defun tinyezip-prin1cat (separator &rest args)
  "Cats elements separated by single space or with SEPARATOR.
The ARGS can be anything."
  (let* (ret
         (sep  (concat "%s" (or separator " ")))
         )
    (if (null sep) (setq sep sep))	;XEmacs 19.14 ByteCom silencer

    (mapcar
     (function
      (lambda (x)
        (setq ret
              (concat
               ret

               (cond
                ((integerp x)
                 (format
                  (concat "%d" (or separator " "))
                  x))

                ((stringp x)
                 (format
                  (concat "%s" (or separator " "))
                  x))

                ((symbolp x)
                 (format
                  (concat "'%s" (or separator " "))
                  x))

                ((and (not (null x))
                      (listp x))
                 (prin1-to-string
                  (eval-expression
                   (quote x))))
                (t
                 (format
                  (concat "%s" (or separator " "))
                  x)
                 ))))))
     args)
    ret
    ))

;;; ----------------------------------------------------------------------
;;; - The world's oldest way to debug program by inserting breakpoints...
;;; - This has to be defmacro, because of the &rest arguments
;;;   are passed "as is" to next function.
;;;
(defmacro tinyezip-d! (&rest args)
  "Print Debug with ARGS. Shown any information and wait keypress.
Use inside lisp code to debug any variables."
  (` (read-from-minibuffer (tinyezip-prin1cat "|" (,@ args)))))



;;; ----------------------------------------------------------------------
;;;
(defun tinyezip-function-args-p (symbol)
  "Return function SYMBOL's argument string or nil.
Works for byte compiled functions too.

Notes:
  if function is alias, the real function behing it is examined.
  if function is in autoload state, \"(autoload-args)\" is returned."
  (let (str
        ret
        )
    (if (tinyezip-autoload-p symbol)
        ;;  We can't know the args. And we don't want to find out,
        ;;  since it would load the package unnnecessarily
        (setq ret "(autoload-args)")

      (setq str (prin1-to-string
                 (symbol-function
                  (or (tinyezip-defalias-p symbol) symbol))))

      (if (or (string-match "^(lambda +nil" str)
              (string-match "^#\\[nil" str)
              (and
               (setq ret (tinyezip-strmatget "[^(]+\\([^)]+)\\)" 1 str))
               ;;  Empty arg list
               (string= ret "")))
          (setq ret nil)
        ))
    ret
    ))



;;}}}
;;{{{ zip

;;; ...................................................... &c-tinyezip ...
;;; Functions/macros belonging to this package only
;;;

(defun tinyezip-execute-extended-command-p ()
  "Check if we need to redefine `execute-extended-command'."
  (let ((args (tinyezip-function-args-p 'execute-extended-command))
        (case-fold-search t)
	)
    ;; The execcmd.el already does this, so if command is already
    ;; redefined, skip it
    ;;
    (or (null args)			;Original subr
	(not (string-match (regexp-quote "COMMAND &optional") args)))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tinyezip-remove-extension  (file)
  "Return filename without extension from FILE."
  (let* (pos
         )
    (cond
     ((setq pos (tinyezip-ti::s-index file ?. 'reverse))
      (setq file (substring file 0 pos))
      ))
    file
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tinyezip-autoload-handle (func-sym)
  "Load autoloaded FUNC-SYM.
Return:
 t	function was in autoload state
 nil	not an autoload function, or not a zipped elisp package for autoload.
        (ie. normal function in normal package)"
  (let* (file
	 path
         )
    (if (null  (setq func-sym (tinyezip-autoload-p func-sym)))
	nil
      (setq file    (elt (symbol-function func-sym) 1))
      (setq path    (tinyezip-zip-elisp-p file))

      (if (null path)
	  nil
	(tinyezip-file-eval path)
	t
	))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tinyezip-filename-hook-init (&optional var1 require-flag)
  "Initalize some variables before `tinyezip-filename-hook' is called.
VAR1 is put to `tinyezip-:require-file' and REQUIRE-FLAG is put
to `tinyezip-:require-flag'."
  (setq tinyezip-:require-file var1)
  (setq tinyezip-:require-flag require-flag)
  )


;;; ----------------------------------------------------------------------
;;;
(defun tinyezip-filename-get (file)
  "Call possible `tinyezip-load-hook' to get the FILE."
  (let* ((hook	tinyezip-filename-hook)
	 (ptr	(if (listp hook)
		    hook (list hook)))
	 (ret	file)
	 elt
	 status
	 )

    (while ptr
      (setq elt (car ptr))

      (setq status
	    (and (fboundp elt)
		 (funcall elt file)))

;;;      (tinyezip-d! (fboundp elt) elt status)
      (if (stringp status)
	  (setq ptr nil   ret status)	;stop the loop
	(tinyezip-NEXTP ptr)
	))
    ret
    ))



;;c ;; expertimental
;;c ;; I and Hans were talking about this approach, but the solution is not
;;; ;; adequate yet.
;;c
;;c (defmacro tinyezip-ad-do-it (func args &rest interactive-form)
;;c   "Preserves interactive property when calling original function..
;;c Should be used in every 'around advice.
;;c
;;c Replaces advice command:
;;c   ad-do-it
;;c
;;c The ARGS must be declared (rest args) in the advice functions input
;;c parameters.
;;c "
;;c   (`
;;c     (cond
;;c      ((interactive-p)        ; handle interactive specially, ignore advice..
;;c       (ad-with-originals ((, func))
;;c            (call-interactively (quote (, func)))))
;;c       (t                      ; let advice handle this
;;c        (setq (, args)
;;c              (,@ interactive-form))
;;c       ad-do-it
;;c       ))
;;c     ))


;;; ----------------------------------------------------------------------
;;; ** do not use this in your own code
;;; ** it won't work in general.
;;;
(defmacro tinyezip-ad-do-it (func-sym)
  "Preserve FUNC-SYM's interactive property when calling original function.
Should be used in every 'around advice that calls interactive funcs.

Replaces advice command:
  ad-do-it"
  (`
    (cond
     ((interactive-p)        ; handle interactive specially, ignore advice..
      (setq ad-return-value
	    (call-interactively (ad-make-origname (, func-sym)))))
     (t                      ; let advice handle this
      ad-do-it
      ))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tinyezip-jka-compress-ext ()
  "Return possible compression endings handled by jka.

Return:
 (ext ext ext ..)       like (\".gz\" \".Z\" ..)"
  (let* ((table jka-compr-compression-info-list)
	 z gz
	 ret
	 re
	 )
    (mapcar
     (function
      (lambda (x)
	(setq re (elt x 0))

	;;  Pick only characters ... ".Z\\([0-9]+\\)" --> ".Z"
	(if (string-match "[.a-zA-Z]+" re)
	    (setq re (substring re (match-beginning 0) (match-end 0)))
	  (setq re nil))

	;;  put extensions .Z and .gz first, and all other extensions
	;;  to "rest" list. It's important to search Z and GZ first
	;;  because eg. XEmacs has HUGE load-path with zillions
	;;  of direstories.
	;;  --> takes time to search them all.
	;;
	(if re
	    (cond
	     ((string= re ".Z")
	      (setq z t))
	     ((string= re ".gz")
	      (setq gz t))
	     (t
	      (push re ret)
	      )))
	))
     table)

    ;; Prepend these to list, put ".gz" first !
    ;;
    (if z   (setq ret (cons ".Z" ret)))
    (if gz  (setq ret (cons ".gz" ret)))

    ret
    ))


;;; ----------------------------------------------------------------------
;;; just testing ... (tinyezip-zip-elisp-p "term")
;;;
;;; Remark:
;;;   Yeah, it stupidily adds \".el\" every time, even though file were
;;;   "my.el" already.
;;;   This way the internal flow is more straight forward.
;;;
(defun tinyezip-zip-elisp-p (file)
  "Find compressed file name for elisp FILE.
If FILE does not have directory portion, `load-path' will be searched.
If file does not have .el extension, it will be added and tried.

    <input>           <tried>
    file.el       --> [load-path]/file.el.gz , [load-path]/file.el.el.gz
    /dir/file.el  --> /dir/file.el.gz ,/dir/file.el.el.gz
    /dir/file     --> /dir/file.gz   , /dir/file.el.gz


If .el file is newer, function will return it instead of existing zip file:

    file.el           ;; return this if newer than zipped one
    file.el.gz

.elc files are not supported.

Returns:
  str     ,full filename with path
  nil     ,not found"
  (let* (;;    Table of valid ZIP format's jka-compr can handle
	 table
	 otable

	 fdir				;dir
	 fndir				;non-dir
	 try1 try2
	 D				;debug flag
	 ret
	 )
    (setq table (or tinyezip-:jka-compress-ext
		    (setq
		     tinyezip-:jka-compress-ext
		     (tinyezip-jka-compress-ext))))
    (setq otable table)

;;;    (if (string-match "." file) (setq D t))

    (cond
     ((not (string-match "[.]el" file))
      ;;  since file does not have .el, we add suffixes
      ;;  .el.gz, .el.Z to the TABLE
      ;;
      (mapcar
       (function
	(lambda (x)
	  ;;  cons adds the .el.gz extensions to the first in the table
	  ;;
	  (setq table (cons (concat ".el" x) table))))
       otable)
      ))


    (if D (tinyezip-d! "TABLE" table))

    (tinyezip-file-components file fdir fndir)

    (while (and (null ret) table)	;every extension is tried

	;;  Read possible extension and cat to filename
	;;  a. file.el                   <-- search this first
	;;  b. file.el.gz
	;;
	(setq try2 (concat fndir (car table))
	      try1 (tinyezip-remove-extension try2))

	(cond ;; ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ cond beg ^^^
	 (fdir				;user gave directory
	  (setq try1 (concat fdir try1)	;add dir part
		try2 (concat fdir try2))

	  (setq ret (tinyezip-ti::f-newer-exist try1 try2))
	  (if D (tinyezip-d! ret "dir>>" try1 try2 ))
	  )

	 (t				;no directory, search path then

;;;	  (cond (D (tinyezip-d! "no-dir>>" try1 try2) (setq T1 try1 T2 try2)))

	  (cond
	   ((setq ret    (or (tinyezip-ti::f-get-load-path try1 load-path)
			     (tinyezip-ti::f-get-load-path try2 load-path)))
	    (if D (tinyezip-d! "no-dir 1>>" ret))
	    ;; - Okay, one these files was found. See if there is
	    ;;   newer compressed/uncompressed file.
	    ;; - pick the other one that wasn't found.
	    ;;
	    (setq try1 (car (delete ret (list try1 try2))))

	    (setq ret (tinyezip-ti::f-newer-exist ret try1))

	    (if D (tinyezip-d! "1>" ret))
	    ))
	  )) ;; ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ cond end ^^^

	(tinyezip-NEXTP table))

    (if D (tinyezip-d! ">>RET" ret))
    ret
    ))


					;
;;; ----------------------------------------------------------------------
;;; Keypressed command come here; this is installed into pre-command-hook
;;; #pre        ,bookmark
;;;
(defun tinyezip-pre-command ()
  "Check some things before executing command.
If command is in autoload form, handle it specially.

Adds support for loading compressed elisp files."
  (let* ((arg     this-command)
	 file
	 path
	 sym prev
	 )

    ;;   Use as fast check as possible..
    (if (or (sequencep arg)		;lambda cmd form, usual in kbd case...
	    (not (symbolp arg))		;not function call ?
	    (not (fboundp arg))
	    )
	nil				;nothing to do for us

      ;; ... ... ... ... ... ... ... ... ... ... ... ... ...  defalias . .
      ;; The tinyezip-defalias-p macro is "opened" and slimmed
      ;; here for speed reasons
      ;;
      (setq sym (symbol-function arg))
      (while (and
	      (not (sequencep sym))	;symbolp cracks otw
	      (symbolp sym)		;discard lambda forms
	      )				;was alias, go into nesting levels
	(setq prev sym)
	(setq sym (symbol-function sym))
	)

      ;;  set SYM = function name
      (if prev
	  (setq sym prev)		;alias was found
	(setq sym arg))			;wasn't an alias

      ;; ... ... ... ... ... ... ... ... ... ... ... ... ... . autolad . .
      ;; Now test if it is in autoload form
      ;;
      (if (not (and (sequencep (symbol-function sym)) ;compiled ?
		    (eq 'autoload (car (symbol-function sym)))))
	  nil				;do nothing

	(setq file    (elt (symbol-function sym) 1))
	(setq path    (tinyezip-zip-elisp-p file))

	(if path	      ;;  Okay, it's compressed, let's start rockin'
	    (tinyezip-file-eval path))
	;;  Make sure it's same when we exit
	(setq  this-command arg)
	))


    ))




;;}}}
;;{{{ Advice

;;; ........................................................ &c-advice ...

;;; ----------------------------------------------------------------------
;;; - M-x COMMAND
;;;
;;; - This function is handled specially, because someone may overload
;;;   completely execute-extended-command Eg. I use package execcmd.el,
;;;   which will take over it and print all keys the command is bound
;;;   to. The keys are shown only when the call is made interactively, so
;;;   this advise must pass the interactive property to that function too.
;;; - This advice MUST be compiled, because it includes special ad-do-it
;;;   macro, hence the 'compile' flag.
;;;
;;; 19.28: (execute-extended-command COMMAND &optional PREFIX-ARGUMENT)
;;;        when execcmd.el has been loaded
;;;
;;; 19.34: (execute-extended-command PREFIXARG)
;;;        default function

;;; Do We have to redefine this command?
;;;
(if (tinyezip-execute-extended-command-p)
    (defun execute-extended-command (command &optional prefix-argument)
      "Redefined in tinyezip.el"
      (interactive
       (list (read-command "M-x ") current-prefix-arg))

      (let ((prefix-arg prefix-argument))
	(command-execute command t)
	)))





(defadvice execute-extended-command (around tinyezip-zip-load act compile)
  "Make it possible to load zipped autoload form elisp code."
  ;;  This prevents Advice from using original function's interactive
  ;;  forms, we handle it elswhere later on.
  (interactive (list nil))

  (let ((arg		(ad-get-arg 0))
	(this-cmd	'execute-extended-command)
        path
        file
        )
    (if (null (setq arg (tinyezip-autoload-p arg)))
	(tinyezip-ad-do-it this-cmd)	;preserve interactive flag
      (setq file    (elt (symbol-function arg) 1))
      (setq path    (tinyezip-zip-elisp-p file))

      (if (null path)
          (tinyezip-ad-do-it this-cmd)
        (tinyezip-file-eval path)
	(ad-set-arg 0 arg)		;make sure it's original
	(tinyezip-ad-do-it this-cmd)
	))
    ))

;; just for emergencies, while I tested
;; (ad-unadvise 'execute-extended-command)




;;; ----------------------------------------------------------------------
;;; (ad-unadvise 'command-execute)
;;;
;;; 19.28: (command-execute CMD &optional RECORD)
;;; 19.34: (command-execute CMD &optional RECORD-FLAG KEYS SPECIAL)
;;;
(defadvice command-execute (around tinyezip-zip-load act)
  "Load functions from compressed elisp files."
  (let* ((arg  (ad-get-arg 0))
	 )
    (tinyezip-autoload-handle arg)
    (ad-set-arg 0 arg)			;make sure it's original
    ad-do-it
    ))



;;; ----------------------------------------------------------------------
;;; (ad-unadvise 'run-hooks)
;;;
(defadvice run-hooks (around tinyezip-zip-load dis)
  "Load functions from compressed elisp files."
  (let* ((hook  (ad-get-arg 0))
	 hooks
	 elt
	 )
    (cond
     ((boundp hook)			;variable must exist
      (setq hooks (eval hook))

      (if (and (not (null hooks))	;single function --> make a list
	       (not (listp hooks)))
	  (setq hooks (list hooks)))

      ;;  functions may be in compressed packages...
      ;;  require every autoload package
      ;;
      (while hooks
	(setq elt (car hooks))
	(tinyezip-autoload-handle elt)
	(tinyezip-NEXTP hooks))
      ))

    ad-do-it
    ))



;;; ----------------------------------------------------------------------
;;; #req   ,bookmark
;;;
(defadvice require (around tinyezip-zip-load act)
  "Load functions from compressed elisp files."
  (let* ((arg     (ad-get-arg 0))
	 (opt     (ad-get-arg 1))	;the optional "file" parameter
	 (lib     (or (and (stringp opt)
			   (expand-file-name opt))
		      (symbol-name arg)))
	 path
	 )
    (if (featurep arg)
	ad-do-it

      ;;  Okay, it may be  compressed...

      (tinyezip-filename-hook-init opt 'require-called)
      (setq lib (tinyezip-filename-get lib))

      (setq path    (tinyezip-zip-elisp-p lib))
      (if (null path)
	  ad-do-it
	(tinyezip-file-eval path)))
    ))
;; (tinyezip-zip-elisp-p "tinylib")

;;; ----------------------------------------------------------------------
;;; #load
;;;
(defadvice load (around tinyezip-zip-load act)
  "Load functions from compressed elisp files."
  (let* ((fn     (ad-get-arg 0))	;filename
	 path
	 )

    (tinyezip-filename-hook-init)
    (setq fn     (tinyezip-filename-get  fn)
	  path   (tinyezip-zip-elisp-p fn))

    (if (null path)
	ad-do-it
      (tinyezip-file-eval path)
      ;; Load command must return t on success.
      ;;
      (setq ad-return-value t)
      )
    ))

;;; ----------------------------------------------------------------------
;;; #load-library
;;;
(defadvice load-library (around tinyezip-zip-load act)
  "Load functions from compressed elisp files."
  (interactive "slibrary: ")

  (let* ((fn  (ad-get-arg 0))
	 path
	 )
    (tinyezip-filename-hook-init)
    (if (not (interactive-p))
	(setq fn (tinyezip-filename-get  fn)))

    (setq path  (tinyezip-zip-elisp-p	fn))

    (if path
	(tinyezip-file-eval path)
      ad-do-it
      )
   ))


;;; ----------------------------------------------------------------------
;;;
(defadvice load-file (around tinyezip-zip-load (file) act)
  "Load functions from compressed elisp files."
  (interactive
   (list
    (read-file-name "Load file: " default-directory )))

  (let* ((fn  (ad-get-arg 0))
	 path
	)

    ;;   The hooks are allowed only non-interactively
    (tinyezip-filename-hook-init)
    (if (not (interactive-p))
	(setq fn (tinyezip-filename-get  fn)))


    (setq path  (tinyezip-zip-elisp-p	fn))

    (if path
	(tinyezip-file-eval path)
      ad-do-it
      )
    ))

;;}}}

;;{{{ example setup

;;; ........................................................ &example1 ...
;;; - It's possible to keep this file in compressed form too. This is
;;;   how you load the file to emacs, before you start loading
;;;   any other files with normal load-file command.

;;* _
;;* _
;;* (defun my-zip-eval (file)
;;*   "Evals compressed file. Don't use this after startup, intead look
;;* at tinylibm::tinyezip-file-eval. This isn't reliable way to do it."
;;*   (interactive "Ffile to eval: ")
;;*   (require 'jka-compr)
;;*   (find-file "~/elisp/tinyezip.el.gz")
;;*   (eval-current-buffer)
;;*   (kill-buffer (current-buffer)))
;;* _
;;*     (progn
;;*       (if (not (file-exists-p "~/elisp/tinyezip.el.gz"))
;;* 	  (load "~/elisp/tinyezip")		;compressed elisp files
;;* 	(my-zip-eval "~/elisp/tinyezip.el.gz")
;;* 	)



;;; ........................................................ &example2 ...
;;; ** WARNING ** DO NOT USE THIS IF YOU DON'T UNDERSTAND IT
;;; ** WARNING ** you may loose your files, try it on ~/tmp/ first

;;; Requirements before you use these functions
;;; ........................................
;;; - You need tinylibXXX.el files, because library funcs are used here
;;; - Code can be extracted with function tinylib.el/ti::m-pkg-rip-magic

;;; Short intro
;;; ........................................
;;;   Okay, these are the functions I use to automatically compress
;;;   files everytime I start emacs in my University account.
;;;   The function (uta) detects if I'm running inside (system-name)
;;;   kielo.uta.fi  xxx.uta.fi ..., that is, one of my  University's machines
;;;
;;;   I put single command at the end of my emacs
;;;		(my-compress-household)
;;;   and it will keep selected files always compressed.
;;;
;;;   The reason I have autocompression, is, that I "shadow" all my
;;;   files to my University account. I do not edit files there, instead I do
;;;   all my editing at work, and let the shadow.el (19.28 std distrib.)
;;;   to transfer modified files to the University. This way, when I develop
;;;   lisp packages, my University account is constantly up to date to latest
;;;   changes. Everytime I change specific .el file at work, it's automatically
;;;   copied to University. Naturally the copied file is in .el form,
;;;   not in .el.gz form, what I would want.
;;;
;;;   When I start my University's emacs, the my-compress-household checks if
;;;   there are newer .el files and compresses them into .el.gz. it
;;;   will also remove all old files.
;;;
;;;   ** ONCE MORE ** Do not use this if it's not clear what it does.
;;;   ** THIS WILL DELETE files **
;;;

;;* ;;; ----------------------------------------------------------------------
;;* ;;;
;;* (defun my-compress-household ()
;;*   "updates shadowed elisp files."
;;*   (interactive)
;;*   (cond
;;*    ((uta)
;;*     (my-file-compress "~/elisp/" ".*el$" "#")
;;*     (my-file-compress "~/" ".emac" "init\\|default\\|emacs$\\|\\.ja.el$")
;;*     ))
;;*   )
;;* _
;;* ;;; ----------------------------------------------------------------------
;;* ;;;
;;* (defun my-file-compress (dir re &optional not-re)
;;*   "Compresses files in direcotry DIR and files matching RE.
;;* NOT-RE excludes files.
;;* _
;;* Drops away files that are read-only, backups [~#] or already zipped.
;;* Deletes older file if comressed one is newer.
;;* "
;;*   (interactive "ddir: \nsre: ")
;;*   (let* ((files   (directory-files dir t re))
;;* 	 (not-re  (if not-re
;;* 		      (concat not-re "\\|g?[Z#~]$")
;;* 		    "g?[Z#~]$"))
;;* _
;;* 	 (files   (if (null not-re)	;exclude files
;;* 		      files
;;* 		    (list-find
;;* 		     files not-re
;;* 		     '(not (string-match arg (car table)))
;;* 		     'all)))
;;* 	 (ext     ".gz")		;gzip used for compression
;;* 	 (bp      (generate-new-buffer "*tmp*"))
;;* 	 (bpn     (buffer-name bp))
;;* 	 (obp     (current-buffer))
;;* _
;;* 	 fn-el fn-gz
;;* 	 bup				;backup ?
;;* 	)
;;*     (require 'jka-compr)
;;*     (set-buffer bpn)			;avoid set-excursion indentation
;;* _
;;*     (list-loop files fn-el
;;*       (erase-buffer)
;;*       (setq fn-gz (concat fn-el ext))
;;* ;;;      (d! fn-el fn-gz)
;;* _
;;* _
;;*       ;;  Check some things first, do not touch #files#
;;*       ;;  And those which are read-only
;;* _
;;*       (setq bup
;;* 	    (concat
;;* 	     (file-name-directory fn-el)
;;* 	     "#"
;;* 	     (file-name-nondirectory fn-el)
;;* 	     "#"))
;;* _
;;* _
;;*       (if (or (file-exists-p bup)
;;* 	      (and (file-exists-p  fn-el)
;;* 		   (not (file-writable-p fn-el)))
;;* 	      (and (file-exists-p  fn-gz)
;;* 		   (not (file-writable-p fn-gz)))
;;* 	      )
;;* 	  (message (concat "Skipped ..." fn-el))
;;* 	(cond
;;* 	 ((and (file-exists-p fn-el)	;.el and el.gz
;;* 	       (file-exists-p fn-gz)
;;* 	       (file-newer-than-file-p fn-el fn-gz);; is .el newer ?
;;* 	       )
;;* 	  (delete-file fn-gz)
;;* 	  (insert-file-contents fn-el)
;;* 	  (write-file fn-gz)
;;* 	  (rename-buffer bpn)
;;* 	  (delete-file fn-el)
;;* 	  )
;;* 	 ((and (file-exists-p fn-el)
;;* 	       (file-exists-p fn-gz)	;.el is older ...
;;* 	       )
;;* 	  (delete-file fn-el)
;;* 	  )
;;* 	 ((file-exists-p fn-el)
;;* 	  (insert-file-contents fn-el)
;;* 	  (write-file fn-gz)
;;* 	  (rename-buffer bpn)
;;* 	  (delete-file fn-el)
;;* 	  )
;;* 	 (t
;;* 	  nil)				;only .gz left
;;* 	 ))
;;*       )
;;* _
;;* _
;;*     (kill-buffer bpn)
;;*     (set-buffer obp)			;restore
;;*     ))


;;; ......................................................... example2 ...
;;; - This shows how to use the hook to convert the file names
;;;   to something else.
;;; - You can't use this function right from the box, because it contains
;;;   functions and code that is specific to my environment only.
;;;
;;; - The function (uta) detects if I'm currently inside my
;;;   University's machine ([U]niversity of [Ta]mpere)
;;; - The conversion goes like this:
;;;   any file that is in RE should have ".sun" extension in
;;;   load commands. In general I can write same commands for my work
;;;   machine like this:
;;;
;;;	    (load "~/.emacs.path")
;;;
;;;    And I don't have to worry about which file gets called in different
;;;    systems. In university; this call is translated into file with
;;;    extension ".sun" as you can see from the code.

;;* _
;;* ;;; ----------------------------------------------------------------------
;;* ;;;
;;* (defun my-tinyezip-filename-hook (file)
;;*   "Changes the load filename if needed for config files."
;;*   (let* (
;;* _
;;* 	 ;;  Handle require call differently; the file may be in
;;* 	 ;;  parameter 2
;;* 	 ;;
;;* 	 (flag	tinyezip-:require-flag)
;;* 	 (req   tinyezip-:require-file)
;;* 	 (fil   (if (or (null flag)
;;* 			(not (stringp req)))
;;* 		    file
;;* 		  req))
;;* 	 ret
;;* 	 )
;;*     (cond
;;*      ((uta)				;SunOS files
;;* _
;;*       (cond
;;*        ;;  This file is special; the funcs reside normally along
;;*        ;;  the path; but here we tell the absolute location
;;*        ((string-match "funcs.ja" file)
;;* 	(setq ret "~/.emacs.funcs.sun"))
;;* _
;;* _
;;*        ;;  converts these files to have extension .sun (SunOS)
;;*        ((and (string-match "emacs.\\(ja\\|path\\|start\\|tips\\)" file)
;;* 	     ;;  already has the extension ?
;;* 	     (not (string-match "sun" file)))
;;* 	(setq ret (concat file ".sun"))
;;* 	)
;;* _
;;*        )))
;;* _
;;*     ret
;;*     ))
;;* _


;;}}}


(tinyezip-pre-command-install)		; Final touch... activate kbd checks

(provide   'tinyezip)
(run-hooks 'tinyezip-load-hook)

;;; tinyezip.el ends here
