;; @(#) tinyfh.el -- Collection of file handling functions

;; @(#) $Id: tinyfh.el,v 1.28 1997/05/21 12:25:22 jaalto NotReleased $
;; @(#) $Keywords: tools $
;;
;; $KnownCompatibility: 19.28, XE 19.14 $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of GNU Emacs

;;{{{ Id

;; Copyright (C) 1995,96,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Mar 29 1995
;;
;; To get information on this program use ident(1) or do M-x tifh-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el

;; LCD Archive Entry:
;; tinyfh|Jari Aalto|jari.aalto@poboxes.com|
;; file handling: easy read-only edit, powerfull find-file-at-point|
;; 21-May-1997|$Revision: 1.28 $|~/misc/tinyfh.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;;; Intallation:
;; ........................................................ &t-install ...
;; Put this file on your Emacs-Lisp load path, add following into your
;; ~/.emacs startup file. Rip code with with tinylib.el/ti::m-pkg-rip-magic
;;
;;*_
;;*     (require 'tinyfh)
;;
;; or use this, your .emacs starts faster, prefered method
;;
;;*_
;;*     (autoload 'tifh-ro-edit                 "tinyfh" t t)
;;*     (autoload 'tifh-ro-edit-once            "tinyfh" t t)
;;*     (autoload 'tifh-ff-at-point             "tinyfh" t t)
;;*     (autoload 'tifh-ff-display-last-ref     "tinyfh" t t)
;;
;; And to make best use of this, define some key binding
;;
;;*_
;;*     (global-set-key "\C-c\C-q" 'tifh-ro-edit)
;;*     (global-set-key "\C-c\C-f" 'tifh-ff-at-point)
;;*_
;;
;; This file automatically installs itself when loded or
;; evaled. If you want to uninstall the automatic read only part:
;;
;;	(tifh-ro-install nil)
;;
;; I would suggest adding this to your .el too, it turns write-once
;; feature automatically on for ro-buffers
;;
;;* 	(setq-default  tifh-:ro-edit-once t)
;;_
;;
;; ULR AND HTTP INSTALLATION
;; ........................................
;; This package does not deal with WWW. Use your favourite URL handler
;; for those purposes. Like this:
;;
;;*	;;  using browse-url.el
;;* 	(setq tifh-:ff-url-hook 'browse-url-at-point)
;;


;;}}}
;;{{{ Documentation


;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "tifh-" in front of
;;   every function & variable. It stands for '(ti)ny (f)file (h)andling
;; - variable names contain letters 'tifh-:'

;;  Preface, 1995
;;
;;	You guessed it, this is again a tiny tool. It was inpired by post
;;	to gnu.emacs.help in Mar 15 1995. The poster had sent amazing post,
;;	which only said on the subject line: 'Q, editing write-protected
;;	files'.  Nothing else, no more information on the body of the
;;	message. I wondered what was this all about, why didn't he just do
;;	C-x C-q to toggle the read only flag? And I posted that answer, but
;;	the sender complained I didn't answer his question! He explained,
;;	that he didn't want to confirm all those silly questions emacs
;;	asked while savind the read-only file.. Well, why didn't he say so
;;	in the first place ...
;;
;;	[Later note: When we went peer to peer with the poster it turned
;;	out that the news server had cut off the body of his message, so
;;	it really wan't his fault that I saw only that short message. :-> ]
;;
;;	I call this .el FH, file handling, not read-only handling, since
;;	I'm looking to the future where I can add more, interesting, functions
;;	here if I come up with something else later.
;;
;;  Overview of features
;;
;;	o   File handling functions that make your day in emacs shine :'/
;;	o   easy read-only file edit/save feature
;;	o   Powerfull find file at point ftp://, ange@type, archie listing
;;	    elisp LCD listing parsing, relative ~/ completion
;;	    simtel/ completions ... all this user defineable.
;;	o   background file loading to download dir.
;;	o   emails address at current point
;;
;;  Read-only editing
;;
;;	There are two different approaches to read only editing.
;;
;;	o   you can 'avoid' emacs confimation with tifh-ro-edit
;;	    and save the file, allthough it was _read_only_ when loaded
;;
;;	Normally I use tifh-ro-edit for temporary changes.
;;	The idea to avoid emacs confirmations is quite simple, we just need
;;	make the file writable before the emacs runs save-buffer, where the
;;	complaints are coming. After the file is saved we restore the
;;	original permissions of file.  That's it. You only have to tell the
;;	buffer that it should be written quietly by turning on the
;;	tifh-:read-only-edit flag
;;
;;  Important note, hooks when editing read-only files
;;
;;	If you load some other .el that does file handling and hooks itself
;;	into one of these hooks:
;;
;;	    [1] write-file-hooks
;;	    [2] after-save-hook
;;
;;	It propably hooks itself at the front of these hooks. It is
;;	prefered that the funcs installed by this el are [1] FIRST and
;;	[2] LAST, because this .el directly touches the file. To make sure
;;	the hooks are in proper order, please load this file as last statement
;;	in your el, or if you notice anything weird, just run the install
;;	again. It will re-order the hooks properly.
;;
;;	    (tifh-ro-install t)
;;
;;	The good news is that the hooks' func order is propably not a problem,
;;	and you may ignore this warning. If all works fine, then there's
;;	nothing to worry about.
;;
;;  Finding file on the line
;;
;;	There exists several file finding packages. Here are few:
;;
;;	    browse-url.el		-- for URL addresses
;;	    ffap.el			-- many features!
;;	    hyberbole.el		-- massive package
;;
;;	The ffap.el looks like quite comprehensive for any kind of
;;	addresses, it even handles URL's, which leaves question, how
;;	the browse-url.el fits into picture, unanswered.
;;
;;	The find file function in this package were developed for my own
;;	needs at the time I didn't even know the ffap.el existed. So, if
;;	you want full blown gun, I'd suggest you to have a look at ffap.el
;;
;;	But for simple file operations (I don't need URL things in emacs or
;;	any other fancy file address parsing.), you want to take a look at
;;
;;	    M-x tifh-ff-at-point
;;
;;	It recognizes few common file addresses, at least those are the one
;;	I point and shoot 90% of my time. Especially it is fun to get file
;;	from emacs shell buffer where you have just started archie.
;;	Even more fun is to fetch new elisp file, when someone
;;	sends the LCD entry in the .emacs groups.
;;
;;	Have a look at function
;;
;;	    M-x tifh-ff-display-last-ref
;;
;;	If you're in doubt where did you landed, when following the reference
;;
;;
;;	This file belongs to my 'tiny tool' series, so it is unlikely that
;;	I ever start programming parse routines similar to ffap.el.
;;
;;  Finding files -- be careful
;;
;;	You propably start adding many sites in the default variables offered
;;	to user. One word about the ange syntax:
;;
;;	    /site.xx.net/foo-dir/bar-dir/
;;	    ^^          ^^
;;
;;	There is two errors in this reference
;;
;;	o   the beginning should be full login reference /loginName@site.xx.net
;;	    or it should not contain  reference at all: site.xx.net
;;	o   You forgot the _colon_. This is important, because the ange doesn't
;;	    recognize the name otherwise
;;
;;	Correct version is:
;;
;;	    site.xx.net:/foo-dir/bar-dir/			or
;;	    /anonymous@site.xx.net:/foo-dir/bar-dir/
;;
;;  Bugs
;;
;;	If the reference under cursor has invalid directory, the ange
;;	can't know that. Say, you try to find:
;;
;;	    Simtel/foo/xdir
;;
;;	and use default completion to pick ftp.coast.net/SimTel/ reference.
;;	This would then expand to full reference:
;;
;;	    ftp.coast.net/SimTel/foo/xdir
;;
;;	Now, if there is no such directory "/foo/xdir", you'll be displayed
;;	an empty buffer and you propably may see message flashing below
;;	"Use keys C-M-xx xxx... to create directory." This error in
;;	unavoidable.
;;
;;	You can easily get back to the reference by saying
;;
;;	    C-u M-x tifh-ff-at-point
;;
;;	It offers the last reference tried. Now you can remove paths
;;	from it and try again. Best advice would be to start from
;;	the root directory, and discard every subdir.
;;
;;  Advertise: tinydired.el
;;
;;	You must get this dired extension package if you're
;;	serious at all with the downloading of files in the backgroud
;;
;;  Advertise: ssz-mini.el
;;
;;	Standard emacs distribution package at least in 19.28.  Put
;;	(require 'rsz-mini) into the ~/.emacs and you can see the full load
;;	path in the echo area. This is a must for long ftp filenames.
;;


;;}}}
;;{{{ history

;; ....................................................... &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; Apr 21	1997	[jari]		19.28	v1.25-1.28	NotReleased
;; - Cosmetic changes. XE20 char handling corrections.
;; - moved buffer chmod command to 'tinymy' module.
;;
;; Jan 15	1997    [jari]		19.28   v1.23-1.24	NotReleased
;; - Corrected "re-ange" reading in cond. Didn't read reference
;;   /ftp@ftp.sunet.se:/pub/lang/perl/CPAN/.
;; - Reordered library loading.
;; - Corrected call mesage --> message. Added browse-url test.
;;
;; Aug 30	1996    [jari]		19.28   v1.21		NotReleased
;; - Cleared all XE 19.14 reported byteComp errors.
;;
;; Jun 19	1996    [jari]		19.28   v1.20		NotReleased
;; - Some minor fixes in the main find-file function. It didn't detect
;;   simple addresses like /fpt@ifi.uio.no:/
;; - Todo: rewrite the whole thing and crack it to small functions instead.
;;
;; May 3	1996    [jari]		19.28   v1.19		NotReleased
;; - Corrected bug in Archie listing. It didn't recognize "DIRECTORY"
;;   only lines. Now it handles them too.
;;
;; Apr 14	1996    [jari]		19.28   v1.18		NotReleased
;; - Changed modeline update command, because XE labels it as obsolete.
;;
;; Apr 4	1996    [jari]		19.28   v1.17		NotReleased
;; - Corrected bug that prevented mailing a person when cursor was in the
;;   <foo@site.com>
;;   tifh-:ff-no-site-regexp		:! serious, missing \\)
;;   tifh-ff-at-point			:! mail,
;;
;; Mar 28	1996    [jari]        19.28   v1.16		NotReleased
;; - Main function now better detects remote file names. defconst --> defvar.
;;   Now allows editing the command if the site is not recognized.
;; - Example updated.
;;   tifh-:ff-no-site-regexp		:+
;;   tifh-:ff-elisp-site		:+
;;   tifh-ff-ange-padd-reference	:!
;;
;; Feb 16	1996    [jari]		19.28   v1.15		NotReleased
;; - Small error in file-charset, the colon was mistakenly put there and
;;   this virtually prevented using ftp:// like addresses.
;; - ByteCompiler checked in fresh emacs. Errors found.
;;
;; Feb 16	1996    [jari]		19.28   v1.14		NotReleased
;; - Added Email capability. I forgot that alltogether. Run hooks
;;   function modified.
;;   tifh-ff-at-point			:!
;;   tifh-mail-at-point			:+
;;
;; Feb 14	1996    [jari]		19.28   v1.13		NotReleased
;; - URL handling was missing. Now it's ensured that packege won't
;;   handle HTTP or WWW related lines --> passes control to someone else.
;; - added 'http installation' 'exaples' sections
;;   tifh-:ff-url-hook			:+
;;   tifh-:ff-not-found-hook		:! removed example --> now at eof
;;   tifh-ff-at-point			:! URL handle, rewrote a lot else
;;
;; Feb 14	1996    [jari]		19.28   v1.12		NotReleased
;; - Added bug reporting function. Added "minimal" ftp reference support.
;;   Converted some function names to more standardized form of "ff"
;; - Now it's possible to edit last REF, arg C-u added, broke the
;;   main ff into half for easier handling.
;; - Wrote sections "Finding files -- be careful", "Bugs"
;;   tifh-submit-bug-report		:+
;;   tifh-ff-display-last-ref		:+
;;   tifh-ff-ref-syntax-ok-p		:+
;;   tifh-ff-1				:+
;;
;; Feb 14	1996    [jari]		19.28   v1.11		NotReleased
;; - Radical rewrote of the "guessing": word --> to reference. This
;;   improves this package light years compared to 1.10 version.
;; - Now the /ftp@ tag is not needed in completion lists any more.
;;   tifh-:ff-file-charset		:+ not hard-coded any more
;;   tifh-:ff-word-to-site-ref-list	:+ the "core" of guessing. Wow!
;;   tifh-:ff-simtel-completion-list	:+
;;   tifh-ff-check-home-file		:+
;;   tifh-ff-ange-padd-reference	:+
;;   tifh-ff-guess		:+
;;   tifh-ro-save-buffer-after		:! rewritten. Shrinked.
;;
;; Feb 10	1996    [jari]		19.28   v1.10		NotReleased
;; - Small error in tifh-ff-at-point. Ehen there were empty
;;   line, the word-now was nil. corrected to read "" instead.
;;
;; Feb 11	1996    [jari]		19.28   v1.9		NotReleased
;; - Improved the site reading furher. Now it understands puctuations in the
;;   end of site name and discards them. Also doesn't get upset with the
;;   url like <> tags so much. Uses file-re and file-charset to filter
;;   words better.
;;
;; Feb 11	1996    [jari]		19.28   v1.8		NotReleased
;; - Improved the file-find-at-point code light years! Now there is
;;   bunch of configuration variables, completion, supports newsgroup
;;   based ange ftp, user's default ftp (run-hooks), more smarted
;;   line detecting, single site address is recognized...
;; - This version requires new tinylib for ange support functions.
;;   tifh-:ff-not-found-hook		:+
;;   tifh-:ff-bg-load-maybe		:+
;;   tifh-:ff-bg-default-download-dir	:+
;;   tifh-:newsgroup-to-archive		:+
;;
;; Feb 10	1996    [jari]		19.28   v1.7		NotReleased
;; - Greatly enchanced the ff-at-point. Added relative ~/ path support
;;   via possible completion.
;;
;; Feb 6	1996    [jari]		19.28   v1.6		NotReleased
;; - Added folding marks.
;;
;; Feb 6	1996    [jari]		19.28   v1.5		NotReleased
;; - Moved the find file functions from my private 'funcs.el' library
;;   to this file. generalized all function to go through
;;   tifh-ff-at-point.
;;
;; Jun 29	1995    [jari]		19.28   v1.4		NotReleased
;; - Now this file depends on the tinylibs.
;; - added tifh-:load-hook , tifh-:ro-vc-check , tifh-ro-chmod ,
;;   tifh-ro-chmod-toggle.
;; - With additional macros, cleaned tifh-ro-edit-once, tifh-ro-edit.
;; - Hard coded 644 write mode changed to 'soft' 600 in
;;   tifh-ro-save-buffer-before. This should be reliable now, because
;;   now raise only user's write bit.
;;
;; May 11	1995    [jari]		19.28   v1.3		NotReleased
;; - Moved the 'provide' at the end, also added one more line to
;;   autoload installation for 'toggle-line'
;;
;; Apr 28	1995    [jari]		19.28   v1.2		NotReleased
;; + Added vercion control check for RCS files, so that you can't save them
;;   automatically.
;;
;; Apr 28	1995    [jari]		19.28   v1.2		NotReleased
;; - Changed all names. Now the read-only functions and variables have
;;   common base tifh-ro or tifh-:ro , so that adding new functions
;;   for other purposes are easy.
;; + Added read-only-once capability, which I find a bit more usefull
;;   than constant read-only edit mode.
;;
;; Mar 29	1995    [jari]		19.28   v1.1		Release_1
;; - Collected simple functions from my .el library to here and made
;;   fully functional .el out of them.


;;}}}


;;; Code:
;;; ......................................................... &require ...

(require 'tinylibm)
(eval-when-compile (use-dynamic-compilation))

;;{{{ setup: -- variables

;;; ......................................................... &v-hooks ...

(defvar after-save-hook nil
  "This hook seems to be undocumented in 19.28, but it's there [files.el]")

(defvar tifh-:load-hook 'tifh-ro-install
  "Hook run when file is loaded.")

(defvar tifh-:ff-url-hook
  (if (featurep 'browse-url)
      'browse-url-at-point)
  "Hook run if URL or HTTP tag has been detected on the line.")


;;  See at the end, how this hook can be used.
;;
(defvar tifh-:ff-not-found-hook  nil
  "*Hooks run if tifh-ff-at-point doesn't succeed.
If any of the functions return non-nil, the rest of the functions
are not run, so be carefull with the function order.")


;;; ... private variables ................................. &v-private ...


(defvar tifh-:ro-edited nil
  "Buffer local variable. It is set/cleared by program to signify
read-only editing is in process. Holds original file
permissions of the file.")
(make-variable-buffer-local 'tifh-:ro-edited)

(defvar tifh-:ff-last-ange-reference nil
  "Contains ange-ftp reference last used.")


(defvar tifh-:ff-at-point-arg nil
  "Contains the ARG passed to function tifh-:ff-at-point-arg.
You can check this argument in your own function put to
tifh-:ff-not-found-hook")



;;}}}
;;{{{ setup: -- USER config

;;; .............................................. &v-read-only-config ...

(defvar tifh-:ro-verbose t
  "*Allows or inhibits informational messages .")


(defvar tifh-:ro-edit nil
  "*Buffer local variable. When set to t, allows writing read-only files
easily without extra confirmations")
(make-variable-buffer-local 'tifh-:ro-edit)



(defvar tifh-:ro-edit-once nil
  "*If this variable is non-nil, the write of read-only file is done only
_once_. After the file is written, the file is again marked with
%% and you have to manually turn on the tifh-:ro-edit to t
to suppress emacs confirmations.

Basicly this says: \"I'm in temporary mode, every write must be handled
individually.\"")
(make-variable-buffer-local 'tifh-:ro-edit-once)



(defvar tifh-:ro-vc-check t
  "*If this variable is non-nil, it ensures that the file you are editing
isn't version controlled.. It won't permit you to accidentally edit
read-only & vc file.")

;;; .............................................. &v-find-file-config ...

;;; The ":" must not be put there, it's needed for Ftp:// etc.
;;;
(defvar tifh-:ff-file-charset  "^,;!|<>(){}'\\\" \t\n"
  "*How the file in the line should be read. This is exclusive syntax,
ie. which characters do not belong to filename. Do not remove
beginning ^ mark. It's important!")


(defvar tifh-:ff-no-site-regexp
  "\\.\\(el\\|cc\\|hh\\|zip\\|exe\\|gz\\|tgz\\|tar\\|ps\\|doc\\)$"
  "*Regexp that says: cursor in this word isn't a site reference.
Normally if the words has ending

    .[a-z][a-z][a-z]?$		-- eg. ftp.cs.uta.fi

It's a possible candidate for site name. However, to prevent some false
hits, this regexp is needed.")


(defvar tifh-:ff-elisp-site
  "/ftp@cis.ohio-state.edu:/pub/gnu/emacs/elisp-archive"
   "*Full ange reference to site that has \"LCD Archive Entry\"
File structure. This is the location where character ~ referes
in LCD entry.")





(defvar tifh-:ff-word-to-site-ref-list
  '(("~"	"^~/"	    tifh-:ff-completion-list tifh-ff-check-home-file)
    ("simtel"	"^simtel/"  tifh-:ff-simtel-completion-list)
    )
  "*Format is.
((PRETTY-NAME REGEXP COMPLETETION-LIST CHECK-FUNC) .. )

PRETTY-NAME
  is string showed to user if completion takes place.

REGEXP
  is tried on current word under cursor. If this matches, then
  possible check function is run.

CHECK-FUNC [optional]
  If there is no check-func, then assume that the match
  is already ok and offer COMPLETETION-LIST.

  should accept 1 arg:
  current-word   ,string under cursor

  should return:
  t		,if the completion list can be used.
  string	,this file is used, no user interaction takes place
  nil		,reject. Do not offer completion list

If the REGEXP matched to the word and check result is true, user
is offered completion list of site names, where the current word
referes to. Eg if user is sitting on word below:

	SimTel/vendors/gnuish

The possible simtel sites, that are prepended to the word are offered.
")



(defvar tifh-:ff-completion-list
  (list
   "archive.cis.ohio-state.edu:/pub/gnu/emacs/elisp-archive"
    (or (getenv "HOME")
	(error "Please setenv HOME"))
    )
  "*When there is ~ at the beginning of file read from buffer,
it is first assumed to reference to user's home. If that directory and
filename doesn't exist, user is asked, where should the filename
be looked.

This is list of expansions offered to user.

ELT NOTE: if there is no \"/ftp@\"  login string in the element,
it will be automatically added. Leaving it out from the list eases picking
up the right completion.
")




(defvar tifh-:ff-simtel-completion-list
  (list
   "oak.oakland.edu:/simtel/"
   "ftp.coast.net:/SimTel/"
   )
  "*The Simtel completion sites. If the word under cursor has
word \"simtel\" ask user what simtel site to use.

ELT NOTE: if there is no \"/ftp@\"  login string in the element,
it will be automatically added.
")


;; Set this variable to "." if you always want to be asked about bg load
;; .doc  --> Suppose Microsoft Word file
;;
(defvar tifh-:ff-bg-load-maybe
  "[.]\\(z|zip\\|tar\\|tgz\\|gz\\|ps\\|doc\\)$"
  "*If the ange FTP reference is created, this variable should be
a REGEXP that tries to match files, which are loaded in the backgroup.

Eg. files with .tar .gz. .Z .zip are best loaded in the background,
not directly into emacs.

However, setting this variable just triggers confirmation for
these files, the user can choose if he wants to do the bg loading.
")


(defvar tifh-:ff-bg-default-download-dir  "~/ftp"
  "*Directory which is proposed if file is downloaded in the background
A good value would be root directory of your ftp storage.")




;;; ..  ..  ..  ..  ..  ..  ..  ..  ..  ..  ..  ..  ..  ..  ..  ..  ..  ..
;;; For newsgroup handling you need tinylibmt.el
;;;
(defvar tifh-:ff-newsgroup-to-archive
  '(
    ;;  The Emacs elisp site
    ("emacs"
     "archive.cis.ohio-state.edu:/pub/gnu/emacs/elisp-archive/")

    ;;  The one and only SIMTEL
    ("msdos"
     "ftp.coast.net:/SimTel/msdos/" )

    ;;  One CICA site
    ("windows"
     "ftp.funet.fi:/pub/msdos/windows")

    ;; The perl CPAN achive
    ("perl"
     "ftp.delphi.com:/pub/mirrors/packages/perl/CPAN/" )
    )
  "*If your're in newsgroup that matchs RE and the filename at point is
not regognized to be in absolute URL or ange ftp form, your're
offered ftp site read from this list.

Format is '((NEWSGROUP_REGEXP ANGE_REFERENCE) .. )
")






;;}}}
;;{{{ setup: -- version

;;; ....................................................... &v-version ...


(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinylisp.el"
 "tit"
 tili-:version-id
 "$Id: tinyfh.el,v 1.28 1997/05/21 12:25:22 jaalto NotReleased $"
 '(tifh-version-id

   after-save-hook
   tinyfh-:load-hook
   tifh-:ff-url-hook
   tifh-:ff-not-found-hook
   tifh-:ro-edited
   tifh-:ff-last-ange-reference
   tifh-:ff-at-point-arg
   tifh-:ro-verbose
   tifh-:ro-edit
   tifh-:ro-edit-once
   tifh-:ro-vc-check
   tifh-:ff-file-charset
   tifh-:ff-no-site-regexp
   tifh-:ff-elisp-site
   tifh-:ff-word-to-site-ref-list
   tifh-:ff-completion-list
   tifh-:ff-simtel-completion-list
   tifh-:ff-bg-load-maybe
   tifh-:ff-bg-default-download-dir
   tifh-:ff-newsgroup-to-archive
   )))





;;}}}

;;; ########################################################### &Funcs ###

;;{{{ ro: install


;;; ----------------------------------------------------------------------
;;;
(defun tifh-hook-it (hook func &optional last)
  "Moves the named FUNC in HOOK to the end or beginning. Adds the FUNC if not
already there. if LAST is non-nil, puts the function at the end."
  (if (memq func (eval hook)) (remove-hook hook func))
  (add-hook hook func last))


;;; ----------------------------------------------------------------------
;;;
(defun tifh-ro-install (&optional uninstall)
  "Installs or uninstalls TIFH package."
  (interactive "P")
  (cond
   (uninstall
    (remove-hook 'write-file-hooks 'tifh-ro-save-buffer-before)
    (remove-hook 'after-save-hook 'tifh-ro-save-buffer-after)
    )
   (t
    (tifh-hook-it 'write-file-hooks 'tifh-ro-save-buffer-before)
    (tifh-hook-it 'after-save-hook 'tifh-ro-save-buffer-after t)
    )))


;;}}}
;;{{{ ro: edit

;;; ----------------------------------------------------------------------
;;;
(defun tifh-ro-off ()
  "Turns ro-edit feature off and makes buffer read-only."
  (tifh-ro-edit 0)			;turn off
  (toggle-read-only 1))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tifh-ro-edit (&optional arg )
  "Toggless buffer local tifh-:ro-edit flag."
  (interactive "P")
  (let* ((verb (interactive-p))
	 )
    (if (null (buffer-file-name (current-buffer)))
	(if verb (message "read-only: not vising a file"))

      (var-toggle  tifh-:ro-edit (prefix-numeric-value arg))
      (if (interactive-p)
	  (message (concat "read only edit = "
			   (if tifh-:ro-edit "t" "nil"))))
      )
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tifh-ro-edit-once (&optional arg )
  "Toggless buffer local tifh-:ro-edit-once flag.
If buffer is not visitin a file, does nothing.
"
  (interactive "P")
  (let* ((verb (interactive-p))
	 )
    (if (null (buffer-file-name (current-buffer)))
	(if verb (message "read-only: not vising a file"))

      (var-toggle  tifh-:ro-edit-once (prefix-numeric-value arg))
      (if (interactive-p)
	  (message (concat "read only edit-once = "
			   (if tifh-:ro-edit-once "t" "nil"))))
      )
    ))

;;}}}
;;{{{ ro: hook funcs
;;; ...................................................... &hook-funcs ...

;;; ----------------------------------------------------------------------
;;;
(defun tifh-ro-save-buffer-before ()
  "Handles read-only files differently. Kinda wrapper before save-buffer.

If the buffer is marked so, that it allows saving read-only file, the
file permissions are changed to +w just before saving. This disables
all confirmations about the save. The permissions are restored after
the file has been written. In order to this function to work with
other hooks placed into write-file-hooks, it should be the first
in the list.

References:
  tifh-toggle-ro-edit         ,func
  tifh-save-buffer-after      ,func
"
  (let* ((vc   tifh-:ro-vc-check)
	 (file (buffer-file-name (current-buffer)))
	 (edit (if (boundp 'tifh-:ro-edit) tifh-:ro-edit nil))
	 (verb tifh-:ro-verbose)
	 orig-mode
	 )

    (setq tifh-:ro-edited nil)	;inform 'after' hook

    (if (or (null file)			;let emacs handle this
	    (null edit)
	    (file-writable-p file)
	    (null (file-exists-p file))	;cannot read modes
	    ;;   If the file is owned by someone else, let the emacs handle.
	    (not (file-owned-p file))
	    )
	nil
      ;; suppose it was read-only, at least it exist, owned by us
      ;; - Make sure it's our file..
      (if (null vc)
	  nil				;do not touch vc controlled
	(tifh-ro-off)			;cancel
	(error "*err read-only: this is vc controlled file")
	)

      (setq orig-mode (file-modes file))
      (setq tifh-:ro-edited orig-mode) ;set >>GLOBAL<<

      ;; 128 d , 200 oct, we turn on user's write bit.
      (set-file-modes file (logior orig-mode 128))

      (if verb 	(message "read-only file"))
      )
    nil					;hook must return nil
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tifh-ro-save-buffer-after ()
  "Handles read-only files differently. Kinda wrapper _after_ save-buffer.

If the file beeing saved was red-only AND that file was temporarily
changed to writable, this function will restore the default permissions
of the file. It is highly recommended that this function is _always_
last in the after-save-hook.

References:
  tifh-:ro-edited
"
  (let* ((file (buffer-file-name (current-buffer)))
	 (verb tifh-:ro-verbose)

	 ;;  The original file modes are put into this var, if
	 ;;  the file was read-only during save sequence.
	 (orig-mode (if (boundp 'tifh-:ro-edited)
			tifh-:ro-edited nil))
	 )
    (if (null orig-mode)		;wasn't read-only
	nil				;just pass-by
      (set-file-modes file orig-mode)

      (if (null verb) nil
	(message "Read only file saved.")
	(sleep-for 1))

      (if tifh-:ro-edit-once (tifh-ro-off))
      )
    ))

;;}}}

;;{{{ find: misc


;;; ----------------------------------------------------------------------
;;; - This is handy. Sometimes, when empty buffer pops up,
;;;   I wonder where did I jumped with ftp? Was the reference wrong
;;;   somehow...
;;;
;;;###autoload
(defun tifh-ff-display-last-ref  ()
  "Display last age-ftp reference used."
  (interactive)
  (let* ((ref  (if (stringp tifh-:ff-last-ange-reference)
		   tifh-:ff-last-ange-reference
		 "<no last ange ref>"))
	 )
    (message ref)))


;;; ----------------------------------------------------------------------
;;;
(defun tifh-ff-check-home-file  (word)
  "Checks if the WORD referes to home file."
  (interactive)
  (cond
   ((file-exists-p word)	;in user's dir
    word)
   (t				;ask completion
    t
    )))


;;; ----------------------------------------------------------------------
;;; - This is handy. Sometimes, when empty buffer pops up,
;;;   I wonder where did I jumped with ftp? Was the reference wrong
;;;   somehow...
;;;
(defsubst tifh-ff-ref-syntax-ok-p  (ref)
  "Check that ange reference syntax is ok."
  (string-match "^/.*@.*:" ref))



;;; ----------------------------------------------------------------------
;;;
(defun tifh-ff-ange-padd-reference (ref)
  "Padds \"/ftp@\" to the beginnig of REF if it's not there
already."
  (let* (ret
	 )
    (cond
     ((var-nil-p ref)			;invalid !
      nil)
     ((string-match "@" ref)		;ok?
      (setq ret ref))			;yes
     (t
      ;;  /ftp.com:/  --> /ftp@ftp.com:/
      (if (string= (str-left ref 1) "/")
	  (setq ref (substring ref 1)))
      (setq ret (concat "/ftp@" ref))	;return this
      ))
    ret
    ))


;;; ----------------------------------------------------------------------
;;; test: (setq a (tifh-ff-guess "SimTel/vendors/gnuish"))
;;;
(defun tifh-ff-guess (word-now)
  "Tries to guess what sitye word refers to."
  (let* ((alist     tifh-:ff-word-to-site-ref-list)
	 (file-set  tifh-:ff-file-charset)
	 (file-re   (concat "\\([" file-set "]+\\)"))
	 (fold-case-search t)		;non sensitive

	 elt ans
	 re
	 complete
	 name

	 stat				;status of check func
	 word
	 file
	 )
    (if (null fold-case-search)		;no-op, Shut up XE ByteComp
	(setq fold-case-search t))


    (while (and (not file)
		alist)

      (setq elt		(car alist)
	    name        (nth 0 elt)
	    re		(nth 1 elt)
	    complete	(nth 2 elt)
	    )


      ;;  Is there match ?
      (setq word (strmatget
		  (concat re file-re)
		  1 word-now))

;;;      (d! re word-now word)

      ;; Is there check func for this match ?
      ;;
      (if word
	  (if (not (eq 4 (length elt)))
	      (setq stat t)
	    (setq stat (funcall (nth 3 elt) word-now))))

      (cond
       ((stringp stat)			;check func gave us filename ?
	(setq file stat))
       ((and word
	     (setq ans
		   (completing-read
		    (concat name " refers to: ")
		    (list-to-assoc-menu (eval complete))))

	     (if (var-nil-p ans)
		 nil			;can't use the answer
	       (setq file
		     (concat
		    (ti::s-verify-ends (ti::s-sqz ans) "/")
		    word
		    )))
	   )))
      (NEXTP alist)
      )
    file
    ))

;;}}}
;;{{{ find: archie handling



;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tifh-ff-archie-at-point ()
  "This command makes sense to be called inside archie listing only.
Gets file on the line where cursor is sitting on. Looks back for lines

    Host emory.mathcs.emory.edu    (128.140.110.1)
    Location: /pub/mic/emacs

To get the hostname and directory information.
"
  (interactive)
  (let* ((dir-re	"^[ \t]+Location:[ \t]+\\(.*\\)")
	 (host-re	"^Host[ \t]+\\([^ \t]+\\)[ \t]+[(]")
	 file dir host
	 ange
	 p1 				;points
	 )
    (save-excursion

      ;;   The filename is always at the _right_ in the archie listing.
      (end-of-line)
      (setq p1 (point))

      (skip-chars-backward "^ \t")
      (setq file (buffer-substring (point) p1))

      (beginning-of-line)
      (if (looking-at ".*DIRECTORY")
	  (strcat file "/"))



      (if (re-search-backward dir-re nil t)
	  (setq dir (read-match 1)))

      (if (re-search-backward host-re nil t)
	  (setq host (read-match 1)))

      (cond
       ((not (and file dir  host))
	(message-quick "Sorry, couldn' parse this archie output.")
	nil)
       (t
	(setq ange (concat "/ftp@" host ":" dir "/" file))
	ange
	))
      )))

;;}}}
;;{{{ find: lcd handling

;;; ----------------------------------------------------------------------
;;;
(defun  tifh-ff-lcd-make-filename ()
  "Makes Ange-frp filename from the LCD archive entry that looks
like this:

    ;; LCD Archive Entry:
    ;; tinytab.el|Jari Aalto|jari.aalto@poboxes.com|
    ;; TAB handling minor mode, eg. step 4 chars bck/fwd,insert/del 4 spaces |
    ;; 20-Jan-1996|$Revision: 1.28 $|~/misc/tinytab.el.Z|

Searches from current point forward to find words \"LCD Archive Entry\"
If cannot find forward, searches backward.

Returns
  string  ,ange-filename
  nil     ,no entry detected.
"
  (let* ((case-fold-search  nil)	;case is important
	 (re		    "LCD Archive Entry")
	 (path              tifh-:ff-elisp-site)
	 ret
	 )
    (save-excursion
      (cond
       ((re-search-forward  re nil t)
	;;  Look for last line where the location is told
	;;
	(if (and (re-search-forward "199.\|[^|]+\|~\\([^|]+\\)" nil t)
		 (setq ret (read-match 1)))
	    (setq ret (concat path ret ))
	  )))
      ret
      )))

;;}}}
;;{{{ find: main


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tifh-ff-1  (file &optional no-confirm)
  "Low level find file function."
  (let* ((re-bg-load		tifh-:ff-bg-load-maybe)
	 (down-dir		tifh-:ff-bg-default-download-dir)
	 to-dir				;where to download when bg
	 )

    (if (not (tifh-ff-ref-syntax-ok-p file))
	(error "invalid ref: %s" file))

;;;    (setq F file RE re-bg-load)
;;;    (d! no-confirm (string-match re-bg-load file) file)

    (cond
     (no-confirm
      (find-file file))
     (t
      (if (not (and (string-match re-bg-load file)
		    (y-or-n-p "load in background? ")))
	  (progn
	    (find-file file)		;normal load
	    )
        ;; ................................................... bg load ...
	(setq to-dir
	      ;;  Hack, read directory easily
	      (let ((default-directory down-dir))
		(call-interactively
		 '(lambda (dir)
		    (interactive "Ddownload dir: ")  dir))))

	    (if (not (file-exists-p to-dir))
		(message "Download dir invalid")
	      (ti::f-ange-download-file file to-dir))
	    )))
    ))


;;; ----------------------------------------------------------------------
;;; - You just sit on current line and shoot... voila!
;;; - Geex, this function is long. But the cond cases are short...
;;;
;;;###autoload
(defun  tifh-ff-at-point (&optional arg)
  "Reads filename at current line/point and uses ange-ftp to get the file.

_Clear_the_line_ and hit RET if you're offered a prompt and you wan't to
ignore the action. If there is more choices available they are
offered after that.

ARG
 C-u		,repeat and edit last file reference
 nbr		,do not confirm loading

Rerefences:
  tifh-:ff-*	,variables


Detects following syntaxes:

*)

        /ftp.netcom.com/pub/pe/pearl
	 ftp.netcom.com:/pub/pe/pearl
        /ftp.netcom.com:/pub/pe/pearl
	/anonymous@ftp.netcom.com:/pub/pe/pearl
	/anonymous@ftp.netcom.com:          /pub/pe/pearl


	ftp://ftp.netcom.com/pub/pe/pearl
	ftp://anonymous@ftp.netcom.com/pub/pe/pearl


*) For these references, point must be at the beginning of
   site name. And there can be only whitespace at the end.

	ftp.netcom.com	       /pub/pe/pearl
	ftp.netcom.com	       /pub/pe/pearl   file.xx

*) Relative filenames are detected.
   See variable tifh-:ff-word-to-site-ref-list for more about setup.

	~/user/mike/file.xx
	SimTel/vendors/gnuish

*) If cursor under word has: CHARS DOT CHARS, then it is assumed
   that file _may_ be a site name. User is then asked for possible
   destination with the word under cursor converted to ange reference.

	ftp.netcom.com  --> /ftp@ftp.netcom.com:/

   To cancel this, just empty the line and press return.


*) Possible archie listing output is also detected. The listing should
   be in following format:


  Host ftp.difi.unipi.it    (131.114.8.130)
  Last updated 05:13  6 Feb 1996

    Location: /pub/Win3
      FILE    -rwxr-----   77179 bytes  21:00 16 Nov 1994  mapedit.zip

   The cursor must over the line \"FILE\"


*) The elisp style lcd headers are rocognized too. The point must be
   before the words 'LCD Archive Entry:'

    ;; LCD Archive Entry:
    ;; tinytab.el|Jari Aalto|jari.aalto@poboxes.com|
    ;; TAB handling minor mode, eg. step 4 chars bck/fwd |
    ;; 20-Jan-1996|$Revision: 1.28 $|~/misc/tinytab.el.Z|


"
  (interactive "P")
  (let* ((nlist			tifh-:ff-newsgroup-to-archive)
	 (file-charset		tifh-:ff-file-charset)
	 (re-no-site		tifh-:ff-no-site-regexp)
	 (bfn			(buffer-name)) ;shorter name
	 (loop                  t)
	 (case-fold-search	t)



	 ;;  - It's possible that some poster has mispelled the URL, so
	 ;;    that it isn't recognized.
	 ;;  - As a last resort, we try to match current word against
	 ;;    this. If match is found offer user the string for
	 ;;    modifying.
	 ;;
	 (edit-re		(concat
				 "ftp:\\|//\\|"
				 "tar\\|gz\\|zip\\|\.ps"))


	 ;;  Form site name with these chars.
	 (site-word-re          "-a-z.0-9")

	 ;;  The ange-ftp like address is not valid in some buffers,
	 ;;  like gnus *Summary*
	 ;;       428: [ 20:merlyn@stoneh]
	 ;;
	 (ign-ange		"Summary\\|Newsgroup\\|Buffer menu")

	 ;;   - Reading filename must not match some characters, eg. the
	 ;;     reference may be inside quotes...
	 ;;   - Too bad that using NOT operator ^ makes regexps slow...
	 ;;
	 (file-re	(concat "\\([" file-charset "]*\\)"))

         ;; .................................................. @ and : ...

	 ;;  Accept spcaes too, site and dir is more easily seen...
	 ;;     /anonymous@ftp.netcom.com:          /pub/pe/pearl
	 ;;
	 ;;
	 (re-ange  (concat
		    "\\(/[a-zA-Z0-9]+@[^ \t]+:\\)[ \t]*"
		    "[/~]" file-re ))

	 ;;    without the @ tag
	 ;;    /ftp.netcom.com:/pub/pe/pearl
	 ;;    /ftp.netcom.com:~ftp/pub/pe/pearl
	 ;;    /ftp.netcom.com:          /pub/pe/pearl
	 ;;
	 (re-ange2  (concat
		     "\\(/?[a-zA-Z0-9]+[^ \t]+:\\)[ \t]*"
		     "\\([/~]\\)" file-re))


	 ;;     ftp.netcom.com       /pub/pe/pearl
	 (re2   "\\([-0-9a-z.]+\\)[ \t]+\\(/[-/0-9a-z.]+\\)")

	 ;;	ftp.netcom.com       /pub/pe/pearl   file.xx
	 (re3   (concat
		 "\\([-0-9a-z.]+\\)[ \t]+"
		 "\\(/[-/0-9a-z.]+\\)[ \t]+"
		 "\\([-/0-9a-z.]+\\)"))

	 ;;	possible ange name at front?
	 ;;	ftp@ftp.netcom.com       /pub/pe/pearl
	 ;;
	 (re2a  "\\([-0-9a-z]+@[-0-9a-z.]\\)[ \t]+\\([-/0-9a-z.]+\\)")

	 ;;	ftp@ftp.netcom.com       /pub/pe/pearl   file.xx
	 (re3a  (concat
		 "\\([-0-9a-z]+@[-0-9a-z.]\\)[ \t]+"
		 "\\(/[-/0-9a-z.]+\\)[ \t]+"
		 "\\([-/0-9a-z.]+\\)"))


	 ;;	/ftp.netcom.com/pub/pe/pearl
	 ;;	There must be 2-3 chars at the end of site name
	 ;;
	 (re-minimal	(concat
			 "/\\([a-z]+\\.[a-z0-9]+\\.[^/]+[a-z][a-z][a-z]?\\)"
			 "/" file-re))


	 (re-archie	(concat "^[ \t]+\\(FILE\\|DIRECTORY\\)"
				"[ \t]+[-dlrwx]+[ \t]+[0-9]+"))

	 ;;		The syntax of this line is _very_ strict
	 ;;		User has error in his lcd entry if it isn't
	 ;;		exactly like this.
	 (re-lcd	"^;;+ LCD Archive Entry:[ \t]*$")
	 (re-mail	"<\\(.*@[^/]+\\)>")

	 (line		(read-full-line))
	 (line2		(buffer-substring (point) (EOLP)))
	 (word-now	(save-excursion
			  (or (ti::b-read-word file-charset) "" )))
	 (word-now2	(save-excursion (or (ti::b-read-space-word) "")))

	 (D		"")		; for tracking down errors
	 (no-confirm    (integerp arg))
	 point
	 tmp
	 word
	 file
	 done				;a flag
	 dir
	 )
;;;    (d! "WN" word-now word-now2)

    ;; ... ... ... ... ... ... ... ... ... ... ... ... ... save values ...
    ;; You can look at this in you own ff-hook handling function.
    ;;

    (setq tifh-:ff-at-point-arg arg)


    ;; ... ... ... ... ... ... ... ... ... ... .. step through choices ...
    ;; - The order of this COND is very important, the most restrictive
    ;;   matches should be first.
    ;; - The while gives a chance to whirl this cond as many
    ;;   times as needed

    (while loop
      (setq loop nil)

    (cond
     ((string-match "h[ht]tp[ \t]*://" word-now)
      ;;  <URL:hhtp://www.ifi.uio.no/~larsi/ding-old>
      ;;  - This reference does not belong to our league. We handle
      ;;    ftp pointers only.
      ;;  - the http: or hhtp: mispelling is recognized
      (if tifh-:ff-url-hook
	  (run-hooks 'tifh-:ff-url-hook)
	(message "tifh-:ff-url-hook is nil."))
      (setq done t)
      )

     ;; ... ... ... ... ... ... ... ... ... ... ... . re-edit last ref  ..
     ((and (listpp arg)			;use last one
	   (stringp tifh-:ff-last-ange-reference)) ;first time, no ref
      (setq D (concat D "last "))
      (setq file tifh-:ff-last-ange-reference))

     ;; ... ... ... ... ... ... ... ... ... ... ... ... ... ...  email  ..
     ((and (string-match re-mail word-now2)
	   (setq done  (tifh-mail-at-point))))


     ;; ... ... ... ... ... ... ... ... ... ... ... ... ... completion  ..
     ((setq file (tifh-ff-guess word-now))
      (setq D (concat D "fancy ")))


     ;; ... ... ... ... ... ... ... ... ... ... ... ... ... ... ftp://  ..
;;;     ((progn
;;;	(setq L line A  (concat "ftp://" file-re))
;;;	(d! (strmatget A 0 line))
;;;	nil))

     ((setq word (strmatget (concat "ftp://" file-re)   ;; it's url
			    0 line))
      (setq file (ti::c-url-to-ange-ftp word))
      (setq D (concat D "ftp:// "))
      )

     ;; ... ... ... ... ... ... ... ... ... ... ... ... ... ... archie  ..
     ((and (string-match re-archie line)	;pass control to another func
	   (setq file (tifh-ff-archie-at-point)))
      (strcat D "archie ")
      )

     ;; ... ... ... ... ... ... ... ... ... ... ... ... ... ... . ange  ..
     ((and (not (string-match ign-ange bfn))
	   (string-match re-ange line))	;ange ftp type filename
      (setq D (concat D "re-ange "))
      (setq file (concat (read-match 1 line)
			 (read-match 2 line)
			 (read-match 3 line)))
      )


     ;; ..................................................................
     ((and (not (string-match ign-ange bfn))
	   (string-match re3a line2))
      (setq D (concat D "re3a "))
      (setq file (concat (read-match 1 line2)  ":"
			 (ti::s-verify-ends
			  (read-match 2 line2) "/")
			 (read-match 3 line2)
			 )))

     ;; ..................................................................
     ((and (not (string-match ign-ange bfn))
	   (string-match re2a line2))
      (setq D (concat D "re2a "))
      (setq file (concat  (read-match 1 line2)  ":"
			  (read-match 2 line2)
			  )))

     ;; ..................................................................
     ((and (not (string-match ign-ange bfn))
	   (string-match re3 line2))
      (setq D (concat D "re3 "))
      (setq file (concat "/ftp@"
			 (read-match 1 line2)  ":"
			 (ti::s-verify-ends
			  (read-match 2 line2) "/")
			 (read-match 3 line2)
			 )))

     ;; ..................................................................
     ((and (not (string-match ign-ange bfn))
	   (string-match re2 line2))
      (setq D (concat D "re2 "))
      (setq file (concat  "/ftp@"
			  (read-match 1 line2)  ":"
			  (read-match 2 line2)
			  )))


     ;; ..................................................................
     ((and (not (string-match ign-ange bfn))
	   (string-match re-ange2 line))
      (setq D (concat D "re-ange2 "))
      (cond
       ((match-end 3)			;has spaces ?
;;;	(d! (read-match 0 line) ) (d! (read-match 1 line) )
;;;	(d! (read-match 2 line) ) (d! (read-match 3 line) )
;;;	(d! (read-match 4 line) )

	;;  leave "/" out from the site name
	;;
	(setq tmp (read-match 1 line))	;site name
	(setq file (concat
		    (read-match 2 line)		;; [/~]
		    (read-match 3 line)))	;; dir

;;;	(d! "a2-/" tmp)

	;;  Remove "/" at front.
	(if (and (string= "/" (substring tmp 0 1))
		 (not (string-match "@" tmp)))
	    (setq tmp  (substring tmp 1)))

	;;  /ftp.netcom.com: + ~ftp/pub/pe/pearl
	;;  -->             "/"
	;;
	(if (and (not (string= "/" (str-right tmp 1)))
		 (not (string= "/" (str-left file 1))))
	    (setq tmp  (concat tmp "/")))


	;;  Add missing "/" the beginning of string if needed
	;;  ftp.com:/dir  --> /ftp.com:/dir
	;;
	(if (not (string= "/" (str-left tmp 1)))
	    (setq tmp  (concat "/"  tmp)))



	(setq file (concat tmp file ))

;;;       (d! "a2-2" tmp file)
       )

       (t
	;;  does it have leading "/site.xx.com"
	;;
	(if (char= ?/ (substring line 0 1))
	    (setq line  (substring line 1)))
;;;	(d! "a22" line)
	(setq file (concat "/ftp@" line))
	)))


     ;; ..................................................................
     ((and (not (string-match ign-ange bfn))
	   (string-match re-minimal line))
      (setq D (concat D "mini "))
      (setq file (concat
		  (read-match 1 line) ":/" (read-match 2 line)
		  )))



     ;; ......................................................... site ...
     ;;  - Now read site name under cursor and ask user about the ftp
     ;;  - The sitename has 3 chars at the end:  .com or .fi
     ;;    otherwise it can't be valid site.
     ;;  - Notice the last .? regexp, it is needed due to puctuation
     ;;    in spoken language:
     ;;
     ;;      I could not find it at ftp.cs.indiana.edu.
     ;;
     ;;  - The site ".el" is not counted because it's normally word
     ;;    referring to elisp packages.
     ;;
     ((and  (string-match "[a-z]+\\.[a-z][a-z][a-z]?[.,]?$" word-now)
	    (not (string-match re-no-site word-now))
	    (setq word (ti::b-read-word site-word-re))

	    (progn			;drop punctuation
	      (if (memq (string-to-char (str-right word 1))
			'(?. ?,))
		  (setq word (substring word 0 (1- (length word)))))
	      t)
	    ;;  Sometimes user may write
	    ;;    is available from space.mit.edu in pub/davis/jed.
	    ;;                                  ^^^^^
	    ;;  Lets' peek if there is directory name after site name.
	    (progn
;;;	      (d! (buffer-substring (point) (EOLP)))
	      (cond
	       ((and (setq dir (bufmatget  ".*in +\\([^., \t]+\\)" 1))
		     (setq point (string-match "/" dir)))

		;;  Remove beginning slash, see the prompt...later
		(if (eq point 0)
		    (setq dir (substring dir 1)))
		)
	       (t
	       ;;  it didn't contain directory slash... discard
		(setq dir "")))
	      t)


;;;	    (d! word-now word (point))
	    (not (var-nil-p
		  (setq word
		       (read-from-minibuffer
			"Ftp to this site ? "
			 (concat "/ftp@" word ":/" dir)))))
	    )
      (setq file word))


     ((save-excursion
	(re-search-forward re-lcd nil t))
      (setq D (concat D "LCD "))
      (if (null (setq file (tifh-ff-lcd-make-filename)))
	  (message-quick "LCD entry is invalid...")))


     ((and (not (string-match ign-ange bfn))
	   (string-match "news\\|gnus" (symbol-name major-mode))
	   (require 'tinylibmt)
	   (setq word (list-find
		       nlist
		       (or (ti::mt-news-group) "#arggh")
		       '(and (string-match (car element) arg)
			     (setq element (nth 1 element))
			     ))))
      ;; We found associated archive for this newsgroup
;;;      (d! bfn ign-ange)
      (setq D (concat D "news "))
      (setq file word)
      )

     ;; ................................................. editing word ...
     ((and (string-match edit-re word-now)
	   (y-or-n-p (concat "Missed it, edit [" word-now "] " )))
      (if (var-nil-p (setq word
			   (read-from-minibuffer "correcting: " word-now)))
	  (setq file nil)
	;;  Re-evaluate it !
	(setq word-now word word-now2 word line word)
	(setq loop t))
      )

     )) ;; COND END, LOOP END


    ;; ===================================================== Main load ===
;;;    (setq L line W word-now) (d! word-now D file line )

    (cond
     (done				;This just stops the COND statement.
      nil
      )

     ((stringp file)
      (setq file (tifh-ff-ange-padd-reference file)) ;make sure it has "/ftp@"

      ;;  This is no-op if file was not found
      (if (or no-confirm		;go straight ahead
	      (not (var-nil-p
		    (setq file (read-from-minibuffer "find: " file)))))
	  ;; .............................................. do file load ...
	  (progn
	    (setq file (ti::s-sqz file)) ;no blanks around, avoid mistakes
	    ;;  Save the reference
	    (setq tifh-:ff-last-ange-reference file)
	    (tifh-ff-1 file no-confirm)
	    )))
     (t
      ;; ............................................. no file, no url ...
      (if (null tifh-:ff-not-found-hook)
	  (message (format "Can't detect full remote location in line.")))


      (run-hooks-until 'tifh-:ff-not-found-hook)
      ))

    ))

;;}}}

;;{{{ mail

;;; ............................................................ &mail ...


;;; ----------------------------------------------------------------------
;;; - This is tried if find-file fails
;;;
(defun tifh-mail-at-point  ()
  "Mails author, who's mail address is at current line.

Return:
 nil	,not mailed
 t
"
  (interactive)
  (let* ((re		"<\\(.*@[^/]+\\)>")
	 (line		(read-full-line))
	 (email		(strmatget re 1 line))
	 )
    (cond
     ((and email
	   (y-or-n-p (concat "Really mail " email "  ")))
      (mail) (PMIN)  (re-search-forward "To: *")
      (insert email)
      (PMIN) (re-search-forward "Subject: *")
      t					;ret val
      )
     (t
      nil
      ))
    ))

;;}}}
;;{{{ examples


;;; ........................................................ &examples ...
;; example for to use tifh-:ff-not-found-hook
;;
;;
;;* _
;;* (defconst my-:ange-ff-completions
;;*   '(
;;*     ;;  The Emacs elisp site
;;*     "archive.cis.ohio-state.edu:pub/gnu/emacs/elisp-archive/"
;;* _
;;*     ;;  The one and only SIMTEL
;;*     "ftp.coast.net:/SimTel/msdos/"
;;* _
;;*     ;; The perl CPAN achive
;;*     "ftp.delphi.com/pub/mirrors/packages/perl/CPAN/"
;;* _
;;*     ;; huge CICA site
;;*     "ftp.funet.fi:/pub/msdos/windows"
;;* _
;;*     )
;;*   "*My possible find file completions.")
;;* _

;;* _
;;* (defun  my-find-file-interactive (&optional file)
;;*   "Completes Ange files"
;;*   (interactive
;;*    (list
;;*     (let ((f   (my-guess-find-file))
;;* 	  )
;;*       (setq f (completing-read
;;* 	       "%find file: "
;;* 	       (list-to-assoc-menu my-:ange-find-file-completions)
;;* 	       nil nil
;;* 	       (or "/o" f)
;;* 	       ))
;;*       f
;;*       )))
;;* _
;;*   (defvar my-:my-find-file-interactive nil
;;*     "Last argument of function.")
;;* _
;;* _
;;*   ;;  This allows hook to call us, force interactive
;;*   (cond
;;*    ((null file)				;not already set by interactive?
;;*     (setq file (completing-read
;;* 		"%find file: "
;;* 		(list-to-assoc-menu my-:ange-find-file-completions)
;;* 		nil nil
;;* 		(or (and (not (null tifh-:ff-at-point-arg))
;;* 			 my-:my-find-file-interactive)
;;* 		    (my-guess-find-file)
;;* 		    "/o")
;;* 		))))
;;* _
;;*    (cond
;;*     ((not (var-nil-p file))
;;*      (setq my-:my-find-file-interactive file) ;save last arg
;;*      (if (featurep 'tinyfh)
;;* 	 (setq tifh-:ff-last-ange-reference file))
;;*      (find-file file)
;;*      ))
;;*    )

;;}}}

(provide   'tinyfh)
(run-hooks 'tifh-:load-hook)

;;; tinyfh.el ends here
