;; @(#) tinylibmt.el -- Library of (m)ail and news (t)ool functions

;; @(#) $Id: tinylibmt.el,v 1.87 1997/06/30 13:20:37 jaalto NotReleased $
;; @(#) $Keywords: mail library, general functions $
;;
;; $KnownCompatibility: 19.28 $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1995,96,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Sep 2 1995
;;
;; To get information on this program use ident(1) or do M-x tilibmt-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el


;; LCD Archive Entry:
;; tinylib|Jari Aalto|jari.aalto@poboxes.com|
;; library of usefull (M)ail/News (T)ool functions|
;; 30-Jun-1997|$Revision: 1.87 $|~/misc/tinylibmt.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ........................................................ &t-install ...
;; Put this file on your Emacs-Lisp load path, add following into your
;; ~/.emacs startup file
;;
;;     (require 'tinylibm)
;;
;; No, there is no mistake here. The 'm' lib will have all autoloads
;; to this package. If you use macros, then you must require this
;; library manually.

;;}}}
;;{{{ Documentation

;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "ti::mt" in front of
;;   every function, variables have "ti:mt-". It stands for '(ti)ny
;;   library for (m)ail (t)ool fuctions.

;; Preface, 1995
;;
;;	o   This is library, so the package itself does nothing.
;;	o   Collection of functions to deal with Mail/News specific tasks.
;;
;;	After my own mail handling package [tinyreply] grow into
;;	considerable size I started thinking that there was no point of
;;	stuffing everything in it, instead I should transfer all general
;;	functions to common library.
;;
;;	I hope this library will make it easier to pick up the
;;	function user wants to use. Some of the functions may seem
;;	alike with the std distrib mail-utils.el, but if you look closely,
;;	you'll see that my functions are much more powerfull than them.
;;	I would have wanted to use tha std functions, but since they didn't
;;	offer what I needed, I had to make my own.
;;
;;  Code policy
;;
;;	All functions are written by [jari], unless otherwise notified.
;;	If I have found interesting function from some other .el I have
;;	written a reference to that .el. If the code is directly taken
;;	'almost as is' all the credit belongs to the Author of the
;;	original code. Cheers for him, to write so recpectable code, that
;;	it can be used right from the box!
;;
;;	If someone contributes good, general purpose function, all the
;;	credit belongs to him automatically. Persons name and date and so
;;	on is included.
;;
;;	See the main library [tinylib] for more docs.

;;}}}
;;{{{ history

;; ........................................................ &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; Jun	30	1997	[jari]		19.28	v1.86-1.87	NotReleased
;; - Removed lisp loop calls and used dolist.
;;   ti::mt-pgpr-close			:+
;;   ti::mt-sendmail-reset-send-hooks	:+
;;   ti::mt-sendmail-pure-env-macro	:+
;;   ti::mt-kill-field-in-body		:+
;;
;; Jun	26	1997	[jari]		19.28	v1.85		NotReleased
;;   ti::mt-pgp-block-area-kill-forward :+
;;   ti::mt-pgp-block-area-kill-forward :! fixed.
;;
;; Jun	25	1997	[jari]		19.28	v1.82-1.84	NotReleased
;; - Renamed all PGP stream functions.
;;   ti::mt-pgp-stream-forward-info	:+
;;   ti::mt-mime-maybe-p		:+
;;   ti::mt-pgp-signature-remove	:! now removes signed headers too.
;;   ti::mt-pgp-stream-data-elt		:! updated
;;   ti::mt-message-length		:! corrected (ignores empty lines)
;;
;; Jun	25	1997	[jari]		19.28	v1.81		NotReleased
;;   ti::mt-body-empty-p		:+
;;   ti::mt-body-clear			:+
;;   ti::mt-pgp-pkey-read		:+
;;   ti::mt-pgp-stream-study-pring	:+ a bit incomplete
;;   ti::mt-pgp-pkey-read		:! fine tuning
;;   ti::mt-pgp-stream-study	:! cursor positioning error
;;   ti::mt-pgp-signed-p		:! added 'pkeye choice
;;
;; Jun	24	1997	[jari]		19.28	v1.80		NotReleased
;; - Still more correction to text-start and hmax functions
;;   ti::mt-pgp-header-kill-in-body	:+
;;   ti::mt-pgp-signature-header-info	:+
;;   ti::mt-pgp-signed-xpgp-p		:+
;;   ti::mt-pgp-signature-remove	:+
;;   ti::mt-pgp-signed-p		:! split
;;   ti::mt-text-start			:! now check mail-p
;;   ti::mt-hmax			:! now check mail-p
;;   ti::mt-pgp-signed-p		:! added 'sige choice
;;
;; Jun	18	1997	[jari]		19.28	v1.77-1.79	NotReleased
;;  ti::mt-pgp-stream-data-elt			:+
;;  ti::mt-next-field-start		:! didn't work right in the last fld
;;  ti::mt-hmax				:! wrong end position.
;;  ti::mt-text-start			:! wrong position
;;
;; Jun	16	1997	[jari]		19.28	v1.76		NotReleased
;;   ti::mt-hmax			:! now returns point-min
;;
;; Jun	11	1997	[jari]		19.28	v1.75		NotReleased
;;   ti::mt-pgp-stream-study-1-ver	    :+
;;   ti::mt-pgp-stream-study-1-key-id  :+
;;   ti::mt-pgp-stream-study-1-time    :+
;;   ti::mt-pgp-stream-study-signed    :+
;;
;; Jun	9	1997	[jari]		19.28	v1.72-1.74	NotReleased
;; - Spent some time last weekend on the lwn, a sunny day, 30C and hot,
;;   and read the PGP data packet documentation. I finally decided
;;   to start writing support fucntions for it.
;;   ti::mt-sendmail-macro		    :+ set-buffer correction
;;   ti::mt-pgp-data-string-to-bin-string   :+
;;   ti::mt-pgp-data-bin-string-to-int-list :+
;;   ti::mt-pgp-data-ascii-armor-convert    :+
;;   ti::mt-pgp-data-study-ctb-byte	    :+
;;   ti::mt-pgp-stream-study-enc	    :+
;;   ti::mt-pgp-stream-study	    :+
;;   ti::mt-pgp-signed-conventional-p	    :+
;;   ti::mt-message-length		    :! Now sees PGP too
;;
;; Jun	6	1997	[jari]		19.28	v1.71		NotReleased
;;   ti::mt-signature-insert-break	:+
;;   ti::mt-pgp-signature-header-info-v3xx	:! Now return "Comment" too
;;   ti::mt-sendmail-macro		:! added save window excursion
;;
;; Jun	5	1997	[jari]		19.28	v1.67-1.70	NotReleased
;; - I removed some functions from 'm' lib and neede to make the
;;   modifications here too. Internal changes.
;; - Checkdoc 1.42 clear.
;;   ti::mt-add-field			:! corrected. More consistent FIELD in
;;
;; May	22	1997	[jari]		19.28	v1.65-1.66	NotReleased
;; - Checkdoc 1.22 documentation string corrections.
;;   ti::mt-mh-macro			:! running in MH message buffer
;;   ti::mt-get-field			:! added option 'pure
;;
;; May	20	1997	[jari]		19.28	v1.64		NotReleased
;; - Removed all nreverse calls and rearranged som PGP functions.
;; - I have a feeling that some functions broke during adjustment.
;;   ti::mt-pgp-signature-normal-info		:!
;;   ti::mt-pgp-signature-header-info-v3xx	:!
;;   ti::mt-mime-parse-header		:!
;;   ti::mt-trim-buffer			:! makes sure there is newline
;;
;; May	14	1997	[jari]		19.28	v1.62-1.63	NotReleased
;; - internal changes.
;; - Changed all non standard boolean tests to use when and unless
;;   ti::mt-trim-buffer			:+
;;
;; Mar	24	1997	[jari]		19.28	v1.61		NotReleased
;;   ti::mt-rmail-copy-message		:+
;;   ti::mt-parse-received		:+
;;   ti::mt-rmail-do-message-macro	:! new mode parameter
;;
;; Mar	23	1997	[jari]		19.28	v1.58-1.60	NotReleased
;; - Changed ti:mt-required-headers to include references.
;;   ti::mt-pgp-comment-file-p		:+
;;   ti::mt-add-field			:! bug, corrected
;;
;; Mar	19	1997	[jari]		19.28	v1.55-1.57	NotReleased
;; - XE20 character handling changes.
;;   ti::mt-gnus-macro			:+
;;   ti::mt-kill-field-elt		:+
;;   ti::mt-email-find-region		:! new parameter no-dupes
;;
;; Mar	7	1997	[jari]		19.28	v1.53-1.54	NotReleased
;;   ti::mt-vm-macro			:+
;;   ti::mt-rmail-macro			:+
;;   ti::mt-rmail-do-message-macro	:+
;;   ti::mt-header-area-size		:+
;;
;; Feb	25	1997	[jari]		19.28	v1.51-1.52	NotReleased
;;   ti::mt-pgp-exe-version-number	:+
;;   ti::mt-field-start			:! bug if on f-start line already
;;   ti::mt-sendmail-macro		:! now resets hooks
;;
;; Feb	16	1997	[jari]		19.28	v1.44-1.50	NotReleased
;;   defsubst --> defun			: so that they can be autoloaded
;;   ti::mt-yank			:+ general yank function.
;;   ti::mt-add-to-field-string		:+
;;   ti::mt-pgp-data-type		:+
;;   ti::mt-nymserver-email-convert		:+
;;   ti::mt-pgp-signed-p		:! still corrections
;;   ti::mt-hmax			:! infinit loop
;;
;; Jan	17	1997	[jari]		19.28	v1.42-1.43	NotReleased
;;   ti::mt-sendmail-macro		:+
;;   ti::mt-pgp-v3xx-p			:! bugs
;;   ti::mt-pgpr-anonymize-headers	:! bugs
;;   ti::mt-text-start			:! bugs
;;
;; Dec	16	1996	[jari]		19.28	v1.40		NotReleased
;;   ti::mt-pgpr-parse-levien-list	:+
;;   ti::mt-get-field-1			:+ faster way to get field content
;;   ti::mt-kill-non-rfc-fields		:! corrected
;;   ti::mt-field-start			:! corrected
;;   ti::mt-next-field-start		:! corrected
;;   ti::mt-pgpr-anonymize-headers	:! new parameters
;;   ti::mt-pgpr-block			:! lot of bugs
;;
;; Dec	3	1996	[jari]		19.28	v1.39		NotReleased
;;   ti::mt-pgpr-reply-type		:+
;;   ti::mt-pgpr-latent-p		:+
;;   ti::mt-pgpr-parse-levien-list	:+
;;   ti::mt-pgpr-block			:! args changed
;;
;; Dec	1	1996	[jari]		19.28	v1.38		NotReleased
;;   ti::mt-pgp-signature-header-info-v3xx	:+ now ready
;;   ti::mt-mime-parse-header		:+ now ready
;;   ti::mt-pgp-v3xx-p			:+
;;   ti::mt-pgp-signed-p		:! bug corrected
;;
;; Nov	27	1996	[jari]		19.28	v1.34-1.37	NotReleased
;; - Optimized all let* --> let. This gains speed by 0.7 - 2X
;; - There was serious bug in one function that caused infinite loop.
;;   I returned back to old version, but this really doesn't solve
;;   the problem of too far lookup that this function suffers.
;;   #todo: fix it some other day better
;;   ti::mt-pgp-block-area		:! new parameter 'signed
;;   ti::mt-pgp-signed-p		:! X-Pgp detection bug
;;   ti::mt-field-start			:!
;;   ti::mt-pgp-encrypted-p		:! new parameter
;;   ti::mt-email-find-region		:! now returns them as read.
;;
;; Nov	25	1996	[jari]		19.28	v1.33		NotReleased
;;   ti::mt-pgp-public-key-p		:+
;;   ti::mt-email-from-string		:! bug, updated regexp
;;   ti::mt-field-start			:! bug, now limited to hdrs
;;
;; Nov	20	1996	[jari]		19.28	v1.32		NotReleased
;;   ti::mt-pgp-trim-buffer		:+
;;   ti::mt-pgp-chop-region		:+
;;   ti::mt-pgp-re-search		:! corrected
;;   ti::mt-pgpk-public-get-region	:! now has RELAX
;;   ti::mt-pgp-msg-end-line		:! bug in regexp
;;   ti::mt-pgp-signed-p		:! failed in X-Pgp headers case
;;   ti::mt-pgp-re			:! arg 'all --> 'any
;;   ti::mt-pgp-block-area		:! bug in finding the area
;;
;; Nov	14	1996	[jari]		19.28	v1.31		NotReleased
;;   ti::mt-mail-p			:! bug
;;   ti::mt-pgp-sig-header-info		: X-pgp old v2.xx
;;   ti::mt-mime-parse-header		: X-pgp new v3.xx, unfinished
;;
;; Nov	10	1996	[jari]		19.28	v1.30		NotReleased
;;   ti::mt-mail-p			:+
;;   ti::mt-pgpk-public-get-region	:+
;;   ti::mt-email-find-region		:! bug
;;   ti::mt-hmax			:! serious bug
;;   ti::mt-pgp-p			:! serious bug
;;   ti::mt-pgp-block-area		:! now supports "- -" TAGS
;;   mt-pgp-re-search			:! new features
;;
;; Nov	8	1996	[jari]		19.28	v1.29		NotReleased
;;   ti::mt-pgp-pkey-begin-line		:+
;;   ti::mt-pgp-pkey-end-line		:+ ... and other tags added
;;   ti::mt-message-length		:+
;;   ti::mt-pgp-signed-p			:+
;;   ti::mt-pgpk-id-lines-in-region	:+
;;   ti::mt-pgpr-block			:+
;;   ti::mt-pgpr-reply-block		:+
;;   ti::mt-email-from-string		:+
;;   ti::mt-email-find-region		:+
;;   ti::mt-pgp-p			:! corrected, bug
;;   ti::mt-pgp-encrypted-p		:! corrected
;;   ti::mt-pgp-normal-p		:! rewritten
;;   ti::mt-pgp-search-begin		:! docs
;;   ti::mt-pgp-block-area		:! rewritten
;;   ti::mt-pgp-re-search		:! rewritten
;;
;; Nov	3	1996	[jari]		19.28	v1.28		NotReleased
;;   ti::mt-pgp-sig-header-info		:! small error in reading Signed:
;;
;; Nov	1	1996	[jari]		19.28	v1.27		NotReleased
;; - Converted almost all defsubsts to defun, because defsubst cannot
;;   be set to autoload state.
;;   ti::mt-pgp-headers-p		:+
;;   ti::mt-pgp-re-search		:+
;;   ti::mt-pgp-signature-normal-do-region	:+
;;   ti::mt-pgp-signature-normal-info		:+
;;   ti::mt-pgp-sig-header-info		:+
;;   ti::mt-hmax			:! corrected
;;   ti::mt-text-start			:! corrected
;;
;; Nov	1	1996	[jari]		19.28	v1.24-26	NotReleased
;; - Added lib m require due to macros.
;;   ti::mt-pgp-p			:! more robust
;;   ti::mt-t-parse-name		:! added more intelligency
;;
;; Sep	25	1996	[jari]		19.28	v1.22-1.23	NotReleased
;; - Rearranged loading. Now uses autoloads.
;;
;; Sep	20	1996	[jari]		19.28	v1.21		NotReleased
;;   ti::mt-get-all-email-addresses	:+
;;
;; Sep	19	1996	[jari]		19.28	v1.20		NotReleased
;;   ti::mt-mime-p			:+
;;
;; Sep	4	1996	[jari]		19.28	v1.19		NotReleased
;; - Corrected bug in ti::mt-parse-name, where the "[]" characters
;;   in users From address caused lisp error. They are now filtered out.
;;
;; Sep	4	1996	[jari]		19.28	v1.17-1.18	NotReleased
;; - Added new pgp functions from my tinypgpg.el
;;   ti::mt-pgp-p			:+
;;   ti::mt-pgp-re			:+
;;   ti::mt-pgp-remail-p		:+
;;   ti::mt-pgp-encrypted-p		:+
;;   ti::mt-pgp-normal-p		:+
;;
;; Aug	22	1996	[jari]		19.28	v1.16		NotReleased
;; - Corrected few typos reported by ville@cena.dgac.fr. No more
;;   calls to tinylibmt via ti::l-get-match. Substituted it with
;;   list-find command.
;; - Corrected all XE 19.14 ByteCompiler errors.
;;
;; Jun	17	1996	[jari]		19.28	v1.13-15	NotReleased
;; - Made chnages according to Gnus5. I'm not sure if I got all right,
;;   and nicely, but at least it works.
;;   ti::mt-news-group			:!
;;   ti::mt-news-buffer-p		:!
;;
;; Jun	11	1996	[jari]		19.28	v1.12		NotReleased
;; - Made hexl autoload.
;;   ti::mt-add-field			:! corrected
;;
;; Jun	6	1996	[jari]		19.28	v1.11		NotReleased
;; - Bug in kill-field, due to previous version's changes.
;;   ti::mt-mail-buffer-name		:! rewrote, due to byteComp
;;
;; Jun	6	1996	[jari]		19.28	v1.10		NotReleased
;; - Supports 19.29 dynamic loading. This reduces memory consumption.
;;   ti::mt-rename-buffer		:! regexps corrected
;;   ti::mt-add-field			:! changed arg parameters
;;   ti:mt-pgp-signed-begin-line	:+
;;   ti:mt-pgp-signature-begin-line	:+
;;
;; Apr	12	1996	[jari]		19.28	v1.9		NotReleased
;; - More customizable parsing.
;;   ti:mt-parse-name-not-accept	:+
;;   ti::mt-t-parse-name		:!
;;
;; Apr	12	1996	[jari]		19.28	v1.8		NotReleased
;; - Corrected small cosmetic error in ti::mt-get-field, which moved the
;;   point unnecessarily. Doesn't move any more.
;; - removed (require gnus) and used the symbol-value trick to
;;   silence stupid byte compiler that can't see boundp tests.
;; - Very serious error corrected. There was some garbage variables
;;   from my tinyreply.el that were sneaked into this code when I moved
;;   general functions here. Corrected, ByteCompiler is now happy.
;;   eval-when-compile			:! corrected
;;   ti::mt-t-parse-name		:! new test
;;   ti::mt-parse-name			:! corrected. Added email support.
;;
;; Feb	20	1996	[jari]		19.28	v1.7		NotReleased
;; - Added two mime functions and one nice mail buffer rename.
;;   ti::mt-rename-buffer		:+
;;   ti::mt-news-reply-p		:+
;;   ti::mt-mime-qp-decode		:+
;;   ti::mt-mime-qp-prepare		:+
;;
;; Feb	10	1996	[jari]		19.28	v1.6		NotReleased
;; - Bytecomp silenced. Added eval-when-compile.
;;
;; Feb	8	1996	[jari]		19.28	v1.5		NotReleased
;; - bug in add-field. Now the "\n" is automatically added if needed.
;;
;; Feb	7	1996	[jari]		19.28	v1.4		NotReleased
;; - Rewrote feedback fuctions. Added news, anon handling functions
;;   ti::mt-get-2re			:+ this is special to this module
;;   ti::mt-news-buffer-p		:+
;;   ti::mt-news-group			:+
;;   ti::mt-set-article-buffer		:+
;;   ti::mt-get-article-re		:+
;;   ti::mt-anon-penet-p		:+
;;   ti::mt-anon-penet-to-p		:+
;;   ti::mt-anon-penet-convert		:+
;;   ti::mt-kill-field			:! bug in ret value
;;
;; Nov	15	1995	[jari]		19.28	v1.3		NotReleased
;;   ti::mt-kill-field			:! bug, limit missed
;;
;; Oct	15	1995	[jari]		19.28	v1.2		NotReleased
;;   ti::mt-TMIN			:- deleted
;;   ti::mt-narrow			:+ hdr, or text narrow
;;   ti::mt-hmax			:+
;;   ti::mt-t-parse-name		:+
;;   ti::mt-parse-name			:+
;;   ti::mt-parse-email			:+
;;   ti::mt-parse-date			:+
;;   ti::mt-add-field			:+
;;   ti::mt-required-headers		:+
;;   ti::mt-text-start			:! Corrected, new parameters
;;   ti::mt-kill-field			:! serious bug, used field-re
;;
;;
;; Sep	4	1995	[jari]		19.28	v1.1		Created

;;}}}

;;; Code:

;;{{{ setup: -- require

;;;  ....................................................... &v-require ...

(require 'tinylibm)

(eval-and-compile

  (autoload 'mail-yank-original			"sendmail")
  (autoload 'mail-send-and-exit			"sendmail")
  (autoload 'mail-setup				"sendmail")
  (autoload 'mail-mode				"sendmail")

  (autoload 'delete-rectangle			"rect")
  (autoload 'string-rectangle			"rect")
  (autoload 'hexl-hex-string-to-integer		"hexl" t nil)

  (defvar   rmail-current-message nil)
  (autoload 'rmail-msgbeg			"rmail" t t)
  (autoload 'rmail-msgend			"rmail" t t)
  )

(eval-when-compile (use-dynamic-compilation))


;;}}}
;;{{{ setup: -- private

;;; ......................................................... &v-hooks ...

(defvar ti:mt-load-hook nil
  "Hook that is run when package is loaded.")

;;; ....................................................... &v-private ...

(defvar ti:mt-ret nil
  "Global return value of this package.")

(defvar ti:mt-tmp-buffer " *ti::mt-tmp*"
  "*Temporary buffer name.")

(defvar ti:mt-mail-buffer " *ti::mt-mail*"
  "*Temporary mail buffer name.")

;;;  Variables could be modified. This makes them persistent
;;;

(defsubst ti::mt-pgp-signature-begin-line  ()
  "Signature start line."
  "-----BEGIN PGP SIGNATURE-----")

(defsubst ti::mt-pgp-signature-end-line  ()
  "Signature end line."
  "-----END PGP SIGNATURE-----")


;; Signed message has:
;;
;; -----BEGIN PGP SIGNED MESSAGE-----
;; -----BEGIN PGP SIGNATURE-----
;; -----END PGP SIGNATURE-----
;;
(defsubst ti::mt-pgp-signed-begin-line ()
  "Text for start of PGP signed messages."
  "-----BEGIN PGP SIGNED MESSAGE-----")

(defsubst ti::mt-pgp-signed-end-line ()
  "Text for start of PGP signed messages."
  (ti::mt-pgp-signature-end-line))



(defsubst ti::mt-pgp-pkey-begin-line  ()
  "PGP public key begin line."
  "-----BEGIN PGP PUBLIC KEY BLOCK-----")

(defsubst ti::mt-pgp-pkey-end-line  ()
  "PGP public key end line."
  "-----END PGP PUBLIC KEY BLOCK-----")


(defsubst ti::mt-pgp-msg-begin-line  ()
  "PGP message, typically base64 signed, begin line."
  "-----BEGIN PGP MESSAGE-----")

(defsubst ti::mt-pgp-msg-end-line  ()
  "PGP message, typically base64 signed, end line."
  "-----END PGP MESSAGE-----")


(defsubst ti::mt-pgp-any-pgp-line-regexp  ()
  "Return regexp that matched any pgp TAG."
  ;;  The lines can be broken, ehen there is encrypted/signed message
  ;;
  (concat
   "- -----\\(BEGIN\\|END\\) PGP.*-----"
   "\\|-----\\(BEGIN\\|END\\) PGP.*-----"
   ))



;;; ........................................................ &v-public ...
;;; User configurable -- but not recommended.

;;  See gnus.el or gnus-msg.el  gnus-required-headers
;;  The 'in-reply-to is for mail messages (additional)
;;
(defconst ti:mt-required-headers
  '(from date newsgroups subject path message-id in-reply-to references)
  "*All required fields that RFC977 and RFC1036 requires.
Make sure symbol names are all in lowercase.")

(defvar ti:mt-parse-name-not-accept
  (concat
   "[A-Z][/][A-Z]"	;company division PMR/TMS ?
   "\\|^[A-Z]+\\'"	;all in capitals
   "\\|^[-]$"	;single '-' word
   "\\|[.0-9]"	;maybe phone number ?
   "\\|com\\|org\\|edu"
   )
  "*Regexp to exclude non-valid people names.
We can't be sure that the names are really good names when we parse the
sender's From field. Let's see an example

	\"Person Someone p. nnn-nnn-nnn\"

There obviously isn't 3rd name, it's used for phone abbrev. And the last
word is the actual phone number.

This regexp tells which word matches are false name hits.
In this example it'd leave:
	\"Person Someone\"

See `ti::mt-parse-name'")

;;}}}
;;{{{ setup: -- version

;;; ....................................................... &v-version ...

(defconst tinylibmt-version
  "$Revision: 1.87 $"
  "Latest version number.")


(defconst tinylibmt-version-id
  "$Id: tinylibmt.el,v 1.87 1997/06/30 13:20:37 jaalto NotReleased $"
  "Latest modification time and version number.")


;;; ----------------------------------------------------------------------
;;;
(defun tinylibmt-version (&optional arg)
  "Show version information. ARG tell to print message in echo area only."
  (interactive "P")
  (ti::m-pkg-version-info "tinylibmt.el" arg))

;;; ----------------------------------------------------------------------
;;;
(defun tinylibmt-submit-feedback ()
  "Submit suggestions, error corrections, impressions, anything..."
  (interactive)
  (ti::m-pkg-submit-feedback "tinylibmt.el"))

;;}}}

;;; ########################################################### &funcs ###

;;{{{ misc


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-body-empty-p ()
  "Check if there is nothing in the body or if whole buffer is empty."
  (save-excursion
    (ti::mt-text-start 'move)
    (eq (point) (point-max))))

;;; ----------------------------------------------------------------------
;;;
(defun ti:mt-body-clear ()
  "Delete message body."
  (ti::mt-text-start 'move)
  (delete-region (point) (point-max)))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-message-length  ()
  "Return message's body length, not including the headers.

Exceptions:

  The start point defaults to `point-min' if body can't be found.

  If there is PGP signed block, then the body length is the text inside
  PGP signed block, not the whole message body."
  (let* ((end (ti::mt-pgp-signed-conventional-p))
	 beg
	 )
    (save-excursion
      (cond
       (end
	;; Do not count empty lines at the end of body
	;;
	(goto-char end) (skip-chars-backward " \t\n") (setq end (point))

	(PMIN)
	(re-search-forward (ti::mt-pgp-signed-begin-line))
	(forward-line 1)
	;;  now we're inside BODY of text, but it's not that simple yet. User
	;;  may have signed headers and they are inserterted into body
	;;  like this:
	;;
	;;  -----BEGIN PGP SIGNED MESSAGE-----
	;;
	;;  ##
	;;  Subject: See this ma!
	;;
	;;  Body text starts here.
	;;

	;;  Note, there is no spaces, becasue the body is trimmed
	;;
	(when (looking-at "\n##\n")
	  (goto-char (match-end 0))
	  (re-search-forward "^$")
	  (forward-line 1)
	  )

	(- end (point))
	)
     (t
      (setq beg (ti::mt-text-start))
      (goto-char (point-max))
      (if (eq beg (point))		;Empty buffer
	  0
	(skip-chars-backward " \t\n") (setq end (point))
	(- (point) (ti::mt-text-start))
	))))))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-get-2re (re str)
  "Use RE and match STR. Return list ('' '') if not matched."
  (let ((m1 "")
	(m2 "")
	)
    (if (eq nil (o-string-match re str))
        t                               ;do nothing, not matched
      (if (match-end 1)
          (setq m1 (substring str (match-beginning 1)
                                 (match-end 1))))

      (if (match-end 2)
          (setq m2 (substring str (match-beginning 2)
                                 (match-end 2))))
      )
    (list m1 m2)))




;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-required-headers ()
  "Return standard RFC header required when posting.

References:
  `ti:mt-required-headers'
  `gnus-required-headers'

Returns:
  list		,(header-name-symbol .. )
  nil		,gnus not loaded ?"
  (cond
   ((listp ti:mt-required-headers)
    ti:mt-required-headers)
   ((boundp 'gnus-required-headers)
    (symbol-value 'gnus-required-headers))
   (t
    nil)))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-mail-p  ()
  "Check if first line contain left flushed header 'Header:'.
This is a sign that current buffer is in mail-like.
You should also check the mode name to get more reliable results."
  (save-excursion
    (PMIN) (looking-at "^[-A-Za-z0-9][-A-Za-z0-9]+:")))



;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-header-area-size ()
  "Count size of header area.
This function can only be called from a buffer that has
`mail-header-separator'. Function count the characters in the header  area.
You can use this information to determine if the headers have been
changed after the last check.

Return:
 nbr
 nil	    can't find `mail-header-separator'"
  (save-excursion
    (PMIN)
    (when (re-search-forward (regexp-quote mail-header-separator) nil t)
      (- (point) (point-min))
      )))

;;; ----------------------------------------------------------------------
;;; - This is suitable for RMAIL, GNUS and for individual buffers
;;;   holding mail or news messages.
;;;
(defun ti::mt-hmax  (&optional move noerr)
  "Search max point of header, optionally MOVE and NOERR.
Order is: `mail-header-separator' or find all \"Headers:\" and then
return next line after them. The header must start at `point-min'.

If no point can be found, return `point-min'."
  (let ((point (point-min))
	)
    (when (ti::mt-mail-p)
      (save-excursion
      (PMIN)

      ;;  VM's "t" key shows all headers, including the
      ;;  "From xxxx"foo.com" line which is not actual header, because
      ;;  it has no colon. Skip ovber it if we see it.
      (if (looking-at "From")
	  (forward-line 1))


      ;;  GNUS 4 sets the mail-header-separator to "" ??
      ;;
      (if (and (not (var-nil-p mail-header-separator))
	       (re-search-forward (regexp-quote mail-header-separator) nil t))
	  (setq point (match-beginning 0))
	;;  Header:
	;;    Continuing line here
	;;  Header2:
	;;
	(while (and (looking-at "^[0-9a-zA-z-]+:")
		    (progn
		      (end-of-line)
		      (not (eobp)))
		    )
	  ;;  If this fucntion doesn't move anuy more, then the headers
	  ;;  have ended.
	  ;;
	  (if (null (ti::mt-next-field-start 'move))
	      (forward-line 1))
	  (setq point (point))
	  ))))

    (if (and move point)
	(goto-char point))
    point
    ))



;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-text-start (&optional move)
  "Return starting point or text. Optionally MOVE to it.
If buffer is not mail-like, then return `point-min'.

References:
  `mail-header-separator'"
  (let ((re	    (regexp-quote mail-header-separator))
	(point      (point-min))
	)
    (when (ti::mt-mail-p)
      (cond
       ((save-excursion			;Do we find the separator?
	  (PMIN)
	  (when (re-search-forward re nil t)
	    (forward-line 1)
	    (setq point (point)))))
       ((setq point (ti::mt-hmax))
	(save-excursion
	  (goto-char point)
	  (forward-line 1)
	  (setq point (point)))
	)
       (t
	(error "Can't find position.")
	))

      (if (eq point (point-min))		;Not found
	  (error "mail-header-separator not found or headers not found."))
      )

    (if (and move point)
	(goto-char point))
    point
    ))


;;; ----------------------------------------------------------------------
;;; - Many std emacs dist. functions work so that you have to narrow
;;;   to headers before you can call the functions.
;;;
(defun ti::mt-narrow (&optional text)
  "Narrows to the headers only. Optionally to TEXT portion."
  (if text
      (narrow-to-region (ti::mt-text-start 'move) (point-max))
    (narrow-to-region (point-min) (ti::mt-hmax)))
  )


;;; ----------------------------------------------------------------------
;;; - This is for both GNUS and RMAIL
;;;
(defun ti::mt-mail-buffer-name ()
  "Find original mail buffer whether in GNUS or in RMAIL.

Returns
   string       ,buffer name
   nil		,if not exist."

  ;;  It's amazing that GNUS uses pointers and RMAIL uses string ...
  (let ((buffer (if (boundp 'mail-reply-buffer)
		    (symbol-value 'mail-reply-buffer)))
	)
    (cond			;what is the mail YANK buffer name?
     ((stringp buffer)
      buffer)

     ((and (not (null buffer))
	   (bufferp buffer))
      (buffer-name buffer))

     (t
      nil
      ))
    ))


;;; ----------------------------------------------------------------------
;;; 27 Jul 1995 Jonathan "Jack" Vinson       jvinson@cheux.ecs.umass.edu
;;;             suggested by John Doppke <doppke@cs.colorado.edu>
;;; 22 Feb 1996 Rewritten by Jari Aalto <jaalto@tre.tele.nokia.fi>
;;;
;;; - put this into your mail setup hook:
;;;   (add-hook 'mail-setup-hook 'ti::mt-rename-buffer)
;;;
(defun ti::mt-rename-buffer (&optional arg)
  "Rename the *mail* buffer to \"*mail* SENDER\" or ARG.
If the function is run with a prefix, it will ask for a new buffer name,
rather than giving an automatic name.

If not called from mail buffer, acts like no-op."
  (interactive "Pbuffer name: ")
  (let ((to   (ti::mt-get-field "To" nil 'nil-mode))
	str
	name
	)

    (cond
     (to				;are we in mail ?

      (cond
       ((setq str (strmatget "\\([^@<]+\\)," 1 to))
	(strcat str ", ..."))
       ((setq str (strmatget "\\([^@<]+\\)" 1 to)))
       (t
	(setq str to)
	))

      (setq name (concat "*mail*" str))
      (if arg
	  (setq name (read-string "New buffer name: " name)))

      (rename-buffer name t)
      ))))




;;; ----------------------------------------------------------------------
;;; - The idea is to find three fields and see what they contain,
;;;   and do they exist?
;;; - What's the use of this function? Well, when you post an article
;;;   or mail it, you can call this function from some of those
;;;   posting hooks to determine what to do with the buffer.
;;;
;;; - code lines disabled now so that it buffer can be checked any time

(defun ti::m-mail-simple-p ()
  "Check if buffer contain headers belonging to simple \\[mail].
You can call this only once, just after the buffer is initially created"
  (require 'mail-utils)
  (let* (
         (sub (mail-fetch-field "Subject"))
         ;;   mail-fetch-field doesn't return nil if field is empty.
         (to  (mail-fetch-field "to"))
         (news (mail-fetch-field "Newsgroups"))
         )
    ;;  When you're replying to message in NEWS, RMAIL, the SUBJ and
    ;;  TO fields are already filled.
    ;;
    ;;  That's why you can only call this function once.
    ;;  When you use C-x m, and fill the fields, there is no way
    ;;  to detect afterwards if the mail buffer was simple mail or not
    ;;
    (cond
     ((or (string-match "news" (symbol-name 'major-mode))
          news)
      nil)
     ((and (var-nil-p sub)
           (var-nil-p to)
           )
      t
      ))
    ))

;;}}}
;;{{{ macros: VM, RMAIL, GNUS




;;; ----------------------------------------------------------------------
;;;
(put 'ti::mt-vm-macro 'lisp-indent-function 0)
(defmacro ti::mt-vm-macro (&rest body)
  "Do BODY in VM's active buffer.
The `save-excursion' -- set buffer form is executed."
  (`
   (let* ((BuffeR-S  (when (boundp 'vm-mail-buffer)
			(symbol-value 'vm-mail-buffer)))
	  )
     (if (or (null BuffeR-S)
	     (and (bufferp BuffeR-S)
		  (killed-buffer-p BuffeR-S)))
	 (error "vm-mail-buffer invalid")
       (job-buffer BuffeR-S
	 (,@ body)
	 )))))

;;; ----------------------------------------------------------------------
;;;
(put 'ti::mt-mh-macro 'lisp-indent-function 0)
(defmacro ti::mt-mh-macro (&rest body)
  "Do BODY in MH's active buffer.
The `save-excursion' -- set buffer form is executed."
  (`
   (let* ((BuffeR-S  (when (boundp 'mh-show-buffer)
			(symbol-value 'mh-show-buffer)))
	  )
     (if (or (null BuffeR-S)
	     (and (bufferp BuffeR-S)
		  (killed-buffer-p BuffeR-S)))
	 (error "mh-show-buffer invalid")
       (job-buffer BuffeR-S
	 (,@ body)
	 )))))


;;; ----------------------------------------------------------------------
;;;
(put 'ti::mt-gnus-macro 'lisp-indent-function 0)
(defmacro ti::mt-gnus-macro (&rest body)
  "Do BODY in GNUS's active buffer.
The `save-excursion' -- set buffer form is executed."
  (`
   (let* ((BuffeR-S  (when (boundp 'gnus-article-buffer)
			(symbol-value 'gnus-article-buffer)))
	  )
     (if (or (null BuffeR-S)
	     (and (bufferp BuffeR-S)
		  (killed-buffer-p BuffeR-S)))
	 (error "gnus-article-buffer invalid")
       (job-buffer BuffeR-S
	 (,@ body)
	 )))))

;;; ........................................................... &rmail ...

;;; ----------------------------------------------------------------------
;;;
(put 'ti::mt-rmail-macro 'lisp-indent-function 0)
(defmacro ti::mt-rmail-macro (&rest body)
  "Do BODY in RMAIL's active buffer. You have be in RMAIL summary."
  (`
   (let* ((BuffeR-R
	   ;;  This variable is available in Rmail-summary
	   ;;
	   (or (if (boundp 'rmail-buffer) (symbol-value 'rmail-buffer))
	       (get-buffer "RMAIL"))
	   ))
     (if (or (null BuffeR-R)
	     (and (bufferp BuffeR-R)
		  (killed-buffer-p BuffeR-R)))
	 (error "rmail-buffer buffer invalid")
       (job-buffer BuffeR-R
	 (,@ body)
	 )))))


;;; ----------------------------------------------------------------------
;;;
(put 'ti::mt-rmail-do-message-macro 'lisp-indent-function 2)
(defmacro ti::mt-rmail-do-message-macro (nbr mode &rest body)
  "Go to message without showing it and execute body.
Must be in RMAIL buffer already.

Input:
 NBR	message number, like `rmail-current-message'
 MODE	if non-nil then the area narrows to full stored message
        with original headers. If nil, then area narrows to displayed
        message.
 BODY	forms to execute in are narrowed to message."
  (`
   (let ((beg (rmail-msgbeg (, nbr)))
         (end (rmail-msgend (, nbr)))
         )
     (save-window-excursion
       (safe-widen
         (goto-char beg)
         (forward-line 1)
	 (if (null (, mode))
	     (search-forward "\n*** EOOH ***\n" end t))
         (narrow-to-region (point) end)
         (goto-char (point-min))
         (,@ body)
         )))))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-rmail-copy-message  (&optional nbr separate)
  "Copy message NBR with header. Defaults to `rmail-current-message'.
Current buffer must me in RMAIL already.

Input:
  NBR	    message number
  SEPARATE  if non-nil, then the headers and message body are returned
	    separately in format (hdr-string . body-string)

Return:
 string
 list	    see mode."
  (interactive)
  (let* (beg
	 end
	 hdr
	 ret
         )
    (setq nbr  (or nbr rmail-current-message)
	  beg  (rmail-msgbeg nbr)
	  end  (rmail-msgend nbr)
	  )

    (or (integerp nbr)
	(error "NBR %s" nbr))

    ;; The BEG isn't exactly the message beginning, skip 3 lines,
    ;; also don't copy the original heades only.
    ;;
    ;; 
    ;; 1, answered,,
    ;; Summary-line: 23-Mar #Please Help Yourself, Help Ot...
    ;; <ORIGINAL HEADERS>
    ;;
    ;; *** EOOH ***
    ;; <HEADERS SHOWN IN RMAIL>
    ;;
    ;; <BODY OF MESSAGE>
    ;;
    (safe-widen
      (goto-char beg) (forward-line 3)
      (setq beg (point))
      (re-search-forward "^[ \t]*$")
      (setq hdr (buffer-substring beg (point)))

      ;;  Already sitting at empty line, move away.
      (forward-line 1)
      (re-search-forward "^[ \t]*$")
      (setq beg (point))

      ;;  Now make HDR + BODY of message
      ;;
      (if separate
	  (setq ret (cons hdr (buffer-substring beg end)))
	(setq ret (concat hdr (buffer-substring beg end))))
      )
    ret
    ))



;;}}}

;;{{{ PGP general, tests

;;; ............................................................. &pgp ...


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-v3xx-p  ()
  "Check if X-Pgp v3.xx header signing is in use.
It should have VALUE = KEYWORD; statement."
  (save-excursion
    (PMIN)
    (when (re-search-forward "X-Pgp-signed" nil t)
      (forward-line 1)
      ;;
      ;; KEYWORD=VALUE;
      (looking-at "^[ \t]+.*=.*;")
      )))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-p  ()
  "Check if buffer contain PGP. It must have left flushed regexp:
\"^-----BEGIN.*PGP +\\(SIGNATURE\\|SIGNED\\\\|MESSAGE)\", otherwise this
string may be inside quoted text.

If there is X-pgp-sig.*: header, then it's also considered PGP message."
 (let ((max (ti::mt-hmax))	;headers ?
       )
   ;;  if headers was found use that.
   (setq max (if (eq (point-min) max)
		 nil max))

   (save-excursion
     (PMIN)
     (or (let (case-fold-search)
	   (re-search-forward
	    "^-----BEGIN PGP \\(SIGNATURE\\|SIGNED\\|MESSAGE\\)"
	    nil t))
	 (progn

	   ;;  The New PGP in headers standard.
	   ;;
	   (re-search-forward "^X-Pgp-Sig.*:" max t)
	   ))
     )))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-signed-conventional-p  ()
  "Return t if message is conventionally signed."
  (save-excursion (PMIN) (ti::mt-pgp-re-search 'sig)))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-signed-conventional-multi-p  ()
  "Return t if message is signed conventionally multiple times."
  (save-excursion
    (PMIN)
    (ti::mt-pgp-re-search 'sig 'move)
    (forward-line 1)
    (ti::mt-pgp-re-search 'sig 'move)
    ))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-signed-xpgp-p  ()
  "Return t if message is X-pgp signed.
There may be X-Pgp headers, but if the message is already
verified, that removes the signature around encrypted
message  \"- -----BEGIN PGP MESSAGE-----\"
       --> \"-----BEGIN PGP MESSAGE-----\"
In this case the message is no more in signed format,
but in encrypted format."
  (and (ti::mt-pgp-headers-p)
       ;;  See documentation above
       (save-excursion
	 (PMIN)
	 (null (re-search-forward
		(concat "^" (ti::mt-pgp-msg-begin-line))
		nil t))
	 )))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-signed-p  ()
  "Return t is message is conventionally or X-pgp signed."
  (or (ti::mt-pgp-signed-xpgp-p)
      (ti::mt-pgp-signed-conventional-p)))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-public-key-p (&optional point)
  "Find public key delimiter from current point forward or using POINT."
  (save-excursion
    (goto-char (or point (point)))
    (re-search-forward (ti::mt-pgp-pkey-begin-line) nil t)
    ))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-remail-p  ()
  "Check if This is remailer message."
  (save-excursion
    (PMIN)
    (re-search-forward "[:#][:#]+\nReply-To" nil t)
    ))



;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-comment-file-p  (&optional point)
  "You can send binary files with base64 signing.
This function checks if comment block has have words 'File: FILE '.

Example:

 -----BEGIN PGP MESSAGE-----
 Version: 2.6.3ia
 Comment: Base64 signed. File: tm.tar uncompresses to approx. 20K

Input:
  POINT	    ,search start point

Return:
  nil
  file"
  (let* (
         )
    (save-excursion
      (if point (goto-char point))
      (when (re-search-forward "^Comment:.*File:? +\\([^ \t,]+\\)" nil t)
	(read-match 1))
      )))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-encrypted-p  (&optional check-pgp-dash-line)
  "Check if there is encrypted PGP message.
It must have left flushed tag. The start point of match is returned.
The following tag will tell if if the message is encrypted.

  ::
  Encrypted: PGP

Input:
  CHECK-PGP-DASH-LINE   if the tag is not found, message _could_ be signed
			if there is -----BEGIN PGP MESSAGE----- tag.
			When this flag is non-nil, it also checks this
			case. Beware: message could be base64 signed too,
			so the encrypted-p test may not be exactly right."
  (save-excursion
    (PMIN)
    (if (re-search-forward "::[ \t]*\nEncrypted:[ \t]*PGP" nil t)
	(match-beginning 0)
      (when check-pgp-dash-line
	(PMIN)
	(car-safe (ti::mt-pgp-block-area 'msg))
	))
    ))



;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-normal-p  (&optional point)
  "Check if there is any PGP in current buffer from POINT forward.
The beginning point of PGP is returned."
  ;; Must find at least two lines, maybe BEG and END
  (let ((re   (concat "^" (ti::mt-pgp-any-pgp-line-regexp)))
	ret
	)
    (save-excursion
      (PMIN)
      (when (re-search-forward re nil t)
	(setq ret (match-beginning 0))
	(if (null (re-search-forward re nil t))
	    (setq ret nil)))
      ret
      )))



;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-headers-p  ()
  "Return t if PGP information is in headers.
Searches string 'X-Pgp-Signed:' and return end of match or nil."
  (let ((psig		"^X-Pgp-Signed:")
	(hmax		(ti::mt-hmax))
	)
    (save-excursion
      (PMIN)
      (re-search-forward psig hmax t))
    ))



;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-re  (str)
  "Add possible beginning anchor if STR doesn't have one."
  (if (not (char= (aref str 0) ?^))
      (concat "^" str)))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-block-area-kill-forward  (mode &optional move)
  "Search PGP block forward and kill it. If no block found, do nothing.

Input:
  MODE	    choices are explained in `ti::mt-pgp-block-area'.
  MOVE      if non-nil, move to killed region begin point."
  (let* (reg
         )
    (when (setq reg (ti::mt-pgp-block-area mode))
      (delete-region (car reg) (cdr reg))
      (when move (goto-char (car reg)))
      )))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-block-area (mode &optional inside max)
  "Return (beg . end) of PGP block from current point forward.

Input:
  MODE      nil   search signed start line..
	    'sig  search signature start instead.
	    'signed  search signed message area
	    'pkey search public key block start instead.
	    'msg  search for pgp base64 signed \"message\"
	    'any  go to `point-min' and search beginning of any
                  PGP line, then go to the end of buffer and search
	          backward any PGP line.  The lines must not be at
	          same position.  This gives you the whole PGP
	          region.

  INSIDE    if non-nil, beg and end are 'inside' without the PGP tags.
  MAX       max point to search

Return:
  (beg . end)
  nil"
  (let ((re1
	  (cond
	   ((null mode)	     (ti::mt-pgp-signed-begin-line))
	   ((eq 'sig  mode)  (ti::mt-pgp-signature-begin-line))
	   ((eq 'pkey mode)  (ti::mt-pgp-pkey-begin-line))
	   ((eq 'msg  mode)  (ti::mt-pgp-msg-begin-line))
	   ((eq 'any  mode)  (ti::mt-pgp-any-pgp-line-regexp))
	   ((eq 'signed  mode)  (ti::mt-pgp-signed-begin-line))
	   (t
	    (error "unknown mode"))))
	 (re2
	  (cond
	   ((null mode)	     (ti::mt-pgp-signed-end-line))
	   ((eq 'sig  mode)  (ti::mt-pgp-signature-end-line))
	   ((eq 'pkey mode)  (ti::mt-pgp-pkey-end-line))
	   ((eq 'msg  mode)  (ti::mt-pgp-msg-end-line))
	   ((eq 'signed  mode)  (ti::mt-pgp-signed-end-line))
	   ))
  	 ret
  	 beg end
         )
    (save-excursion
      (cond
       ((eq mode 'any)

	(when (re-search-forward re1 max t)
	  (setq beg (match-beginning 0))
	  (when (re-search-forward re1 max t)
	    (beginning-of-line)
	    (when (not (eq beg (point)))
	      (forward-line 1)
	      (setq ret (cons beg (point)))))
	  ))

       (t
	(setq re1    (concat "^-? ?" re1)
	      re2    (concat "^-? ?" re2))
	(when (re-search-forward re1 max t)
	  (if inside
	      (forward-line 1)
	    (beginning-of-line))
	  (setq beg (point))

	  (when (re-search-forward re2 max t)
	    (if inside
		(beginning-of-line)
	      (forward-line 1))

	    (setq end (point))
	    (setq ret (cons beg end))
	    )))
       ))
    ret
    ))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-re-search (&optional mode move end no-anchor)
  "Re-search-forward to find -----BEGIN.*SIGNED.

Input:
  MODE		nil	search signed start line.
		'sig	search signature start.
		'sige	search signature block end.
		'pkey	search public key block start.
		'pkeye	search public key block end.
		'msg	search for pgp base64 signed \"message\"
		'kid	search for  'Key for user ID: '
		'kpub	search for 'pub   512/47141D35 1996/06/03 ...'
			note: match level 1 matches 0x code 47141D35
  MOVE		flag,	non-nil moves point to found point
  END		flag,	use `match-end' instead of math-beginning.
  NO-ANCHOR	flag,	non-nil disables '^' anchor using.

Return:
  point		,beginning of line
  nil		,if not found"
  (let ((re   (cond
	       ((null     mode)  (ti::mt-pgp-signed-begin-line))
	       ((eq 'sig  mode)  (ti::mt-pgp-signature-begin-line))
	       ((eq 'sige mode)  (ti::mt-pgp-signature-end-line))
	       ((eq 'pkey mode)  (ti::mt-pgp-pkey-begin-line))
	       ((eq 'pkeye mode) (ti::mt-pgp-pkey-end-line))
	       ((eq 'msg  mode)  (ti::mt-pgp-msg-begin-line))
	       ((eq 'kid  mode)  "Key for user ID: ")
	       ((eq 'kpub mode)
		"pub[ \t]+[0-9]+/\\([A-Z0-9]\\)+[ \t]+.*/.*/[0-9]")
	       (t
		(error "unknown mode"))))
	point
	)
    (if (and (null no-anchor)
	     (not (memq mode '(kid))))
	;;  suppose encrypted and signed message
	;;  - -----END PGP MESSAGE-----
	;;
	(setq re (concat "^-? ?" re)))

    (save-excursion
      (if (re-search-forward re nil t)
	  (if end
	      (setq point (match-end 0))
	    (setq point (match-beginning 0)))))

    (if (and move point)
	(goto-char point))
    point
    ))

;;}}}
;;{{{ PGP misc

;;; ........................................................ &pgp-misc ...

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-exe-version-number  (&optional exe-file-location)
  "Call pgp executable to find out it's version number.
EXE-FILE-LOCATION defaults to \"pgp\" but can also be absolute pgp exe path."
  (let ((buffer  (use-tmp-buffer ti:mt-tmp-buffer 'clear))
	ret
	)
    (shell-execute (format "%s -k" (or exe-file-location "pgp"))
		   buffer)
    (job-buffer buffer
      (PMIN)
      (if (re-search-forward "Pretty Good Privacy(tm) +\\([^ ]+\\)" nil t)
	  (setq ret (read-match 1)))
      )
    ret
    ))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-data-type  ()
  "Examine pgp data packet type by searching _forward_.
Return:
  'base64 'pgp 'conventional or nil"
  (let ((re (concat "^" (ti::mt-pgp-any-pgp-line-regexp)))
	char
	)
    (save-excursion
      (when (and (re-search-forward re nil t)
		 (re-search-forward "^$" nil t))
	;;  #todo: Check first character. Actually we should check bit mask...
	;;
	;;  -----BEGIN PGP MESSAGE-----
	;;  Version: 2.6.2
	;;  Comment: Encrypted by xxx
	;;
	;;  hEwDYCggxO/bFq0
	;;
	(forward-line 1)
	(setq char (following-char))
	(cond
	 ((char= char ?p) 'conventional)
	 ((char= char ?h) 'pgp)
	 ((char= char ?o) 'base64)
	 )))))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-trim-buffer  ()
  "Trim buffer: pgp blocks are left flushed and junk around them is removed."
  (let ((stat  t)
	region
	)
    (save-excursion
      (PMIN)
      (while (and stat
		  (setq region (ti::mt-pgp-block-area 'any)))

	(when (setq stat (ti::mt-pgp-chop-region (car region) (cdr region)))
	  (goto-char (cdr stat)))
	))))


;;; ----------------------------------------------------------------------
;;; - This is needed after finger or http call to clean up all unnecessary
;;;   tags around the PGP key.
;;;
(defun ti::mt-pgp-chop-region (beg end)
  "Delete junk around BEG END from pgp public key block.
Area BEG END that correspond to pgp begin and end
lines (call `ti::mt-pgp-block-area' with argument 'any),
then we chop the public key region so that only the pgp area
is left without additional garbage.

Return
 (beg .end)	    the canonilized area of PGP block

Example:

<PRE>
<SAMP>      -----BEGIN PGP PUBLIC KEY BLOCK-----</SAMP>
<SAMP>      Version: 2.6.3ia</SAMP>
</PRE>
<PRE>
<SAMP>      mQBNAzGzQ2MAAAECAM4p2THKCpNjYXDLpsg4sLHyEiNxJwQuEYfipdTj</SAMP>
<SAMP>      p5CPHN+0LkphcmkgQWFsdG8sIEZpbmxhbmQgPGphcmkuYWFsdG9AbnRj</SAMP>
<SAMP>      LmNvbT6JAFUDBRAxs0O+wLrt1UcUHTUBAbMhAf9Qgh6EznEcY2OUOIPg</SAMP>
<SAMP>      =46gx</SAMP>
<SAMP>      -----END PGP PUBLIC KEY BLOCK-----</SAMP>


This is converted into

-----BEGIN PGP PUBLIC KEY BLOCK-----
Version: 2.6.3ia</SAMP>


mQBNAzGzQ2MAAAECAM4p2THKCpNjYXDLpsg4sLHyEiNxJwQuEYfipdTj
p5CPHN+0LkphcmkgQWFsdG8sIEZpbmxhbmQgPGphcmkuYWFsdG9AbnRj
LmNvbT6JAFUDBRAxs0O+wLrt1UcUHTUBAbMhAf9Qgh6EznEcY2OUOIPg
=46gx
-----END PGP PUBLIC KEY BLOCK-----"
  (save-excursion
    (goto-char beg) (beginning-of-line)
    (safe-narrow (point) (progn
			   (goto-char end)
			   (end-of-line)
			   (point))

      (ti::m-add-spaces-80 (point-min) (point-max))
      (PMIN)
      (re-search-forward "-----END")

      (goto-char (match-beginning 0))
      (if (> (current-column) 0)	;Nothing to do, it's left flushed
	  (delete-rectangle (point-min) (point)))

      (PMIN)
      (ti::b-replace-regexp "<.*$" 0 "")

      ;;  Because the last line does not have newline, the
      ;;  previous regexp doesn't match. Fix the last line too.
      ;;
      (goto-char (point-max))
      (beginning-of-line)

      (let (case-fold-search)		;be sensitive
	;;  -----END PGP PUBLIC KEY BLOCK-----
	(if (and (looking-at ".*[A-Z]-----\\(.*\\)")
		 (match-end 1))
	    (kill-match 1)
	  ))
      (setq beg (point-min)
	    end (point-max)
	    )

      ))

  (cons beg end)
  )

;;}}}
;;{{{ PGP signed headers

;;; ...................................................... &pgp-header ...

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-header-kill-in-body  ()
  "Kill headers that are inserted into the body of message.
If there is no headers, this function does nothing.

--text follows this line--  [or empty line after headers]
##<no spaces>
Header1: content
Header2: content
<empty line>
BODY"
  (let* (beg
         )
    (save-excursion
      (ti::mt-text-start 'move)
      (setq beg (point))
      (when (and (looking-at "^##\n")
		 (re-search-forward "^$" nil t))
	(delete-region beg (point))
	))))



;;}}}
;;{{{ PGP ASCII armor

;;; ....................................................... &pgp-armor ...


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-data-char-to-int  (char)
  "Process PGP ascii armor data.
Input is ASCII armor CHAR (as one string). Function return respective int."
  (let* ((table (concat "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
			"abcdefghijklmnopqrstuvwxyz0123456789+/"))
	 case-fold-search
	 str
         )
    (if (null (setq str (strmatget (concat ".*" (regexp-quote char)) 0 table)))
	(error "Armor char is invalid %s " char)
      (1- (length str)))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-data-string-to-bin-string (string)
  "Process PGP ascii armor data.
Convert quoted printable ASCII armor STRING into binary string."
  (let* ((len (length string))
	 (i    0)
	 (ret  "")
	 ch int bin
         )
    (while (< i len)
      (setq ch (substring string i (1+ i)))
      (setq int (inline (ti::mt-pgp-data-char-to-int ch)))
      (setq bin (inline (int-to-bin-string int 6)))
      (setq ret (concat ret bin))
      (inc i)
      )
    ret
    ))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-data-bin-string-to-int-list(string)
  "Process PGP ascii armor data.
Convert 8bit binary byte string \"000001...\" into list of ints."
  (let* ((len (/ (length string) 8))
	 (i   0)
	 ret
	 bin int
	 )
    (while (< i len)
      (setq bin (substring string (* i 8) (+ 8 (* i 8))))
      (setq int (inline (bin-string-to-int bin)))
      (inc i)
      (push int ret)
      )
    (nreverse ret)
    ))


;;; ----------------------------------------------------------------------
;;;
(defsubst ti::mt-pgp-data-ascii-armor-convert  (string)
  "Convert PGP ascii armor STRING(quoted printable) into list of ints."
  (ti::mt-pgp-data-bin-string-to-int-list
   (ti::mt-pgp-data-string-to-bin-string string)))



;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-data-study-ctb-byte  (int)
  "From single INT, examine the PGP CTB structure.
Return
 nil    ,input was not CTB byte
 '(ctb-type length-field)
        ctb-type  is
        'enc      (pgp encrypted message)
        'signed   (signed message)
        'secring  (secret keyring)
	'pring    (public keyring)
	'base64   (base 64 signed)
	'crypt    (conventionally crypted)
	'raw      (raw literal plaintext)
	'trust    (keyring trust packet)
	'uid      (user id packet)
	'comment  (comment packet)
        'unknown  (none of the above...)

	length is
	nil	  no length, unknown length
	1 2 4	  byte packet length"
  (let* ((length-mask   3)     ;; 00000011b
	 (type-mask     60)    ;; 00111100b
	 (ctb-mask      128)   ;; 10000000b
	 (table
	  '((1  . enc)
	    (2  . signed)
	    (5  . secring)
	    (6  . pring)
	    (8  . comp)
	    (9  . crypt)
	    (11 . raw)
	    (12 . trust)
	    (13 . uid)
	    (14 . comment)
	    ))
	 (type 'unknown)
	 val
	 ret
         )
    (when (logand int ctb-mask)
      ;; shift to the right 2 bits
      ;;
      (when (setq val (assq (lsh (logand int type-mask) -2) table))
	(setq type (cdr val)))

      (setq val (logand int length-mask))
      (cond
       ((eq 0 val)  (setq val 1))
       ((eq 1 val)  (setq val 2))
       ((eq 1 val)  (setq val 4))
       ((eq 3 val)  (setq val nil)))

      (setq ret (cons type val))
      )
    ret
    ))


;;; ----------------------------------------------------------------------
;;;
(defsubst ti::mt-pgp-stream-study-1-ver  (int)
  "Return pgp version string from stream INT."
  (cond
   ((eq 2 int) "2.5")
   ((eq 3 int) "2.6")
   (t (error "Invalid Data format."))))


;;; ----------------------------------------------------------------------
;;;
(defmacro ti::mt-pgp-stream-study-1-key-id (stream result)
  "Read MSB and LSB key-id from STREAM to RESULT.
STREAM will be advanced during read."
  (`
   (let* ((i 0)
	  )
     (while (< i 4)
       (setq (, result) (concat (or (, result)  "")
				(format "%x" (car (, stream))))
	     (, stream) (cdr (, stream))
	     i      (1+ i)
	     ))
     (setq (, result) (upcase (, result)))
     )))



;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-stream-study-1-time (stream)
  "Read TIME from STREAM to RESULT."
   (let* (val1 val2
	  )
     ;;  There must be easier way to do, but right now it goes like this
     ;;  '(51 158 95 145)
     ;;  --> hex 339E 5F91
     ;;  --> int 13214  24464  which is in (current-time) format
     ;;

     (setq val1 (hexl-hex-string-to-integer
		 (concat
		  (int-to-hex-string (car stream))
		  (int-to-hex-string (car (cdr stream)))
		  ))

	   stream (cdr (cdr stream))
	   val2  (hexl-hex-string-to-integer
		 (concat
		  (int-to-hex-string (car stream))
		  (int-to-hex-string (car (cdr stream)))
		  ))
	   )
     (ti::d-standard-date nil (list val1 val2))
     ))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-stream-study-enc  (length stream)
  "Study the 'enc packet, which has known LENGTH.
STREAM is list of ints, minimum 11 integers, 13 is the full 'enc packet.

Return:
 '(version-string
   key-msb-hex-string
   key-lsb-hex-string

   rsa-algorithm                 ;; nil if stream is not long enough
   rsa-int (encrypted integer)   ;; nil if stream is not long enough
   )"
  (let* ((msb "")
	 (lsb "")
	 ver
	 val
	 rsa-alg
	 rsa-int
         )
    ;;  Skip to begin of real data
    ;;  CTB   LENGTH     VERSION KEY-MSB KEY-LSB
    ;;  1byte 1-4bytes   1byte   4bytes  4bytes
    (setq stream (nthcdr (1+ length) stream))
    (setq val    (car stream)   stream (cdr stream))
    (setq ver (ti::mt-pgp-stream-study-1-ver val))

    (ti::mt-pgp-stream-study-1-key-id stream msb)
    (ti::mt-pgp-stream-study-1-key-id stream lsb)

    (setq rsa-alg (car stream)
	  rsa-int (cadr stream))

    (list ver msb lsb rsa-alg rsa-int)
    ))



;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-stream-study-signed  (length stream)
  "Study the 'sign packet, which has known LENGTH. STREAM is list of ints.

Return:
 '(version-string
   md-length
   sig-class
   timestamp

   key-msb-hex-string
   key-lsb-hex-string

   alg-rsa		    ;; nil if stream is not long enough
   alg-md5		    ;; nil if ...
   digest '(int int);; nil if ...
   )

   rsa-algorithm                 ;; nil if stream is not long enough
   rsa-int (encrypted integer)   ;; nil if stream is not long enough
   )"
  (let* ((msb "")
	 (lsb "")
	 ver
	 md-length
	 sig-class
	 timestamp
	 alg-rsa
	 alg-md5
	 digest
         )
    ;;  Skip to begin of real data
    ;;  CTB   LENGTH     VERSION KEY-MSB KEY-LSB
    ;;  1byte 1-4bytes   1byte   4bytes  4bytes
    (setq stream (nthcdr (1+ length) stream))
    (setq ver	    (ti::mt-pgp-stream-study-1-ver (car stream))
	  stream    (cdr stream)
	  md-length (car stream)
	  stream    (cdr stream)
	  sig-class (car stream)
	  stream    (cdr stream)
	  )
    (setq timestamp
	  (ti::mt-pgp-stream-study-1-time stream))
    (setq stream (nthcdr 4 stream))

    (ti::mt-pgp-stream-study-1-key-id stream msb)
    (ti::mt-pgp-stream-study-1-key-id stream lsb)


    (setq alg-rsa (car stream)
	  stream  (cdr stream)
	  alg-md5 (car stream)
	  stream  (cdr stream)
	  digest  (list (car stream) (car (cdr stream)))
	  )

    (list ver md-length sig-class timestamp msb lsb
	  alg-rsa alg-md5 digest)
    ))



;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-stream-study-pring  (length stream)
  "Study the 'pring packet, which has known LENGTH. STREAM is list of ints.

Return:
 '(version-string
   timestamp

   key-msb-hex-string
   key-lsb-hex-string
   )"
  (let* ((msb "")
	 (lsb "")
	 ver
	 timestamp
	 validity
         )
    ;;  Skip to begin of real data
    ;;  CTB   LENGTH     VERSION TIME    VALIDITY RSA-ALG
    ;;  1byte 1-4bytes   1byte   4bytes  2bytes   1byte
    (setq stream (nthcdr (1+ length) stream))
    (setq ver	    (ti::mt-pgp-stream-study-1-ver (car stream))
	  stream    (cdr stream))

    (setq timestamp
	  (ti::mt-pgp-stream-study-1-time stream))
    (setq stream    (nthcdr 4 stream)
	  validity  (car stream)
	  stream    (cdr stream)
	  )

    ;;  PGP format spec is not clear enough here!
    ;;  Don't know where the User ID is...

;;;    (ti::mt-pgp-stream-study-1-key-id stream msb)
;;;    (ti::mt-pgp-stream-study-1-key-id stream lsb)

    (list ver timestamp validity msb lsb)
    ))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-stream-study  (ctb stream)
  "Study PGP data.

Input:

 CTB		in format `ti::mt-pgp-data-study-ctb-byte'
 STREAM		dearmored int stream (list of ints including ctb-byte)

Return:
  LIST		depends on the ctb, see conversion functions."
  (let* ((type  (car ctb))
	 (len   (cdr ctb))
	 )

;;;    (d! type)
    (cond
     ((eq type 'enc)
      (ti::mt-pgp-stream-study-enc len stream))

     ((eq type 'signed)
      (ti::mt-pgp-stream-study-signed len stream))

     ((eq type 'base64)
      )
     ((eq type 'crypt)
      nil
      )
     ((eq type 'pring)
      (ti::mt-pgp-stream-study-pring len stream))

      )))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-stream-forward-xpgp  ()
  "If there is X-Pgp-Signed field, goto signature stream."
  (let* ((point (ti::mt-pgp-headers-p))
	 )
    (when point
      (goto-char point)
      ;;  must exist, this call dies if not found
      (re-search-forward "Signature[ \t]*=[ \t\n\"]+")
      )))



;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-stream-forward (&optional any)
  "Find PGP data stream block start forward. PGP block must be left flushed.

Input:
  ANY	    if non-nil, then find any stread (not necessarily left flushed)

Return:
  point	    cursor is placed in front of stream
  nil       If there is no PGP stream block, do nothing."
  (let* ((beg   (ti::mt-pgp-msg-begin-line))
	 (sig   (ti::mt-pgp-signature-begin-line))
	 (pkey  (ti::mt-pgp-pkey-begin-line))
	 (re    (concat (if any "" "^") "-----BEGIN"))
	 (point (point))
	 (loop  t)
	 col
	 ret
	 )

    ;; base64
    ;; -----BEGIN PGP MESSAGE-----
    ;; Version: 2.6.3ia
    ;;
    ;; owEBbACT/4kAVQMFATOb

    ;; normal signature
    ;; -----BEGIN PGP SIGNATURE-----
    ;; Version: 2.6.3ia
    ;; Charset: noconv
    ;;
    ;; iQBVAwUBM55fkcC67dVHFB01AQGnHwIAqe2OfkdcnQviGzCmy3KddnsE8uFkAeaV

    ;; conventional crypt
    ;; -----BEGIN PGP MESSAGE-----
    ;; Version: 2.6.3ia
    ;;
    ;; pgAAACN9WXlrJFURU5Xgi+YyN

    ;; encrypted
    ;; -----BEGIN PGP MESSAGE-----
    ;; Version: 2.6.3ia
    ;;
    ;; hEwDwLrt1UcUHTUBAf9
    ;;

    ;; Extracted public key
    ;; -----BEGIN PGP PUBLIC KEY BLOCK-----
    ;; Version: 2.6.3ia
    ;;
    ;; mQBNAzOW770AAAECANDkXBfEbJk0gW41o52nLiktpThcBY+BMQCY5zyGCyUIbrDp

    (while (and loop (re-search-forward re nil t))
      (goto-char (match-beginning 0))

      (when (or (looking-at beg)
		(looking-at sig)
		(looking-at pkey)
		)
	(setq col (current-column))
	(when (re-search-forward "^[ \t]*$" nil t)
	  (setq loop nil)
	  (forward-line 1)
	  (move-to-column col)
	  (setq ret (point))
	  ))

      (if loop (end-of-line))		;wrong match, Continue search
      )

    (unless ret
      ;;  none found, return to original position.
      (goto-char point))

    ret
    ))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-stream-forward-and-study  (&optional search any)
  "Find PGP data stream forward and study it.

If normal search fails, then find X-Pgp-Signed field's first
data stream.

Input:

  SEARCH    if non-nil, then search PGP starting from `point-min' if
	    forward lookup fails.
  ANY	    if non-nil, find also non-left flushed stream.

Return:
  '(CTB . (INFO-LIST))	the CTB type and data for CTB
  nil			no stream found forward."
  (interactive)
  (let* ((point (point))
	 ctb
	 line
	 list
	 data
	 ret
         )
    (when (or (ti::mt-pgp-stream-forward any)
	      (and search
		   (PMIN) (ti::mt-pgp-stream-forward any)
		   )
	      (ti::mt-pgp-stream-forward-xpgp)
	      )
      ;;  Will match all base64 characters (approx.)
      (setq line (bufmatget "[^ \t\n\"\']+" 0)
	    list (ti::mt-pgp-data-ascii-armor-convert line)
	    ctb  (ti::mt-pgp-data-study-ctb-byte (car list))
	    data (ti::mt-pgp-stream-study ctb list)
	    )
      (setq ret (cons (car ctb) data))
      )
    (unless ret (goto-char point))	;Nothing found, return to point
    ret
    ))



;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-stream-forward-info  (&optional search any)
  "Find PGP data stream and read some information. Return string.

Input:

  SEARCH    if non-nil, then search PGP starting from `point-min' if
	    forward lookup fails.
  ANY	    if non-nil, find also non-left flushed stream."
  (let* (ver
	 time key-id
	 data
	 type
	 ret
         )
    (when (setq data (ti::mt-pgp-stream-forward-and-study search any))
      (setq type (car data))
      (cond
       ((eq type 'signed)
	(setq ver	(ti::mt-pgp-stream-data-elt data 'ver)
	      time	(ti::mt-pgp-stream-data-elt data 'time)
	      key-id	(ti::mt-pgp-stream-data-elt data 'key-id)
	      )
	(setq ret (format "Signed by 0x%s %s [v%s.x]" key-id time ver))
	)
       ((eq type 'enc)
	(setq ver	(ti::mt-pgp-stream-data-elt data 'ver)
	      key-id	(ti::mt-pgp-stream-data-elt data 'key-id)
	      )
	(setq ret (format "Encrypted to 0x%s [v%s.x]" key-id ver))
	)
       ((eq type 'pring)
	(setq ver	(ti::mt-pgp-stream-data-elt data 'ver)
	      time 	(ti::mt-pgp-stream-data-elt data 'time)
	      )
	(setq ret (format "Public key %s [v%s.x]" time ver))
	)))
    ret
    ))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-stream-data-elt  (data elt)
  "Study DATA and Return packet ELT.
DATA must be in the format of `ti::mt-pgp-stream-forward-and-study'
ELT  can be 'ver 'time 'key-id"
  (let* ((type (car data))
	 (pos  (assq
		elt
		(nth
		 1
		 (assq type
		       '(
			 (signed ((ver 0) (time 3) (key-id 5)))
			 (pring  ((ver 0) (time 1) (key-id 4)))
			 (enc    ((ver 0) (key-id 2))) ;No TIME field
			 ;; #todo, not ready
			 (base64 ((ver 0) (time 3) (key-id 5)))
			 (crypt  ((ver 0) (time 3) (key-id 5)))
			 )))))
         )
    (if (null pos)
	(error "Wrong specification %s %s %s" type elt data)
      (nth (nth 1 pos) (cdr data)))
    ))


;;; Test suite with live data: first ASCII armor bytes
;;
;; (setq list (ti::mt-pgp-data-ascii-armor-convert "hEwDwLrt1UcUHTUBA"))
;; (setq ctb  (ti::mt-pgp-data-study-ctb-byte (car list)))
;; (setq data (ti::mt-pgp-stream-study ctb list))

;; Sig data
;; (setq s "iQBVAwUBM55fkcC67dVHFB01AQGnHwIAqe2OfkdcnQviGzCmy3KddnsE8uFkAeaV")
;;
;; (setq list (ti::mt-pgp-data-ascii-armor-convert s))
;; (setq ctb  (ti::mt-pgp-data-study-ctb-byte (car list)))
;; (setq data (ti::mt-pgp-stream-study ctb list))


;;}}}

;;{{{ PGP key info

;;; ......................................................... &pgp-key ...


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgpk-id-lines-in-region  (beg end)
  "Search all lines in BEG END matching pgp -kvc lines.
pub   512/47141D35 1996/06/03 Jari Aalto, Finland <jaalto@tre.tele.nokia.fi>

And return list of those lines."
  (ti::b-grep-lines "pub[ \t:]+[0-9]+/[A-Z0-9]+[ \t:]+.*/.*/[0-9]" beg end))




;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgpk-public-get-region  (&optional beg end buffer relax)
  "Get all keys in BEG END from BUFFER and build list of key data.
The blocks searched are in following format.

  Key for user ID: Jari Aalto, Finland <jaalto@tre.tele.nokia.fi>
  512-bit key, key ID 47141D35, created 1996/06/03
  Also known as: Jari Aalto, Finland <ssjaaa@uta.fi>
  Also known as: Jari Aalto, Finland <jari.aalto@poboxes.com>

  -----BEGIN PGP PUBLIC KEY BLOCK-----
  [...]
  -----END PGP PUBLIC KEY BLOCK-----


Note1:

  If there _no_ 'Key for user ID:' string in the buffer, this function
  can't find the public key block while it may be there. It is
  assumed that each p-key block is _preceded_ by that string.

  All anonymous p-key block are skipped.

Note2:

  If there are two sequential key-id strings, like
  Key for user ID:       <<A
  Key for user ID:       <<B
  -----BEGIN PGP PUBLIC KEY BLOCK-----


  The p-key block in the list for A will be nil.

Note3:

   If RELAX argument is non-nil, then the 'Key for user ID:'
   must not exit. Only the Public key tags are searched.

   Recommended way of informing public keys is however displaying
   full public key information and not just PK block

Return:
  '((KEY-ID-STRING PUBLIC-KEY_BLOCK)"
  (let ((opt  (if relax 'pkey 'kid))
	id
	block
	region
	ret
	max
	)
    (job-buffer (or buffer (current-buffer))
      (safe-narrow (or beg (point-min)) (or end (point-max))
	(PMIN)
	(while (ti::mt-pgp-re-search opt 'move)
	  (setq id (read-full-line))

	  ;;  If there are two
	  ;;    Key for user ID:
	  ;;    Key for user ID:
	  ;;
	  ;;  And there is no public key between these two, set the
	  ;;  search limit to stop to next Key-id line.
	  ;;
	  (setq max
		(save-excursion
		  (end-of-line)
		  (setq max (ti::mt-pgp-re-search 'kid))))

;;;	  (d! ">>" id ">>" max (ti::mt-pgp-block-area 'pkey nil max))

	  (cond
	   ((setq region (ti::mt-pgp-block-area 'pkey nil max))
	    (setq block (buffer-substring (car region) (cdr region)))
	    (goto-char (cdr region))
	    )
	   (t
	    (end-of-line)		;Continue search
	    ))
	  (push (list id block) ret)
	  (setq id nil   block nil)

	  )))
    (reverse ret)
    ))

;;}}}
;;{{{ PGP signature info

;;; ................................................... &pgp-signature ...

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-signature-remove (&optional add no-cnv)
  "Remove PGP signature (and headers that have included in there).
Below, only lines 7 and 8 are left in buffer.

  1   -----BEGIN PGP SIGNED MESSAGE----
  2
  3   ##			    << Header start mark
  4   Header1: content
  5   Header2: Content
  6				    << space here
  7   test
  8
  9   -----BEGIN PGP SIGNATURE-----

With ADD flag

  The tag lines are reassembled and point sits at the begining of line 6
  and whitespaces around (email) buffer text are deleted.
  If tag lines are found while ADD, this function does nothing.

With NO-CNV

  When removing signature, do not convert '- -' back into '-'.
  Eg. If message is encrypted and signed; it is not desireable to
  do this conversion if you just want to strip out the signature to Xpgp.
  The '- -' lines must stau there."
  (let* ((beg (save-excursion (PMIN) (ti::mt-pgp-re-search)))
	 )
    (cond
     ((and add (null beg))
      (ti::mt-trim-buffer)
      (ti::mt-text-start 'move)
      (insert (ti::mt-pgp-signed-begin-line) "\n\n")
      (PMAX)
      (insert "\n" (ti::mt-pgp-signature-begin-line) "\n")
      )
     ((null add)
      ;;  there is one thing to fix first, PGP converts lines that have
      ;;  double '--' at front
      ;;
      ;;		--text follows
      ;;	-->
      ;;		- --text follows
      ;;
      ;;	Let's correct those lines too.
      (when (null no-cnv)
	(save-excursion (ti::b-replace-regexp "^- -" 0 "-")))

      (when beg				;Thre is regular PGP sig
      ;;  note: We don't trim BODY here, we only remove
      ;;  the pgp tag lines. The receiving end should do
      ;;  the trimming. (we save one function call)
      ;;
      (goto-char beg)			;One newline at beg
      (ti::b-kill-line 'del 2)		;TWO lines; important

      ;;  Kill the included headers
      ;;
      (when (and (looking-at "##\n.*: ")
		 (re-search-forward "^$" nil t))
	(delete-region beg (1+ (point))))

      (when (and (prog1 (setq beg (ti::mt-pgp-re-search 'sig 'move))
		   (ti::b-kill-line))
		 (ti::mt-pgp-re-search 'sige 'move))
	(forward-line 1)
	(delete-region beg (point))
	)
      )))
    ))


;;; ----------------------------------------------------------------------
;;;
(defmacro ti::mt-pgp-signature-normal-do-region (&rest body)
  "Execute BODY and calculate pgp signature region.
In the macro you can use following variables:
  'limits'      (area-beg . area-end)
  'area-beg'
  'area-and'

This macro does nothing if there is no normal PGP signature."
  (`
   (let (limits
	 area-beg
	 area-end
	 )
     (setq limits (ti::mt-pgp-block-area 'sig))
     (when limits
       ;;   Set values
       (setq area-beg (car limits)
	     area-end (cdr limits))

       ;;  If these are no used in BODY: no-op Quiet XE ByteCompiler
       ;;
       (if (null area-beg)  (setq area-beg nil))
       (if (null area-end)  (setq area-end nil))

       (,@ body)
       ))
   ))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-signature-normal-info  ()
  "Return signature information from normal PGP format.
Return:
 ((beg . end) (fld fld ..) (signarure-data sig ..))"
  (let (sig-list
	info-list
	ret
	)
    (ti::mt-pgp-signature-normal-do-region
     (save-excursion
       (goto-char area-beg)

       (forward-line 1)

       ;;  Here are the comments and other PGP headers
       ;;
       (while (looking-at "^[^ \t]+:+ .*")
	 (nconcc info-list (read-full-line))
	 (forward-line 1)
	 )
       ;; Here is the signature itself
       ;;
       (while (not (>= (point) (cdr limits)))
	 (if (looking-at "^[^ \t\n]+$")
	     (nconcc sig-list (read-match 0)))
	 (forward-line 1))

       (setq ret (list limits info-list sig-list))
       ))
    ret
    ))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-sig-header-info-v2xx  ()
  "Return signature information from X-pgp v2.xx headers.

Reads format:

X-Pgp-Comment: Processed by TinyPgp.el 1.56
X-Pgp-Version: 2.6.3ia
X-Pgp-Charset: noconv
X-Pgp-Signed:
	iQBVAwUBMoijBMC67dVHFB01AQGf3QH/dmgc47fx1tvHYPcuKWIz0Fe7HnWXmd63
	3IBA6vhSqzbUT4nkKL2QJQX/0Z8I9dkmOahSQNKvU/7qsB9Iw8JwpQ==
	=9yu9

Return:
 ((beg . end) (fld fld ..) (signature-data sig ..))"
  (let* ((case-fold-search t)
	 (pbase		"X-Pgp-")
	 (p-re          (concat "^" pbase)) ;pgp regexp for hdrs
	 (psig		(concat p-re "Signed:"))
	 (fld-re        (concat
			 p-re
			 "\\(Version:\\|Charset:\\|Comment:\\|Signed:\\)"
			 ))
	 (hmax          (ti::mt-hmax))
	 val
	 sig-list
	 info-list
	 beg end
	 ret
         )

    (save-excursion
      (PMIN)
      (while (and
	      hmax
	      (< (point) hmax)		;those fwl-line calls may go past...
	      (re-search-forward fld-re hmax t))
	(beginning-of-line)
	(if (null beg)			;record it NOW
	    (setq beg (point)))

	(cond
	 ((looking-at (concat psig "[ \t]*\\([^ \t\n]*\\)"))

	  ;;  Is this the signature itself ? Special handling,
	  ;;  because spreads multiple lines.
	  ;;
	  (setq val (remove-properties (read-match 1)))
	  (if (not (string= "" val))
	      (nconcc sig-list val))

	  (forward-line 1)

	  (while (looking-at "^[ \t]+\\([^ \t\n]+\\)")
	    (nconcc sig-list (remove-properties (read-match 1)))
	    (forward-line 1))
	  )

	 ;; Nope, some additional PGP header
	 (t
	  (nconcc info-list (remove-properties (read-full-line)))
	  ))

	;;  Because there is already one while loop that says fwd-line,
	;;  we don't want to go furher if it stopped us.
	;;
	(if (looking-at (concat p-re  "\\|^\t+"))
	    (forward-line 1))
	)
      (beginning-of-line)
      (setq end (point))
      )

    (if sig-list
	(setq ret (list (cons beg end) info-list sig-list)))

    ret
    ))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-signature-header-info-v3xx ()
  "Return signature information from X-pgp v3.xx headers.

Return:
 '((nil . nil)
   (\"Version: x.x.x\" \"Charset: xxxx\" ...)
   (signature-string sig-string ..)
   )"
  (let ((field  (ti::mt-get-field "X-Pgp-signed"))
	info-list
	sig-list
	elt
	list
	)
    (when field
      (setq list (ti::mt-mime-parse-header field 'downcase))

      ;;  Push adds to the front of list, so beware order of elements
      ;;
      (if (setq elt (assoc "signature" list))
	  (setq sig-list  (cdr elt)))

      (if (setq elt (assoc "comment" list))
	  (push  (concat "Comment: " (car (cdr elt))) info-list  ))

      (if (setq elt (assoc "charset" list))
	  (push (concat "Charset: " (car (cdr elt))) info-list))

      (if (setq elt (assoc "version" list))
	  (push  (concat "Version: " (car (cdr elt))) info-list  ))

      (if info-list
	  (list (cons nil nil) info-list sig-list))
      )))



;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-signature-header-info  ()
  "Return X-pgp header info if X-Pgp header exist."
  (if (ti::mt-pgp-v3xx-p)
      (ti::mt-pgp-signature-header-info-v3xx)
    (ti::mt-pgp-sig-header-info-v2xx)))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-mime-parse-header (header-string &optional downcase)
  "Parse Variable=value HEADER-STRING like and optionally DOWNCASE keywords.

Header-this: var1=value2; var2= val2; var3=\"starts here \"
  \" continues here\"; var4= v1,v2,v3;

The VAL returned is different for continued string. It is a list of
individual parts in the parsed. In this case the whole returned value
would be:

'((var1 . (val1))
  (var2 . (val2))
  (var3 . (\"starts here \" \" continues here\"))
  (var4 . (\" v1,v2,v3\"))
  )

Return:
  ((var . VAL) (var . VAL) ..)"
  (let ((tag-re  "^[ \t]*\\([^ \t\n]+\\)[ \t\"]*=")
	(val-re  "[ \t\"]*\\([^\n\"]+\\)")
	(buffer  (use-tmp-buffer "*tmp*" 'clear))
	name val
	ret
	)

    (job-buffer buffer
      (insert header-string)
      ;; put into same line
      ;;
      (PMIN) (ti::b-replace-regexp "[ \t]*;[ \t]*" 0 "\n")

      ;; Okay now it's in canonical format. First
      ;; pick up signature, then delete it and parse other fields.
      ;;  Version=2.6.3ia
      ;;  Charset=noconv
      ;;
      (PMIN)
      (while (re-search-forward tag-re nil t)
	(setq name (read-match 1)   val nil)
	(cond
	 ((looking-at val-re)		;VALUE at the same line
	  (nconcc val (read-match 1))
	  (forward-line 1)
	  )
	 (t
	  ;;  Multiline
	  ;;
	  (while (progn (forward-line 1)
			(looking-at val-re))
	    (nconcc val (read-match 1))
	    )))

	(if downcase
	    (setq name (downcase name)))

	(push (cons name val) ret)
	))

    (reverse ret)
    ))

;;}}}
;;{{{ PGP public key

;;; ........................................................ &pgp-pkey ...

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgp-pkey-read  (&optional raw kill-file)
  "Read public key block from current point forward. Point is moved.

Input:

  RAW		If non-nil, return only raw public key block.
  KILL-FILE	if non-nil, kill temprory file after statement
		'Key extracted to file ...' Once the file is killed the
		message will be removed from buffer."
  (let* (beg end
	 file
	 ret
         )
    ;;  No temp files are left on disk
    ;;  Remove also the file message from buffer before we read the
    ;;  content.
    ;;
    ;;       Extracting from key ring: '/users/xxx/.pgp/pubring.pgp',\
    ;;       userid "xxx".
    ;;
    ;;       Key for user ID: <xxx@some.fi>
    ;;       512-bit key, key ID 8125CAAA, created 1997/06/05
    ;;
    ;;       -----BEGIN PGP PUBLIC KEY BLOCK-----

    (when (and kill-file
	       (re-search-forward "Key extracted to file.*'\\(.*\\)'" nil t))
      (setq file (read-match 1))
      (ti::b-kill-line)
      (file-delete-safe file)
      )

    (goto-char (point))
    (when (re-search-forward (ti::mt-pgp-pkey-begin-line) nil t)
      (re-search-backward "Key for user ID:") (beginning-of-line)
      (when raw
	(re-search-forward "^-----BEGIN")
	(beginning-of-line)
	)
      (setq beg (point))
      (when (ti::mt-pgp-re-search 'pkeye 'move)
	(forward-line 1)
	(setq end (point))
	)

      (when (and beg end) (setq ret (buffer-substring beg end)))
      )
    ret
    ))


;;}}}
;;{{{ PGP remail

;;; ...................................................... &pgp-remail ...


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgpr-close ()
  "Close reply block by adding '**' to the end.
If there already is '**', do nothing."
  (save-excursion
    (PMAX)
    ;;  Remailers need "**" at the end of encrypted block
    ;;
    (if (not (re-search-backward "^\\*\\*" nil t))
	(insert "\n**\n"))))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgpr-anonymize-headers  (mode &optional no-ins arg1 arg2 hash)
  "Destroy header information according to mode and move it to message body.
This function does nothing if first line is not header.

Input:
  MODE	    'move-to-body moves, all headers to body
	    'move-to-body-maybe, all headers to body only if
	     there is not already hashmarks.
	     arg1 is used for subject       defaults to 'dummy'
	     arg2 is used for organisation  defaults to 'dummy'
  NO-INS     Do not insert the hash headers into body, but return them
             as list instead.
  ARG1 ARG2  used by MODE
  HASH       Use hash marks string other that \"##\"

Return:
  list"
  (let ((hlist  '("In-reply-to" "Organization"
		  "Subject"
		  ))
	(empty  " dummy")
	(full-string "")
	done
	ptr
	list
	str
	ret
	)
    (setq hash (or hash "##")
	  arg1 (or arg1 empty)
	  arg2 (or arg2 empty)
	  )

    (save-excursion
      (when (ti::mt-mail-p)
      (cond
       ((memq mode '(move-to-body move-to-body-maybe))

	;;  First check if hash mark is already there
	;;  If mode is "maybe" we don't add new headers.
	;;
	;;  The regexp matches to the end of line, because you may have
	;;  quoted the message
	;;
	;;  jerry>  ##
	;;  jerry>  Subject:  this here
	;;
	;;
	(PMIN)

	(unless (and (eq mode 'move-to-body-maybe)
		     (re-search-forward (concat hash "[ \t]*$") nil t))
	  (setq ptr hlist)

	  (dolist (elt ptr)
	    (setq str (ti::mt-get-field elt))
	    (when (and str (not (string= empty str)))
	      (setq elt (format "%s: %s\n" elt str))
	      (push elt list)

	      ;;  so that we can match against this later
	      ;;
	      (setq full-string (concat full-string elt))
	      ))

	  (ti::mt-text-start 'move)

	  (when list
	    (setq ret list  done t)
	    (unless no-ins
	      (insert hash "\n")))	;Remailer hash mark

	  )

	;;  Anonymize some headers
	;;
	(if arg1
	    (ti::mt-kill-field "^subject"  arg1))
	(if arg2
	    (ti::mt-kill-field "^organization" arg2))

	(when (and done (null no-ins))
	  (dolist (elt list)
	    (insert elt))) ;  Copy headers inside message
	)

       (t
	(error "Invalid mode [%s]" mode)
	))))
    ret
    ))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgpr-reply-type (property-string)
  "Return remailer reply block type from PROPERTY-STRING.
The 'post' type is not checked, because it relates to Usenet
and can be mixed with other types."
  (if (string-match "cpunk\\|eric\\|penet" property-string)
      (read-match 0 property-string)))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgpr-block (mode &optional type email key latent)
  "Return remailer header string defined by mode.
be sure to have <> in the email, which defaults to `user-mail-address'.

Input:

  MODE	    'epgp -- return encrypted pgp tag.
	    'post -- return simple Newsgroup post block. 'email'
	    contains the address of post remailer.
	    If there is not enough
	    parameters, say for 'tk, the previous one is used: 't

  TYPE	    cpunk   Request-Remailing-To
	    eric    Anon-Send-To
	    penet   X-Anon-To
	    post    Anon-Post-To Usenet
  EMAIL	    Parameter for type
  KEY	    Parameter for type
  LATENT    Parameter for type"
  (let* ((reply
	  (cond
	   ((string= type "cpunk")  "Request-Remailing-To")
	   ((string= type "eric")   "Anon-To")
	   ((string= type "penet")  "X-Anon-To")
	   ((string= type "post")   "Anon-Post-To")
	   ((memq mode '(epgp post)))	;Ok; skip
	   ((error "Unknown type '%s'" type)
	    )))
	 )

    (setq email (or email user-mail-address))

    (cond
     ((equal mode 'epgp)
      "::\nEncrypted: PGP\n\n")

     ((equal mode 'post)
      (concat
       "::\n"
       "Anon-Post-To: " (or email (error "invalid args.")) "\n"
       "Cutmarks: --\n"))


     ((and (stringp email) (stringp key) (stringp latent))
      (format "::\n%s: %s\nEncrypt-Key: %s\nLatent-Time: %s\n"
	      reply email key latent))

     ((and (stringp email) (stringp latent))
      (format "::\n%s: %s\nLatent-Time: %s\n" reply email latent))

     ((and (stringp email) (stringp key))
      (format "::\n%s: %s\nEncrypt-Key: %s\n" reply email key))


     ((or (stringp email))
      (format "::\n%s: %s\n" reply email))

     (t
      (error "Wrong args '%s' '%s'" mode type )

      ))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgpr-reply-block (pgp-email)
  "Return reply block header.
Should be inserted just before PGP crypted message to PGP-EMAIL."
  (format "Reply-Block:\n::\nAnon-To: %s\n\n" pgp-email))




;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-pgpr-parse-levien-list  (&optional buffer control-list)
  "Parse remailer list finger <remailer-list@kiwi.cs.berkeley.edu>.
The parsing starts from current point forward.

Input:

  BUFFER	    defaults to current buffer
  CONTROL-LIST	    '(remailer-alias  (prop-no-list) [(prop-add-list)])
		    This control list says 'if REGEXP matches the
		    email address, remove all properties listed in
		    prop-no-list and add all properties listed in
		    prop-add-list.

		    So, if you're sure that the levien-list has some
		    faulty entries, eg. say remailer@replay.com doesn't
		    have feature 'ek' although levien list contains that,
		    your control-list is like this. The ek property
		    is removed even if the list says otherwise.

		    '(\"replay\" '(\"ek\"))

Return:

  '((alias remailer property_string (property property ...))
    (alias remailer property_string (p p p ..))
    )

  The properties are sorted: cpunk mix pgp..."
  (let ((re  (concat
	      "^[ \t]*$remailer{[\"']\\(.*\\)[\"']}.*=[ \t]*[\"']"
	      "<\\(.*\\)>[ \t]+\\(.*\\)[\"']"
	      ))
	a r p
	blocks
	ret
	elem
	list
	)
    ;;  The list is in Perl hash array format in case you're interested...
    ;;
    (job-buffer (or buffer (current-buffer))
      (while (re-search-forward re nil t)
	(setq a (read-match 1) r (read-match 2)  p (read-match 3))
	(setq blocks (ti::s-get-blocks p "[ \t]+"))
	(setq blocks (sort blocks 'string<))

	(when (and control-list
		   (setq elem (assoc a control-list)))
	  (setq list (nth 1 elem))
	  (dolist (elt list) (setq blocks (delete elt blocks)))

	  (setq list (nth 2 elem))
	  (dolist (elt list) (push elt blocks))

	  ;;  We used this now, remove from list
	  (setq control-list (delete elem control-list)
		p            (mapconcat 'concat blocks " "))
	  )

	;; features In alphabetic order
	(setq p (mapconcat 'concat blocks " "))

	(push (list a r p blocks) ret)
	))
    ret
    ))

;;}}}

;;{{{ email addresses

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-email-from-string  (string)
  "Return email addresses from STRING.
The addresses must have @ character. Surrounding <> characters are removed.
If STRING is nil this function does nothing."
  (let (tmp
	ret
	)
    ;; Using buffer is faster that reading string
    (when string
      (setq tmp (use-tmp-buffer ti:mt-tmp-buffer 'clear))
      (job-buffer tmp
	(insert string)
	(setq ret (ti::mt-email-find-region))
	))
    ret
    ))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-email-find-region (&optional beg end no-dupes)
  "Find all email addresses within region BEG END (defaults to buffer).
The email addresses must contain @. Surrounding <> characters are removed.

Input:
  BEG	    ,region start; defaults to `point-min'
  END	    ,region end; defaults to `point-min'
  NO-DUPES  ,flag; if non-nil then cache only unique entries."
  (let (list
	elt
	)
    (save-excursion
      (setq beg (or beg (point-min))
	    end (or end (point-max)))

      (keep-lower-order beg end)

      (goto-char beg)
      (while (re-search-forward
	      "[^ '\",:<\t\n(]+@[^ '\">:,\t\n)]+"
	      end
	      t)
	(setq elt (remove-properties (read-match 0)))
	(if (or (null no-dupes)
		(and no-dupes
		     (not (member elt list))))
	    (push elt list))
	)
      list
      )))

;;}}}
;;{{{ parsing

;;; ......................................................... &parsing ...

;;; ----------------------------------------------------------------------
;;;   (ti::mt-t-parse-name)
;;;
(defun ti::mt-t-parse-name ()
  "This is a test function, do not call from programs.

Because the `ti::mt-parse-name' is quite complicated,
and slightest modification may render it, this functions tests
that the old functionality is preserved in spite of changes."
  (let (list
	e1 e2 stat ptr
	)

    (setq list
	  '(
	    "<jari.aalto@poboxes.com> (Finland, pgp id 512/47141D35)"
	    "(Rune Juntti[FRONTEC Pajala]) <Rune.Juntti@pjl.frontec.se>"
	    "shahramn@wv.mentorg.com (shahram_najm@mentorg.com)"
	    "(santini@chambord.univ-orleans.fr)"
	    "Jerome Santini <santini@chambord.univ-orleans.fr>"
	    "halladay@stlhp56.ug.eds.com (Harry Halladay - EDS St. Louis)"
	    "etxaks@aom.ericsson.se (Ake Stenhoff TM/PMD 83442 3003)"
	    "Jari-Teppo Aalto jari.aalto@poboxes.com"
	    "TOMMI <\"TNCLUS::TSYVANEN\@tne01.tele.nokia.fi\""
	    "\"NTC01::KARINIEMI\"@tne01.tele.nokia.fi"
	    "Job.Ganzevoort@cwi.nl"
	    "Marla=Bush%aoa.rdt%OS.DC@Ban-Gate.AoA.DHHS.EDU"
	    "mernst@theory.lcs.mit.edu (Michael Ernst)"
	    "\"/G=Jamie/S=Lokier/OU=comlab/O=oxford/PRMD=UK.AC/ADMD= /C=GB/\"@research.nokia.fi\""
	    "\"wayne (w.d.) bell\" <wbell@bnr.ca>"
	    "Ted Stern <stern@amath.washington.edu>"
	    "\"Joseph B. Ottinger\" <joeo@freenet.scri.fsu.edu>"
	    "\"Mia Ingstr|m puh. 2407 600\" <INGSTROM@tne01.tele.nokia.fi>"
	    "\"stephane (s.) boucher\" <sbo@bnr.ca>"
	    "stern@amath.washington.edu (L.G. \"Ted\" Stern)"
	    "eh@eclipse.in-berlin.de (H.J.Ehling)"
	    "jaalto@mac.uu.net.fi \(Jari Aalto\)"
	    "jaalto@mac.uu.net.fi \(Jari S. Aalto\)"
	    "\"Mia Ingstr|m puh. 2407 600\" Mia@mac.uu.net.fi"
	    "\"John B. Doe\" <foo@mac.uu.net.fi>"
	    ))

    (setq ptr list)
    (dolist (n ptr)
      (setq stat  (ti::mt-parse-name n))
      (setq e1 (nth 0 stat)) (setq e2 (nth 1 stat))
      (read-from-minibuffer (concat "TEST>>" e1 "," e2 "<"))
      )
    ))



;;; ----------------------------------------------------------------------
;;; (ti::mt-t-parse-name)
;;;
(defun ti::mt-parse-name (line)
  "Try to parse various formats of 'From:' fields.
Supposes that the 'From:' keyword is removed from the LINE.

Returns:
  list		,(firstname surname)
  nil           ,if cannot parse both"
  (let* ((re-A		"[-a-zA-Z0-9.{|]")
	 (re-AG		(concat "\\("  re-A "+\\)"))

	 ;;  'From: Jari-Teppo Aalto jari.aalto@poboxes.com'
	 (fs-re2  (concat re-AG " +" re-AG))

	 ;;  'TOMMI <\"TNCLUS::TSYVANEN\@tne01.tele.nokia.fi\"'
	 (fs-vax  (concat "^" re-AG "[ \t<\"]+[A-Z]+::" re-AG))

	 ;;  '\"NTC01::KARINIEMI\"@tne01.tele.nokia.fi'
	 ;;  This is incomplete Name, it does not contain NAMES at all, but
	 ;;  we consider mail name as surname. The first group-RE is dummy.
	 (fs-vax2 (concat re-AG "::" re-AG))


	 ;;  'Job.Ganzevoort@cwi.nl', where person's name is complete
	 ;;  address
	 (fs-fse    (concat re-AG "\\." re-AG "@" ))

	 ;;  matches gateway-type addresses
	 ;;  'Marla=Bush%aoa.rdt%OS.DC@Ban-Gate.AoA.DHHS.EDU'
	 (gtw-re1    (concat re-AG "=" re-AG "%" ))


	 (q-no-re   ti:mt-parse-name-not-accept)

	 (mail      (or (ti::mt-parse-email line) ""))
	 (account   (if (= 2 (length mail))
			(nth 0 mail)
		      "#@$@#$@#$"))	;just some dummy

	 fn sn				;first, surname
	 pick
	 w w1 w2
	 D				;debug
	 beg end beg1 end1 beg2 end2
	 tmp
	 list

	 )
    (if D (setq D D))			;XE 19.14 ByteComp silencer, no-op


    (catch 'found
      ;;  It's most important that the match test are made IN THIS ORDER
      ;;  - Quote test cannot precede vax name test.
      ;;  - Try most restrictive first.

      ;; ..............................................................
      ;;  VAX is identified by "::" marks
      ;;
      (when (o-string-match "::" line)

	(setq list (ti::mt-get-2re fs-vax line))
	(when (not (string= "" (nth 0 list)))
	  (setq D "vax1")
	  (throw 'found t))

	(setq list (ti::mt-get-2re fs-vax2 line))
	(when (not (string= "" (nth 0 list)))
	  (setq D "vax2")
	  (throw 'found t))
	)

      ;; ............................................................
      ;; Try gateway addresses, rare, but seen in net still
      ;;
      (when (o-string-match "%" line)
	(setq list (ti::mt-get-2re gtw-re1 line))
	(when (not (string= "" (nth 0 list)))
	  (setq D "gtw1")
	  (throw 'found t)
	  ))

      ;; X.400 address
      ;;
      (when (o-string-match "/G=\\(.*\\)/S=\\([^/]+\\).*C=" line)
	(setq fn (read-match 1 line)
	      sn (read-match 2 line)
	      )
	(when (and fn sn)
	  (setq list (list fn sn)   D "gateX400")
	  (throw 'found t)
	  ))


      ;; .................................................................
      ;; mike.gordon@bix.com
      ;;
      (when (o-string-match fs-fse line)
	(setq list (ti::mt-get-2re fs-fse line))
	(when (not (string= "" (nth 0 list)))
	  (setq D "mike.gordon")
	  (throw 'found t)
	  ))


      ;; ............................................................
      ;; And the rest , is there paren or ""  somewhere ?
      ;;

      ;;  If this is a full email string Joe@foo.com
      ;;  then get only the first part.
      ;;
      (when (and (setq tmp (strmatget "^\\([^ \t]+\\)[@%][^ \t]+$" 1 line))
		 (setq tmp (strmatget re-AG 1 tmp)))
	(setq D "email")
	(setq list (list tmp ""))
	(throw 'found t)
	)

      ;;   - if we get multiple match "stephane (s.) boucher" ,
      ;;     (L.G. \"Ted\" Stern) , pick the one that's longer.
      ;;
      (if (o-string-match "\"\\(.*\\)\"" line)
	  (setq beg1 (match-beginning 1)  end1  (match-end 1)))

      (if (o-string-match "[(]\\(.*\\)[)]" line)
	  (setq beg2 (match-beginning 1)  end2  (match-end 1)))

      (cond
       ((and beg1 beg2)
	(if (> (- end1 beg1) (- end2 beg2))
	    (setq beg beg1  end end1)
	  (setq beg beg2  end end2)))
       (beg1
	(setq beg beg1  end end1))
       (beg2
	(setq beg beg2  end end2)))

      ;; ...  ...  ...  ...  ...  ...  ...  ...  ...  ...  ...  ...  ...

      (cond
       (beg
       	;;   - Get list of words into W
	;;   - Someone wrote M. "Mack" Monroe, so the " is included
	;;     in words separate list
	;;   - The latter picks only NON-ABBREVIATED names, non-phones..
	;;     M. "Mack" Monroe --> Mack Monroe
	;;

	(setq pick (substring line beg end))
	(setq w (ti::s-get-blocks pick "[][@%. \"]+" "[.]" ))


	(setq D "standard")
;;;	(d! "w-1" w)

	(let ((case-fold-search nil))	;case is important !!
	  (setq w			;returned word list
		(list-find
		 w q-no-re
		 '(not (string-match arg (car table)))
		 'all-items
		 )))

;;;	(d! "w-2" w)

	(cond
	 ((> (length w) 2)		;too much abbrev names
	  ;;  pick first and account or last word

;;;	  (setq W w AC account)

	  (setq w1 (nth 0 w)  w2 (nth (1-(length w)) w)  )

	  (setq tmp (list-find
		     w account
		     '(o-string-match (car table) arg)))
	  (if tmp			;account name found
	      (setq w2 tmp))

	  (setq list (list w1 w2)))

	 ((= 2 (length w))
	  (setq w1 (nth 0 w)  w2 (nth 1 w))
	  (setq list (list w1 w2)))

	 ((eq 1 (length w))
	  (setq list w))

	 (t
	  nil))

	(if list (throw 'found t))
	;; ...  ...  ...  ...  ...  ...  ...  ...  ...  ... ConditionEnd .
	))

      ;; .................................................................

      (setq list (ti::mt-get-2re fs-re2 line))
      (when (not (string= "" (nth 0 list)))
	(setq D "2.1")
	(throw 'found t))



      ;; ... ... ... ... ... ... ... ... ... ... ... ... ... CatchEnd ..
      )


;;;    (d! "parsed" D  list)





    ;;   what should we return ?
    (if (and (string= (nth 0 list) "")
	     (string= (nth 1 list) ""))
	nil
      list)
    ))





;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-parse-email (line)
  "Try to parse various formats of 'From:' field from LINE.
Input is full LINE, possibly containing 'From' keyword.

Returns:
  list		,(usrname site)
  nil		,if cannot parse."
  (let* (
	 account
	 site
	 tmp

	 ;; '.' is for firstname & surname combination
	 ;; '=' is for gateway form
	 ;; '|{' are scandinavian characters in name
	 ;; '+' Believe or not, but I just saw account name like
	 ;;     "Stephen M. Lacy" <sl31+@andrew.cmu.edu>

	 (A "[-a-zA-Z|{0-9_=.+]+")	; alphabet
	 (As "[-a-zA-Z0-9.%]+")		; site name

	 ;;  Note that username can have scandinavian {| marks
	 ;;  Normal site name
	 ;;  o   Simon.Marshall@mail.bar.foo.fi (Simon Marshall)
	 (re1 (concat "\\(" A "\\)@\\(" As "\\)"  ))

	 ;;  Marla=Bush%aoa.rdt%OS.DC@Ban-Gate.AoA.DHHS.EDU
	 (re2 (concat "\\(" A "\\)\\(%" As "\\)"  ))


	 ;;  VAX address <"TNCLUS::TSYVANEN"@mailer.foo.fi>
	 (re-vax (concat "\\(\"" A "::" A "\"\\)@\\(" As "\\)"  ))
	 em				; email

	 ;;  "/G=Jamie/S=Lokier/OU=comlab/O=oxford/PRMD=UK.AC...
	 (re-x400
	  (concat "/G=\\([^/]+\\)/S=\\([^/]+\\)" ;fn sn
		  "/OU=\\([^/]+\\)/O=\\([^/]+\\)"
		  "/PRMD=\\([^/]+\\)"
		  ))
	 )
    (catch 'found
;;;      (setq LINE line RE re-x400)

      (if (null (o-string-match re-x400 line)) nil
	(setq account (concat (read-match 1 line) "." (read-match 2 line)))
	(setq site    (concat (read-match 3 line) "." (read-match 4 line)))

	;;  Now switch the last items PRMD=UK.AC --> ac.uk
	(setq tmp (read-match 5 line))
	(setq tmp (ti::s-get-blocks tmp "[.]"))
	(setq site (downcase (concat site "." (nth 1 tmp) "." (nth 0 tmp))))
	(setq em (list account site))
	(throw 'found t)
	)

      (setq em (ti::mt-get-2re re-x400 line))
      (if (not (string= "" (nth 0 em))) 	  (throw 'found t))

      (setq em (ti::mt-get-2re re1 line))
      (if (not (string= "" (nth 0 em))) 	  (throw 'found t))

      (setq em (ti::mt-get-2re re2 line))
      (if (not (string= "" (nth 0 em))) 	  (throw 'found t))


      (setq em (ti::mt-get-2re re-vax line))
      (if (not (string= "" (nth 0 em))) 	  (throw 'found t))
      )

    (if (< (length (nth 0 em)) 1)
	(setq em nil))
    em
    ))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-parse-received (&optional not-matching)
  "Search all 'Receive' fields and read site names followed by 'from' 'by'.
Duplicate entries are not added.

Point must be at the beginning of headers to search, and
point movement is not preserved `save-excursion'.

Comment

    Where do you need this? If you got spam mail, you can use
    this function to find out from where the mail originated and
    send complaint to postmasters of all those sites.

Input:

    NOT-MATCHING

    If read entry matches this regexp it is not included in
    return list

Return:

    list

Example of headers:

Received: from mailhost.worldnet.att.net ([206.85.117.127])
          by mtigwc02.worldnet.att.net (post.office MTA v2.0 0613 )
          with SMTP id AAD8244; Sun, 23 Mar 1997 23:03:10 +0000
Received: from mail.msss.v.com [atl.asd.com [234.454.54]]
	  by mediabrokers.cobracomm.com (8.8.5/8.6.5) with
	  SMTP id GAA07901 for <box17@mediabrokers.cobracomm.com>"
  (let* (ret
	 str
	 beg end
         )
    (while (re-search-forward "^Received:" nil t)
      (setq beg (point))
      (setq end (ti::mt-next-field-start))
      (while (and beg end
		  (re-search-forward
		   "\\(from\\|by\\) +\\([^] \t(]+\\)"
		   end t))
	(setq str (read-match 2))
	(if (and (or (null not-matching)
		     (not (string-match not-matching str))
		     )
		 (not (member str ret)))
	    (push str ret))
	))
    (reverse ret)
    ))

;;;  ............................................................ &date ...

;;; ----------------------------------------------------------------------
;;; Try this:  (ti::mt-parse-date "Wed, 21 Jul 93 09:26:30 EST")
;;;
(defun ti::mt-parse-date (str)
  "Try to parse date field.
Supposes that keyword 'date: ' -- ie. garbage is removed from the STR.
No leading spaces allowed. The STRING must be in regular mailer format.

Returns:
  list		,(dd mm yy tt wd m yy tz)
		 \"\" in fields which weren't identified.

  list members:
  0 dd     day		,nbr
  1 mm     month	,nbr
  2 yy     year		,2 last nbrrs
  3 tt     hh:mm	,nbr:nbr
  4 wd     week day	,str eg. \"Mon\"
  5 m      month	,str eg. \"Jun\"
  6 yy     year		,2 last nbrs
  7 tz     time zone	,eg, [+-]nnnn, where n = nbr"
  (let* ((wd "")			; inital values
	 (dd "")
	 (mm "")
	 (m  "")
	 (yy "")
	 (tt "")
	 (tz "")
	 p
	 pyy
	 (ofs 0) (ofsdd 0)

	 (rAaa	 "[A-Z][a-z][a-z]")
	 (rd	 "[0-9][0-9]?")                 ; typical day nbr
	 (rd4	 "[0-9][0-9][0-9][0-9]")	; typical year nbr (regexp day)
	 (rt	 "[0-9:]+")			; time
	 (rz	 "\\([+-][0-9]+\\|[A-Z]\\)+")	; zone

	 (re-y4				       ; - year 4
	  (concat rd4 " " rt))		       ; 1994 08:52:25
	 (re-y2				       ; - year 2
	  (concat rd " " rt))		       ; 94 08:52:25

	 (re-wd1			       ; - week day format
	  (concat "^" rAaa ", " rd " " rAaa))  ; Mon, 24 Oct
	 (re-wd1-1 (concat re-wd1 " " re-y4)) ; Mon, 24 Oct 1994 08:52:25 +0200
	 (re-wd1-2 (concat re-wd1 " "re-y2))  ; Mon, 24 Oct 94 08:52:25 GMT

	 ;;  Wed Oct 14 22:21:05 1987
;;;	 (re-wd2
;;;	  (concat "^" rAaa " " rAaa " " rd)) ;  Wed Oct 14
;;;	 (re-wd2-1			     ;  'current-time-string'
;;;	  (concat re-wd2 " " rt " " rd4))    ;  Wed Oct 14 22:21:05 1987

         (re-dd1 (concat "^" rd " " rAaa))   ; 24 Oct
	 (re-dd1-1			     ;
	  (concat re-dd1 " " re-y4))	     ; 24 Oct 1994 00:28:04 GMT
	 (re-dd1-2			     ;
	  (concat re-dd1 " " re-y2))	     ; 24 Oct 94 00:28:04 GMT


	 )
    ;;   We get rid of all the extra spaces just leaving one to separate
    ;;   items..
;;;    (setq str (ti::s-sqz (ti::s-1-space str)))

    (catch 'out

      ;;   "Tue, 1 Nov 1994 8:52:36 +0300 (EET)"
      (cond
       ((o-string-match re-wd1 str)
	(if (o-string-match re-wd1-1 str) ; select YEAR format
	    (setq pyy 4  ofs 2))	; offset

	(if (o-string-match re-wd1-2 str)	; set year pointer
	    (setq pyy 2))

	(if (null pyy)
	    (throw 'out t))

	(setq wd (substring str 0 3))
	;;    we remove trailing space  "2 " --> 2 --> "2"
	;;    it's funny that the day can also be "02" ,which
        ;;    is checked with the 2nd OR case
	(setq dd (int-to-string (string-to-int (substring str 5 7))))

	(if (or (> (string-to-int dd) 9)
		(string= (substring str 5 6) "0"))
	    nil
	  (setq ofsdd -1)
	  (setq ofs (- ofs 1)))

	(setq m (substring str (+ 8 ofsdd) (+ 11 ofsdd) ))
	(setq mm (int-to-string (ti::d-s-month-n m)))

	(setq p (+ 12 ofs))		; start of -[YY] format
	(setq yy (substring str p  (+ p 2)))
	(setq tt (substring str (+ p 3)  (+ p 8)))

	(if (> (length str) 25)        ;it's only "Wed, 14 Dec 1994 08:35:51"
	    (setq tz (substring str (+ p 12)  nil)))
	)

       ;; ... ... ... ... ... ... ... ... ... ... ... ... ... ... ... ...
       ((o-string-match re-dd1 str)
	(if (o-string-match re-dd1-1 str) ; select YEAR format
	  (setq pyy 4   ofs 2))		; offset

	(if (o-string-match re-dd1-2 str)
	    (setq pyy 2))

	(if (null pyy)
	    (throw 'out t))


        (setq dd (int-to-string (string-to-int (substring str 0 2))))
	(if (or (> (string-to-int dd) 9)
		(string= (substring str 0 1) "0"))
	    nil
	  (setq ofsdd -1)
	  (setq ofs (- ofs 1)))

        (setq m (substring str (+ 3 ofsdd) (+ 6 ofsdd) ))
        (setq mm (int-to-string (ti::d-s-month-n m)))

	(setq p (+ 7 ofs))	; start of -[YY] format
	(setq yy (substring str p  (+ p 2)))
	(setq tt (substring str (+ p 3)  (+ p 8)))

	(if (or (< (length str) 25)
	     (null (o-string-match (concat rz "$") str)))
	    nil				;skip
	  (setq tz (substring str (+ p 12)  nil)))

	))				; cond end
      )
    ;; ... ... ... ... ... ... ... ... ... ... ... ... ... ... . catch ...
    (list dd mm yy tt wd m yy tz)
    ))

;;}}}
;;{{{ misc

;;; ............................................................ &misc ...


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-signature-insert-break   (&optional point)
  "Insert RFC signature break to current point or POINT if no sig break exist.
According to RFC there must be \"-- \\n\" before signature. The extra space
separates the signature from eg. digest messages that come with \"--\\n\"

We try to find this string forward and it is not there we add one."
    (save-excursion
      (if point (goto-char point))
      (if (null (re-search-forward "^-- \n" nil t))
	  (insert "-- \n"))
      ))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-yank (&optional prefix)
  "Yank message to current point and add optional PREFIX. GNUS/RMAIL."
  (let* (p
	 (yb (ti::mt-mail-buffer-name))	;where is the yank buffer ?
	 ;;  See this mail is called from GNUS
	 ;;
	 ;;  - If GNUS isn't loaded, set buf name to nil
	 ;;
	 (gnus-buf (and (boundp 'gnus-article-buffer)
			(symbol-value 'gnus-article-buffer)))

	 ;;  Test if gnus-reply; the buffers are the same
	 ;;
	 (gnus-r (and gnus-buf
		      (string= gnus-buf yb)))
	 )
    (save-excursion
      (setq p (point))
      ;;  (mail-yank-original '(4))     ; mimic C-u C-c C-y == no indent
      ;;  - bypass all, see sendmail::mail-yank-original
      ;;    this is more robust, and runs no extra hooks
      ;;  - If in GNUS, the buffer will be *Article*, which is
      ;;    narrowed to headers...widen the buffer before yanking.

      (if (null gnus-r)
	  (progn                        ; normal mail
	    (mail-yank-original '(4)))
	(save-excursion (set-buffer yb) (widen))
	(insert-buffer yb))

      (PMAX)
      (delete-blank-lines)

      (if prefix
	  (string-rectangle p (point-max)  prefix))

      )))



;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-trim-buffer  ()
  "Trim email message so that there are no trailing white spaces.
- at the beginning of message
- at the end of each line
- at the end of message.

If cannot find text start point, uses `point-min'. The point is not preserved.

Return:
  t	    ,there is no text. All white spaces were removed
  nil	    ,trimming done."
  (let ((beg  (ti::mt-text-start))
	ret
	)
    (goto-char beg)
    (ti::b-replace-regexp "[ \t]+$" 0 "") ;right hand spaces (ragged lines)
    (goto-char beg)
    ;;   Beginning of email message
    (ti::b-trim-blanks beg (point-max))
    (ti::b-delete-until-non-empty-line nil beg)

    (ti::b-delete-until-non-empty-line 'backward (point-max))
    (forward-line 1)

    ;;  Any text left ? Signing empty file is not sensible...
    ;;
    (if (eq (point) beg)
	(setq ret t)
      ;;  Note: User may write message "123" to the body, but we must
      ;;  require final newline every time: "123\n", the trim
      ;;  command will remove any exeessive newlines.
      ;;
      (PMAX)
      (if (not (char= (preceding-char) ?\n))
	  (insert "\n"))
      )
    ret
    ))




;;}}}

;;{{{ fields, headers

;;;  .......................................................... &fields ...


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-field-start (field-re &optional move max)
  "Return starting point of FIELD-RE or nil. Optionally MOVE to it.

Supposes that field has following format, the cursor -!- position
signifies returned point.

  field:-!-

Input:
  FIELD-RE	,field regexp
  MOVE		,should we move to found point? (beginning-of-line)
  MAX		,search until MAX point"
  (let (ret
	)
    (save-excursion
      (PMIN)
      (if (null (re-search-forward field-re max t))
	  nil
	(beginning-of-line)
	(if (null (re-search-forward ":" max t))
	    nil
	  (setq ret (point))
	  )))
    (if (and move ret)
	(goto-char ret))
    ret
    ))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-next-field-start (&optional move back max)
  "Return starting point of next field or nil. Optionally moves to field.

Note:

 If you're somewhere else than inside header area, the return value
 is not defined.


Input:

  MOVE		move to point
  BACK		move backward (field start)
  MAX		search until this point. PLEASE USE THIS TO LIMIT SEARCH

Return:

  point
  nil"
  (let ((func (if back 're-search-backward 're-search-forward))
	opoint point
	ret
	)
    (save-excursion
      (if (null back)
	  (end-of-line))

      (if (and (bobp) back)		;first field
	  (setq ret (point))

	;;   Next line must have text, otherwise the headers have ended
	;;   alredy
	;;
	;;   Header1:
	;;   Header2:
	;;
	;;   BODY-OF-TEXT
	;;
	(cond
	 ((save-excursion
	    (forward-line 1)
	    (looking-at ".*[a-zA-Z0-9]")
	    )
	  (setq opoint (point))
	  ;;  In the last field, the previsu regexp skips too much,
	  ;;  see where the cursor (*) is. We search backward if possible
	  ;;  to find header separator (empty line)
	  ;;
	  ;;  Header:
	  ;;    last header text
	  ;;
	  ;;  *BODY
	  ;;
	  (when (progn
		  (when (and (setq point (funcall func "^[^ \t]" max t))
			     (eq func  're-search-forward))
		    (goto-char opoint)	;Try again
		    (if (re-search-forward "^$" nil t)
			;; no, it was further ahead, use previous search pos
			(if (< (point) point)
			    (setq point (point)))))
		  point
		  )
	    (goto-char point)
	    (beginning-of-line)
	    (setq ret (point))
	    ))
	 (t
	  ;;  Hm, next line is empty line, not a field for us any more.
	  ;;
	  ))))

    (if (and move ret)
	(goto-char ret))

    ret
    ))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-kill-field-in-body   (list)
  "Kill list of field that are inserted into BOdY of message."
  (safe-narrow (ti::mt-text-start) (point-max)
    (dolist (header list)
      (ti::mt-kill-field header)
      )))

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-kill-field (field-re &optional replace-str)
  "Delete header field. Remember to supply Anchor '^' in FIELD-RE.

Input:

  FIELD-RE	,Any regexp matching a line
  REPLACE-STR	,replace field content with this

Return:
  t		,field changed or killed
  nil		,nothing done [field not exist]"
  (let ((hdr-end  (ti::mt-hmax))
	beg
	end
	)

    (if replace-str
	(setq replace-str (ti::s-verify-ends replace-str " " nil 'beg)))

    (save-excursion
      (when (and hdr-end
		 (setq beg  (ti::mt-field-start field-re 'move))
		 (setq end  (ti::mt-next-field-start))
		 (<= end hdr-end)
		 )
;;;      	(setq F field-re B beg E end)

	(if replace-str
	    (progn
	      (delete-region beg end)
	      (insert (concat replace-str "\n")))
	  (beginning-of-line)
	  (delete-region (point) end)
	  t
	  )))
    ))



;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-get-field-1  (field)
  "Read FIELD by finding regexp matching '^FIELD:'.
Starting searching from the beginning of buffer. You are encouraged to call
this function instead of `ti::mt-get-field' if you want to get the
field information fast eg. in `post-command-hook'.

This function is not as reliable as `ti::mt-get-field', because
the search is not limited to header area, but for regular headers
you can use this function safely."
  (let ((re  (format "^%s:" field))
	beg end
	)
    (save-excursion
      (PMIN)
      (if (and (re-search-forward re nil t)
	       (setq beg (point)
		     end (ti::mt-next-field-start)))
	  (buffer-substring beg (1- end)))
      )))

;;; ----------------------------------------------------------------------
;;; - This is almost the same as mail-utils.el/mail-fetch-field,
;;;   but offers more control. It can get citated fields too, if
;;;   ANY parameter is non-nil.
;;; - And it returns _strict_ content of the field, fetch-field strips
;;;   spaces away.
;;;
(defun ti::mt-get-field (field &optional any mode)
  "Return field content.

Input:

   FIELD         field name without anchor '^' and char ':'
   ANY           return any field. When non-nil, drops anchor ^
                 from the  ^field: criteria
   MODE		 nil	read the field as is, returning all chars
			after the ':'
		 t	If field has only spaces, Return nil
		 'pure	Include header name as well as content.

Return:
   nil or contents of field."
  (let ((case-fold-search t)		;ignore case = t
	(re (if any
		(concat field ":")  ; pick first one met
	      (concat "^" field ":"))); require STRICT HEADER

	(hmax (if any nil (ti::mt-text-start)))
	beg end
	ret
	)



    (save-excursion
      (when (and (setq beg (ti::mt-field-start re 'move hmax))
		 (setq end (ti::mt-next-field-start nil nil hmax)))

	(when (and (eq mode 'pure)
		   (looking-at "[\t ]*[^\n\t ]+")) ;not empty
	  (beginning-of-line)
	  (setq beg (point)))

	(setq ret (buffer-substring beg (1- end)))
	))

    (cond
     ((and mode
	   (stringp ret)
	   (string-match "^[ \t\n\r]*\\'" ret)
	   )
      (setq ret nil)))

    ret
    ))



;;; ----------------------------------------------------------------------
;;; - If you want simple filed adding to your mail, then have a look
;;;   at this instead:
;;;
;;;     (defconst my-mail-info-string "Emacs RMAIL in 19.28")
;;;	(setq mail-default-headers
;;;	      (concat
;;;	       "X-info: " my-mail-info-string "\n"))
;;;
;;;
;;;
(defun ti::mt-add-field (field text &optional look-field mode replace)
  "Add FIELD and puts TEXT into it.
If field already exist, replaces field text.
By default, field is added to the end of header.

Input:

  FIELD
  TEXT		\\n at end is optional. _No_ colon and _no_ spaces.
  LOOK-FIELD	new field will be added after this. _No_ colon at end.
		if MODE is non-nil, field is added before this field

		If there is no LOOK-FIELD, nothing is done and nil
		is returned.

  MODE		see look-field
  REPLACE	if non-nil Any previous field is removed. You probably
		want to set this flag to non-nil if you only want unique
		field names.

Return:
  t          ,something done
  nil        ,nothing done, maybe look-field doesn't exist ?"
  (let* ((field-re (concat "^" field ":"))
	 )
    (when text
      (setq text (ti::s-verify-ends text "\n"))
      ;;  There must be leadig space after field keyword.
      ;;
      (setq text (ti::s-verify-ends text " " nil 'beg))
      )
    (save-excursion
      (cond
       (look-field

	(if replace (ti::mt-kill-field field-re)) ;Remove

	(when (and look-field
		   (ti::mt-field-start (concat "^" look-field) 'move))
	  (unless mode			;use only forward
	    (ti::mt-next-field-start 'move mode))

	  (beginning-of-line)
	  (insert (concat field ":" text))
	  t
	  ))
       (t				;add to the end
	(if (ti::mt-get-field field)
	    (ti::mt-kill-field field-re text)
	  (ti::mt-hmax 'move)
	  (insert (concat field ":" text)))
	t
	))
      )))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-add-to-field-string  (field string &optional sep)
  "Find FIELD and add STRING to the. Field is created if it does not exist.

Input:
  FIELD	    ,string WITHOUT colon or spaces.
  STRING    ,added text
  SEP	    ,defaults to comma"
  (or sep
      (setq sep ","))

  (save-excursion
    (PMIN)
    (if (null (re-search-forward (concat "^" field ":") nil t))
	(ti::mt-add-field field (concat " " string))
      (ti::mt-hmax 'move)
      (skip-chars-backward " \n\t")
      (insert sep string)
      )))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-kill-field-elt (re &optional field)
  "Kill all elts matching RE from FIELD, which defaults to cc.
Elements are supposed to be separated by commas.

Example:

   To: him@site.com
   CC: me@site.com, you@site.com

   ;;  If called with this
   ;;
   (ti::mt-kill-field-elt \"me\")
   --> To: him@site.com
   --> CC: me@site.com

   ;; If called with this; all elts are matched and thus the
   ;; field is removed
   ;;
   (ti::mt-kill-field-elt \".\")
   --> To: him@site.com"
  (let* (flag
	 str fld
         )
    (setq field (or field "CC"))

    (when (setq fld    (ti::mt-get-field field))
      (setq fld    (ti::s-del-re "[\n\f\t ]+" fld)) ;remove spread lines
      (setq fld    (ti::s-get-blocks fld "[,]+")) ; divide into items

      ;; ... ... ... ... ... ... ... ... ... ... ... ...  remove items . .
      (setq fld
            (list-find  fld re
			'(not (string-match arg (car table)))
                        'all-items))

      ;; ... ... ... ... ... ... ... ... ... ... ... . build up string . .
      (dolist (elt fld)
        (if (null flag)
            (setq flag  t               ;done 1st line
                  str   (concat " " elt))
	  (setq str (concat
		     str ", " elt
		     (if (> (+ (length str) (length elt))  70)
			 "\n  "  "")
		     ))
	  ))

      ;; ... ... ... ... ... ... ... ... ... ... ... ...  write new fld . .
      (if str
          (ti::mt-kill-field (concat "^" field) str) ;replace
	;;  Remove whole field, all entries were discarded.
	;;
	(ti::mt-kill-field (concat "^" field)))
      )
    ))



;;; ----------------------------------------------------------------------
;;; - This is mainly for converting your mail to anon post by
;;;   removing any headers you might have added.
;;;
(defun ti::mt-kill-non-rfc-fields (&optional list)
  "Kill all non RFC fields unless LIST (HEADER-NAME-SYMBOL .. ) list is given.

References
  `ti::mt-required-headers'    ,default rfc headers"
  (let ((ptr (or list
		 (ti::mt-required-headers)
		 (error "(ti::mt-required-headers) returned nil")
		 ))
	(case-fold-search t)
	fld
	list
	)
    ;;  First we gather all valid headers to list
    (dolist (elt ptr)
      (setq fld (symbol-name elt))
      (when (setq elt (ti::mt-get-field fld))
	(nconcc list (format "%s:%s" (capitalize fld) elt)))
      )

    ;; Now we kill all headers and yank the valid ones back.
    ;;
    (ti::mt-hmax 'move)
    (delete-region (point-min) (point))

    (setq ptr list)
    (dolist (elt ptr) (insert elt "\n"))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-get-all-email-addresses (&optional field-list abbrev-alist)
  "Return all email addresses from FIELD-LIST.
Eg. '(\"To\" \"CC\"), which defaults to To, CC and BCC.

Input:
 FIELD-LIST
 ABBREV-ALIST	,see function `ti::xe-expand-mail-aliases'

Return:
 '(str str ..)  ,notice that there may be \"\" empty strings"
  (let (mems
	field
	ret
	)
    (setq field-list (or field-list '("To" "CC" "BCC")))

    (save-with-marker
      (ti::xe-expand-mail-aliases
       (point-min) (ti::mt-text-start) abbrev-alist)

      (dolist (elt field-list)
	(when (setq field (ti::mt-get-field elt))
	  (setq mems (ti::s-get-blocks field "[,\n]")) ;members ?
	  (dolist (mem mems)
	    (unless (var-nil-p mem)
	      (push mem ret)))
	  )))
    (reverse ret)
    ))

;;}}}
;;{{{ News, articles

;;; ............................................................ &News ...


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-news-buffer-p  ()
  "Check if current buffer is news post, followup or the like."
  (interactive)
  (or (o-string-match "news" (symbol-name major-mode))
      ;;  Gnus5 keeps it in 'message-mode', so search this header then
      ;;
      (save-excursion
	(PMIN)
	(re-search-forward "Newsgroups\\|References:\\|Gcc:" nil t))))


;;; ----------------------------------------------------------------------
;;;
;;; - Some special GNUS/NEWS macros that you can call.
(defun ti::mt-news-group ()		;ding & gnus compatible
  "Return newsgroup name if group exists."
  (if (boundp 'gnus-newsgroup-name)
      (symbol-value 'gnus-newsgroup-name)
    ))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-set-article-buffer ()
  "Do `set-buffer' to *Article* if it exists. Return nil if no buffer."
  (if (boundp 'gnus-article-buffer)
      (set-buffer-safe
       (symbol-value 'gnus-article-buffer))))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-get-article-re (re &optional level)
  "Switch to article buffer; match RE at LEVEL and return match."
  (let (line
	)
    (setq level (or level 0))
    (save-excursion
      (cond
       ((ti::mt-set-article-buffer)
	(PMIN)
	(if (re-search-forward re nil t)
	    (setq line (read-match level)))
	)))
      line
      ))


;;; ----------------------------------------------------------------------
;;; - This is usefull if you use same hook for both
;;;   regular mail posting AND for gnuis posting.
;;; - It makes it possible to decicede inside hook, which post
;;;   type this is. Eg. setting extra headers for NEWS and not
;;;   different for regular Mail
;;;
(defun ti::mt-news-reply-p  ()
  "Return type of message being composed.
This function is meaningful _only_ when you use it inside
some GNUS or mail hook.  The buffer must be current mail buffer.

Return:
  'news		,news
  nil"
  (let* ((mode		(symbol-name major-mode))
	 ;;		GNUS might not be loaded in this emacs
	 (gnus-buf 	(if (boundp 'gnus-article-buffer)
			    ;; normally name is "Article"
			    (symbol-value 'gnus-article-buffer)
			  ""))
	 (mail-buf	(ti::mt-mail-buffer-name))   ;YANK buffer name?

	 ;; ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ GNUS
	 ;; - Detect news posting mode. The mail program uses
	 ;;   YANK from the gnus buffer "*Article*" So we can detect
	 ;;   if this is gnus post
	 ;; - gnus 'news-mail-reply-->rnewspost.el
	 ;;
	 (gnus		(string= gnus-buf mail-buf))
	 (news-mode	(or gnus
			    (o-string-match "news" mode)
			    (save-excursion ;Gnus5
			      (PMIN)
			      (re-search-forward "^References:" nil t))
			    ))
	 )
    (cond				;what's the verdict ?
     (news-mode
      'news)
     (t
      nil))
    ))







;;}}}
;;{{{ anon.penet.fi anon-nymserver.com

;;; ............................................................ &anon ...

;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-anon-penet-p (email)
  "Check if EMAIL is pent's anon address."
  (o-string-match "[an][an][0-9]+@.*penet.fi" email))



;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-anon-penet-to-p ()
  "Check if the TO: field contain anon.penet.fi addr.

Return:
  nil
  email       ,if it's anon addr."
  (let  ((to (ti::mt-get-field "to"))
	 )
    (if (and to (ti::mt-anon-penet-p to))
	to nil)))



;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-nymserver-email-convert  (email &optional na-mode)
  "Convert penet EMAIL address.

If NA-MODE is nil: do 'an' conversion
	anXXX@site.com   --> anXXX
	naXXX@site.com   --> anXXX
       VANITY@site.com   --> VANITY@site.com
    VANITY.an@site.com   --> VANITY@site.com
    VANITY.na@site.com   --> VANITY@site.com

If NA-MODE is non-nil:
    Then do opposite 'na' conversion"
  (cond
   (na-mode
    (if (string-match "\\(an\\)[0-9]+@\\|\\.\\(an\\)@" email)
	(setq email (kill-match 1 "na" email))
      ;; the email is VANITY@site.com
      (if (string-match "\\(.*\\)\\(@.*\\)" email)
	  (setq email (concat
		       (read-match 1 email) ".na"
		       (read-match 2 email)))
	)))
   (t
    (cond
     ((string-match "\\(na\\)[0-9]+@" email)
      (setq email (kill-match 1 "an" email)))
     ((string-match "\\(\\.na\\)@" email)
      (setq email (kill-match 1 "" email)))
     )
    ))
  email
  )


;;}}}
;;{{{ mime

;;; ............................................................ &mime ...


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-mime-maybe-p  ()
  "Check if buffer possibly contain MIME sections, not just plain text."
  (let* ((ret (ti::mt-mime-p))
         )
    (unless ret
      (save-excursion
	(cond
	 ((featurep 'tm-edit)		;TM.el
	  ;;   TM puth these markes to MIME section; Try to find one.
	  ;;   This can be only found if the mimi-edit mode is not
	  ;;   yet exited. Upon exit the message will match true
	  ;;   MIME (ti::mt-mime-p).
	  ;;
	  (setq ret (re-search-check (concat "^" (regexp-quote "--[[") 0)))
	  ))
	))
    ret
    ))


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-mime-p  ()
  "Check if buffer has mime message. You probably want `ti::mt-mime-maybe-p'.
It must contain boundary field in the headers and the boundary
must be found from the message body itself.
Only the header is not enough to say it's a composed mime mail.

Content-Type: ...
  boundary=\"Multipart_Thu_Sep_19_12:46:36_1996-1\"
           ^^^^^^^^^

This text must be found after the headers until the MIME criteria is
satisfied."
  (interactive)
  (let ((field (ti::mt-get-field "Content-Type" 'any))
	re
	)
    (when (and field
	       ;;  Content-Type field may not include "boundary"
	       ;;  --> it's not multipart mime.
	       (setq re (strmatget ".*boundary=.\\(.*\\)\"" 1 field))
	       )
      (setq re (regexp-quote re))
      ;;   start finding the boundary text after the headers.
      (save-excursion
	(PMIN) (re-search-forward re)	;This is the header, ignore it
	(forward-line 1)
	(re-search-forward re nil t)
	))))



;;; ----------------------------------------------------------------------
;;; 19 Feb 1996, elgaard@diku.dk (Niels Elgaard Larsen), comp.emacs
;;; #not tested
;;;
(defun ti::mt-mime-qp-decode(from to)
  "Mime. Decode quoted-printable from region between FROM and TO."
  (save-excursion
    (goto-char from)
    (while (search-forward "=" to t)
      (cond ((char= (following-char) ?\n)
	     (delete-char -1)
	     (delete-char 1))
	    ((o-looking-at "[0-9A-F][0-9A-F]")
	     (delete-char -1)
	     (insert (hexl-hex-string-to-integer
		      (buffer-substring (point) (+ 2 (point)))))
	     (delete-char 2))
	    ((message "Malformed MIME quoted-printable message"))))))


;;; ----------------------------------------------------------------------
;;; 19 Feb 1996, elgaard@diku.dk (Niels Elgaard Larsen), comp.emacs
;;; You can do:
;;;
;;;    (add-hook 'vm-select-message-hook 'ti::mt-mime-prepare-qp)
;;;
;;;
(defun ti::mt-qp-mime-prepare ()
  "Mime. Unquote quoted-printable from mail buffers.
Searches for tag:

content-transfer-encoding: quoted-printable"
  (interactive)
  (save-excursion
    (let ((case-fold-search t)
	  (type (mail-fetch-field "content-transfer-encoding"))
	  buffer-read-only
	  )
      (cond
       ((and (stringp type)
	     (o-string-match "quoted-printable" type))
	(PMIN)

	(search-forward "\n\n" nil 'move)
	(message  "MIME Unquoting printable...")
	(ti::mt-mime-qp-decode (point) (point-max))
	(message  "MIME Unquoting printable...done"))
       ))))

;;}}}

;;{{{ Mail sending

;;; .................................................... &mail-sending ...


;;; ----------------------------------------------------------------------
;;;
(defun ti::mt-sendmail-reset-send-hooks  ()
  "Make `mail-send-hook' et al. buffer local and set to nil."
  (dolist (sym '(mail-send-hook
		 message-send-hook
		 mh-before-send-letter-hook
		 ))
    (when (boundp sym)
      (make-local-hook sym)
      (set sym nil))
    ))

;;; ----------------------------------------------------------------------
;;;
(put 'ti::mt-sendmail-pure-env-macro 'lisp-indent-function 0)
(defmacro ti::mt-sendmail-pure-env-macro (&rest body)
  "Reset all mail/message hooks/vars locally to nil and run BODY."
  (`
   (let* (mail-mode-hook
	  mail-setup-hook
	  mail-archive-file-name
	  mail-default-headers
	  mail-default-reply-to
	  )
     ;; byteComp silencer: "Not used variables." These are no-op
     ;;
     (if mail-mode-hook		(setq mail-mode-hook		nil))
     (if mail-setup-hook	(setq mail-setup-hook		nil))
     (if mail-archive-file-name	(setq mail-archive-file-name	nil))
     (if mail-default-headers	(setq mail-default-headers	nil))
     (if mail-default-reply-to  (setq mail-default-reply-to 	nil))
     (,@ body)
     )))


;;; ----------------------------------------------------------------------
;;;
(put 'ti::mt-sendmail-macro 'lisp-indent-function 3)
(defmacro ti::mt-sendmail-macro (to subject send &rest body)
  "(to subject send &rest body) Send/construct mail according to parameters.
Use TO, SUBJECT and If SEND if non-nil, send mail after BODY finishes.

When the body starts. The point is at text beginning.
If you want to send plain mail, bind locally necessary variables like
`mail-setup-hook' , `mail-mode-hook' to nil.

Note:

    `mail-mode-hook' `mail-setup-hook' `mail-archive-file-name'
    `mail-default-headers'

    are set to nil. If you need these, please copy them before calling this
    macro and restore their values in BODY, possibly calling
    and using them as sendmail normally would.

    The hooks are set to nil so that mail buffer is created fast and
    that nothing causes trouble when mail buffer is ready."
  (`
   (let* ((BuffeR	(use-tmp-buffer ti:mt-mail-buffer 'clear))
;;;	  (ObuffeR	(current-buffer))
;;;	  (visible       (get-buffer-window ObuffeR))
	  )
     (ti::mt-sendmail-pure-env-macro
       (save-window-excursion
	 (save-excursion
	   (set-buffer BuffeR)
	   ;;   to subject in-reply-to cc replybuffer actions
	   ;;
	   (mail-setup (, to) (, subject) nil nil nil nil)
	   (mail-mode)
	   (ti::mt-kill-field "^fcc")
	   (ti::mt-text-start 'move)
	   (,@ body)

;;;	 (d!! (current-buffer) mail-send-hook (buffer-string) )
;;;      (d! 1)

	   (PMIN)
	   (kill-buffer-safe " sendmail temp") ;See sendmail-send-it
	   (when (, send)
	     (mail-send-and-exit nil))
	   )))
     )))

;;}}}

;;{{{ provide

(provide   'tinylibmt)
(run-hooks 'ti:mt-load-hook)

;;}}}

;;; tinylibmt.el ends here
