;; @(#) tinylibo.el -- Library for handling (o)verlays easily

;; @(#) $Id: tinylibo.el,v 1.19 1997/06/04 06:44:59 jaalto NotReleased $
;; @(#) $Keywords: extensions lisp matching $
;; $KnownCompatibility: 19.28 $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1995, 96, 97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      May 3 1995
;;
;; To get information on this program use ident(1) or do M-x tinylibo-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el

;; LCD Archive Entry:
;; tinylibo|Jari Aalto|jari.aalto@poboxes.com|
;; List handling library|
;; 04-Jun-1997|$Revision: 1.19 $|~/misc/tinylibo.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;;; Intallation:

;; ........................................................ &t-install ...
;; Put this file on your Emacs-Lisp load path, add following into your
;; ~/.emacs startup file
;;
;;     (require 'tinylibo)
;;

;;}}}
;;{{{ Documentation


;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - The name convention of the TINY LIB functions are described and defined
;;   in main module: tinylib.el
;; - To avoid collisions to other modules "ti::o-" is used in front of
;;   every function. It stands for (ti)ny library for (o)verlays
;; - variable names contain letters 'ti:o', except the version
;;   variables.

;;  Preface 1995
;;
;;	The functions were developed to ease the usage of highlighting,
;;	which really seemed "inside" stuff when I first wanted to use
;;	colored text in emacs.
;;
;;	o   This is LIBRARY module, it does nothing on its own.
;;	o   Offers functions for 19.xx overlay handling
;;
;;
;;  Code policy
;;
;;	All functions are written by [jaalto], unless otherwise notified.
;;	If I have found interesting function from some other .el I have
;;	written a reference to that .el. If the code is directly taken
;;	'almost as is' all the credit belongs to the Author of the
;;	original code. Cheers for him, to write so recpectable code, that
;;	it can be used right from the box!
;;
;;	If someone contributes good, general purpose function, all the
;;	credit belongs to him automatically. Persons name and date and so
;;	on is included.


;;}}}
;;{{{ history

;; ........................................................ &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; May	28	1997	[jari]		19.28	v1.13-1.19	NotReleased
;; - some interactive specs deleted.
;; - This package can't be loaded in XE if the overlay.el is not present.
;; - Changed all non standard boolean tests to use when and unless
;; - Checkdoc 1.42 clear.
;; - Removed version require function (Rick Flower <flower@ms31.sp.trw.com>)
;;
;; Nov	6	1996    [jari]		19.28   v1.11-1.12	NotReleased
;;   ti::o-make-match			:+
;;
;; Sep	25	1996	[jari]		19.28	v1.10		NotReleased
;; - Cosmetic changes.
;;
;; Aug	30	1996	[jari]		19.28	v1.9		NotReleased
;; - Fixed all XE 19.14 ByteComp reported errors.
;;
;; May	26	1996	[jari]		19.28	v1.8		NotReleased
;;   ti::o-remove-area			:! parameter order, interactive
;;
;; Apr	19	1996	[jari]		19.28	v1.7		NotReleased
;; - Supports 19.29 dynamic loading. This reduces memory consumption.
;;
;; Apr	10	1996	[jari]		19.28	v1.6		NotReleased
;; - lisp-mnt.el's lm-verify checked. Bytecomp ok. Rewrote feedback funcs.
;;
;; Jan	22	1996	[jari]		19.28	v1.5		NotReleased
;; - Bug in tinylibo-require, replaced with totally new function.
;; - Bug in ti::o-re-search, the reuse parameter wasn't working
;;   correctly. New function ti::o-buffer-substring
;;
;; Aug	30	1995	[jari]		19.28	v1.4		NotReleased
;; - ti::o-re-search-move added, version query corrected
;;
;; Aug	7	1995	[jari]		19.28	v1.3		NotReleased
;; - Corrected function o-re-search-move so that OV moves to the current
;;   buffer, no matter where it were.
;;
;; Jun	22	1995	[jari]		19.28	v1.2		NotReleased
;; - Haven't looked this for a long time. Some debugging commented.
;;
;; May	10	1995	[jari]		19.28	v1.1		Created
;; - realized that functions that were originally develooped for
;;   tinycerr.el can be used by any program as long as they are
;;   generalized. Effectively this library was born.


;;}}}


;;; Code:

;;{{{ setup: -- require

;;; ......................................................... &require ...

(require 'tinylibm)			;macro package


(eval-and-compile
  (when (and (xemacs-p)
	     (null (load "overlay" 'noerr))) ;XE overlay compatibility
    (message "\n\
tinylibo: ** You need 19.15+ overlay.el package in order to use this library.")
    (error "tinylibo: Aborted loading tinylibo.el"))
    )


(eval-when-compile (use-dynamic-compilation))


;;}}}
;;{{{ setup: -- vars

;;; ....................................................... &v-private ...


(defconst ti:o-package "ti::o"
  "Librarys identifier. For error messages etc...")


(defmacro ti::o  ()
  "Return package Id string."
  ti:o-package)

(defvar ti:o-ret nil
  "Global return value of this package.")

(defvar ti:o-tmp-buf "*tmp*"
  "*Temporary buffer name")

;;; ........................................................ &v-public ...

;;}}}
;;{{{ setup: -- version

;;; ....................................................... &v-version ...

(defconst tinylibo-version
  (substring "$Revision: 1.19 $" 11 15)
  "Latest version number.")


(defconst tinylibo-version-id
  "$Id: tinylibo.el,v 1.19 1997/06/04 06:44:59 jaalto NotReleased $"
  "Latest modification time and version number.")



;;; ----------------------------------------------------------------------
;;;
(defun  tinylibo-version (&optional arg)
  "Show version information. ARG will instruct to print message to echo area."
  (interactive "P")
  (ti::m-pkg-version-info "tinylibo.el" arg))

;;; ----------------------------------------------------------------------
;;;
(defun  tinylibo-feedback ()
  "Submit suggestions, error corrections, impressions, anything..."
  (interactive)
  (ti::m-pkg-submit-feedback "tinylibo.el"))


;;}}}

;;; ########################################################### &funcs ###

;;{{{ macros

;;; .......................................................... &macros ...

(defsubst ti::o-make (level)
  "Make overlay according to match in buffer at LEVEL.
The match is NOTchecked. Returns new overlay."
  (make-overlay
   (match-beginning level)
   (match-end level)))


(defsubst ti::o-makec (level)
  "Make overlay according to match in buffer at LEVEL.
The match _is__checked_ . Returns new overlay or nil."
  (if (match-end level)
      (make-overlay
       (match-beginning level)
       (match-end level))))

;;}}}
;;{{{ funcs


;;; ----------------------------------------------------------------------
;;;
(defun ti::o-make-match  (level plist)
  "Make overlay over the matched text portion. The match level is checked.

Input:
  LEVEL	    ,match level
  PLIST	    ,property list '(PRO-NAME PROP-VAL)

Return:
  ov	    ,overlay or nil"
  (let* ((ov   (ti::o-makec level))
	 prop propv
         )
    (when ov
      (while plist
	(setq prop (nth 0 plist)  propv (nth 1 plist))
	(setq plist (cdr (cdr plist)))  ;go 2 fwd
	(overlay-put ov prop propv)
	))
    ov
    ))

;;; ----------------------------------------------------------------------
;;;
(defun ti::o-buffer-substring (ov &optional no-properties)
  "Read `buffer-substring' underneath overlay OV.

Input:
  OV			,overlay to read
  NO-PROPERTIES		,remove all possible properties

Return:
  string
  nil"
  (let* (ret
	 )
    (setq ret
	  (buffer-substring
	   (overlay-start ov) (overlay-end ov)))
    (if no-properties
	(setq ret (remove-properties ret)))
    ret
    ))

;;; ----------------------------------------------------------------------
;;;
(defun ti::o-mouse-on-p (ov)"Check if overlay OV has mouse-face on.
If mouse-face contains 'default, it's treated to mean same as nil.

Return:
  nil or property value of mouse-face"
  (let* (prop propl)
    (if (null ov) nil
      (setq propl (overlay-properties ov)
	    prop  (if (null (memq 'mouse-face propl)) nil
		    (overlay-get ov 'mouse-face))
	    )
      (if (or (null prop) (eq prop 'default ))
	  nil
	prop				;it had some property
	)
      )))



;;; ----------------------------------------------------------------------
;;;
(defun ti::o-get-mouse ()
  "Check if the point has 'mouse-face overlay.

Return:
  nil          ,no overlay at the point found
  t            ,no mouse face
  ov           ,overlay"
  (let* (ovl				;overlay list
	 ov
	 )
    (setq ovl (overlays-at (point)))
    (if (null  ovl) nil
      (setq ov (ti::o-get-prop ovl (list 'mouse-face)))
      (if (null ov) (setq ov t))	;no mouse
      )
    ov
    ))




;; ----------------------------------------------------------------------
;; 'prop'   means parameter form
;; - There should only one unique...
;;
(defun ti::o-get-prop (ovl prop-list)
  "Read OVL and return first _overlay_ where is property list PROP-LIST.

Input:
  OVL		,overlay list
  PROP-LIST	,list of properties (PROP PROP .. )"
  (let (ov
	ovx
	ptr propl
	(len (length prop-list))
	i
	)
    (if (not (and ovl  prop-list))
	(error "Invalid parameters")
      (while (and ovl			;until list end
		  (null ov))		;until found
	(setq ptr prop-list
	      ovx (car ovl)
	      propl (overlay-properties ovx)
	      i 0)

	(while ptr			;check all properties
	  (if (not (memq (car ptr) propl)) nil
	    (inc i))			;hit count
	  (setq ptr (cdr ptr)))

	(if (= i len)
	    (setq ov ovx))	;found all matches
	(setq ovl (cdr ovl))
	))
    ov
    ))

;; ----------------------------------------------------------------------
;; 'prop-val'   means parameter form
;; - This is more heavier function
;;
(defun ti::o-get-prop-val (ovl prop-list)
  "Read OVL and find first overlay where are properties PROP-LIST.

Input:
  OVL		,overlay list
  PROP-LIST	,list of properties (PROP VAL PROP VAL ..)"
  (let (ov
	ovx
	ptr propl
	prop propv
	(len (length prop-list))
	i
	)
    (if (or (not (and ovl prop-list))
	    (not (= 0 (% len 2))))		;must go paired
	(error "Invalid parameters" ovl prop-list)

      (setq len (/ (length prop-list) 2))
      (while (and ovl			;until list end
		  (null ov))		;until found
	(setq ptr prop-list
	      ovx (car ovl)
	      propl (overlay-properties ovx)
	      i 0)
	(while ptr
	  (setq prop  (car ptr)   ptr (cdr ptr)
		propv (car ptr)   ptr (cdr ptr))
;;;	  (d! prop propv propl (memq prop propl) (overlay-get ovx prop) )
	  (if (and (memq prop propl)
		   (eq (overlay-get ovx prop) propv))
	      (inc i))			;hit count
;;;	  (d! i)
	  )

	(setq ovl (cdr ovl))
	(if (not (= i len)) nil
	  (setq ov ovx))			;all matched
	))
;;;    (d! "~out" i len (prin1-to-string ov))
    ov
    ))





;;; tested..
;;; ----------------------------------------------------------------------
;;;
(defun ti::o-re-search (re level list &optional max back reuse reuse-t)
  "Search for RE at LEVEL by creating overlay and its property LIST.
Assigning LIST (PROP PROP_VAL) to the overlay. The search is repeated
until no more hits or up till MAX point is reached.

Input:
  RE	   ,regexp
  LEVEL	   ,subexpression level in regexp
  LIST	   ,list of (PROP PROP_VAL)
  MAX      ,if non-nil, searches up till MAX point.
  BACK     ,search backward
  REUSE    ,(PROP PROP PROP ..) or (PROP_SYM PROP_VAL ..)

	   When re match is found it looks overlays underneath the
	   point, and the first overlay that satisfies list, will
	   be reused, instead of creating new one. Note that _first_
	   overlay matched is used, if none is found, new is created
  REUSE-T
	   Specifies the list _type_ that was given in REUSE.
	   nil = first type , non-nil = second type."
  (let* ((func (if back 're-search-backward 're-search-forward))
	 (max  (if max
		   max			;it's given
		 (if back (point-min) (point-max))))
	 ov
	 ovl
	 prop
	 propv
	 ptr
	 mb 				;match pointers
	 )
    (if (and list
	     (listp list)
	     (= 0 (% (length list ) 2))) nil
      (error "Parameter LIST invalid" re level list))

    (save-excursion
      (while (funcall func re max t  )
	(setq mb  (match-beginning level))

;;;	(d! level (read-match level) mb)

	(if (null mb)
	    nil				;no match at this level
	  (setq ovl (overlays-at mb)) 	;try finding overlays
	  (if (or (null reuse)
		  (null ovl))
	      (setq ov (ti::o-make level))

;;;	    (d! "r" reuse ovl)  (setq OVL ovl RE reuse)
	    (if reuse-t			;what type the list is ?
		(if ovl (setq ov (ti::o-get-prop-val ovl reuse)))
	      (if ovl (setq ov (ti::o-get-prop ovl reuse))))

	    (d! "after" ov)

	    (if (null ov)		;none satisfies us
		(setq ov (ti::o-make level)))
	    )

	  ;; Now we should have overlay in a way or other
	  (setq ptr list)
;;;	  (d! list)
	  (while ptr
	    (setq prop (nth 0 ptr)  propv (nth 1 ptr))
	    (setq ptr (cdr (cdr ptr)))	;go 2 fwd
;;;	    (d! "put" prop propv ov)
	    (overlay-put ov prop propv)
	    ))
	))
    ))

;;; ----------------------------------------------------------------------
;;; - Try following example, cool!
;;;
;;; (setq OV (ti::o-make-overlay (point) (point)))
;;; (overlay-put OV 'face 'highlight)
;;; (ti::o-re-search-move OV "ti::o")
;;;
;;;
(defun ti::o-re-search-move (ov re &optional level back max)
  "Maove OV to Search match RE at LEVEL.
Default level is 0, full match. if BACK is given, search is done
backward. MAX is last position to search.

If overlay OV is currently in some other buffer, it will be transferred
to the current buffer.

Input:
  OV	,overlay
  RE	,regexp
  LEVEL	,subexpression level in regexp
  BACK	,flag, is non-nil, go backward
  MAX	,max point of search

Return:
  nil	,if not moved.
  nbr   ,overlay end position [matched portion end]"
  (let* ((cmd (if back 're-search-backward 're-search-forward))
	 (max (or max
		  (if back (point-min) (point-max))))
	 (level (or level 0))		;default is full string
	 )
    (if (null ov)  (error "invalid overlay, nil"))

    (if (not (and (funcall cmd re max t)
		  (match-end level)))
	nil
      (move-overlay ov (match-beginning level)
		    (match-end level)
		    (current-buffer)
		    )
      (match-end level)			;return value
      )
    ))

;;; ----------------------------------------------------------------------
;;;
;;;
(defun ti::o-get-within-area (propl &optional propl-t beg end)
  "Return all overlays which match property list PROPL.
If PROPL is t then returns all overlays. Default is to search from
current point forward.

Input:
  PROPL		,property list, see next
  PROPL-T	,if nil the propl is of type    (PROP PROP .. )
		 if non-nil                     (PROP VAL PROP VAL ..)
  BEG		,region beginning
  END		,region end"
  (let* (
	 (p   (or beg (point)))
	 (max (or end (point-max)))
	 (all (eq t propl))
	 ovl ovx
	 l
	 )
    (save-excursion
      (goto-char p)
      (while (< (setq p (next-overlay-change p)) max)
	(goto-char p)
	(setq ovl (overlays-at p))
	(if (null ovl)
	    nil				;no overlays here
;;;	  (d! "~go" p max (prin1-to-string ovl))
	  (if all
	      (mapcar
	       (function
		(lambda (x)
		  (if x (push x l))))
	       ovl)
;;;	    (d! "prop type" propl-t  (prin1-to-string propl) )

	    (if propl-t
		(setq ovx (ti::o-get-prop-val ovl propl))
	      (setq ovx (ti::o-get-prop ovl propl)))
	    (if ovx  (push ovx l))))
	)) ;; while, save-excur
    l
    ))





;;; ----------------------------------------------------------------------
;;; If you're in trouble, call this function interactively
;;; and it wipes out all overlays.
;;;
(defun ti::o-remove-area (&optional beg end propl propl-t)
  "Remove all matched overlays within area.
Default is from point forward. Ignores buffer read-only status.

Input:
  BEG	    ,area beginning
  END	    ,area end
  PROPL	    ,(PROP PROP ..) or
	     (PROP VAL PROP VAL ..)
	     If this value is t, removes all overlays

  PROPL-T   Specifies the list type given. nil = first list type."
  (interactive "r")
  (let* (buffer-read-only
	 (p     (or beg (point)))
	 (max   (or end (point-max)))
	 (propl (if propl propl t))	;set to t is not given
	 ovl
	 ovx
	 )

;;;    (d! propl p max)

    (setq ovl  (ti::o-get-within-area propl propl-t p max))
;;;    (d! "FULL" (length ovl) (prin1-to-string ovl))
    (while ovl
      (setq ovx (car ovl)  ovl (cdr ovl))
;;;      (d! "del" (prin1-to-string ovx))
      (if ovx (delete-overlay ovx))
      )
    ))



;;}}}
;;{{{ provide

(provide 'tinylibo)

;;}}}

;;; tinylibo.el ends here
