;; @(#) tinylibxe.el -- Emulation library for Emacs and XEmacs

;; @(#) $Id: tinylibxe.el,v 1.25 1997/06/04 06:48:45 jaalto NotReleased $
;; @(#) $Keywords: library, macros $
;; $KnownCompatibility: 19.28+ XEmacs19.14+ $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1996,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Apr 19 1996
;;
;; To get information on this program use ident(1) or do M-x tilibxe-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el

;; LCD Archive Entry:
;; tinylibxe|Jari Aalto|jari.aalto@poboxes.com|
;; Compatibility for both XEmacs and Emacs and possible for 18.xx Emacs|
;; 04-Jun-1997|$Revision: 1.25 $|~/misc/tinylibxe.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ........................................................ &t-install ...
;; Put this file on your Emacs-Lisp load path, add following into your
;; package that you're currently developing. This ensures compatibility
;; for some extent to XEmacs and Emacs.
;;
;;
;;     	(require 'tinylibxe)


;;}}}
;;{{{ Documentation

;; ..................................................... &t-commentary ...
;;; Commentary:
;;; Briefly:
;; o   This is library, package itself does nothing.
;; o   Compatibility for both XEmacs and Emacs and possible for 18.xx Emacs

;;  Preface 1996
;;
;;	I saw some incompatibilities reported for XE and Emacs, which
;;	prevented writing portable code. I started collecting some
;;	problematic things here. Please send me more lisp code if you know
;;	more imcompatibilities, so that developers don't have to struggle
;;	so much.
;;
;;  Overlay.el in XEmacs19.15
;;
;;	Good news; XEmacs 19.15 will have new package that emulates
;;	Emacs overlay functions. Beware, at least VM goes nuts in XEmacs if you
;;	load this...
;;	The author is Joe Nuspl <nuspl@nvwls.cc.purdue.edu>
;;
;;  What you should know -- keep this in mind
;;
;;	This library's intention is to make it possible to use some package
;;	that is written only in Emacs or only in XEmacs in mind. Normally
;;	You couldn't use that package in another emacs, because there may
;;	be functions that are not implemented.
;;
;;	When you load this library, it emulates those unknown functions
;;	as much as it can, so that that you can use the package for
;;	both emacs. However, it may not be possible to reproduce
;;	exacttly the same behavior in the Emacs that wa not the primary
;;	target for the package. The emulation may in worst case be
;;	only so, that you are able to load the package without errors,
;;	but the functionality of the package doesn't correspond
;;	to the original version.
;;
;;	When you write your own packages, do not try to make use of the
;;	latest glory details of the Emacs. If you do, you can be
;;	sure that your package won't work anywhere else.
;;
;;  Coding to specific Emacs
;;
;;	I'd like propose that you don't test against missing fetatures
;;	when it comes to XEmacs and Emacs compatibility, but instead use
;;	direct tests
;;
;;	    (emacs-p)		;; from 'M' lib, tinylibm.el
;;	    (xemacs-p)
;;
;;	In these those specific cases where you would like to use "special"
;;	code for either Emacs version.


;;}}}
;;{{{ history

;; ........................................................ &t-history ...
;;; Change log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; Jun	4	1997	[jari]		19.28	v1.22-1.24	NotReleased
;;   button-release-event-p		:! byteComp error removed.(sfuncall)
;;
;; May	14	1997	[jari]		19.28	v1.22-1.23	NotReleased
;; - Added dired some functions.
;; - Changed all non standard boolean tests to use when and unless
;;   button-release-event-p		:+
;;   error-message-string		:+
;;
;; Mar	5	1997	[jari]		19.28	v1.21		NotReleased
;; - jon.connell@hl.siemens.de reported that event-start fucntion emulation
;;   in XEmacs wasn't right. There missed "event" parameters. Corrected.
;;
;; Feb	28	1997	[jari]		19.28	v1.20		NotReleased
;; - the overlay.el XEmacs lisbrary does not work correcly; because loading it
;;   will trash VM. The load command for that file was removed.
;;
;; Feb	27	1997	[jari]		19.28	v1.19		NotReleased
;; - disable the overlay and glyph support fucntions, because
;;   I got reports that VM broke after this library was loaded. This
;;   was because VM uses fboundp tests to see if emacs has
;;   overlay or extent functions. The unfortunate thing is that
;;   1) you load VM; it makes extents in XEmacs 2) then you load libxe
;;   and suddenly VM thinks we can use overlays. But since it has
;;   already created extents, it passes them to overlay fucntions and
;;   here happens the clash.
;;
;; Feb	14	1997	[jari]		19.28	v1.18		NotReleased
;; - Now has call (load "overlay" 'noerr) at the end of file so that
;;   XEmacs19.15 library is loaded _after _ this pacakge. We want to use
;;   the official XEmacs functions if they are available and loading the
;;   package last wipes functions defined in this module.
;;
;; Nov	14	1996	[jari]		19.28	v1.17		NotReleased
;; - set-glyph-face was mistakenly looped to itself. Fixed.
;;   (XEmacs redGnus error)
;;
;; Nov	1	1996	[jari]		19.28	v1.16		NotReleased
;; - Rewrote the dependencies between emacs versions. Now uses
;;   nafboundp, so that the autoloads are handled too.
;;
;; Oct	30	1996	[jari]		19.28	v1.15		NotReleased
;;   window-edges			:+
;;   extentp				:+
;;
;; Sep	25	1996	[jari]		19.28	v1.14		NotReleased
;; - Rearranged loading. now uses autoloads.
;;
;; Aug	29	1996	[jari]		19.28	v1.12-13	NotReleased
;; - Cleared all XEmacs/Emacs byteCompiler errors messages.
;; - Added 'overlayp' emulation
;;
;; Aug	23	1996	[jari]		19.28	v1.7-1.11	NotReleased
;; - Wrote section 'what you should know'.
;; - Reordered all overlay code and some more overlay compatibilities.
;; - Added event support, frame cmd support, more overlay commands,
;;   corrected run-at-time emulation for XEmacs.
;;
;; Jun	28	1996	[jari]		19.28	v1.5-1.6       	NotReleased
;; - added 'region face which is missing from XEmacs.
;; - Added run-at-time emulation in XEmacs.
;;
;; Jun	24	1996	[jari]		19.28	v1.4		NotReleased
;; - Commented the error message, so that XEmacs users can use this lib.
;;
;; Jun	19	1996	[jari]		19.28	v1.3		NotReleased
;; - No code changes, but put to RCS tree as a first non-official release.
;; - The XEmacs is not supported yet, ie. loading this file intentionally
;;   fails in XEmacs, since I don't know all the overlay equivalent
;;   in XEmacs yet.
;;
;; Apr  19      1996    [jari]          19.28   v1.1-1.2	Created



;;}}}

;;; Code:

;;{{{ setup: require

;;; ......................................................... &require ...

;;; Never advice "2.15" uses this call, make sure it exist.
;;;
(if (not (fboundp 'byte-code-function-p))
    (fset 'byte-code-function-p 'ignore))


(require 'advice)
(require 'tinylibm)


;;}}}
;;{{{ setup: -- variables

;;; .......................................................... &v-bind ...

;;; ......................................................... &v-hooks ...

(defvar ti:xe-load-hook nil
  "*Run when file has been loaded.")


;;; ....................................................... &v-private ...

(defconst ti:xe-emacs (not (boundp 'xemacs-logo))
  "Flag if this is file is loaded under GNU Emacs.")

;;}}}
;;{{{ setup: -- version

;;; ....................................................... &v-version ...

(defconst tinylibxe-version
  "$Revision: 1.25 $"
  "Latest version number.")

(defconst tinylibxe-version-id
  "$Id: tinylibxe.el,v 1.25 1997/06/04 06:48:45 jaalto NotReleased $"
  "Latest modification time and version number.")


;;; ----------------------------------------------------------------------
;;;
(defun tinylibxe-version (&optional arg)
  "version information."
  (interactive "P")
  (ti::m-pkg-version-info "tinylibxe.el" arg))


;;; ----------------------------------------------------------------------
;;;
(defun tinylibxe-submit-bug-report ()
  "Submit bug report."
  (interactive)
  (ti::m-pkg-submit-bug-report
   "tinylibxe.el"
   tinylibxe-version-id
   '(tinylibxe-version-id
     )))


;;}}}


;;; ########################################################### &funcs ###

;;{{{ events, window, frames, misc
;;; .......................................................... &events ...


(cond
 (ti:xe-emacs
  (fset 'event-window	    'posn-window)
  (fset 'event-point        'posn-point)
  (fset 'event-timestamp    'posn-timestamp)
  (fset 'window-pixel-edges 'window-edges)
  )

 (t
  (fset 'posn-window	    'event-window)
  (fset 'posn-window	    'event-window)
  (fset 'posn-point	    'event-point)
  (fset 'posn-timestamp	    'event-timestamp)
;;;  (fset 'posn-col-row    ')
  (fset 'window-edges       'window-pixel-edges)

  ))


;;; From wid-edit.el by Per Abrahamsen <abraham@dina.kvl.dk>
;;;
(unless (fboundp 'button-release-event-p)
  ;; XEmacs function missing from Emacs.
  (defun button-release-event-p (event)
    "Non-nil if EVENT is a mouse-button-release event object."
    (and (eventp event)
	 (memq (sfuncall 'event-basic-type event) '(mouse-1 mouse-2 mouse-3))
	 (or (memq 'click (event-modifiers event))
	     (memq  'drag (event-modifiers event))))))



(nafboundp 'event-start "tinylibxe"
  (defun event-start (event)
    "tinylibxe.el"
    ;; In Emacs (WINDOW BUFFER-POSITION (X . Y) TIMESTAMP)
    (list
     (sfuncall 'event-window event)
     (sfuncall 'event-point event)
     (sfuncall 'posn-x-y event)
     (sfuncall 'event-timestamp event)
     )))

(nafboundp 'event-x "tinylibxe"
  (defun event-x (event)
    "tinylibxe.el"
    (let* ((data (sfuncall 'event-start event))
	   )
      (car data))))

(nafboundp 'event-y "tinylibxe"
  (defun event-y (event)
    "tinylibxe.el"
    (let* ((data (sfuncall 'event-start event))
	   )
      (cdr data))))

(nafboundp 'posn-x-y "tinylibxe"
  (defun posn-x-y (event)
    "tinylibxe.el"
    (cons (sfuncall 'event-x event) (sfuncall 'event-y event))))



;;; .......................................................... &frames ...

(cond
 (ti:xe-emacs
  nil
  )
 (t
  (and (not  (fboundp 'frame-parameters)) ;obsolete in 19.14
       (fboundp 'frame-property)
       (fset 'frame-parameters 'frame-property))
  ))


;;}}}
;;{{{ faces

;;; ........................................................... &faces ...

;;; XE/Emacs doesn't have compatible faces

(unless (face-p 'region)
  (make-face 'region)
  (set-face-foreground 'region "white")
  (set-face-background 'region "black")
  )

;;}}}
;;{{{ dired

(cond
 (ti:xe-emacs
  (defalias  'dired-unmark-subdir-or-file   'dired-unmark)
  (defalias  'dired-mark-subdir-or-file	    'dired-mark)
  (defalias  'dired-mark-get-files	    'dired-get-marked-files)
  (defalias  'dired-mark-map		    'dired-map-over-marks)
  )
 (t
  (defalias 'dired-unmark	    'dired-unmark-subdir-or-file)
  (defalias 'dired-mark		    'dired-mark-subdir-or-file)
  (defalias 'dired-get-marked-files 'dired-mark-get-files)
  (defalias 'dired-map-over-marks   'dired-mark-map)
  ))

;;}}}
;;{{{ ovcerlays -- extents

;;; ......................................................... &overlay ...
;;; Idea by Kyle Jones,  kyle@wonderworks.com, in setnu.el
;;; 27.02.97 Cancelled, because 19.15 will have overlay emulation

(when nil

(cond
 (ti:xe-emacs
  (fset 'extent-live-p		'overlayp)
  (fset 'extentp		'overlayp)
  (fset 'make-extent		'make-overlay)
  (fset 'delete-extent		'delete-overlay)
  (fset 'extent-property	'overlay-get)
  (fset 'set-extent-property	'overlay-put)
  (fset 'set-extent-endpoints	'move-overlay)
  (fset 'extent-end-position	'overlay-end)
  (fset 'extent-start-position	'overlay-start)
  (fset 'extent-start-position	'overlay-buffer)
  (fset 'extent-start-position	'overlay-buffer)
  (fset 'next-extent-change	'next-overlay-change)
  (fset 'extent-properties      'overlay-properties)


  (defun extent-list (buffer point)
    "tinylibxe.el -- arg3 not supported."
    (save-excursion
      (if buffer
	  (set-buffer buffer))
      (sfuncall 'overlays-at point)))

  (defun extent-length (e)
    "tinylibxe.el -- return overlay length."
    (- (sfuncall 'overlay-end e) (sfuncall 'overlay-start e)))


  )
 ;; ............................................................... xe ...
 (t
  (fset 'overlayp               'extent-live-p)
  (fset 'make-overlay		'make-extent)
  (fset 'delete-overlay		'delete-extent)
  (fset 'overlay-get		'extent-property)
  (fset 'overlay-put		'set-extent-property)
  (fset 'move-overlay		'set-extent-endpoints)
  (fset 'overlay-end		'extent-end-position)
  (fset 'overlay-start		'extent-start-position)
  (fset 'overlay-buffer		'extent-start-position)
  (fset 'overlay-buffer		'extent-start-position)
  (fset 'next-overlay-change    'next-extent-change)
  (fset 'overlay-properties     'extent-properties)

  (defun overlays-at (point)
    "tinylibxe.el -- return overlay at POINT."
    (sfuncall 'extent-list (current-buffer) point))

  (fset 'overlay-length		'extent-length)

  ))

)



;;}}}
;;{{{ glyphs

;;; ........................................................... glyphs ...
;;; Thanks to Kyle Jone,  kyle@wonderworks.com, in setnu.el

(when nil

(defvar ti:xe-begin-glyph-property (if (fboundp 'extent-property)
                                       'begin-glyph
                                     'before-string)
  "Property name to use to set teh begin glyph of an extent.")


(defun set-overlay-begin-glyph (e g)
  "tinylibxe -- Set glyph G in overlay E."
  (sfuncall 'overlay-put e ti:xe-begin-glyph-property g))


(nafboundp 'make-glyph "tinylibxe"
  (fset 'make-glyph 'identity))


(cond
 (ti:xe-emacs
  (nafboundp 'set-glyph-face "tinylibxe"
    (defun set-glyph-face (g face)
      "tinylibxe -- Set glyph G to FACE"
      (put-text-property 0 (length g) 'face face g))))

 (t
  (fset 'set-glyph-face 'ignore)))

)

;;}}}
;;{{{ misc

;;; ............................................................ &misc ...

;;; ----------------------------------------------------------------------
;;; From wid-edit.el by Per Abrahamsen <abraham@dina.kvl.dk>
;;;
(unless (fboundp 'error-message-string)
  ;; Emacs function missing in XEmacs.
  (defun error-message-string (obj)
    "Convert an error value to an error message."
    (let ((buf (get-buffer-create " *error-message*")))
      (erase-buffer buf)
      (display-error obj buf)
      (buffer-string buf))))


;;; ----------------------------------------------------------------------
;;; XE, labels this obsolete
;;; In older Emacs it does not exist
;;;
(nafboundp 'force-mode-line-update "tinylibxe"
  (defun force-mode-line-update  ()
    "tinylibxe -- XE & Emacs  Compatibility, Please use ti::xe-modeline-update
and remove references to this function."
    (set-buffer-modified-p (buffer-modified-p))))



;;; ----------------------------------------------------------------------
;;; XE doesn't have 'timer package
(nafboundp 'run-at-time "tinylibxe"
  (defun run-at-time  (time repeat function &rest args)
    "tinylibxe -- XE & Emacs Compatibility."
    (require 'itimer)

    ;;  We can't use following Emacs arguments: ARGS
    ;;  (run-at-time TIME REPEAT FUNCTION &rest ARGS)
    ;;  start-itimer: (name function value &optional restart)
    ;;
    (sfuncall
     'start-itimer
     (cond				;ARG1
      ((symbolp function)
       (symbol-name function))
      (t
       "itimer-with-no-name"
       ))
     function				;ARG2

     (if (integerp repeat)		;ARG3
	 repeat 10)

     (if (integerp repeat)		;ARG4
	 repeat 10)
     ))
  )

;;}}}
;;{{{ advice: code from XE --> Emacs

;;; ----------------------------------------------------------------------
;;;  This is same as 'beep'
;;;  Emacs, subr.el:(defalias 'beep 'ding) ;preserve lingual purity
;;;
(defadvice ding (around tinylibxe (&optional arg &rest args) act)
  "tinylibxe -- Define Xemacs compatible ding comamnd. Ignores arg 2."
  ad-do-it
  )


;;; ----------------------------------------------------------------------
;;; by Anders Lindgren <andersl@csd.uu.se>
;;;
(defadvice make-sparse-keymap (before tinylibxe (&optional no-op) act)
  "tinylibxe -- This advice does nothing except adding an optional argument
to keep the byte compiler happy when compiling Emacs specific code
with XEmacs.")

;;}}}

(provide   'tinylibxe)
(run-hooks 'ti:xe-load-hook)

;;; tinylibxe.el ends here
