;; @(#) tinymail.el -- Some mail additions: dynamic Fcc, Cc, hilit, X-info

;; @(#) $Id: tinymail.el,v 1.91 1997/06/23 11:37:53 jaalto NotReleased $
;; @(#) $Keywords: mail $
;;
;; $KnownCompatibility: 19.28+ XEmacs 19.14+ $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1996,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Apr 10 1996
;;
;; To get information on this program use ident(1) or do M-x timi-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el

;; LCD Archive Entry:
;; tinymail|Jari Aalto|jari.aalto@poboxes.com|
;; Dynamic Fcc, Cc, hilit, while you type msg. Powerfull TAB completion modes|
;; 23-Jun-1997|$Revision: 1.91 $|~/misc/tinymail.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ....................................................... &t-install ...
;;   Put this file on your Emacs-Lisp load path, add following into your
;;   ~/.emacs startup file.
;;   Code can be easily ripped with ti::m-pkg-rip-magic from tinylib.el
;;
;;* _
;;*     (autoload 'timi-mail            "tinymail" t t)
;;*     (add-hook 'mail-setup-hook      'timi-mail)
;;*     (add-hook 'mail-setup-hook      'my-mail-setup-hook)
;;*     (add-hook 'message-mode-hook    'timi-mail)
;;*     (add-hook 'message-mode-hook    'my-mail-setup-hook)
;;* _
;;* _
;;*     (setq timi-:passwd-cat-cmd "cat /etc/passwd") ;; or "ypcat passwd"
;;*
;;*     (defun my-mail-setup-hook ()
;;*      "Define my mail things."
;;*      ;;  This key completes anywhere in the mail buffer.
;;*      ;;  We Pick faster key than the default installation
;;*      ;;
;;*      (define-key mail-mode-map [f5] 'timi-guess-complete)
;;*      )
;;*
;;*	 ;;  Set this to shell command which prints the password file.
;;*      (setq timi-:passwd-cat-cmd "cat /etc/passwd")
;;
;;   If you don't want to use default bindings provided, add following
;;   statement to your startup file too:
;;
;;* _
;;*     (add-hook 'timi-:load-hook 'timi-cancel-bindings)
;;
;;   If you have any questions, use 'submit' function. In case of error
;;   or misbehavior, turn on the debug too and send the debug result and
;;   describe what you did and where went wrong.
;;
;;	M-x timi-debug-toggle
;;      M-x timi-submit-bug-report

;;}}}
;;{{{ Documentation

;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "timi-" in front of
;;   every function & variable. It stands for '(ti)ny (m)a(i)l add-on'.
;; - variable names contain letters 'timi-:', excecpt version/hook/mode vars.

;;  Preface, Apr 1996
;;
;;      This mail composing helper functions have been with me over a year
;;      or so but and I finally decided to put them to a nice package.
;;      Hope you find use for this 'tiny tool' too and don't
;;      hesitate to drop me a mail if you know more nifty things that
;;      I could add here to improve this package.
;;
;;  Overview of features
;;
;;      o  Some handy additions to mail sending interface.
;;      o  Adds automatically Cc field when you type the To: address.
;;      o  Adds X-info dynamically while you type To: address.
;;      o  Changes Fcc dynamically according to header content.
;;      o  Very easy TAB completion: two modes, alias and definition string.
;;         Also completes password file entries if your .mailrc doesn't
;;         have match.
;;      o  Reminders: remind you about old messages where you expected an
;;         answer but got none.
;;      o  Fcc folder can have compression extension .gz or .Z
;;	o  Mail, GNUS 4/5 message, VM, MH compatible.
;;
;;  Installation note
;;
;;	This package installs itself to `mail-setup-hook' and you should
;;	know why if you try to get the package running for some other
;;	mail agent that Emacs mail, RMAIL, GNUS and VM where this pacakge
;;	has been tested.
;;
;;	The `mail-setup-hook' is called *after* the basic headers, like
;;	To and Subject, are already in the buffer. Function `timi-mail'
;;	needs to read the contents of To in order to determine how
;;	it starts (it puts 1 or 2 spaces at the beginning of To field at
;;	the initial start, so that the pacakges Cc control is started
;;	right. When you call simple mail, C-x m, the auto Cc feature is on
;;	(1 space) and when you hit reply, the auto Cc feature is off
;;	(2 spaces).
;;
;;	So, if your mail agent doesn't call `mail-setup-hook', find similar
;;	hook that runs after the headers are in the buffer and install
;;      to that hook `timi-mail'.
;;
;;  Guess Completion feature
;;
;;      There are two basic completion modes: 'alias and 'string,
;;      which refer to your ~/.mailrc definitions:
;;
;;          alias Peter-Pan     "Disney character -- <ppan@disney.com>
;;                |             |
;;                alias mode    string match mode
;;
;;      when you hit the completions key (TAB in headers) the current
;;      word is picked at point and searched from either of these two
;;      definition lists.
;;
;;      Since the alias expansion is familiar to everyone, I'll only explain
;;      the latter. It is very powerfull completion indeed. Say you
;;      have this in your ~/.mailrc
;;
;;          alias tweedle "Mister Foo, Skyscraper Doing co. <foo@company.com>
;;
;;      Then you start composing a message. The lines 1-4 present
;;      what possibly you can type into the To: field before you
;;      hit the comletion key, TAB.
;;
;;          1  To: company
;;          2  To: Foo
;;          3  To: sky
;;          4  To: mister
;;
;;      It doesn't matter what you type initially; it can be anything you
;;      remember from the person's definition string. At this point you
;;      call timi-guess-complete key, TAB, and any of those lines, 1-4,
;;      will be replaced with
;;
;;          To: Mister Foo, Skyscraper Doing co. <foo@doing.com>
;;
;;      Easy? Yep, and if there are more matches to, say "foo", you're offered
;;      a completion list of alias names, where you can pick exact match.
;;
;;      You no longer have to remember "exact" alias names, It's enough
;;      if you remember "something" that belongs to person's definition:
;;      email, surname, firstname, domain part....
;;
;;  Password table completion mode
;;
;;	In addition to .mailrc completion, there is support for completing
;;	entries found from passwd file. If the Guess complete above fails
;;	the password file is examined if the mode is turned on. See
;;	varaible
;;
;;	    timi-:password-mode
;;
;;	Which is set to t by default. When you complete password entries
;;	for the first time, building all necessary variables will take
;;	some time. This is normal and you have to live with that.
;;
;;	Once the password file completions have been parsed, the content
;;	is written to cache file
;;
;;	    timi-:passwd-file
;;
;;	Next time you need password completions, if this file exists,
;;	it will be read insted of heavy /etc/passwd file parsing. If you
;;      want to force reading the /etc/passwd again, just delete that file
;;      and it will be recreated next time passwd completion is used.
;;
;;  Basic functioning information
;;
;;      When you load this package, it adds some keybindings to the
;;      `mail-mode-map'. You need `timi-mail' command is, because
;;      Emacs clears the `post-command-hook' if you press
;;      C-g too much ==> timi is deactivated too.
;;
;;          Press C-c t m to activate TIMI again.
;;
;;      Another key C-c t -, toggles if timi should take control of
;;      changing the Cc headers.
;;
;;          You need to turn off the TIMI if you add/change
;;          Cc field content manually.
;;
;;  To field tracking: Automatic Cc and X-info field adding
;;
;;      While you're composing message, this package awakens in
;;      regular basis and checks your To: field's content. If it finds
;;      anything to do, it adds some more fields to your message.
;;
;;          To: foo@bar.com             <-- let's assume this triggers
;;
;;      Result is:
;;
;;          To: foo@bar.com
;;          CC: bix@helios.org
;;          X-info: Jerry Foo. 21st, NY, p. 222-333-444.
;;
;;      If you now want to add more Cc people, you can do it because one
;;      the Cc has been added TIMI starts sleeping provided that you don't
;;      touch the TO field again. If you do touch the TO field, then TIMI
;;      tries to follow you and determine new Cc fields.
;;
;;      You can disable the automatic Cc field addition by adding *two*
;;      spaces in the TO field. The hotkey C-ct- will also toggle this.
;;
;;          To:  foo@bar.com    <-- disables Cc feature
;;             ^^
;;
;;      This still leaves the X-info adding feature on and if you want to
;;      disable that too you add one more space to the To field. Now there
;;      is 3 spaces and _No_ To field tracking is done any more. You can
;;      give prefix argument 3 to hotkey C-ct-
;;
;;          To:   foo@bar.com   <-- disables both Cc and X-info feature
;;             ^^^
;;
;;      Note: There is one thing you should know about the implementation.
;;      As soon as the Cc or X-info tracking is done, the To-field content
;;      is saved. When the tracking awakens again, it *checks* if the
;;      To-field content has changed against this saved value. If it hasn't,
;;      nothing is done.
;;
;;      This means, that you *can* modify the contents of X-info and Cc
;;      field after the post command function has finished. If you later
;;      want to force it to track the headers again, change the To-field
;;      contents somehow (add space to the end...)  and this gives a signal
;;      that to-field has changed.
;;
;;  Fcc field tracking: automatic Fcc change
;;
;;      By default the Fcc is not added in your mail message, thus the
;;      automatic Fcc tracking doesn't get activated. Add following entry
;;      into your emacs to record your outgoing mail messages
;;
;;          (setq mail-archive-file-name "~/.RMAIL.out")
;;
;;      When Emacs sees that you have set this, it adds the Fcc field to
;;      your mail message. Alternatively you can press keys
;;
;;          C-c C-f C-f    ;; mail-fcc
;;
;;      in *mail* buffer and it asks you to insert the Fcc field.  Only
;;      now, when the Fcc is in the message, the automatic Fcc handling
;;      starts snooping around your headers and changing it if it finds a
;;      match from variable
;;
;;          timi-:table-fcc
;;
;;      If you want to disable Fcc changing (and edit it by hand),
;;      put two spaces at front of the Fcc. like this:
;;
;;          FCC:  ~/.RMAIL.secondary
;;              ^^
;;
;;  Fcc and saving outgoing copy in compressed format
;;
;;	If you have an ccount that has quota limits, you want to save space
;;	as much as possible. You can save your outgoing mail copy in
;;	compressed format if you prepend the filename with ".gz" or
;;	".Z". TinyMail will automatically load jka-compr if it sees any of
;;	those extensions. The fcc folder definition then looks like this.
;;
;;          (defconst timi-:table-fcc
;;             (list
;;              (list "elisp-archive"    (my-z " ~/.mail.elisp-post"))
;;              (list "bug-gnu"          (my-z " ~/.mail.bug"))
;;              (list "."		 (my-z " ~/.RMAIL.out")) ;; general out
;;              ))
;;
;;	The `my-z' function adds ".gz" into the filename if you have enabled
;;	the compression. The implementation of the `my-z' function is
;;	presented in the end of the file. You can use it to construct other
;;	filenames as well. You _have_ to add (require 'jka-compr) is you
;;	want to use compresses RMAIL file.
;;
;;	    ;;  first one is defined in paths.el
;;	    (setq rmail-file-name        (my-z "~/RMAIL")
;;	    (setq mail-archive-file-name (my-z "~/.RMAIL.out")
;;
;;
;;  Gcc archiving in GNUS 5
;;
;;	Gcc featuire is similar to Fcc, but the Gcc is special in
;;	GNUS 5 newsreader. All instruction you read above about Fcc are
;;	same for Gcc tracking feature. The only table you have to
;;	configure is
;;
;;	    timi-:table-gcc
;;
;;	Before you start defining Gnus folders; you must create them
;;	to Gnus from *Group* buffer with commands G m. Eg. I have
;;	created following Gnus folders for newsgroup posting
;;
;;	    nnfolder+archive:post-pgp
;;	    nnfolder+archive:post-emacs
;;	    nnfolder+archive:post-gen
;;
;;	Which are used when I hit F key to answer to message. [See
;;	setup below] You might wonder "...if I use code below and
;;	TIMI; what happens?". The code below sets the Gcc folder only
;;	once when you start composing message, propably a followup and
;;	there is a `Newsgroups' header in the buffer. But if you hit R
;;	r to reply directly to person (or use message-mode for
;;	mailing), there is `To' header in the buffer. Only now TIMI
;;	changes the Gcc field according to `To' field contents.
;;
;;	The code below is for Newsgroup posting, TIMI, in the other
;;	hand is for private mail handling.
;;
;;	    (setq gnus-message-archive-group 'my-gnus5-archive)
;;
;;  	    (defun my-gnus5-archive  (group)
;;  	      "Achive outgoing mail to right group: Create the group by G m"
;;  	      (interactive)
;;  	      (let* (
;;  		     )
;;  		(or (stringp group)                 ;No accidents...
;;  		    (setq group ""))
;;  		(cond
;;  		 ((string-match "pgp\\|anon\\|privacy" group)
;;  		  (my-z "nnfolder+archive:post-pgp"))
;;  		 ((string-match "emacs\\|gnu" group)
;;  		  (my-z "nnfolder+archive:post-emacs"))
;;  		 (t
;;  		  (my-z "nnfolder+archive:post-gen")
;;  		  ))
;;  		))
;;  	    ;;
;;
;;
;;  Saving mail buffers on Emacs exit
;;
;;	This file installs one function to `kill-emacs-hook' that loops
;;	through all mail buffers and appends the buffer content to
;;
;;	    timi-:dead-mail-file
;;
;;	If you had some unfinished messages that you didn't yet send, you
;;	can restore the copy from this file when you restart emacs again.
;;	By default function `mail-send-and-exit' is adviced so that
;;	when you send mail and hit C-c C-c  the *mail* buffer is not
;;	buried, but killed (ie. there is no dead mail in your emacs when
;;	you send them). In gnus 5 message-mode, you hsould also use
;;	following to trash sent mail
;;
;;	    (setq message-kill-buffer-on-exit t)
;;
;;	If you don't want to use this feature, add following command to
;;	your .emacs
;;
;;	    (add-hook 'timi-:load-hook 'my-timi-:load-hook)
;;	    (defun my-timi-:load-hook  ()
;;	      (remove-hook 'kill-emacs-hook 'timi-save-dead-mail))
;;	    ;; End
;;
;;
;;  Reminders
;;
;;      Has it happened to you that you have sent mail to someone and then
;;      you didn't get answer although you expected it?  If the mail is
;;      very old, you propably don't even remember that you had sent an
;;      important mail to someone. I have included simple reminder to this
;;      package. When you are about to send a message you call
;;
;;          C-c t a     timi-remind-add
;;
;;      to add current mail entry to your reminder list. What this command
;;      does, is, that it writes following entry into reminder file. The
;;      field separator is string "-|- and the fields contain date, to
;;      field and subject field content.
;;
;;          1997-1-13-|-Mr.Doodle@zlimit.com-|- Stock markets are doing good
;;
;;      You can peek the reminder list with following commands. The check
;;      command shows you the active reminder list in separate buffer for
;;      current person/subject and the the brief command tells you the
;;      count of reminders in echo area.
;;
;;          C-c t c     timi-remind-check
;;          C-c t i     timi-remind-info-brief
;;
;;      When you later compose another mail messages, the active reminders
;;      for current person is checked and you'll we warned if too many days
;;      have passed since you sent previous mail. See variable
;;
;;          timi-:remind-days-old
;;
;;      When you answer to a message, a possible reminder matching the
;;      subject and the TO field is silently deleted, so you seldom need to
;;      call delete command by hand
;;
;;          C-c t d     timi-remind-delete
;;
;;  Managing mailing lists
;;
;;      If you want to send mail to several people in some list that may
;;      shrink or grow, please look at the example at the end of this file.
;;      You find "t00" mailing list where are members 1..8 and entries in
;;      .mailrc look like this
;;
;;          t00  mailingList-A@nowhere.com
;;          t01  jack@fex.com
;;          t02  mike@foo.it
;;          t03  duke@bar.edu
;;          ...
;;
;;      You start posting to the mailing list by invoking M-x mail and adding
;;
;;          To: t00
;;
;;      Soon timi notices "t00" and adds all those people to the Cc.
;;
;;          To: mailingList-A@nowhere.com
;;          CC: jack@fex.com,
;;              mike@foo.it,
;;              duke@bar.edu
;;              ...
;;
;;      The TO field doesn't actually point to anywhere, it was just some
;;      stupid expansion of the first item t00. You cannot define t00 to be
;;      You.surname@site.com, because look what happens then:
;;
;;      o   The t00 is expanded --> To: You.surname@site.com and
;;          the additional Cc fields are added.
;;
;;      o   after a while TIMI awakens again and sees changed To: field
;;          "You.surname@site.com", it determines that there is no
;;          Cc information associated with it --> The Cc field is deleted!
;;
;;      That's why you must select some dummy address for "t00" and when
;;      the To and Cc are expanded in full, you invoking a hot key C-c t M,
;;      which deactivates TIMI and changes To address to point to you.
;;      This may sound a little complicated at first, but I assure you,
;;      it's quite easy when you try it few times.
;;
;;  Faster response -- but also affects your typing
;;
;;      The post command takes a nap regularly and when it awakens, it ;;
;;      checks what has happened to headers. If you feel that the response
;;      time ;; is too slow, you can change the value
;;
;;          timi-:awake-time
;;
;;      You must remember that the more smaller the time is, the more often
;;      the post command is run. This means that whenever you hit cursor
;;      down to move around the buffer, the more likely you see that the
;;      cursor hangs or doesn't move very quickly. It takes some time to
;;      run the post-command, so it is not recommended that you use too low
;;      value.
;;
;;      Please use PgUp and PgDown keys instead to move over big buffer
;;      portions to avoid cursor hang effect.
;;
;;  Configuring the faces -- colors
;;
;;      The default highlighting is only provided to your convenience. If
;;      you chooce to use lazy-lock or something else, you can disable the
;;      highlight with the following.
;;
;;      You can also copy `timi-hilit' and make modifications to it.
;;
;;	    (add-hook 'timi-:load-hook 'my-timi-:load-hook)
;;
;;	    (defun my-timi-:load-hook ()
;;	      (remove-hook 'timi-:mail-hook 'timi-hilit)
;;	      (remove-hook 'timi-:post-hook 'timi-hilit)
;;	      ;; Use my custom highlighting instead. You don't include this
;;	      ;; if you use font/lazy-lock
;;	      ;;
;;	      (add-hook 'timi-:mail-hook 'my-mail-hilit)
;;	      (add-hook 'timi-:post-hook 'my-mail-hilit)
;;	      )
;;	    ;; end
;;
;;	    (defun my-mail-hilit ()
;;	      ;; Whatever...
;;	      )
;;	    ;; end
;;
;;
;;      To customize default emacs faces, you can do this. See also
;;      command `set-face-foreground'.
;;
;;          (set-face-background 'highlight "blue3")
;;
;;      If you just want to change the default face colors, you can do it
;;      this way: after the file has been loaded.
;;
;;	    (add-hook 'timi-:load-hook 'my-timi-:load-hook)
;;
;;	    (defun my-timi-:load-hook ()
;;	      "override TIMI settings."
;; 	      ;;  - Let's create our own custom face
;; 	      ;;  - The defvar just makes the face once
;; 	      ;;
;; 	      (defvar my-:fc-yellow
;; 		(progn
;; 		  (make-face 'my-face-yellow)
;; 		  (set-face-foreground 'my-face-yellow "yellow")
;; 		  (set-face-background 'my-face-yellow "skyBlue")
;; 		  'my-face-yellow))
;; 	      ;;  Now use my custom face, default Fcc face
;; 	      ;;
;; 	      (setq tim1-:face2  my-:fc-yellow)
;; 	      )
;; 	    ;; end
;;
;;
;;  Examples
;;
;;      Look at the end of this file for other examples I have attached.

;;}}}
;;{{{ history

;; ........................................................ &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; Jun	23	1997	[jari]		19.28	v1.90-1.91	NotReleased
;; - Added function timi-fld-to-move-maybe, which takes care of putting
;;   cursor to right positon in To field.
;; - Forgot to activate advice: changed state from "dis" --> "act"
;;
;; Jun	23	1997	[jari]		19.28	v1.87-1.89	NotReleased
;; - Added passwd file cache, because it trakes too much time to parse
;;   /etc/passwd every time you start emacs and use the passwd completion
;;   feature.
;; - Added section 'Saving mail buffers on Emacs exit'.
;;
;; Jun	23	1997	[jari]		19.28	v1.86		NotReleased
;; - Added GNUS 5 message-mode's Gcc tracking and wrote section
;;   'Gcc archiving in GNUS 5'. New user variable timi-:table-gcc.
;; - moved mail-send-and-exit advice code from tinyad.el to here.
;;
;; Jun	19	1997	[jari]		19.28	v1.82-1.85	NotReleased
;; - Added more debuggging. Updated password command for SGI/irix
;;   Thanks to Jan Braun <c0031015@rzsgi.rz.tu-bs.de>
;;
;; Jun	4	1997	[jari]		19.28	v1.79-1.81	NotReleased
;; - Checkdoc 1.42 clear. defcustom support. Library function changes.
;; - New X-info setup at the end of file.
;; - Rick Flower <flower@ms31.sp.trw.com> told what is the right
;;   passwd read command in SunOs (timi-:passwd-cat-cmd)
;;
;; May	25	1997	[jari]		19.28	v1.74-1.78	NotReleased
;; - Small error in inserting X-info. does not insert filed if the text
;;   is nil. Rewrote the binding fucntions. X-info example setup update.
;; - Added MH support, thank to Rick Flower <flower@ms31.sp.trw.com>
;; - Checkdoc.el 1.22 now gives clear papers. defcustom support
;;
;; May	14	1997	[jari]		19.28	v1.72-1.73	NotReleased
;; - Added more debug calls. Removed narrow function timi-fld-fcc-determine,
;; - The package should now be much faster and occasional cursor hannging
;;   should not be that visible any more.
;; - Changed all non standard boolean tests to use when and unless
;;
;; Apr	18	1997	[jari]		19.28	v1.71		NotReleased
;; - Internal changes.
;;
;; Mar	22	1997	[jari]		19.28	v1.68-1.70	NotReleased
;; - Corrected VM's F f R r space adjustement in To field.
;; - Rewrote timi-fld-cc, timi-fld-fcc-determine, timi-to-field-on
;;   timi-to-field-off and timi-on-off-toggle.
;; - Some library function parameter changes. Text changes.
;;
;; Mar	19	1997	[jari]		19.28	v1.67		NotReleased
;; - Added tipgp-:debug varaible and debug fucntion.
;;
;; Mar	10	1997	[jari]		19.28	v1.65-1.66	NotReleased
;; - Added passwd completion feature from my private libraries to here.
;;   Works in Hp-UX.
;;
;; Mar	8	1997	[jari]		19.28	v1.63-1.64	NotReleased
;; - Adjusted the initial behavior. In VM; the F key inserts Cc headers,
;;   then the space count is 2, r and R doesn't insert Cc, so the space
;;   count is 1. C-x m simple mail, the space count is not touched.
;;
;; Mar	5	1997	[jari]		19.28	v1.62		NotReleased
;; - Fixed problem in installation, the right place to hook is
;;   mail-setup-hook and not mail-mode-hook, because mail-mode-hook
;;   doesn't see the Headers. Thanks to Dan Rich <drich@cisco.com>
;;   for patient debugging for the feature in VM.
;; - Wrote section 'Installation note'. Added single install function
;;   timi-install. Made timi-:message-type buffer local (bug)
;;
;; Feb	25	1997	[jari]		19.28	v1.57-1.61	NotReleased
;; - The completions didn't happen place in continued header line.
;;   Corrected. To field highlighting corrected.
;; - Converted module to use text properties instead of overlays, because
;;   XE doesn't like overlays.
;; - Corrected hilit match regexp calls so that rear-nonsticky
;;   propperty is used.
;;
;; Feb	23	1997	[jari]		19.28	v1.56		NotReleased
;; - Added Gnus 5 message mode support. Updated variables in bug report
;;   function. Now the TAB key is shared between TIMI and message
;;   more transparently.
;; - renamed all variables(hooks). They now have common syntax '-:'.
;;   Notice that load hook is now timi-:load-hook !
;;
;; Feb	20	1997	[jari]		19.28	v1.52-1.55	NotReleased
;; - Changed the post hook a little so that the user hook is run after
;;   the fcc change. This was needed so that fcc highlight is kept
;;   and not wiped.
;; - Explained the slow cursor movement in the which is due to post
;;   command wake time.
;; - Generated error if there was no To field. corrected.
;;
;; Feb  13      1997    [jari]          19.28   v1.50-1.51      NotReleased
;; - If you reased whole mail buffer, you got error, because there was
;;   no TO field. Corrected.
;; - timi-hilit is nowno-op in non window system.
;;
;; Jan  21      1997    [jari]          19.28   v1.47-1.49      NotReleased
;; - Peter Simons <simons@petium.rhein.de> commented that when you reply
;;   a message, all Field tracking is turned off.
;; - Now I have separated Cc,X-info and Fcc tracking to separate places,
;;   and added new feature where 2 spaces inhibit Cc tracking, but leaves
;;   X-info tracking. 3 spaces disables both. Fcc tracking is now aways
;;   done and not bound to to-field any more.
;; - Clarified the docs better to reflect these changes.
;; - Peter Simons noticed that cursor was in the wrong spot in TO
;;   field when timi-mail function was called. Now it is at the eol.
;;
;; Jan  13      1997    [jari]          19.28   v1.44-1.46      NotReleased
;; - Reminder code corrected, thanks to Peter Simons <simons@petium.rhein.de>
;;   for bug reports. Also cleared the documentation.
;; - Now the TIMI Cc field tracking is turned off when you reply to
;;   mail message. This preserves the original Cc fileds.
;; - If you added a reminder to message, it was immediately wiped away when
;;   message was sent by C-c c-c. This was a bug in the code.
;; - In the example installation, corrected call "remove hook"
;;   with "remove 'hook"
;;
;; Dec  17      1996    [jari]          19.28   v1.40-1.43      NotReleased
;; - Added REMINDERS to keep you not forgetting iportant mail messages
;;   where you expect answers. This is really handy feature I think,
;;   because many times I don't even remember I sent a mail to someone
;;   who didn't answer back
;; - 1.41+ have some bug fixes.
;;
;; Nov  27      1996    [jari]          19.28   v1.39           NotReleased
;; - Rearranged code a bit so that highlighting function is now used
;;   in different hooks. Updated documentation a bit and started using
;;   the unused function timi-fld-x-info. That helps tracking down
;;   problems in X info table.
;;
;; Nov  19      1996    [jari]          19.28   v1.36-1.38      NotReleased
;; - Corrected installation instructions
;; - Removed 'error' command call if abbrev was
;;   not found, because it would disable furher processing.
;; - Now the completion happens only in buffer that has first line
;;   KEYWORD: -- Hooking the TIMI to tinytab.el previously caused
;;   undesired results, because the buffer was not checked first.
;; - made guess completion to honor tinytab.el
;;
;; Nov  8       1996    [jari]          19.28   v1.35           NotReleased
;; - Added code to support tinytab.el minor mode simultaneously.
;; - Addd tinytab.el example setup to the end of file.
;;
;; Nov  8       1996    [jari]          19.28   v1.34           NotReleased
;; - The guess complete feature rewritten.
;;   venkatraman_sitaram <sitaramv@loc251.tandem.com> suggested that
;;   there should be also alias completion, because he names hte
;;   aliases in following way (firstname surname pairs)
;;
;;          jari-aalto    <jari.aalto@poboxes.com>
;;          peter-pan     <ppan@disney.com>
;;
;; - Now the installation installs key "\t" to call
;;   timi-guess-complete-in-headers. Prefix arg C-u completes anywhere.
;; - Added abbrev example setup to the end of file.
;;   timi-:complete-mode                :+ user var
;;   timi-guess-complete-in-headers     :+
;;   timi-guess-complete-1              :! rewritten
;;   timi-guess-complete                :! rewritten
;;
;; Nov  7       1996    [jari]          19.28   v1.33           NotReleased
;; - Small problem in guess-complete, it didn't replace the found match
;;   Correctred.
;;
;; Nov  7       1996    [jari]          19.28   v1.32           NotReleased
;; - Wrote section 'guess completion feature'
;; - I got a fancy idea, why not offer completion based on words you just
;;   typed in the TO/Cc/BCC lines? That way you just have to remember
;;   something from a person: email chars, surname chars. Then you hit
;;   the complete to search the match from cached aliases. Voila!
;; - This feature is much more usefull than the alias cmpletion,
;;   because many times you don't remember "what was the alias for person
;;   foo?". At least I don't when 50-100 aliases in my ~/.mailrc, but
;;   I remember something from the person itself :-)
;; - The nice thing is that it matches ANYTHING what you put in the " "
;;   in alias definition.
;;   timi-guess-complete                :+ new user cmd C-c c \t
;;   timi-read-completion-string        :+
;;
;; Nov  5       1996    [jari]          19.28   v1.31           NotReleased
;; - Radical rewrite of the abbrev part, it hung for me in 19.30. This
;;   needed changes to the mail lib too. The code is lot faster now.
;;
;; Nov  1       1996    [jari]          19.28   v1.28-30        NotReleased
;; - Added timi-update-cc-abbrevs call to write-hook, so that
;;   changes in ~/.mailrc are immediately shown in Cc table too.
;; - Added example how to send library information to me automatically.
;;
;; Sep  25      1996    [jari]          19.28   v1.25-1.27      NotReleased
;; - Rearranged loading. Now uses autoloads.
;; - Added x-info example for MIME discussion list.
;; - Moved timi-:post-hook so that it is run every time when awake
;;   time is up. [bug]
;;
;; Sep  11      1996    [jari]          19.28   v1.23-1.24      NotReleased
;; - Wrote section 'managing mailing lists'.
;; - Added new user command C-c t M to be used with mailing lists.
;;   timi-deactivate-and-send-to-you    :+
;;   timi-to-field-on                   :+
;;   timi-to-field-off                  :+
;;   timi-on-off-toggle                 :!
;;
;; Sep  9       1996    [jari]          19.28   v1.22           NotReleased
;; - put into use the "to" field lookup in timi-post-command, so that
;;   if the To fields remains the same, the function does nothing.
;;   This saves great deal of extra work, eg expanding the aliases all
;;   the time. (You would see cursor stopping in a while while writing
;;   your mail message.)
;; - Corrected the timi-write-file-hook: it has to return nil otw
;;   the normal save is not performed.
;;
;; Aug  29      1996    [jari]          19.28   v1.18-21        NotReleased
;; - Rewrote installation. Now just loading this package puts keybindings
;;   in mail-mode-map. no need to copy anything extar into one's .emacs
;; - New write-file-hook takes care of keeping things straight when you
;;   modify your ~/.mailrc
;;
;; Aug  28      1996    [jari]          19.28   v1.17           NotReleased
;; - Doesn't use expand-mail-aliases any more because XE doesn't
;;   have such function. Now uses my main libs ti::xe-expand-mail-aliases
;;   which works both for XE and Emacs. No ByteComp warnings any more.
;; - Seem working in XE 19.14 without problems now. (Not fully tested)
;;
;; Jun  28      1996    [jari]          19.28   v1.16           NotReleased
;; - XE, Used ti::xe-mail-abbrev-table function.
;;
;; Jun  26      1996    [jari]          19.28   v1.11-1.15      NotReleased
;; - Some textual corrections, like installation keybinding info.
;; - The mailabbrev pkg is not in XE, instead it uses mail-abbrevs.
;;   Corrected the require statement.
;; - Removed the idle-hook, because it wasn't a good idea. Now uses
;;   post command hook in 19.30+ too. The post-command is now
;;   run only when the To: field changes.
;; - Corrected point-moving bug, that was due to missing save-ex
;;   arounf the expand-mail-aliases command.
;;
;; Jun  24      1996    [jari]          19.28   v1.10           NotReleased
;; - Small minor corrections. Eg Cc field handling corrected due
;;   to libmt changes.
;;
;; Jun  19      1996    [jari]          19.28   v1.9            NotReleased
;; - Now Uses post-command-idle-hook in emacs 19.30. Added examples
;;   at the end of file.
;;
;; Jun  12      1996    [jari]          19.28   v1.8            NotReleased
;; timi-update-cc-abbrevs               :! infinite loop corrected.
;;
;; Jun  7       1996    [jari]          19.28   v1.7            NotReleased
;; - Corrected some errors that were caused by changes in lib 'mt'.
;; - Now supports dynamic compilation
;;   timi-on-off-toggle                 :+ new user function
;;
;; Jun  2       1996    [jari]          19.28   v1.6            NotReleased
;; - The automatic cc  field adding was totally broke. If you used
;;   mailaabrev to expanf the To: fields content, the alias vanished.
;; - Now if looks behind the alias too and adds cc field in right manner
;;   timi-:abbrevs-table-cc-expanded    :+
;;   timi-update-cc-abbrevs             :+
;;   timi-resolve-abbrevs               :+
;;   timi-fld-cc                        :! rewrote
;;   timi-mail                          :! corrected
;;
;; Apr  19      1996    [jari]          19.28   v1.5            NotReleased
;; - Reverted to previous version 1.19, because I found better way
;;   to cope with XE. There is now new lib tinylibxe.el
;;
;; Apr  18      1996    [jari]          19.28   v1.4            NotReleased
;; - Removed all Emacs specific overlay commands and used portable
;;   lib 'O' commands instead. This should be XE compatible.
;;
;; Apr  11      1995    [jari]          19.28   v1.3            NotReleased
;; - Textual correction. Some functions renamed (1 or 2).
;;
;; Apr  10      1995    [jari]          19.28   v1.2            NotReleased
;; - Added ###autoload commands and changed almost all function names
;;   to reflect better their purpose.
;;
;; Apr  10      1995    [jari]          19.28   v1.1            Created

;;}}}

;;; Code:

;;{{{ setup: require

;;; ......................................................... &require ...

(require 'tinylibm)

(eval-and-compile
  (if (emacs-p) ;;  XE and Emacs differ here
      (require 'mailabbrev)
    (require 'mail-abbrevs)
    )

  (autoload 'message-tab "message" t t)	;GNUS 5
  )


(eval-when-compile (use-dynamic-compilation))

(defgroup-tiny TinyMail timi-: mail
    "Some mail additions: dynamic Fcc, Cc, hilit, X-info.
  Overview of features

      o  Some handy additions to mail sending interface.
      o  Adds automatically Cc field when you type the To: address.
      o  Adds X-info dynamically while you type To: address.
      o  Changes Fcc dynamically according to header content.
      o  Very easy TAB completion: two modes, alias and definition string.
         or password file entry completion.
      o  Reminders: remind you about old messages where you expected an
         answer but got none.
      o  if Fcc folder has .gz or .Z name it automatically triggers
         loading jka-compr."
    )


;;}}}
;;{{{ setup: private


;;; .......................................................... &v-bind ...

;;; ......................................................... &v-hooks ...

(defcustom timi-:load-hook nil
  "*Hook run when package has been loaded."
    :type 'hook :group 'TinyMail
    )

;;  This is run only once, allows you to add headers etc.
;;
(defcustom timi-:mail-hook  '(timi-hilit)
  "*Hook run when simple mail, \\[mail], is detected."
    :type 'hook :group 'TinyMail
    )

;;  Add more dynamic change functions to this hook
;;
(defcustom timi-:post-hook  '(timi-hilit)
  "*Hook run when `timi-:awake-time' is up. This hook is always run."
    :type 'hook :group 'TinyMail
    )

;;; ....................................................... &v-private ...
;;; private

(defvar timi-:message-type  nil
  "Flag. The initial message type in mail buffer.
When TIMI is first turned on, it checks if the message
is composed with \\[mail] or if you have replied to someone
else's message with 'r' from some mail mode. This initial
message type determines how \\[timi-mail] call behaves in the buffer.")

(put 'timi-:message-type 'permanen-local t)
(make-variable-buffer-local 'timi-:message-type)
(setq-default timi-:message-type nil)


(defvar timi-:abbrevs-table-cc-expanded  nil
  "Private. Expanded abbrev list of `timi-:table-cc'.")

(defvar  timi-:time-counter 0
  "Private. Used in post command hook. Activate time.")
(make-variable-buffer-local 'timi-:time-counter)

(defvar timi-:last-to-field  nil
  "Private. Last to: field value.")
(make-variable-buffer-local 'timi-:last-to-field)


(defvar timi-:mail-aliases-alist  nil
  "Private. Cached aliases.
Run function `timi-update-mail-abbrevs' if you change your
~/.mailrc so that this variable gets updated.

Format: ((\"ALIAS\" . \"EXPANDED\") ..)")

(defvar timi-:tmp-buffer  " *timi-tmp*"
  "Temporary buffer.")



(defvar timi-:reminder-printed  nil
  "When `post-command' awakes, it prints the reminders and quits.
This flag is set to t to disable further messages between awakening times.")
(make-variable-buffer-local 'timi-:reminder-printed)


(defvar timi-:last-reminder-entry  nil
  "Last added reminder.")



(defvar timi-:passwd-alist nil
  "Contents of passwd file in assoc form: '((LOGNAME . PASSWD-ENTRY))")

(defvar timi-:passwd-completion-alist nil
  "Completion table of login names.")

(defvar timi-:debug-buffer  "*timi-debug*"
  "Debug buffer.")


;;}}}
;;{{{ setup: config

;;; ........................................................ &v-public ...
;;; User configurable

(defcustom timi-:table-keymap-list
  '(mail-mode-map
    message-mode-map
    mh-letter-mode-map
    )
  "*List of keymaps where to install default bindings. (TIMI pacakge)."
  :type '(repeat (symbol :tag "Keymap variable")) :group 'TinyMail
  )


(defcustom timi-:passwd-file "~/.emacs.timi.passwd.el.gz"
  "*Preparsed password file completions.
If this file does not exist it will be created when passwd
completion is needed. You _can_ keep this file in compressed format by
adding extension .gz to filename."
  :type 'file :group 'TinyMail
  )


(defcustom timi-:dead-mail-file "~/.emacs.timi.dead"
  "*Append all mail buffers to this file on Emacs exit."
  :type 'file :group 'TinyMail
  )

(defcustom timi-:awake-time
    (if (xemacs-p)
        7				;XE needs lower value
      10)
  "*Sleep time of `post-command-hook' before activation."
  :type '(integer :tag "Movement Cycles") :group 'TinyMail
  )


(defcustom timi-:remind-days-old  7
  "*Maximum wait time in days before list of old reminders are displayed.
activates from `post-command' hook.

Use \\[timi-remind-edit] to edit the reminder entries."
  :type '(integer :tag "Days") :group 'TinyMail
  )


(defcustom timi-:x-info-label "X-info"
  "*Extra info field label. There must be NO COLON or SPACES.
User addd header names must start with X-* according to RFC."
  :type '(string :tag "Header name") :group 'TinyMail
  )



(defcustom timi-:password-mode t
  "*Should we try to complete passwd entries?.
if normal .mailrc completion fails then  non-nil enables completion."
  :type 'boolen :group 'TinyMail
  )


(defcustom timi-:passwd-cat-cmd
  (cond
   ((string-match "hpux\\|hppa" (emacs-version))
    "ypcat passwd")
   ((string-match "solaris" (emacs-version))
    "cat /etc/passwd")
   ((string-match "sparc\\|sun\\|sunos" (emacs-version))
    "cat /etc/passwd")
   ((string-match "sgi-irix\\|irix" (emacs-version))
    "ypcat passwd")
   (t
    (error "tinymail: timi-:passwd-cat-cmd, Please contact maintainer.")
    ))
  "*Shell command that prints the contents of standard UNIX passwd file.
If your systems shell command isn't seen here, contact maintainer
immedately and report right shell command, so that it is set automatically
right."
  :type '(string :tag "Command")  :group 'TinyMail
  )



(defcustom timi-:complete-mode 'string
  "*Control how completion is done.

'alias

    Means that we should complete alias names and
    that the alias expansion is shown in echo-area.

'string

    Means that _string_, which may include any character including white
    spaces, is searched from the full expansion list of aliases. This way
    you can remember anything from the person itself and it will be
    searched. Found expansion is inserted in place of typed
    string.

    Your ~/.mailrc can have entries like this:

    alias mark   \"Mark Eggert -- Project engineer <meg@twenix.com>\"
    alias mike   \"Michael Lowell  -- SkyTrax consulting <ml@sky.com>\"

    The right hand strings are searched with picked _string_ and
    if there is only one match for the string, the expansion (rh element)
    is inserted into buffer."
  :type '(choice
	  (const alias)
	  (const expansion)
	  )
  :group 'TinyMail
  )



(defcustom timi-:remind-file  "~/.emacs.var.timi"
  "*File where to put reminder information about current message.
The file reminds you about mails where you expect answers."
  :type 'file :group 'TinyMail
  )



;;; ........................................................ &v-tables ...


(defcustom timi-:table-x-info nil
  "*The X-info: field is added if REGEXP entry match To: field.
The LISP-FORM can be string or any form that yields string.
Remember that according to RFC the continuing lines must have
\\t or space at the beginning. Eg.

X-info: line-1-comes-here      [newline]
[ tab  ]line-2-continues-here  [newline]
[ tab  ]line-3-continues-here  [newline]


Format is '((REGEXP LISP-FORM) (REGEXP LISP-FORM) ..)"
  :type  '(repeat
	   (list
	    (string :tag "Regexp")
	    (sexp   :tag "Form")
	    ))
  :group 'TinyMail
  )



(defcustom timi-:table-cc nil
  "*Add Cc and CC-FIELD-STRING if TO-FIELD-REGEXP match To field.
Remember to use most restrictive regexps first, since first match in the
list will be used.

If you change this after this TIMI package has been loaded, you must
call function \\[timi-update-cc-abbrevs]

Format is

'((TO-FIELD-REGEXP CC-FIELD-STRING) ..)

Example:
  '((\"regexp1\"
     \" cc-recipient1, cc-recipient2\"
;;;   ^^Leave emty spaces here.
    ))"
  :type  '(repeat
	   (list
	    (string :tag "To Regexp")
	    (sexp   :tag "CC field string")
	    ))
  :group 'TinyMail
  )



(defcustom timi-:table-fcc nil
  "*Replace Fcc content with FCC-FIELD-STRING when headers match REGEXP.
If there is _two_ spaces in the Fcc field, the Fcc header is not touched.
Format is '((REGEXP FCC-FIELD-STRING) ..)"
  :type  '(repeat
	   (list
	    (string :tag "To Regexp")
	    (sexp   :tag "Fcc field string")
	    ))
  :group 'TinyMail
  )


(defcustom timi-:table-gcc nil
  "*Replace Gcc content with GCC-FIELD-STRING when headers match REGEXP.
If there is _two_ spaces in the Gcc field, the Gcc header is not touched.
Format is '((REGEXP GCC-FIELD-STRING) ..)"
  :type  '(repeat
	   (list
	    (string :tag "To Regexp")
	    (sexp   :tag "Gcc field string")
	    ))
  :group 'TinyMail
  )

;;; .......................................................... &v-misc ...

(defcustom timi-:debug  nil
  "*Debug on/off.
If you find anything weird from the pacakge:  turn on the debug, clear
debug buffer if it exists, do actions, save mail buffer result and
copy debug buffer and contact maintainer.

Debug is directed to `timi-:debug-buffer'."
  :type 'boolean :group 'TinyMail
  )


;;}}}
;;{{{ setup: hilit

;;; ........................................................ &vp-hilit ...
;;; private


(defvar timi-:face1 'highlight
  "Mail face. 'Subject'.")

(defvar timi-:face2 'bold-italic
  "Mail face. 'Fcc' or 'Gcc'.")

(defvar timi-:face3 'highlight
  "Mail face. 'Cc' or 'Newsgroup'.")

(defvar timi-:face4
  (or (and (face-p 'region)			;XE doesn't have this
	   'region)
      'bold)
  "Mail face. 'To' field.")

;;}}}
;;{{{ setup: version

;;; ....................................................... &v-version ...

(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinymail.el"
 "timi"
 timi-:version-id
 "$Id: tinymail.el,v 1.91 1997/06/23 11:37:53 jaalto NotReleased $"
 '(timi-:version-id
   write-file-hooks
   message-mode-hook
   mail-mode-hook
   mail-setup-hook
   mail-send-hook
   mh-letter-mode-hook
   )))


(defmacro timi-debug (&rest args)
  "Generate debug if debug is on with ARGS."
  (`
   (when timi-:debug
     (let* ( ;; write to package's private buffer.
            (ti:m-debug-buffer timi-:debug-buffer)
            )
       (save-match-data
         (d!! (,@ args) "\n")
         )))))

(defun timi-debug-toggle  ()
  "Toggle package debug."
  (interactive)
  (var-toggle timi-:debug)
  (message "Debug is %s" (if timi-:debug "on" "off"))
  )


;;}}}

;;; ########################################################### &Funcs ###

;;{{{ code: install

;;; ----------------------------------------------------------------------
;;;
(defun timi-install-default-bindings-1 (map)
  "Install bindings to MAP."
  (define-key map "\C-ctm"  'timi-mail)
  (define-key map "\C-ctM"  'timi-deactivate-and-send-to-you)
  (define-key map "\C-ct-"  'timi-on-off-toggle)
  (define-key map "\C-ct\t" 'timi-guess-complete)
  (define-key map "\t"      'timi-guess-complete-in-headers)

  (define-key map "\C-cta"  'timi-remind-add)
  (define-key map "\C-ctc"  'timi-remind-check)
  (define-key map "\C-cti"  'timi-remind-info-brief)
  (define-key map "\C-cte"  'timi-remind-edit)
  (define-key map "\C-ctd"  'timi-remind-delete)
  )

;;; ----------------------------------------------------------------------
;;;
(defun timi-install-default-bindings ()
  "Install keybindings to all mail mode maps.
This function should be used in mail mode hook."
  (interactive)
  (let* ((list timi-:table-keymap-list)
	 sym
	 )
    (list-loop list sym
      (when (boundp sym)			;Keymap may not exist
	(let* ((map     (symbol-value sym))
	       (look    (lookup-key map "\C-ct"))
	       )
	  (cond
	   ((not (or (null look)	;not defined
		     (keymapp look)))	;already defined
	    (error "Sorry, the default C-ct prefix already in use.")
	    )
	   (t
	    (timi-install-default-bindings-1 map)
	    )))))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun timi-cancel-bindings (&optional install)
  "Remove `timi-install-default-bindings' function from all mail mode hooks.
Optionally INSTALL function back to hooks."
  (add-hooks
   '(mail-setup-hook
     message-mode-hook
     mh-letter-mode-hook
     )
   'timi-install-default-bindings
   (not install)
   ))


;;; ----------------------------------------------------------------------
;;;
(defun timi-install-hooks  (&optional remove verb)
  "Install needed hooks, optionally REMOVE. VERB."
  (let* ((func  (if remove 'remove-hook 'add-hook))
	 )
    (VERB)
    (timi-cancel-bindings (not remove))	;remove = nil --> install

    (add-hooks
     '(mail-setup-hook message-mode-hook mh-letter-mode-hook)
     'timi-mail
     remove)

    (add-hooks
     '(mail-send-hook message-send-hook mh-before-send-letter-hook)
     'timi-remind-delete
     remove)

    (funcall func 'write-file-hooks		'timi-write-file-hook)

    (when verb
      (if remove
	  (message "timi: hooks removed, timi disabled.")
	(message "timi: hooks installed, timi enabled.")))

    ))

;;; ----------------------------------------------------------------------
;;; - This is the main controller "install" that calls all other
;;;   functions.
;;;
(defun timi-install  (&optional remove verb)
  "Install or REMOVE package. VERB."
  (interactive "P")
  (VERB)
  (timi-install-hooks remove verb))

;;}}}
;;{{{ code: misc


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun timi-hilit ()
  "Mail highlighting. This is no-op in non-windowd system."
  (let* ((fc1 timi-:face1)
         (fc2 timi-:face2)
         (fc3 timi-:face3)
         (fc4 timi-:face4)
         maxp
         )
    (when (and (face-supported-p)
	       (ti::mt-mail-p)
	       (setq maxp (ti::mt-hmax)))

;;; Too slow in post command
;;;      (tima-clear-region-properties (point-min) maxp '(owner tima))

      (tima-clear-region-properties (point-min) maxp)

      (save-excursion
        (PMIN) 	(tima-re-search
		 "^To: *\\(.*+\\)@" nil 1 maxp
		 (list 'owner 'tima 'face fc4 'rear-nonsticky 'face))

        ;; "Gcc" is used by Gnus5
        ;; "Fcc" is normal in Gnus4 and RMAIL
        ;;
        (PMIN) (tima-re-search
		"^[FG]cc: *\\(.*\\)" nil 1 maxp
		(list 'owner 'tima 'face fc2 'rear-nonsticky 'face))

        (PMIN) (tima-re-search
                "^\\(Newsgroups\\|cc\\): \\([^\n]+\\)"
                nil 2 maxp
                (list 'owner 'tima 'face fc3 'rear-nonsticky 'face))

        (PMIN) 	(tima-re-search
		 "^Subject: *\\(.*\\)" nil 1 maxp
		 (list 'owner 'tima 'face fc1 'rear-nonsticky 'face))
        nil
        ))))

;;; ----------------------------------------------------------------------
;;;
(defun timi-write-file-hook ()
  "Rebuild aliases everytime .mailrc is saved."
  (when (string-match "\\.mailrc" (or (buffer-file-name) ""))
    (message "tinymail: Updating mail aliases and abbrevs...")
    (build-mail-aliases)

    (when (fboundp 'build-mail-abbrevs)    ;update abbrevs too
      (sfuncall 'build-mail-abbrevs))

    (timi-update-mail-abbrevs)
    (timi-update-cc-abbrevs)
    (message "tinymail: Updating mail aliases and abbrevs...done")
    nil                                 ;Hook return value
    ))


;;; ----------------------------------------------------------------------
;;;
(defun timi-update-mail-abbrevs  ()
  "Build up all aliases to `timi-:mail-aliases-alist' cache.
You need to run this function if you change your ~/.mailrc."
  (interactive)
  (setq timi-:mail-aliases-alist (ti::m-mail-abbrevs-get-alist))
  )


;;; ----------------------------------------------------------------------
;;;
(defun timi-update-cc-abbrevs  ()
  "Reparse cc abbrevs in `timi-:table-cc'."
  (interactive)
  (let* ((cc-list       timi-:table-cc) ;data list
         pure-list                      ;straight list
         elt elt2
         blocks
         )
    (message "timi: Updating mail aliases 1 ...")
    (or timi-:mail-aliases-alist
        (timi-update-mail-abbrevs))

    ;;  Gather all unique abbrevs to pure-list
    ;;
    (message "timi: Updating mail aliases 2 ...")
    (list-loop cc-list elt

      (if (not (member (car elt) pure-list))
	  (push (car elt) pure-list))

      (setq blocks (ti::s-get-blocks
                    (nth 1 elt) "[ ,\n\t]+")) ;; parse RH elts

      (list-loop blocks elt2
	(if (not (member elt2 pure-list))
	    (push elt2 pure-list)))
      )

    ;; Look behind every abbrev then --> email
    ;;
    (message "timi: Updating mail aliases 3 ...")
    (if pure-list
        (setq timi-:abbrevs-table-cc-expanded
              (timi-resolve-abbrevs pure-list)))
    (message "timi: Updating mail aliases ... done.")
    ))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun timi-deactivate-and-send-to-you  ()
  "Deactivate TIMI and change To field to point to your address.
This function is normally used when you use mailing lists. See
documentation in the tinymail.el or hit \\[timi-version]."
  (interactive)
  (if (var-nil-p user-mail-address)
      (error "Please set user-mail-address")
    (ti::mt-kill-field  "^to:" user-mail-address)

    (timi-to-field-off)
    (if (interactive-p)
        (message "Address changed to point to you. TIMI signs off."))
    ))



;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun timi-mail (&optional disable verb)
  "Install Mail watchdog to `post-command-hook'.
Adds or changes Cc, FF, X-info fields on the fly while you're
composing the message.

Input:
  DISABLE	removes package from `post-command-hook'
  VERB		print verbose message.

References:
  `timi-post-command'
  `timi-:mail-hook'"
  (interactive "P")
  (VERB)

  (cond
   (disable
    (if (memq 'timi-post-command post-command-hook)
        (remove-hook 'post-command-hook 'timi-post-command))
    )
   (t
    ;;  Now install automatic mailing aid
    ;;
    (make-local-hook 'post-command-hook)
    (add-hook 'post-command-hook 'timi-post-command)
    (run-hooks 'timi-:mail-hook)

    ;;  If you're replying to someone else's message, the TO field
    ;;  must have two spaces to turn off TIMI so that the remaining
    ;;  Cc fields are not modified. (VM's F key)
    ;;
    ;;  R and r keys don't add Cc field, so we put there only one space.
    ;;
    ;;  For simple C-x m TO field will be initially empty.
    ;;

    (timi-debug
     "MAIN BEFORE"
     (point)
     "mtype"   timi-:message-type
     "to"      (ti::mt-get-field-1 "to")
     "cc"      (ti::mt-get-field-1 "cc")
     "Subject" (ti::mt-get-field-1 "subject")
     "\n["
     (buffer-substring (point-min) (point-max))
     "]\n"
     )

    (cond
     (timi-:message-type
      ;;  User calls us some other time
      ;;
      (if (var-nil-p (ti::mt-get-field-1 "subject"))
	  (timi-to-field-off))
      )
     (t
      ;;  We're called from some mail setup hook. See what is the initial
      ;;  state of the buffer...
      ;;
      (cond
       ((var-nil-p (ti::mt-get-field-1 "subject"))
	;;  simple mail: there is no subject field filled
	;;
	(setq timi-:message-type 'simple)
	)
       ((not (var-nil-p (ti::mt-get-field-1 "cc"))) ;VM's F inserts Cc
	;;  Deactivate CC field tracking, because there already is CC field.
	;;  Let user change it if needed.
	;;
	(setq timi-:message-type 'reply-cc)
	(timi-to-field-off)
	)
       (t				;R or r; No cc field
	(setq timi-:message-type 'reply)
	(timi-to-field-on)
	))))
    ))


  (timi-fld-to-move-maybe)

  (timi-debug
   "MAIN AFTER"
   (point)
   "disable" disable
   "mtype"    timi-:message-type
   "to"      (ti::mt-get-field-1 "to")
   "cc"      (ti::mt-get-field-1 "cc")
   "Subject" (ti::mt-get-field-1 "subject")
   "\n["
   (buffer-substring (point-min) (point-max))
   "]\n"
   )

  (if verb
      (message
       (format "Simple mail hook %s"
               (if (memq 'timi-post-command post-command-hook)
                   "on" "off"))))
  )



;;; ----------------------------------------------------------------------
;;; - This is just an example function, Demonstrates how you can send
;;;   the version information automatically, See timi-:table-x-info
;;;
(defun timi-mail-info-ange ()
  "Return ange-ftp/efs info."
  (if (boundp 'efs-19-version)
      (concat "efs "
              ;;  This is only needed to shut up the 19.28 Byte Compiler
              ;;  Which can't see the bounndp test...
              (symbol-value 'efs-19-version)
              "1.14\n")
    (concat "ange-ftp," emacs-version)))


;;; ----------------------------------------------------------------------
;;;
(defun timi-resolve-abbrevs (list)
  "Resolves LIST of mail abbrevs in format '(\"abbrav\" \"abbrev\" ..)
Return:
  ((\"alias\" . \"expansion\") (A . E) .. )
  alias        = the alias definition
  expansion    = expanded alias"
  (let* ((abbrevs   timi-:mail-aliases-alist)
         pre-abbrev-expand-hook          ;; prevent recursion
         exp-list
         elt
         hit
         )
    (list-loop list elt
      ;;  Returns (ABBR . ABBR-EXPANDED)
      ;;
      (if (not (setq hit (assoc elt abbrevs)))
          (message "timi: Can't find abbrev '%s', is it in ~/.mailrc ?" elt)
	(if (not (member hit exp-list))
	    (push hit exp-list)))
      )

    exp-list
    ))


;;; ----------------------------------------------------------------------
;;;
(defun timi-passwd-save  (&optional load verb)
  "Save passwd completions to file `timi-:passwd-file'. Optionally LOAD.
If that variable is nil, then do nothing."
  (let* ((file  timi-:passwd-file)
	 (list  timi-:passwd-completion-alist)
	 (list2 timi-:passwd-alist)
	 elt
	 buffer
	 )
    (VERB)
    (if (and file (string-match "\\.gz$\\|\\.Z$" file))
	(require 'jka-compr))

    (when file
      (cond
       (load
	(if (not (file-exists-p file))	;Create file then
	    (timi-passwd-define-variables 'force))
	(load file)
	(if verb (message "TIMI: passwd completions loaded."))
	)
       (t
	(if (null list)
	    (error "timi-:passwd-alist is empty, nothing to save."))

	(setq buffer (use-tmp-buffer "*timi-passwd*" 'clear))
	(job-buffer buffer
	  (insert ";; (#) " file " -- TinyMail.el password completions\n\n")

	  (insert "(defconst timi-:passwd-completion-alist\n  '(\n")
	  (list-loop list elt
	    (insert (format "    (\"%s\" . %d)\n" (car elt) (cdr elt))))
	  (insert "    ))\n\n")


	  (insert "(defconst timi-:passwd-alist\n  '(\n")
	  (list-loop list2 elt
	    (insert (format "    (\"%s\" . \"%s\")\n" (car elt) (cdr elt))))
	  (insert "    ))\n\n")

	  (insert ";; End of file")

	  (write-region (point-min) (point-max) file)
	  )
	(kill-buffer buffer)
	(if verb (message "TIMI: passwd completions saved."))
	)))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun timi-passwd-define-variables (&optional force no-save)
  "Defines passwd variables.
Read definitions from  `timi-:passwd-file' if FORCE is nil.

Input:
  FORCE	    flag, if non-nil, read passwd table and reset all variables.
  NO-SAVE   flag, if non-nil, do not save passwd completions to file."
  (cond
   ((or force
	(and timi-:passwd-file
	     (not (file-exists-p timi-:passwd-file)))
	(and (null timi-:passwd-file)
	     (null timi-:passwd-alist))
	)
    (message "timi: Buildig password completions...")
    (setq timi-:passwd-alist (ti::m-passwd-build-alist timi-:passwd-cat-cmd))
    (message "timi: Buildig password completions...done")
    )
   ((and (file-exists-p timi-:passwd-file)
	 (null timi-:passwd-alist))
    (timi-passwd-save 'load)
    ))

  (when (or force
	    (null timi-:passwd-completion-alist))
    (setq timi-:passwd-completion-alist
	  (let* ((list    timi-:passwd-alist)
		 (i       0)
		 completions
		 x
		 )
	    (list-loop list x
	      (setq i (1+ i))

	      (setq completions (cons  (cons (car x) i) completions)))

	    completions
	    ))
    (if (null no-save) (timi-passwd-save))
    ))


;;}}}
;;{{{ Completion

;;; ----------------------------------------------------------------------
;;;
(defun timi-read-completion-string-read ()
  "Return completion string from current point or nil.
See `timi-guess-complete'.

Return:
 (beg end string)"
  (let* ((point  (point))
         string
         beg end
         )
    (safe-narrow (BOLPP) (EOLPP)
      (goto-char point)

      ;;  First, go away from whitespace so that match-end gets
      ;;  length in newxt case statement
      ;;
      (skip-chars-forward " \t")

      (cond
       ((or (re-search-backward "[:,][ \t]*"  nil t)
	    ;; continued line
	    ;;
	    ;;  Cc: this,
	    ;;
	    ;;
	    ;;
	    ;;      here_is_point
	    ;;
	    (re-search-backward "[:, ][ \t]*"  nil t))
        (setq beg (match-end 0)))

       ((re-search-backward "^[ \t]*" nil t)
        (setq beg (1+ (point)))
        ))

      (goto-char beg)
;;;       (d! beg (looking-at "[^\n\t ,:]+") (buffer-substring beg (EOLP)))


      ;;  There must be somthing, not just empty lines
      ;;
      (when (looking-at "[^\n\t ,:]+")
        (cond
         ((re-search-forward " *,"  nil t)
          (setq end (match-beginning 0)))

         ((re-search-forward "[ \t]*,\\|[ \t]*$"  nil t)
          (setq end (match-beginning 0))
          )))

      (if (and beg end)
          (setq string (buffer-substring beg end)))
      )

    (if string
        (list beg end string)
      nil)
    ))

;;; ----------------------------------------------------------------------
;;;
(defun timi-guess-complete-1  (&optional mode verb)
  "Try to expand using underlying characters.
Look completion from `mail-aliases'. If there is more that 1 match,
ask which one to use.

If MODE is 'string, then text read from buffer must be separated by

    LEFT-COLON:          txt [COMMA,WHITESPACE]
    LEFT-ALL-WHITESPACE  txt [COMMA,WHITESPACE]
    COMMA                txt [COMMA,WHITESPACE]

If MODE is 'alias then text is read direcly under point separated
by spaces. This function does nothing if the first line doesn't contain

  KEYWORD:

Indicating a mail like mode. VERB prints verbose messages.

Return:
  t	    ,completed
  nil"
  (let* ((fid    "timi-guess-complete-1: ")
	 (list   timi-:mail-aliases-alist)
	 (mail   (ti::mt-mail-p))
	 data
         elt
         beg end str
         done
	 ret
         )
    (timi-debug fid "in" mode verb mail)
    (when mail
      ;; ... ... ... ... ... ... ... ... ... ... ... ... ... .. get word ...
      (cond
       ((eq mode 'string)
        (setq data   (timi-read-completion-string-read))
        (if data
            (setq beg (nth 0 data)
                  end (nth 1 data)
                  str (regexp-quote (nth 2 data))))
        )
       ((eq mode 'alias)
        (save-excursion
          (when (< (skip-chars-backward "^ \t\n") 0)
            (setq beg (point))
            (when (> (skip-chars-forward "^ \n\t") 0)
              (setq end (point)))))
        (if (and beg end)
            (setq str (buffer-substring beg end))
          (setq str nil))
        ))

      (timi-debug fid "after type" beg end str)
      ;; ... ... ... ... ... ... ... ... ... ... ... ... .. find matches ...
      (when (and (not (var-nil-p str))
                 (setq elt
                       (list-find
                        list str
                        '(progn
                           (if (eq mode 'string)
                               (string-match arg (cdr (car table)))
                             (string-match (concat "^" arg)
					   (car (car table)))))
                        'all-matches
                        )))
        (setq str nil)

	(timi-debug fid "matches" (length elt) (cdr (car elt)) elt)
        ;;  It makes sense to complete only aliases, because they are short
        ;;
        ;;

        (unless (eq 1 (length elt))
          (setq
           str
           (completing-read
            (format "%d: pick alias: " (length elt))
            elt nil 'match-it
            (car (car elt))
            ))
          (setq elt (list (assoc str elt)))
          )

        (when elt
          (if (eq mode 'string)
              (setq str (cdr (car elt)))
            (setq str (car (car elt)))))

	(timi-debug fid "almost" elt str)
        (cond
         (str
          (goto-char beg) (delete-region beg end)
          (insert str) (setq done t)
	  (setq ret t)
          ))
        )

      (cond
       ((and verb (null done) str)
        (message (format "Timi: no completion match on '%s'" str))
	)
       ((and verb str done (eq mode 'alias))
        (message (cdr (car elt)))
        ))

      (timi-debug fid "ret" ret)
      ret
      )))




;;; ----------------------------------------------------------------------
;;;
(defun timi-guess-complete  (&optional verb)
  "Main function to select complete function mode.
See variable `timi-:complete-mode'. VERB."
  (interactive "*")
  (let ((mode  timi-:complete-mode)
	(pmode timi-:password-mode)
        )
    (VERB)
    (cond
     ((eq  mode 'alias)
      (and (null (timi-guess-complete-1 'alias  verb))
	   pmode
	   (timi-passwd-complete-names verb)
	   ))

     ((eq  mode 'string)
      (and (null (timi-guess-complete-1 'string  verb))
	   pmode
	   (timi-passwd-complete-names verb)
	   ))
     (t
      (error "Unknown mode in timi-:complete-mode.")
      ))))

;;; ----------------------------------------------------------------------
;;;
(defun timi-guess-complete-in-headers  (&optional arg)
  "Like \\[timi-guess-complete], but complete only in headers.
If `last-command-char' is non-nil, call `self-insert-command'
outside of headers. You can eg. hook this function to TAB key.

If optional ARG is '(4), then the completion is forced at current point

Note:

  In case the buffer is GNUS 5 message buffer(news article), the
  function `message-tab' is called inside Newsgroups and Followup-To
  fields."
  (interactive "P")
  (let* ((ch      last-command-char)
         (hmax    (if (ti::mt-mail-p)
		      (or (ti::mt-hmax) 1)))
         )
    (cond
     ((and
       (eq major-mode 'message-mode)
       (save-excursion
	 (and (ti::mt-next-field-start 'move 'back)
	      (looking-at "Newsgroups\\|Followup-To")
	      )))
      (call-interactively 'message-tab)
      t					;return value
      )
     (t

      (if (or (equal '(4) current-prefix-arg)
	      (and hmax
		   (< (point) hmax)))
	  (call-interactively 'timi-guess-complete)

	(setq arg (prefix-numeric-value arg))

	;;  The TinyTab minor mode overrides tab, return nil
	;;  so that it knows to proceed
	;;
	(if (and (boundp 'tit-mode)	;tinytab.el
		 (symbol-value 'tit-mode)) ;is it on?
	    nil
	  (when ch
	    (self-insert-command (or arg 1))
	    t
	    )))
      ))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun timi-passwd-complete-names (&optional verb)
  "Complete passwd file's names in header area, otw do nothing. VERB.
Return:
  t	,completed
  nil	,nothing done"
  (interactive)
  (let* ((end   (ti::mt-hmax))
	 ret
	 table
	 word
	 completions
	 )
    (VERB)
    (save-excursion
      (forward-char -1)		;move ovr some char
      (setq word (ti::b-read-word "[-_+a-zA-Z0-9]" 'strict)))

    (when (and (< (point) end)
	       (not (var-nil-p word)))
      (timi-passwd-define-variables)
      (setq table	timi-:passwd-completion-alist)
      (setq completions (all-completions word table))

      (cond
       ((eq 1 (length completions))
	(insert (substring (car completions) (length word)))
	(if verb (message-quick (cdr completions)))
	)
       (t
	;;   Flash possible completions
	(if verb (message (list-to-string completions)))
	(timi-passwd-grep word 'verb)
	))
      (if verb (message "timi: Passwd entry completed using word %s" word))
      (setq ret t)
      )
    ret
    ))


;;; ----------------------------------------------------------------------
;;;
(defun timi-passwd-grep (user &optional verb)
  "Greps USER names from passwd. VERB."
  (interactive "sUser regexp: ")
  (let ((buffer   timi-:tmp-buffer)
	table
	alist
	elt
	)
    (VERB)

    (if (null timi-:passwd-alist)
	(timi-passwd-define-variables))

    (setq table timi-:passwd-alist)

    ;;   Force loading it if not exist
    (if verb (message "Grepping passwd contents..."))
    (setq alist (ti::m-passwd-grep-user-alist user nil table))
    (if verb (message "Grepping...done"))

    (when alist
      (save-excursion
	(use-tmp-buffer buffer 'clear)
	(set-buffer buffer)
	(list-loop alist elt
	  (insert (format "%-10s %s\n"  (car elt) (cdr elt))))
      (display-buffer buffer)
      ))
    ))



;;}}}
;;{{{ reminders

;;; ........................................................ &reminder ...


;;; ----------------------------------------------------------------------
;;;
(defun timi-remind-delete  (&optional verb)
  "Delete reminders for current subject. VERB."
  (interactive)
  (let* ((entry (timi-remind-compose))
         (last   timi-:last-reminder-entry)
         )
    ;;  - User adds reminder to his list, then he presses C-c C-c to
    ;;    send the message --> the mail-send-hook calls this function
    ;;    to see if the message was "reply" to a message that was
    ;;    recorded in reminder list.
    ;;  - We don't want to delete newly added reminder, that's why
    ;;    we test against 'last'
    ;;  - For interactive call, we are permitted to delete the entry.

    (if (or (not (string= entry last))
            (and (string= entry last)
                 (interactive-p)))
        (timi-remind-check 'subject 'kill (or verb (interactive-p))))
    nil					;provide clean return value
    ))


;;; ----------------------------------------------------------------------
;;;
(defun timi-remind-info-brief  (&optional secs)
  "Display SECS brief reminder summary for current TO field.
You must press eg. return key to wipe the message away,"
  (interactive)
  (let* ((data (timi-remind-check 'to))
         (nbr  (car-safe data))
         )
    (cond
     ((and data (> nbr 0))
      (read-char-safe-until
       (format
        "%d reminders matching TO field, %s " nbr (car (cdr data)))))

     (t
      (if (interactive-p)
          (message "No reminders."))
      ))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun timi-remind-edit  ()
  "Show reminder buffer fro editing."
  (interactive)
  (let* (buffer
         )
    (if (not (file-exists-p timi-:remind-file))
        (message "Sorry, no reminder file available.")
      (setq buffer (find-file-noselect timi-:remind-file))
      (if (or (not (empty-buffer-p buffer))
              (y-or-n-p "The reminder file is empty. Still edit? "))
          (pop-to-buffer buffer)
        (message "Delete entries and then save buffer. Modify with care.")
        ))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun timi-remind-older-than (&optional days)
  "Check which reminders that are older that DAYS [default 7].

Return:
  buffer     ,if entries found
  nil"
  (interactive)
  (let* ((buffer (use-tmp-buffer timi-:tmp-buffer 'clear))
         (file   timi-:remind-file)
         (tlist  (ti::c-time-list))
         (yy     (string-to-int (nth 2 tlist)))
         (mm     (nth 1 tlist))
         (dd     (nth 0 tlist))
         tmp
         day
         hit
         ret
         )
    (setq days (or days 7))

    (if (> days 30)
        (error "Sorry, supports only max 30 days."))

    (setq tmp (- dd days))

    (if (> tmp 0)       ;can't substract that much
        (setq dd tmp)
      (if (not (eq mm 1))
          (dec mm)
        (dec yy) (setq mm 12))
      (setq dd (- 31 days))
      )

    (job-buffer (find-file-noselect file)   ;Make copy
      (append-to-buffer buffer (point-min) (point-max)))

    (job-buffer buffer
      (sort-lines 'newest-first (point-min) (point-max))
      ;;  Must not include empty lines.
      (ti::b-cnv-empty-lines (point-min) (point-max))
      (PMIN)

      ;;  Hmm, how do we find which reminders are old?
      ;;  o  search by month
      ;;  o  Compare days, exact or older that given day
      ;;  o  delete rest that are newer when match found
      ;;
      (cond
       ((re-search-forward (format "19%s-%s" yy mm) nil t)
        (beginning-of-line)
        (when (setq day (bufmatget "^19..-..-\\(..\\)" 1))
          (setq day (string-to-int day))
          (while
              (and day
                   (or (when (or (eq dd day)
                                 (< day dd))
                         (setq hit (point))
                         nil
                         )
                       (not (eobp))
                       ))
            (forward-line 1)
            (if (setq day (bufmatget "^19..-..-\\(..\\)" 1))
                (setq day (string-to-int day)))

            ))

        (when hit                       ;found ?
          (goto-char hit)
          (forward-line -1)
          (delete-region (point-min) (point))
          (setq ret buffer)
          )
        )))
    ret
    ))



;;; ----------------------------------------------------------------------
;;;
(defun timi-remind-compose ()
  "Composes reminder from the message.
Return
  nil      ,not enough data
  str"
  (let* ((to        (ti::mt-get-field-1 "To"))
         (subject   (ti::mt-get-field-1 "Subject"))
         (date      (ti::d-standard-date 'short))
         )
    (when (and to subject date)
      (setq subject (ti::s-del-re "^re[0-9]*:" (ti::s-sqz subject)))
      (format "%s-|-%s-|-%s"
              (ti::s-sqz date)
              (ti::s-sqz to)
              subject
              ))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun timi-remind-add (&optional delete verb)
  "Save reminder data for current mesage.
Optional arg DELETE, deletes current message entry from reminder list.

The subject must be the same as you used fro sending th message, the
're:' word is stripped when trying to find the same subject. VERB."
  (interactive "P")
  (let* ((file      timi-:remind-file)
         (count     0)
         (entry     (timi-remind-compose))
         buffer
         )
  (setq verb (or verb (interactive-p)))


  (if (not (ti::mt-get-field-1 "fcc"))
      (error
       "Plese save this message to Fcc folder, otw you can't recall msg."))

  ;;  If there is no previous file, this will create file buffer
  ;;  from scratch
  ;;
  (setq buffer (find-file-noselect file))

  (cond
   ((null entry)
    (if verb
        (message "Timi: Please fill mail fields first. Not enough data.")))

   (t
    (job-buffer buffer
      (PMIN)

      (cond
       (delete
        (setq count (car-safe (timi-remind-check 'to 'kill)))
        (if verb (message "killed %d entries" (or count 0)))
        )

       (t
        (setq timi-:last-reminder-entry entry)
        (ti::b-cnv-empty-lines (point-min) (point-max))
        (if (re-search-forward (regexp-quote entry) nil t)
            (if verb (message "Timi: Reminder entry is already archived."))
          (PMAX)
          (insert entry "\n")
          (save-buffer)
          (if verb       (message entry))
          ))))
    ))

  ))


;;; ----------------------------------------------------------------------
;;;
(defun timi-remind-check (&optional mode kill verb)
  "Check how many reminders hits for this message.
If called interactively, then show all current reminders in separate buffer.

Input:
  MODE  If 'to or C\-u then process according to TO field contents.
        If 'subject or C\-u C\-u then look only subject
        If nil, then look at TO + SUBJECT fields.
  KILL	Delete entries.
  VERB	verbose messages.

Return:
  (count . list)    count = 1...xx
                    list  = (mathed-line mathed-line ...)
  nil"
  (interactive "P")
  (let* ((file      timi-:remind-file)
         (to        (ti::mt-get-field-1 "To"))
         (subject   (ti::mt-get-field-1 "Subject"))
         (obuffer   (current-buffer))
         (count     0)
         list
         entry
         buffer
         )
  (setq verb   (or verb (interactive-p)))
  (setq buffer (find-file-noselect file))

  (cond
   ((not (and to subject))
    (if verb
        (message "Timi: Please to/subject fields first. Not enough data.")))

   (t
    (setq subject (ti::s-del-re "^re[0-9]*:" (ti::s-sqz subject)))

    ;;  forget the time, subject is important
    ;;
    (cond
     ((memq mode '(to (4)))
      (setq entry (format "-|-%s-|-" (ti::s-sqz to))))

     ((memq mode '(subject (16)))
      (setq entry (format "-|-%s" subject )))

     (t
      (setq entry (format "-|-%s-|-%s" (ti::s-sqz to) subject))
      ))

    (job-buffer buffer
      (PMIN)
      (while (re-search-forward (regexp-quote entry) nil t)
        (beginning-of-line)
        (push (read-full-line)  list)
        (if kill (ti::b-kill-line))
        (end-of-line)
        (inc count)
        ))))

  (cond
   ((and verb kill)
    (if (null list)
        (message "No reminders killed.")
      (message "%d reminders killed." (length list)))
    )

   ((and verb (not kill))
    (if (null list)
        (message "No matched entries for '%s'" entry)
      (pop-to-buffer (use-tmp-buffer timi-:tmp-buffer 'clear))
      (list-loop list entry (insert entry "\n"))
      (select-window (get-buffer-window obuffer))
      )))

  (cons count list)
  ))


;;}}}
;;{{{ advice

;;; .......................................................... &advice ...

;;; ----------------------------------------------------------------------
;;; In gnus 5 you can say: (setq ssage-kill-buffer-on-exit 'yes)
;;;
(defadvice mail-send-and-exit (after timi act)
  "Kill mail buffer after send."
  (let* (buffer
         )
    ;;  There is bury buffer command... get that back
    (setq buffer (car (reverse (buffer-list))))

    (if (string-match "mail" (buffer-name buffer)) ;be cautious
        (kill-buffer buffer))
    ))


;;}}}
;;{{{ Extra

;;; ----------------------------------------------------------------------
;;;
(defun timi-save-dead-mail  ()
  "Loop mail buffers and save them to `timi-:dead-mail-file' on emacs exit."
  (ti::b-buffer-list
   '(memq major-mode
	  '(mail-mode
	    message-mode
	    ))
   'loop-temp-buffers
   nil
   '(progn
      (set-buffer-modified-p nil)	;"no changes in this buffer"
      (append-to-file (point-min) (point-max) timi-:dead-mail-file)
      )
   ))




;;}}}

;;{{{ code: Cc, X-info

;;; ............................................................. &fld ...


;;; ----------------------------------------------------------------------
;;;
(defun timi-fld-to-move-maybe ()
  "move cursor to the end of TO field if it is empty."
  (when (save-excursion (beginning-of-line) (looking-at "To: *$"))
    (end-of-line)
    (timi-debug "TO-move:" (read-full-line) (point))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun timi-fld-cc (arg)
  "Add appropriate Cc, if ARG match `timi-:table-cc'."
(let* ((table         timi-:table-cc)
         (exp-list      timi-:abbrevs-table-cc-expanded)
         info
         )
    (cond
     ((and
       (not (var-nil-p arg))
       (or (setq info (list-find table arg))
           (progn
             (when
                 (setq
                  info
                  (list-find
                   exp-list
                   arg
                   '(progn
                      (string= (cdr (car table)) arg))
                   ))
               (setq info (list-find table (car info)))
               ))
           ))
      (setq info (nth 1 info))  ;get second element
      (ti::mt-add-field "CC" info "To" nil 'replace)
      )
     (t
      (ti::mt-kill-field "^CC") ;no info, remove field
      ))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun timi-fld-x-info (&optional to-field)
  "Add info field to post according to TO-FIELD."
  (let* ((fid       "timi-fld-x-info: ")
	 (label     timi-:x-info-label)
         (table     timi-:table-x-info)
         info
         )

    (setq to-field  (or to-field (ti::mt-get-field-1 "To")))

    (cond
     ((and (not (var-nil-p to-field))
	   (setq info (list-find table (ti::s-sqz to-field))))
      (timi-debug fid info)
      (if (setq info (eval (nth 1 info)))
	  (ti::mt-add-field label info "Subject" nil 'replace))
      )
     (t
      (ti::mt-kill-field label)
      ))
    ))

;;}}}
;;{{{ code: Fcc handling

;;; ......................................................... &fld-fcc ...


;;; ----------------------------------------------------------------------
;;;
(defun timi-fld-fcc-determine (&optional type)
  "Look if default folder must be changed.
Tries to find RE given in `timi-:table-fcc' by looking at header area.

Input TYPE:
  nil	    Find Fcc folder
  'gcc	    Find Gcc folder.

Return:
   string       ,suggested folder
   nil"
  (let ((fid	"timi-fld-fcc-determine: ")
	(ptr	(if type
		    timi-:table-gcc
		  timi-:table-fcc
		  ))
	(hmax	(ti::mt-hmax))
        ret
        elt re folder
        )
    (when hmax				;No header end found
      (save-excursion
	(PMIN)
        (list-loop ptr elt
          (setq re (nth 0 elt)   folder (nth 1 elt))
          (if (re-search-forward re hmax t)
              (setq ret folder     ptr nil))
          )))
    ;;  Make sure the compression pacakge is loaded
    ;;
    (if (and (stringp ret)
	     (string-match "gz$\\|Z$" ret))
	(require 'jka-compr))

    (timi-debug fid "hmax" hmax "ret" ret)
    ret
    ))


;;; ----------------------------------------------------------------------
;;;
(defun timi-fld-fcc (&optional type)
  "Set right [GF]cc folder if there is match in `timi-:table-[gf]cc'.
Input TYPE:
  nil	    set Fcc folder
  'gcc	    set Gcc folder."
  (let* ((fid    "timi-fld-fcc: ")
	 (fld    (if type "GCC" "FCC"))
	 (fcc    (ti::mt-get-field-1 fld))
	 folder
	 )
    (timi-debug fid fld type fcc)
    (when (and fcc
               (not (string-match "^  +" fcc))
               (setq folder (timi-fld-fcc-determine type)))
        (ti::mt-kill-field (concat "^" fld) folder)
        )
    ))


;;; ----------------------------------------------------------------------
;;;
(defun timi-to-field-on ()
  "Keep TIMI activated by making sure the To: field has only one space."
  (timi-to-field-off 1))

;;; ----------------------------------------------------------------------
;;;
(defun timi-to-field-off (&optional count field)
  "Disable TIMI by space COUNT for FIELD.
2       ,spaces disables Cc tracking, but allows X-info tracking.
3       ,spaces disables both Cc and X-info tracking."
  (let* (str
         )
    (setq str (make-string (or count 2) ?\ ))
    (setq field (or field "to"))

    (save-excursion
      (PMIN)
      (when (re-search-forward
	     (concat "^" field ":\\([ \t]*\\)")
	     (ti::mt-hmax)
	     t)
	(if (match-beginning 1)
	    ;; There is spaces
	    (kill-match 1 str)
	  ;;  There is no spaces
	  (insert str))
	t
	))
    (timi-fld-to-move-maybe)
    ))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun timi-on-off-toggle  (&optional arg)
  "Toggle TIMI mode on and off by Changing spacing of To field.
This affects automatic Cc and X-info tracking. ARG behaves  like mode argument.

Without arg, this toggless Cc tracking, with prefix argument,
it toggless both Cc and X-info tracking."
  (interactive "P")
  (if arg (setq arg 3))                 ;3 spaces turn off completely.

  (if (null arg)
      (cond
       ((save-excursion
	  (PMIN)
	  (re-search-forward "^to:  " nil t)
	  )
	(timi-to-field-on)
	(message "tinymail: mail field tracking mode on.")
	)
       (t
	(timi-to-field-off arg)
	(message "tinymail: mail field tracking mode off.")
	))
    (timi-to-field-off arg)
    (message "tinymail: mail field tracking mode off.")
    ))

;;}}}

;;{{{ code: main


;;; ----------------------------------------------------------------------
;;;
(defun timi-post-command ()
  "Does some nifty things while user is typing message.
Expands mail aliases and inserts additional info."
  (let* ((fid           "timi-post-command: ")
	 (value         timi-:awake-time)
         (last-to       timi-:last-to-field)
         (alias-alist   timi-:mail-aliases-alist)
         (days-old      timi-:remind-days-old)
         to
         buffer
         )

    (if (not (integerp timi-:time-counter))
        (setq timi-:time-counter 0))


    (inc timi-:time-counter)

    ;;  In Emacs 19.30+ we get activated all the time, in lower,
    ;;  we save some process time.
    ;;
    (when (> timi-:time-counter value)  ;every Nth interval only
      (setq timi-:time-counter 0)       ;reset
      ;;  - If "To:" field content has two spaces at front, this is signal
      ;;    for us to stay away. This post-command is then disabled.
      ;;
      ;;  - If To address has remained the same, we do nothing.
      ;;    If we would aways go into expanding and killing the Cc,Fcc
      ;;    fields blindly, user would notice that while he was
      ;;    writing his message. Avoid that as much as posisble
      ;;
      (setq to (ti::mt-get-field-1 "To"))

      (timi-debug fid "TO" to last-to
		  "P-hook" timi-:post-hook
		  "Remind" timi-:reminder-printed
		  "awake"  value
		  )

      (cond
       ((or (not (stringp to))
	    (string-match "^   " to))    ;DISABLED
	nil
	)

       ((and (string-match "^  "  to)   ;Only X-info tracking
             (not (var-nil-p to))       ;Not just spaces ?
             (not (string= to last-to))
             )
        (timi-fld-x-info to))

       ((and (not (string= to last-to)) ;not same as previously ?
             (not (var-nil-p to)))
        (timi-fld-cc (ti::s-sqz to))    ;should we Cc ?
        (save-excursion
          (ti::xe-expand-mail-aliases
           (point-min) (ti::mt-hmax) alias-alist))
        (setq to (ti::mt-get-field "To")) ;what was expanded
        (timi-fld-x-info to)
        ))

      (timi-fld-fcc)			;handle possible Fcc
      (timi-fld-fcc 'gcc)		;handle possible Gcc

      (run-hooks 'timi-:post-hook)       ;User's things now

      ;;  print reminder only once
      ;;
      (unless timi-:reminder-printed
        (setq timi-:reminder-printed t)
        (timi-remind-info-brief 3)
        (when (setq buffer (timi-remind-older-than days-old))
          (pop-to-buffer buffer)
          (message-flash
            3 'sleep "You have old reminders. Would you check these.")
          ))
      (setq timi-:last-to-field to)     ;update
      )
    ))

;;}}}
;;{{{ examples

;;; ........................................................ &examples ...
;;; Rip code with tinylib.el/ ti::m-pkg-rip-magic

;;* ;;  Remember, first match is taken, so order here is important.
;;* _
;;* (defmacro my-ver (sym)
;;*   "Return version number for variable SYM if it exists."
;;*   (`
;;*    (strmatget
;;*     "[0-9][0-9.]+" 0
;;*     (if (boundp (, sym)) (symbol-value (, sym)) "")
;;*     )))
;;* _
;;* (defconst timi-:table-x-info
;;*   (list
;;* _
;;*    ;;  send 'tiny tool' library information when talking
;;*    ;;  with the maintainer.
;;*    ;;
;;*    (list
;;*     "jaalto\\|ssjaaa\\|jari.aalto"
;;*     '(format (concat "E%s lib%s m%s mt%s xe%s o%s ma%s pg%s \n"
;;* 		     "  mi%s my%s di%s dd%s tf%s li%s tit%s %s "
;;* 		     )
;;* 	     (my-ver 'emacs-version)
;;* _
;;* 	     ;;  Libraries
;;* 	     (my-ver 'tinylib-version)
;;* 	     (my-ver 'tinylibm-version)
;;* 	     (my-ver 'tinylibmt-version)
;;* 	     (my-ver 'tinylibxe-version)
;;* 	     (my-ver 'tinylibo-version)
;;* 	     (my-ver 'tima-:version-id)
;;* _
;;* 	     ;; programs
;;* 	     (my-ver 'tipgp-:version-id)
;;* 	     (my-ver 'timi-:version-id)
;;* _
;;* 	     (my-ver 'timy-:version-id)
;;* 	     (my-ver 'tdi-:version-id)
;;* 	     (my-ver 'tdd-:version-id)
;;* 	     (my-ver 'titf-:version-id)
;;* 	     (my-ver 'tili-:version-id)
;;* 	     (my-ver 'tit-:version-id)
;;* _
;;* 	     ;;  Tell also what mail agent user uses.
;;* 	     ;;  If you're not in this list, contact jari!
;;* 	     ;;
;;* 	     (cond
;;* 	      ((get-buffer "RMAIL") "R")
;;* 	      ((boundp 'vm-version) (concat "VM" vm-version))
;;* 	      ((boundp 'mh-e-version) (concat "MH" h-e-version))
;;* 	      ((boundp 'gnus-version) gnus-version)
;;* 	      (t "?")
;;* 	      ))
;;*     )
;;* _
;;*    ;;  Mail to company workers.
;;*    ;;
;;*    (list "ntc\\|nokia\\|tre\\|tnclus\\|[<][a-z]+[>]"
;;*      "P. +358-31-240 7668, DMTPER, old NNPER group, local server aapo.")
;;* _
;;*    ;;  Send my home address info to my country men
;;*    ;;
;;*    (list "\\.fi"
;;*      " Uudenkylankatu 7 E 41, 33530 Tre, p. tyo 931 240 7668.")
;;* _
;;*    ;;  When asking help from ange (the efs maintainer) send ange
;;*    ;;  version information automatically
;;*    ;;
;;*    (list
;;*     "ange"
;;*     '(progn (timi-mail-info-ange)))
;;* _
;;*    ;;  Send tm version information whenever I post to MIME
;;*    ;;  discussion list tm-en@chamonix.jaist.ac.jp
;;*    ;;  [get this package if you want GOOD mime handling in emacs]
;;*    ;;
;;*    ;;  to your ~/.mailrc
;;*    ;;  alias mime tm-en@chamonix.jaist.ac.jp
;;*    ;;
;;*    (list
;;*     "tm-en"
;;*     '(format "Emacs %s %s"
;;*          emacs-version
;;*          (if (boundp 'mime-editor/RCS-ID)
;;*              mime-editor/RCS-ID "")
;;*          ))
;;*    ))
;;* _



;;  And here is example for automatic Fcc
;;
;;  ;;  This is needed, otw the automatic Fcc doesn't get activated.
;;  (setq mail-archive-file-name "~/.RMAIL.out")
;;
;;* ;;  We want as short function name as possible so that it can be
;;* ;;  called easily.
;;* ;;
;;* (defun my-z  (file)
;;*   "Should compression be used for files? Add .gz if needed."
;;*   ;;
;;*   ;;  uta:  is a function that check if I'm in the
;;*   ;;  (u)niversity of (ta)mpere account. I have quota limit
;;*   ;;  there so I have to save space as much as possible
;;*   ;;
;;*   (if (or (null (uta))
;;* 	  ;;  already has compress extension?
;;* 	  ;;
;;* 	  (string-match "gz$\\|Z$" file))
;;*       file
;;*     (concat file ".gz")))
;;* _
;;* _
;;* (defconst timi-:table-fcc
;;*   (list
;;*    (list "elisp-archive"    (my-z " ~/.mail.elisp-post"))
;;*    (list "bug-gnu"          (my-z " ~/.mail.bug"))
;;*    (list "."		(my-z " ~/.RMAIL.out"))  ;; otw general out
;;*    ))

;; The "jack" and "anders" are in fact mail abbrevs that gets magically
;; expanded too.
;;
;;* _
;;* (defconst timi-:table-cc
;;*   (list
;;*    '("anders"       " jack")
;;*    '("jack" " anders")
;;*    '("mc1"  " mc2")
;;* _
;;*    (list
;;*     "t00"                           ; mailing list t00 ?
;;*     (concat
;;*      " "
;;*      (mapconcat                             ;add Cc alias members t01 .. t08
;;*       'concat                               ;they are automatically expanded
;;*       (let (l)                              ;by the mailabbrev.el
;;*     (loop-for  0 9
;;*       (push (format "t%02d" COUNTER) l))
;;*     l)
;;*       ",\n\t")))
;;*     ))


;;* _
;;* (add-hook 'timi-:post-hook 'my-timi-:post-hook)
;;* _
;;* _
;;* (defun my-timi-:post-hook  ()
;;*   "Add Reply-to to outgoing post."
;;*   (let* ((to  (or (ti::mt-get-field "to") ""))
;;*          )
;;*     ;;  Reply-to to isn't added when doing C-x m: add it to posts which
;;*     ;;  go outside of my company address to make sure person can reply
;;*     ;;  to valid address.
;;*     ;;
;;*     (if (or (string-match "ntc\\|nokia\\|tne" to)
;;*         ;;  internal address does not have "@" site reference.
;;*         (not (string-match "@" to)))
;;*     (ti::mt-kill-field "^reply-to:")
;;*       (if (var-nil-p (ti::mt-get-field "reply-to"))
;;*       (ti::mt-add-field "Reply-to" mail-default-reply-to)))
;;*     ))

;;; ................................................. &example-abbrevs ...
;;; Emacs has handy package called mailabbrevs. If you haven't used it
;;; yet, here is sample installation to get you started.
;;;
;;; This information may be Emacs specific (19.28)

;;* (add-hook 'mail-setup-hook 'my-mail-abbrevs)
;;* _
;;* _
;;* (defun my-mail-abbrevs ()
;;*   "Installs mailabbrev.el"
;;*   (interactive)
;;* _
;;*   (require 'mailabbrev)
;;*   (mail-abbrevs-setup)
;;* _
;;*   (define-key mail-mode-map [down] 'mail-abbrev-next-line)
;;*   (define-key mail-mode-map "\C-n" 'mail-abbrev-next-line)
;;*   (define-key mail-mode-map "\M->" 'mail-abbrev-end-of-buffer)
;;* _
;;*   (define-key mail-mode-map "\C-ca" 'mail-interactive-insert-alias)
;;*   )

;;; ................................................. &example-TinyTab ...
;;; There is TinyTab.el that is a tab minor mode, so turning it on
;;; in buffer would override the key underneath: in this case
;;; the timi-guess-complete-in-headers.
;;;
;;; To allow these two work together, so that you get completion in headers
;;; and leave the tab handling to the body, here is setup for
;;; tinytab.el

;;* (add-hook 'tit-mode-hooks           'my-tit-mode-hook)
;;* _
;;* (defun my-tit-mode-hook ()
;;*   "My Tinytab key additions, override settings."
;;* _
;;*   (if (featurep 'tinymail)
;;*       ;;  make TinyTab.el work with TinyMail so that they share
;;*       ;;  common TAB key.
;;*       ;;
;;*       (add-hook 'tit-:tab-insert-hook 'timi-guess-complete-in-headers))
;;* _
;;*   )

;;}}}

(timi-update-cc-abbrevs)
(provide   'tinymail)

(add-hook 'kill-emacs-hook 'timi-save-dead-mail)
(timi-install)
(run-hooks 'timi-:load-hook)

;;; tinymail.el ends here
