;; @(#) tinymatch.el -- Highlight matched text in a buffer easily

;; @(#) $Id: tinymatch.el,v 1.32 1997/06/19 10:06:00 jaalto NotReleased $
;; @(#) $Keywords: extensions $
;;
;; $KnownCompatibility: 19.28 $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1995,96,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Apr 5 1995
;;
;; To get information on this program use ident(1) or do M-x tima-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el

;; LCD Archive Entry:
;; tinymatch|Jari Aalto|jari.aalto@poboxes.com|
;; Highlight matched text in a buffer easily, re-search-fwd, re-search-bck ..|
;; 19-Jun-1997|$Revision: 1.32 $|~/misc/tinymatch.el.Z|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even tqhe implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ....................................................... &t-install ...
;; Put this file on your Emacs-Lisp load path, add following into your
;; ~/.emacs startup file. Rip code with with tinylib.el/ti::m-pkg-rip-magic
;;
;;
;;* (require 'tinymatch)
;;* _
;;
;; No autoload is suggested, because almost every function would have
;; needed to be in autoload state :-) It's easier to use require.
;; Suggested keybings for interactive use. Bind the mouse functions too
;;
;;* (global-unset-key "\C-z")
;;* (global-set-key "\C-ztm" 'tima-mark-region)   ;; eg. permanent 'mark'
;;* (global-set-key "\C-ztu" 'tima-unmark-region) ;; remove 'mark'
;;* (global-set-key "\C-ztc" 'tima-clear-buffer-properties)
;;* (global-set-key "\C-ztb" 'tima-buffer)
;;* (global-set-key "\C-ztU" 'tima-undo)
;;* _
;;
;; If you have any questions or feedback, use this function
;;
;;      M-x tima-submit-bug-report


;;}}}
;;{{{ Documentation

;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "tima-" in front of
;;   every function & variable. It stands for '(ti)ny (ma)tch
;; - variable names contain letters 'tima-:'

;;  Preface, apr 1995
;;
;;	  [This is primarily a library]
;;
;;	Now, when this highlightging is available in 19.xx, why don't we
;;	use it for marking texts too, especially when we are looking for
;;	documentation of some program and we want to find certain words, or
;;	that we want to see TABS in the buffer. This little .el eases the
;;	job of visualizing items in buffer.
;;
;;	Where I use this most is inside GNUS/DING/RMAIL hooks, to mark
;;	interesting peoples articles and seeing where is my articles.  The
;;	font-lock/lazy-lock are a little bit too heavy for those things,
;;	because these buffers hold "solid" data, text, that does not
;;	change, and I feel that executing arbitrary 'tima-re-search'
;;	command to highligh the headers I want is more easier that with
;;	font-lock, especially when I can change the faces in the fly with
;;	'tima-re-search'.
;;
;;	Yes, I'm aware of the capabilities of DING, which highlights
;;	articles for you already, but sometimes I want more personal
;;	highlighting in summary buffer etc...
;;
;;	But the font-lock/lazy-lock are unreplacable in programming modes,
;;	which has to deal with dynamic text.
;;
;;  Overview of features
;;
;;	o   This package is primary aimed for elisp programmers, but
;;	    interactive users will also find handy functions.
;;	o   Shows matched text with color in the buffer.
;;	o   This is *NOTHING* like font-lock, lazy-lock or hilit19,
;;	    which are demand driven packages intended for certain major modes.
;;	    Use this package to "manually" mark interesting things in
;;	    any buffer.
;;	o   Examples: highlighting on/off tabs, Verifying PGP
;;	    fingerprints against trusted list like http://www.uit.no/
;;	o   UNDO: adjustable stack size. Stack is cleared if
;;	    stack limit reached (stack 'wraps')
;;
;;  User functions
;;
;;	Mostly this package is designed for lisp programmers, who add
;;	some highlight commands in hooks. For quick text highlighting,
;;	you can use these interactive functions:
;;
;;	    tima-re-search-forward
;;	    tima-re-search-backward
;;	    tima-undo
;;
;;  Setting different face (programming)
;;
;;	If you want permanetly change the face, when marking text
;;	use commands
;;
;;	    tima-search-face-set   ;to set
;;	    tima-search-face-reset ;to get default color back
;;
;;	If you want temporarily use some face, supply direct FACE parameter
;;	when you call search functions, like:
;;
;;	    tima-re-search-forward (re &optional level face)
;;
;;  Note
;;
;;	This is for simple text highlighting only. Like finding certain items
;;	or marking something quickly and temporarily (great for text files)
;;
;;	You can mix font-lock/hilit19 and TIMA package, but remember that
;;	these packages have different goals. Use TIMA only for finding
;;	things in buffer, or marking certain articles in gnus...
;;
;;	Be carefull: if you use `tima-clear-buffer-properties', you will
;;	wipe out all text properties.


;;}}}
;;{{{ history

;; ........................................................ &t-history ...
;;; Change Log:
;;
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; Jun	19	1997	[jari]		19.28	v1.29-1.30	NotReleased
;; - Started using new buffer modified FORM from 'm' lib. Text changes.
;;
;; Jun	5	1997	[jari]		19.28	v1.27-1.28	NotReleased
;; - Shen Jinwei <shenjw@wam.umd.edu>reported that the undo feature didn't
;;   work. Corrected.
;; - Changed implementation of all match functions: now the undo infomation
;;   must be REQUESTED by passing non-nil 'save-undo' parameter to functions.
;;   This is t if user calls functions interactively, but if you call
;;   from program, the undo information is not automatically saved.
;; - Corrected couple of stack functions, made defsubst for greated speed.
;; - renamed stack variables. Added view-search advice example to the end.
;;
;; Jun	5	1997	[jari]		19.28	v1.27		NotReleased
;; - Rick Flower <flower@ms31.sp.trw.com> reported that buffer modified flag
;;   wasn't preserved. Thank you, that was clear bug in the function which
;;   unconditionally set the flag to nil after every face change.
;; - Now the original state is preserved.
;; - Checkdoc 1.42 clear.
;;
;; May	14	1997	[jari]		19.28	v1.25-1.26	NotReleased
;; - internal changes.
;; - Changed all non standard boolean tests to use when and unless
;;
;; Mar	26	1997	[jari]		19.28	v1.23-1.24	NotReleased
;;   tima-property-search-and-modify	:+
;;
;; Mar	20	1997	[jari]		19.28	v1.23-1.24	NotReleased
;; - Variable and documentation changes. Range check error corrected
;;   in tima-match-level. Added parameters to tima-match-level.
;;
;; Mar	1	1997	[jari]		19.28	v1.22		NotReleased
;; - corrected serious bug in the tima-re-search function. The mached
;;   level count was wrong, used now < when it had <=.
;; - rear sticky properties are now fixed to to last character.
;;   Should work now ok.
;;
;; Feb	27	1997	[jari]		19.28	v1.19-1.21	NotReleased
;; - corrected tima-clear-region-properties. NEW tima-match-level
;; - rear-nonsticky used, but there is some problem: it doesn't
;;   work right in tima-re-search
;; - Rewrote tima-looking-at.
;;
;; Feb	26	1997	[jari]		19.28	v1.18		NotReleased
;; - Rewrote tima-clear-buffer-properties to accept PROPL list argument.
;;   Only text that matches PROPL will be cleared. This allows using
;;   'owner property efectively.
;; - New function tima-get-mouse-property
;;
;; Jan	21	1997	[jari]		19.28	v1.16-1.17	NotReleased
;; - Text changes. if FACE is list, then function tima-re-search
;;   assigns face property list to mathed text. This way you can
;;   put properties '(owner package-xxx face highlight).
;; - Added undo history. Very easy to use.
;;
;; Sep	25	1996    [jari]		19.28	v1.14-15	NotReleased
;; - Fixed all XE reported byteCompiler errors.
;;
;; Jun	8	1996    [jari]		19.28	v1.13	        NotReleased
;; - Notices that the tima-re-search main function didn't return the
;;   success of search at all. This made it impossible to use it
;;   in lisp programs. Corrected. Now return match point.
;; - Added PGP example. Dynamic byteComp support.
;;
;; Mar	28	1996    [jari]		19.28	v1.12	        NotReleased
;; - Corrected example code my-tabs-hilit
;;
;; Feb	13	1996    [jari]		19.28	v1.10-1.11	NotReleased
;; - Tom Fontaine <fontaine@esd.ray.com> sent valuable comments about
;;   the interactive functions. They should use regexp-history.
;; - Now uses Tom's code for reading and search regexp and puts
;;   it in regexp-history.
;; - Tom also suggested that interactive functions should use
;;   save-excursions. Now they use save excursion.
;; - Rewrote the installation section.
;;
;; Feb	4	1996    [jari]		19.28	v1.9		NotReleased
;; - Corrected file so that lisp-mnt.el's lm-verify returns "ok".
;; - Changed version functions to call tinylib, which has ready
;;   feedback/bug report functions.
;; - Made few minor rearrangements to functions.
;;
;; Nov	8	1995    [jari]		19.28	v1.8		NotReleased
;; - Tim Rueger <rueger@areaplg2.corp.mot.com> reported that the buffer
;;   was marked as modified, when region had been marked. Corrected.
;;   tima-unmark-region		:+ for kbd users
;;   tima-mark-region		:! corrected
;;
;; Nov	8	1995    [jari]		19.28	v1.7		Release_4
;; - Many times I had in mind that undo capability of last acion, now
;;   it's there.
;;   tinymatch-load-hook		-> tima-:load-hook
;;   tima-:last-xxx			:+ new variables to save last data
;;   tima-save-data			:+ store search data
;;   tima-undo				:+ undo last highlight
;;   tima-re-search			:! undo support
;;   tima-looking-at			:! undo support
;;   my-gnus-summary-hilit		:! example updated
;;   my-buffer-menu-hook-func		:! example updated
;;
;; Sep	2521	1995    [jari]		19.28	v1.6		NotReleased
;; - I needed mouse funcs that would mark region of important text easily...
;; - updated installation for new funcs
;;   tima-re-search			:! new parameter MODE
;;   tima-re-search-forward		:! now accepts prefix arg for LEVEL
;;   tima-re-search-backward		:! now accepts prefix arg for LEVEL
;;   tima-mouse-mark-region		:+ new
;;   tima-mouse-unmark-region		:+
;;   tima-mark-region			:+
;;   tinymatch-load-hook		:+
;;
;; Aug	21	1995    [jari]		19.28	v1.5		Release_3
;; - Only small fine tunes to examples, function descriptions adjusted
;;
;; Aug	15	1995    [jari]		19.28	v1.4		Release_2
;; - Added optional 'face' parameter to every function, so that the
;;   tima-search-face-set isn't needed for temporary face changes.
;; - tima-re-search :  Corrected the main function a little.
;; - Added more examples : dired, new-buffer-menu.
;;
;; Aug	10	1995    [jari]		19.28	v1.3		Release_1
;; - Just some cleanup.
;;
;; Apr	5	1995    [jari]		19.28	v1.2		NotReleased
;; - So that we can use several colors, I added restoring macro.
;; - Ready example for using these functions in GNUS/DING is provided
;;   at the end. See also other examples.
;; - Added parameter maxp to tima-re-search to limit the search
;;
;; Apr	5	1995    [jari]		19.28	v1.1		NotReleased
;; - Just what I needed for now

;;}}}

;;; Code:

;;{{{ setup: require

;;; ......................................................... &require ...

(require 'tinylibm)
(eval-when-compile (use-dynamic-compilation))

;;}}}
;;{{{ setup: variables

;;; ......................................................... &v-hooks ...

(defvar timi-:load-hook nil
  "*Hook run when file has been loaded.")

(defvar tima-:stack-size 1000
  "*How big undo history.")

;;; ....................................................... &v-private ...

(defvar tima-:stack-count nil
  "Incremented after every search. Do not touch.")

(defvar tima-:stack-push-flag  nil
  "Non-nil  means ('undo-func) do not record match data to stack.
If this variable has value 'undo-func then the next calls to
`tima-re-search' won't record data to stack.")


(defvar tima-:stack nil
    "Private. Contain last search data.
This is actually property list stack so that undo can be done.

Format:
  '(start-point
    last-func
    last-re
    las-beg-point
    last-level
    last-mode
    )")

;;; ........................................................ &v-public ...
;;; user configurable

(defvar tima-:face-search-default 'highlight
  "*Default face used when marking searched text.")

;;; For now, only search face is used, but maybe in the future the others..
;;;
(defvar tima-:face-table
  (list
   (cons 'search 'highlight)
   (cons 'warn   (if (emacs-p) 'region 'bold)) ;XE doesn't have 'region face
   (cons 'head    'bold)
   )
  "*Faces used for marking text.")

;;}}}
;;{{{ version

;;; ....................................................... &v-version ...

(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinymatch.el"
 "tima"
 tima-:version-id
 "$Id: tinymatch.el,v 1.32 1997/06/19 10:06:00 jaalto NotReleased $"
 '(tima-:version-id
  timi-:load-hook
  tima-:stack-size
  tima-:stack-count
  tima-:stack-push-flag
  tima-:stack
  tima-:face-search-default
  tima-:face-table
  )))

;;}}}

;;; ########################################################### &funcs ###

;;{{{ code: misc funcs

;;; ----------------------------------------------------------------------
;;;
(defmacro tima-search-face-reset ()
  "Reset used face to the default value.
If you use many colours to highlight text. Remember to call this
when you're finished."
  (list
   'setcdr (list 'assq ''search 'tima-:face-table)
  'tima-:face-search-default))


;;; ----------------------------------------------------------------------
;;;
(defmacro tima-search-face-set (face)
  "Change search colour to FACE."
  (list 'setcdr (list 'assq ''search 'tima-:face-table) face))

;;; ----------------------------------------------------------------------
;;;
(defmacro tima-face (face)
  "Return real face when logical FACE is given."
  ;;  This way the global variable does not float around the file
  (list 'cdr (list 'assoc face 'tima-:face-table)))



;;; ----------------------------------------------------------------------
;;;
(defsubst tima-stack-clear  ()
  "Clear undo stack."
  (put 'tima-:stack 'definition-stack nil)
  (setq tima-:stack-count  0
        tima-:stack        nil
	))

;;; ----------------------------------------------------------------------
;;;
(defsubst tima-stack-length  ()
  "Return undo stack length."
  (length (get 'tima-:stack 'definition-stack)))

;;; ----------------------------------------------------------------------
;;;
(defsubst tima-stack-full-p  ()
  "Check if stack is full."
  (eq (tima-stack-length) (1+ tima-:stack-size)))

;;; ----------------------------------------------------------------------
;;;
(defsubst tima-stack-p  ()
  "Check if there is data in undo stack. nil means that stack is empty."
  (or (get 'tima-:stack 'definition-stack)
      ;;  Make sure this is also zero because there is no data
      ;;
      (progn (setq tima-:stack-count 0)
	     nil
	     )))



;;; ----------------------------------------------------------------------
;;;
(defun tima-save-data (re level func mode beg)
  "Save search values RE LEVEL FUNC MODE BEG for undo.
If the stack is full, then Clear the stack before pushing to it."
  (or (integerp tima-:stack-count)
      (setq tima-:stack-count (tima-stack-length) ))

  (when (and (stringp re)
  	     (not (eq tima-:stack-push-flag 'undo-func))
	     )
    ;; The last entry in the stack is always nil, that's why
    ;; 1+.
    ;;
    ;; inital:	    nil
    ;; 1st:	    '(mil)		pushed last data; size 1
    ;; 2nd	    '((..) nil)		pushed next, size 2
    ;;
    ;; As you can see there actually is only one real data;
    ;; the 1st push reads the current calues of tima-:stack
    ;; which is nil; because it was the last element that was poped
    ;;
    ;;
    (when (>= tima-:stack-count
	      (1+ tima-:stack-size))
      (tima-stack-clear)
      (setq tima-:stack-count 0))

    (push-definition 'tima-:stack)

    (setq tima-:stack
	  (list
	   func
	   re
	   beg
	   level
	   mode
	   ))
    (inc tima-:stack-count)
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tima-undo ()
  "Undo last highlighting.
`tima-:stack-push-flag' is set to 'undo-func while this function runs."
  (interactive)
  (let* ((tima-:stack-push-flag  'undo-func)
	 (prev   tima-:stack)
	 func
	 beg
	 re
	 level
	 mode
	 )
    ;; - Hm The stack values are empty now, try to pop last saved values
    ;;   from stack.
    ;; - Actually there should be something in the variable is the stack is
    ;;   not empty and it is an erro condition is variable IS empty AND
    ;;   there is values in the stack! ... We'll we don't nag about that
    ;;   here. I just thought you should know about it.
    ;;
    (unless  prev
      (dec tima-:stack-count)
      (pop-definition 'tima-:stack)
      (setq prev tima-:stack)		;Maybe this time there is something
      )

    (if (not (and (listpp prev) (nth 0 prev)))
	(progn
	  (tima-stack-clear)
	  (error "tinymatch: No more undo information in stack."))
      (setq func    (nth 0 prev)
	    re	    (nth 1 prev)
	    beg	    (nth 2 prev)
	    level   (nth 3 prev)
	    mode    (nth 4 prev)
	    )
      (dec tima-:stack-count)

      (save-excursion
	;;  - What if user has narrowed the buffer
	;;  - Or he has deleted text
	;;
	(or (safe-do (goto-char beg))
	    (error "\
There is no such search point in the buffer any more? %s" beg))

	(cond
	 ((eq func 'looking-at)
	  (tima-looking-at re level 'default))
	 (t
	  (tima-re-search
	   re
	   (if (eq func 're-search-backward)
	       'back nil)
	   level
	   nil
	   'default
	   mode)
	  )))
      ;;  UNDO done; now get next undo information
      ;;
      (pop-definition 'tima-:stack)
      )
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tima-clear-buffer-properties (&optional  propl)
  "Remove all properties from buffer that match property list PROPL.

Input:
  See function `tima-clear-region-properties'"
  (interactive)
  (let* ()
    (tima-clear-region-properties (point-min) (point-max) propl)

    (cond
     ((interactive-p)
     (redraw-display)
     (message "Properties cleared")))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tima-clear-region-properties (beg end &optional propl)
  "Remove properties from BEG END. Affects read only buffers too.

Input:

  PROPL  (PROP VAL PROP VAL ..) Property list to match.
         If nil, then remove all properties."
  (interactive "r")
  (let* (buffer-read-only		;allow writing
	 point
	 prop
	 val
	 list
	 read-list
	 elt
	 rprop				;read property name
	 rval

	 delete				;flag
	 )
    (save-buffer-modified
      (if (null propl)
	  (set-text-properties beg end nil)
;;;	text-properties-at text-property-any
	(keep-lower-order beg end)
	(setq prop (nth 0 propl)  val (nth 1 propl))


	(while (and (> (point-max) beg)
		    ;;  The beg is advanced in loop
		    ;;
		    (<= beg end)
		    (setq point (text-property-any beg end prop val))
		    )

	  (setq read-list (text-properties-at point)
		list      propl
		delete    t
		)

	  (while list
	    (setq rprop (nth 0 list)
		  rval  (nth 1 list)
		  list  (cdr (cdr list))) ;go 2 forward

;;;	  (d! rval list propl)

	    ;;  The memq return the rest of list
	    ;;
	    ;;	    '(owner me face highlight  do-it nil)
	    ;;	    (memq 'face)  --> '(face highlight  do-it nil)
	    ;;
	    ;;  So the (nth 1) gives the value 'highlight which we
	    ;;  test.
	    ;;
	    (if (or (null (setq elt (memq rprop read-list)))
		    (not (eq (nth 1 elt) rval)))
		;;  This doesn'tmatch, stop and cancel delete
		;;
		(setq list nil  delete nil))
	    )

	  ;;  Character by character, this is bit slow...
	  ;;
	  (when delete
;;;	  (goto-char point)  (d! point read-list)
	    (set-text-properties point (1+ point) nil))

	  ;;  Search again
	  (setq beg (1+ point))
	  )

	))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tima-get-mouse-property  ()
  "Check if the point has 'mouse-face text property.
notice that if value read from point is nil,
that means same as no mouse-face property exists.

Return:
  nil          ,no property at point found
  prop         ,mouse-face property value"
  (let* ((prop  (text-properties-at (point)))
         )
    (if (setq prop (memq 'mouse-face prop))
	(cdr prop))			;return value, may be nil
    ))



;;; ----------------------------------------------------------------------
;;;
(defun tima-match-level  (&optional level face-or-propl beg end)
  "Add to match LEVEL a FACE-OR-PROPL in region BEG END.
If no match in that level, do nothing. Property rear-nonsticky is
added to the end of match unless FACE-OR-PROPL contains it.

Input:

  LEVEL		    Defaults to 0
  FACE-OR-PROPL	    Defaults to '(face highlight)
		    If symbol, must be face symbol.
		    Can also be property list '(PROP VAL PROP VAL ..))

  BEG END	    If given, then these are the are matched."
  (let ((add-flag   t)
	)
    (setq beg (or beg (match-beginning (or level 0)))
	  end (or end (match-end       (or level 0))))

    (when (and (and beg end)
	       (not (eq beg end)))	;Nothing to do
      (cond
       ((null face-or-propl)
	(add-text-properties beg end '(face highlight)))

       ((symbolp face-or-propl)
	(add-text-properties beg end (list 'face face-or-propl)))
       (t
	(setq add-flag (null (memq 'rear-nonsticky face-or-propl)))
	(add-text-properties beg end face-or-propl)
	))
      (when add-flag
	(if (eq end 1) (setq end 2))	;(1- 1) = 0, invalid charpos
	(add-text-properties (1- end) end '(rear-nonsticky t)))
      )))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tima-re-search (re &optional direction level maxp face mode save-undo)
  "Highlight found text with search face.

Note:

    The beginning of match and end of match will have
    property 'rear-nonsticky t, so that adding characters before of after
    text, won't inherit the face.

Input:

  RE		str  regexp
  DIRECTION	bool non-nil means backward
  LEVEL		nbr  which subexpression in re to highlight, default is 0
  MAXP		nbr  last search point [default until bob/eob]

  FACE		sym  face symbol
		     if symbol is 'null then set face to value nil
		     or if this is list; then it must be properly list
		     of format '(PROP PROP-VAL  PROP PROP-VAL ..)

  MODE		nbr  signify that func should highlight all matches
                     that occur within LEVEL..NBR
                     if you have lot's of xx(match)yy|zz(match)tt|
                     the subexpression are counted from left to to
		     right: 1,2 ...
  SAVE-UNDO	flag non-nil means that the highlight information is
		     recorded for undo. This flag in on if function is
		     called interactively. NOTE: ther undo information is
		     saved only if somthing was matched.

Return:

 nil		No match
 nbr		start of match at LEVEL."
  (let* ((func		(if direction 're-search-backward 're-search-forward))
	 (start-point	(point))
	 buffer-read-only
	 max-level
	 count
	 bp ep				;beg/end points
	 ret
	 prop val
	 list
	 )
    (save-buffer-modified
      (if (interactive-p) (setq save-undo t))

      ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. set defaults . .
      (or level (setq level 0))
      (or maxp  (setq maxp
		      (if direction
			  (point-min)
			(point-max))))

      (cond
       ((equal 'null face)    (setq face nil))
       ((null face)	    (setq face (tima-face 'search)))
       ;;  Otherwise face is non-nil
       )

      (setq max-level (1+ (or mode level)))

      (if (and (listpp face)
	       (not (eq 0 (% (length face) 2))))
	  (error "face properties are not paired: prop val"))


      ;; .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. .. ..  do matching . .
      (while (funcall func re nil maxp)
	(setq count level)

	(while (< count max-level)
	  (setq  bp (match-beginning count)  ep (match-end count))

	  (if (and bp (null ret))	;do only once
	      (setq ret bp))

;;;	(d! (current-buffer) level (read-match count) (read-match 0) re)
	  (when (and bp (not (eq bp ep))) ;empty string
	    (cond
	     ((or (symbolp face) (null face))
	      (put-text-property bp ep 'face face))

	     ((listp face)
	      (setq list face)
	      (while list
		;;  Read two values at time
		;;
		(setq prop (car list) list (cdr list)
		      val  (car list) list (cdr list))
		(put-text-property bp ep prop val)
		)))

	    ;;  #todo: something is wrong here, investigate..
	    ;;
	    ;;  If a character's `rear-nonsticky'
	    ;;  property is `t', then none of its properties are rear-sticky.
	    ;;
	    ;;  Hmm, doesn't affect 19.28; is there bug in this emacs?
	    ;;  The highlight is still extended If one adds chars after
	    ;;  the matched text.
	    ;;
	    (if (eq bp (1- ep))
		(add-text-properties (1- bp) ep '(rear-nonsticky t))
	      (add-text-properties (1- ep) ep '(rear-nonsticky t)))
	    )

	  (setq count (1+ count))
	  ))
      ;;   Saving the search values for possible undo.
      (if (and save-undo ret)
	  (tima-save-data re level func mode start-point))


      ;; Return the succesfull status
      ret
      )))




;;; ----------------------------------------------------------------------
;;;
(defun tima-property-search-and-modify
  (match-plist set-plist &optional beg end)
  "Search all characters forward, whose text properties match MATCH-PLIST.
Set properties to SET-PLIST. The point moves along the search.

Input:

 MATCH-PLIST	,property list '(prop val prop val ..)
 SET-PLIST	,property list '(prop val prop val ..)
 BEG		,start point of search; defaults to `point-min'
 END		,end point of search; defaults to `point-max'"
  (let* ((sprop (car match-plist))	;serach property
	 (sval  (car (cdr match-plist)))
	 point
	 plist mlist
	 elt
	 ok
         )
    (if (null match-plist) (error "Invalid match-plist"))

    (setq beg (or beg (point-min)))
    (setq end (or end (point-max)))
    (keep-lower-order beg end)

    (goto-char beg)  (setq point (point))

    (while (and (not (eobp))
		(< (point) end)
		(setq point (text-property-any (point) end sprop sval))
		)
      (setq plist (text-properties-at (point))
	    mlist match-plist
	    ok t)

      (while (and ok mlist)
	;; Select 1str PROP fro match-list and see if it is in read PLIST
	;; Continue until all MLIST properties are found from read PLIST
	;;
	(setq elt (memq (car mlist) plist)
	      ;;   first non-match terminates loop immediately
	      ;;
	      ok    (and elt (eq (nth 1 elt) (nth 1 mlist)))
	      mlist (cdr (cdr mlist))
	      )
	)
      (if ok (set-text-properties (point) (1+ (point)) set-plist))
      (goto-char (1+ point))		;Continue search
      )))


;;}}}
;;{{{ code: interactive

;;; ----------------------------------------------------------------------
;;; Mon, 12 Feb 1996,  Tom Fontaine <fontaine@esd.ray.com>
;;; Sent this piece of code.  Thanks Tom!
;;;
(defun tima-read-regexp ()
  "Read regexp using `regexp-history'."
    (let*
        ((default (car regexp-history))
         (input
          (read-from-minibuffer
           (if default
               (format "Highlight matching regexp (default `%s'): " default)
             "Highlight matching regexp: ")
           nil nil nil
           'regexp-history))
	 )
      (if (> (length input) 0)		;the return value
	  input
        (setcar regexp-history default)
	)))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tima-looking-at (re &optional level face-or-plist )
  "Highlight found RE at LEVEL with FACE-OR-PLIST.
The LEVEL is subexpression to highlight. PLIST means property list."
  (interactive "slook at: ")
  (let (buffer-read-only		;allow writing
	)
    (save-buffer-modified
      (setq level		(or level 0)
	    face-or-plist (or face-or-plist (tima-face 'search)))

      (when (and (looking-at re)
		 (match-end level))
	(tima-save-data re level 'looking-at nil (point))
	(tima-match-level level face-or-plist)
	)
      )))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tima-buffer (re &optional level face mode save-undo)
  "Highlight RE and sub LEVEL in whole buffer, starting from `point-min'.
Preserve point.

See `tima-re-search' for descriptions of FACE and MODE."
  (interactive (list (tima-read-regexp)  current-prefix-arg))
  (save-excursion
    (if (interactive-p) (setq save-undo t))
    (goto-char (point-min))
    (tima-re-search re nil level nil face mode save-undo)))


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tima-re-search-forward (re &optional level face mode save-undo)
  "Search RE and highlight forward until `point-max'.
Optional prefix arg tells which subexpression LEVEL to match that
function should highlight. point is preserved during call.

See `tima-re-search' for descriptions of FACE and MODE."
  (interactive (list (tima-read-regexp)  current-prefix-arg))
  (save-excursion
    (if (interactive-p) (setq save-undo t))
    (tima-re-search re nil level nil face mode save-undo)))



;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tima-re-search-backward (re &optional level face mode save-undo)
  "Search RE and highlight backward until `point-min'.
Optional prefix arg tells which subexpression LEVEL to match that
function should highlight. point is preserved during call.

See `tima-re-search' for descriptions of FACE and MODE."
  (interactive (list (tima-read-regexp)  current-prefix-arg))
  (save-excursion
    (if (interactive-p) (setq save-undo t))
    (tima-re-search re 'back level nil face mode save-undo)))


;;; ----------------------------------------------------------------------
;;; - These are handy when you want to "mark" ceratin texts for quick ref.
;;;
;;;###autoload
(defun tima-mouse-mark-region (beg end event)
  "Highlight region BEG END. EVENT is mouse event."
  (interactive "r\ne")
  (tima-mark-region beg end)
  )


;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tima-mouse-unmark-region (beg end event)
  "Remove highlight from region BEG END. EVENT is mouse event."
  (interactive "r\ne")
  (tima-mark-region beg end 'remove))


;;; ----------------------------------------------------------------------
;;; - This is for keyboard users
;;;
;;;###autoload
(defun tima-unmark-region (beg end)
  "Remove highlight from region BEG END."
  (interactive "r")
  (tima-mark-region beg end 'remove))

;;; ----------------------------------------------------------------------
;;;
;;;###autoload
(defun tima-mark-region (beg end &optional remove face)
  "Highlight BEG END. With optional prefix arg REMOVE all matching FACE."
  (interactive "r\nP")
  (let* (buffer-read-only		;set this to nil
	 (face	    (if remove
			'default
		      (or face (tima-face 'search))))
	 )
    (save-buffer-modified
      (put-text-property beg end 'face face)
      )))

;;}}}
;;{{{ examples

;;; ############################################### &example-gnus/ding ###
;;; - This is brief example that you can use staight away in GNUS and in
;;;   any version of DING.
;;;
;;; - The examples can be automatically extracted with function
;;;   ti::m-pkg-rip-magic from tinylib.el, follow the instructions
;;;   there.
;;;
;;;   PMIN is a macro from tinylibm.el to go point-min

;;* ;;  Purposively use SETQ, since I want to have only one main func
;;* (setq gnus-summary-prepare-hook 'my-gnus-summary-prepare-hook)
;;* (setq gnus-select-article-hook  'my-gnus-article-prepare-hook)

;;* ;;; ----------------------------------------------------------------------
;;* ;;;
;;* (defun my-gnus-summary-prepare-hook ()
;;*   "Things to do when entering group."
;;*   (my-gnus-summary-hilit)
;;*   nil
;;*   )

;;* ;;; ----------------------------------------------------------------------
;;* ;;;
;;* (defun my-gnus-article-prepare-hook ()
;;*   "Activates when article is displayed."
;;*   (let* ((n  (buffer-name (current-buffer)))
;;* 	 (art gnus-article-buffer)
;;* 	 )
;;*     (save-excursion
;;*       (set-buffer art)
;;*       (my-gnus-article-hilit)
;;*       (toggle-read-only 1)
;;*       nil
;;*       )))


;;* _
;;* ;;; ----------------------------------------------------------------------
;;* ;;;
;;* (defun my-gnus-summary-hilit ()
;;*   "Hilights some thigs from the Summary buffer"
;;*   (interactive)
;;*   (let* (
;;* 	 (grp gnus-newsgroup-name)
;;* 	 (aid (if (boundp 'my-anon-account) my-anon-account nil))
;;* 	 (an  (if (boundp 'my-anon-nick-name) my-anon-nick-name nil))
;;* 	 (re
;;* 	  (concat
;;* 	   "\\("			;start the regexp
;;* 	   ;; The perl group Gurus!
;;*            "tchrist\\|lwall\\|merlyn\\|ian@pipex\\|Tim.Bu"
;;* _
;;* 	   ;; Games group
;;* 	   "\\|buzz"			;thrustmaster joysticks
;;* 	   "\\|ratava"			;elite WWW page
;;* 	   "\\|microprose"		;any new games ?
;;* 	   "\\)[^\] ]+" ))		;end the regexp
;;* _
;;* 	 (re-my				;my account names
;;* 	  (concat
;;* 	   "\\("
;;* 	   "ssjaaa\\|jaalto\\|jari.aal"
;;* 	   (and aid (concat "\\|" aid))
;;* 	   (and an  (concat "\\|" an))
;;* 	   "\\)[^\] ]+" ))
;;* _
;;* 	 (re-games1
;;* 	  (concat
;;* 	   "C[&]C\\|C +[&] +C"
;;* 	   "\\|\\].*Com.*conq\\(uer\\)?"
;;* 	   "\\|Silent hunter"
;;* 	   "\\|pirates\\|quake\\|Apache"
;;* 	   "\\|\\(Wing Commander\\|WC\\).*\\(III\\|3\\|4\\|IV\\)"
;;* ;;	   "\\|"
;;* _
;;* 	   ))
;;* _
;;* _
;;* 	 (re-hot
;;* 	  (concat
;;* 	   "c[ae][ae]sar\\|jagged.*"
;;* 	   "\\|Pax Imperia"
;;* 	   "\\|moo\\|master.*orion"
;;* 	   "\\|us[mn]f\\|privateer"
;;* 	   "\\|Fleet Defender"
;;* 	   "\\|Master Of Antares\\|Moa"
;;* 	   "\\|Silent Hunter"
;;* 	   ))
;;* _
;;* 	 )
;;* _
;;*     (save-excursion
;;*       (tima-clear-buffer-properties)	;allows successive calls
;;*       (PMIN) (tima-re-search-forward re)
;;*       (PMIN) (tima-re-search-forward re-my 0     'region)
;;*       (PMIN) (tima-re-search-forward re-games1 0 'my-face-yellow)
;;*       (PMIN) (tima-re-search-forward re-hot 0    'region)
;;*       )))
;;* _


;;* ;;; ----------------------------------------------------------------------
;;* ;;;
;;* (defun  my-gnus-article-hilit ()
;;*   "Article hilit."
;;*   ;; Same color is enough this time
;;*   (save-excursion
;;*     (PMIN)
;;*     ;; I'm interested in DOMAIN name only, what's the country ?
;;*     (tima-re-search-forward "From: .*@\\([^ \t\n>]+\\)" 1)
;;*     (PMIN) (tima-re-search-forward  "Subject: \\(.*\\)" 1))
;;*   )

;;; #################################################### &example-tabs ###
;;; - You want to see where the TABS are? This toggles TAB viewing.

;;* ;;  (global-set-key "\C-ct" 'my-tabs-hilit)
;;* ;;

;;* (defun my-tabs-hilit (&optional arg)
;;*   "Toggless hilit/dehiliting tabs in buffer.
;;* If ARG is integer, force highlighting. If ARG is C-u, then
;;* force dehihiliting."
;;*   (interactive "P")
;;*   (let* (prop
;;* 	 )
;;*     (save-excursion
;;*       (goto-char (point-min))
;;*       (if (null (re-search-forward "\t" nil t))
;;* 	  nil                           ;no tabs in buffer
;;* 	;; is the tab marked?
;;* 	(setq prop (get-text-property (1- (point)) 'face))
;;* _
;;* 	(cond
;;* 	 ((or (integerp arg)
;;* 	      (or (eq prop nil)
;;* 		  (eq prop 'default)) )
;;* 	  ;; Do hilit
;;* 	  (beginning-of-line)
;;* 	  (tima-re-search-forward "\t+")
;;* 	  )
;;* 	 (t
;;* 	  ;; Remove hilit
;;* 	  (beginning-of-line)
;;* 	  (tima-re-search-forward "\t+" 0 'default )
;;* 	  ))
;;* 	))
;;*     ))
;;* _


;;; ##################################################### &example-pgp ###

;;* _
;;* ;;; ----------------------------------------------------------------------
;;* ;;; (global-set-key [C-f5] 'my-pgp-fp-certify)
;;* ;;;
;;* (defun my-pgp-fp-certify  ()
;;*   "To certify keys, Eg. get list of remailers  from http://www.uit.no/
;;* - Display in window1 the UIT.NO result file
;;* - Put received key fingerprints in other window (pgp -ka will tell you)
;;* _
;;* Call this function in the Received keys buffer, and it'll highlight
;;* keys that match Fingerprint in uit.no window.
;;* "
;;*   (interactive)
;;*   (let* (elt
;;* 	 (blist  (ti::w-list 'buffers))
;;* 	 (buffer (car (delq (current-buffer) blist)))
;;* 	 ok
;;*          )
;;*     (require 'tinymatch)
;;*     (require 'tinylib)
;;*     (PMIN)
;;*     (while (re-search-forward "Key fingerprint.*= +\\(.*\\)" nil t)
;;*       (setq elt (read-match 1)  ok nil)
;;*       (setq A elt)
;;*       (save-excursion
;;* 	(set-buffer buffer)
;;* 	(PMIN)
;;* 	(setq ok (tima-re-search-forward elt))
;;* 	)
;;*       (when ok
;;* 	(beginning-of-line)
;;* 	(tima-looking-at ".*"))
;;*       (end-of-line)
;;*       )
;;*     ))
;;*


;;; #################################################### &example-less ###

;;* ;;; ----------------------------------------------------------------------
;;* ;;;  19.30: view-search (times regexp)
;;* ;;;
;;* ;;;
;;* ;;;  Note:  With emacs view.el this advice mimics the famous unix
;;* ;;;  'less' command a bit better, however; at least my 19.28 and 30
;;* ;;;  emacs view.el doen't work right:
;;* ;;;
;;* ;;;  I do "/a"  to searc all a's in the buffer. highlights all
;;* ;;;  characters ok, but if I press "n" or "p" to go to next or
;;* ;;;  previous match the view-mode goes entirely wrong location!
;;* ;;;
;;* ;;;  --> I looked the source code and the problematic line that takes
;;* ;;;  cursor too far away is (move-to-window-line (if (< times 0) 0 -1))
;;* ;;;  I don't know why it's implemented this way...
;;* ;;;
;;* ;;;  Try calling M-x view-buffer
;;* ;;;
;;* (defadvice view-search (around tima act)
;;*   "Call TinyMatch.el package to highlight all matches.
;;* If you want to remove highlighting, then search string that will not
;;* be found from the buffer."
;;* _
;;*   ;;  Make private varaible for UNDO information
;;*   ;;
;;*   (unless (boundp 'my-:view-undo-data)
;;*     (defvar my-:view-undo-data nil)
;;*     (make-variable-buffer-local 'my-:view-undo-data))
;;* _
;;*   (when (featurep 'tinymatch)
;;*     ;; If this is same search string, do hothing.
;;*     ;; It means that user is browsing "n" "p"
;;*     ;;
;;* _
;;* ;;;    (d! my-:view-undo-data regexp)
;;* _
;;*     (if (and my-:view-undo-data
;;* 	       (string= (nth 1 my-:view-undo-data) regexp))
;;* 	nil				;skip
;;*       (save-excursion
;;* 	(PMIN)
;;* _
;;* 	;;  Undo previous highlight first
;;* 	;;
;;* 	(if (and my-:view-undo-data
;;* 		 (tima-stack-p))
;;* 	    (tima-undo))
;;* _
;;* 	;;  now highlight and save match data, check first if
;;* 	;;  there is nothing to do..
;;* 	;;
;;* 	(when (re-search-check regexp)
;;* 	  (if (tima-re-search-forward regexp 0 nil nil 'save-undo)
;;* 	      (setq my-:view-undo-data tima-:stack)
;;* 	    (setq my-:view-undo-data nil))
;;* 	  ))))
;;*   ad-do-it
;;*   )


;;}}}

(provide     'tinymatch)
(run-hooks   'timi-:load-hook)

;;; tinymatch.el ends here
