;; @(#) tinyrlog.el -- RCS rlog minor mode. ChecOut, CheckIn...

;; @(#) $Id: tinyrlog.el,v 1.18 1997/05/29 18:29:27 jaalto NotReleased $
;; @(#) $Keywords: tools $
;;
;; $KnownCompatibility: 19.28 XEmacs 19.14 $
;; $outlineRegexp: '@+' $
;; $bookMarkRegexp: '[-~+=*#.,'`^]+ &+\\([^ \t]+\\) [-~+=*#.,'`^]+' $
;; $PackageInstallRe: '^[ \t]*;;+[*] ' $

;; This file is not part of Emacs

;;{{{ Id

;; Copyright (C) 1996,97 Jari Aalto
;; Author:       Jari Aalto <jari.aalto@poboxes.com>
;; Maintainer:   Jari Aalto <jari.aalto@poboxes.com>
;; Created:      Dec 17 1996
;;
;; To get information on this program use ident(1) or do M-x tirl-version
;; Look at the code with folding.el, tinyfold.el, tinybm.el

;; LCD Archive Entry:
;; tinyrlog|Jari Aalto|jari.aalto@poboxes.com|
;; RCS rlog minor mode for vc.el, co, un/lock, status, chmod, hilit, mark...|
;; 29-May-1997|$Revision: 1.18 $|~/misc/tinyrlog.el|

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2 of the License, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with GNU Emacs.  If you did not, write to the Free Software Foundation,
;; Inc., 675 Mass Ave., Cambridge, MA 02139, USA.

;;}}}
;;{{{ Install

;; ....................................................... &t-install ...
;; Put this file on your Emacs-Lisp load path, add following into your
;; ~/.emacs startup file. Before doing require see tirl-:load-hook.
;; Code can be easily ripped with tinylib.el / ti::m-pkg-rip-magic
;;
;;* _
;;*     (require 'tinyrlog)
;;
;; Or prefer autoload: your emacs loads this package only when you
;; need it. This is for 19.30+
;;
;;* _
;;*     (eval-after-load "vc" '(require 'tinyrlog))
;;
;; Older emacs releases have different eval-after-load description,
;; so this use this instead. (You propably need this in XEmacs if there
;; is no eval-after-load command)
;;
;;* _
;;* (defadvice vc-print-log (after tirl act)
;;*   "Run hook tirl-:vc-print-log-hook."
;;*   (require 'tinyrlog)
;;*   (run-hooks 'tirl-:vc-print-log-hook))
;;
;;
;; If you define your own bindings and use enu, Update following variable
;; and call M-x tirl-install-mode.
;;
;;     tirl-:mode-menu
;;
;; If you have any questions, use this function
;;
;;      M-x tirl-submit-bug-report


;;}}}

;;{{{ Documentation

;; ..................................................... &t-commentary ...
;;; Commentary:

;; RESERVED prefix FOR FUNCTIONS IN THIS FILE
;; - To avoid collisions to other modules I use "tirl-" in front of
;;   every function & variable. It stands for '(ti)ny (rl)og'
;; - variable names contain letters 'tirl-:'.

;;  Preface, Dec 1996
;;
;;	In my work I have to manage very complex RCS revision
;;      numbers, multiple branches and I may have several branches
;;      CheckedOut for testing, correcting, and developing new
;;      features. It seemed natural to handle this "multiversioning"
;;      control from rlog output.
;;
;;  Overview of features
;;
;;	o   Companion to vc.el, Minor mode in RCS rlog buffer (C-x v l)
;;	o   Highlighing in windowed emacs in rlog buffer.
;;	o   You can do CheckOut, Lock a file, unclock file(s), show status
;;	    for current rcs file in emacs and ChekOut multiple revision
;;	    for viewing purposes (un/marking viewed versions)
;;
;;  Do you need this package?
;;
;;	If you don't use RCS/CVS don't load this package, it only works
;;	for rcslog output and contains direct 'co' and other
;;	rcs command calls.
;;
;;	And if you don't use vc's rlog output much, this
;;      package may not be essential to you. See if you want
;;      to utilize C-x v l more with this package and decide yourself.
;;
;;      This pacakge uses colors if window system is detected, but
;;	it partially copes with non-window system too, so that
;;	eg. marks appear in the buffer as charaxter codes. But
;;	all in all, you miss the other coloring that is user for
;;	version numbers.
;;
;;  Todo
;;
;;	Make it possible to lock multiple files and load
;;	them into emacs. The directory problem is still there, because
;;	when you save file, it overwrites the version underneath.
;;
;;	Don't expect this multilock feature to appear soon.

;;}}}
;;{{{ history

;; ........................................................ &t-history ...
;;; Change Log:
;; [jari]   = <jari.aalto@poboxes.com>
;;
;; May	29	1997	[jari]		19.28	v1.18		NotReleased
;; - Checkdoc 1.29 clear. defcustom support.
;;
;; May	14	1997	[jari]		19.28	v1.15-1.17	NotReleased
;; - uses new minor mode wizard.
;; - modified tirl-select-backend.
;; - Changed all non standard boolean tests to use when and unless
;;
;; Mar	27	1997	[jari]		19.28	v1.13-1.14	NotReleased
;; - USES OLD MINOR MODE DEFINITION.
;; - Kai Grossjohann <grossjohann@charly.cs.uni-dortmund.de>
;;   Sent CVS instructions. Because CVS's rlog is the same as RCS's,
;;   the needed changed were minor. This verison in not tested in full yet.
;; - Added automatic backend detection. Don't know yet if that works
;; - mistakenly used "cvs status" when the right command is "cvs log"
;;   for printing the rlog.
;; - Now Uses vc-buffer-backend (19.30+) if it is available to determine
;;   backend.
;;
;; Mar	19	1997	[jari]		19.28	v1.10-1.12	NotReleased
;; - internal function call parameter change. XEmacs20 CHAR changes
;;
;; Mar	4	1997	[jari]		19.28	v1.9		NotReleased
;; - Added "h" command to CheckOut th ehead version
;; - corrected some buffer check bugs that prevented using "o" command.
;;   Now there is chmod to -r if file was +r before doing co.
;; - RCS file name was not found correctly. Corrected.
;;
;; Feb	24	1997	[jari]		19.28	v1.8		NotReleased
;; - Corrected keybinding installation. Added menu-bar menu.
;; - Now uses fmacros from the library to create standard functions.
;;
;; Feb	21	1997	[jari]		19.28	v1.7		NotReleased
;; - Prefix key variable name changed.
;;
;; Feb	13	1997	[jari]		19.28	v1.6		NotReleased
;; - The filename read from rcslog was not absolute --> now
;;   tries to find buffer from emacs and then read the absolute filename
;; - Corrected one macro which prevented functions from working allthoug
;;   the file was in read-only-state.
;;
;; Jan	15	1997	[jari]		19.28	v1.2-1.5	NotReleased
;; - Major changes. This package has 70% new code and features and...
;;
;; Dec	16	1996	[jari]		19.28	v1.1		Created


;;}}}

;;; Code:

;;{{{ setup: require

;;; ......................................................... &require ...


(require 'advice)
(require 'tinylibm)

(eval-and-compile (use-dynamic-compilation))


(defgroup-tiny TinyRlog tirl-: tools
    "Electric file minor mode. Designed for minibuffer file prompt editing.
  Overview of features
	o   Companion to vc.el, Minor mode in RCS rlog buffer (C-x v l)
	o   Highlighing in windowed emacs in rlog buffer.
	o   You can do CheckOut, Lock a file, unclock file(s), show status
	    for current rcs file in emacs and ChekOut multiple revision
	    for viewing purposes (un/marking viewed versions)"
    )

;;}}}
;;{{{ setup: hooks

;;; .......................................................... &v-mode ...
;;;  We use name "vc log", bacause this minor mode is called
;;;  mostly from vc


(defcustom tirl-:menu-use-p t
  "*Should we use echo-area menu?"
  :type 'boolean :group 'TinyRlog
  )


;;;###autoload (autoload 'tirl-mode "tinyrlog" t t)
(eval-and-compile

(ti::m-fmacro-minor-mode-wizard
 "tirl-" " vcl" "'" "Vcl" 'TinyRlog "tirl-:"		;1-6

  "RCS log minor mode.
Mode description:
\\{tirl-:mode-map}
"

  "RCS rlog "
 nil


 "RCS Rlog minor mode menu."
 (list
  tirl-:mode-easymenu-name
  ["Do CheckOut at point"	    tirl-do-co		    t]
  ["Do CheckOut at point (lock) " tirl-do-co-l		    t]
  ["Do CheckOut head "	    tirl-do-co-head		    t]
  ["Unlock version"		    tirl-cancel-co	    t]
  ["Unlock unsafely version"	tirl-unlock-unsafely	    t]
  "----"
  ["Find (show) this revision"   tirl-find-file-tmp	    t]
  ["Mark 'find' versions"		    tirl-mark	    t]
  ["Pop to RCS buffer"	    tirl-pop-to-buffer		    t]
  ["Kill temporary files (flush)"  tirl-kill-tmp	    t]
  ["Reload rlog"		    tirl-reload		    t]
  )

 (progn
   (define-key tirl-:mode-map (concat p) 'nil)     ;; conquer prefix
   (if tirl-:menu-use-p
       (define-key tirl-:mode-map (concat p)   'tirl-:mode-menu)
     (define-key tirl-:mode-map (concat p "h") 'tirl-do-co-head)

     (define-key tirl-:mode-map (concat p "k") 'tirl-kill-tmp)
     (define-key tirl-:mode-map (concat p "l") 'tirl-do-co-l)
     (define-key tirl-:mode-map (concat p "m") 'tirl-mark)

     (define-key tirl-:mode-map (concat p "f") 'tirl-find-file-tmp)

     (define-key tirl-:mode-map (concat p "o") 'tirl-do-co)
     (define-key tirl-:mode-map (concat p "p") 'tirl-pop-to-buffer)
     (define-key tirl-:mode-map (concat p "r") 'tirl-reload)

     (define-key tirl-:mode-map (concat p "u") 'tirl-cancel-co)
     (define-key tirl-:mode-map (concat p "U") 'tirl-unlock-unsafely)
     ))
 )

)

;;; ......................................................... &v-hooks ...

(defcustom tirl-:load-hook nil
  "*Hook that is run when package is loaded."
  :type 'hook :group 'TinyRlog
  )


(defcustom tirl-:vc-print-log-hook
  '(turn-on-tirl-mode
    tirl-rename-buffer
    tirl-select-backend
    tirl-highlight
    )
  "*Hook run after `vc-print-log' command.
See also `tirl-:invoked-buffer' what the functions in this hook
can examine."
  :type 'hook :group 'TinyRlog
  )


;;}}}
;;{{{ setup: private

;;; ....................................................... &v-private ...
;;; Private variables


(defvar tirl-:invoked-buffer  nil
  "When you call `vc-print-log', the buffer-pointer is recored here.")
(put 'tirl-:invoked-buffer 'permanent-local t)


(defvar tirl-:shell-buffer "*tirl-tmp*"
  "Shell buffer.")


;;}}}
;;{{{ setup: user config

;;; ........................................................ &v-config ...
;;; User configurable
;;;

(defcustom tirl-:cmd-function 'tirl-cmd-get-rcs
  "*Return RCS executable shell command.
See `tirl-cmd-get' source code. Input parameters are symbols:

  'co 'ci 'rcs 'rcsdiff.

Predefined functions:

  'tirl-cmd-get-rcs
  'tirl-cmd-get-vcs

Note:

  This variable is set to buffer local and one of the above choices is
  set if `tirl-select-backend' function, which is installed in
  `tirl-:vc-print-log-hook' recognized the backend."
  :type 'hook :group 'TinyRlog
  )


(defcustom tirl-:locker-name (user-login-name)
  "*Your RCS locker ID that apperas in the lock statement."
  :type 'string :group 'TinyRlog
  )

;;}}}
;;{{{ version

;;; ....................................................... &v-version ...

(eval-and-compile
(ti::m-fmacro-version-bug-report
 "tinylisp.el"
 "tit"
 tili-:version-id
 "$Id: tinyrlog.el,v 1.18 1997/05/29 18:29:27 jaalto NotReleased $"
 '(tirl-:version-id
   tirl-:load-hook
   tirl-:mode-name
   tirl-:mode-prefix-key
   )))


;;}}}

;;; ########################################################### &Funcs ###

;;{{{ Minor mode

;;; ......................................................... &install ...

(defconst tirl-:mode-menu
  '("uU=unlock l=lock o=co h=co head  p=pop f=find  m=mark k=kill [rsc]"
    (
     (?u  . ( (call-interactively 'tirl-cancel-co)))
     (?U  . ( (tirl-unlock-unsafely timu-:prefix-arg 'verb)))
     (?o  . ( (call-interactively 'tirl-do-co)))
     (?l  . ( (call-interactively 'tirl-do-co-l)))
     (?h  . ( (call-interactively 'tirl-do-co-head)))
     (?f  . ( (tirl-find-file-tmp timu-:prefix-arg 'verb)))
     (?p  . ( (call-interactively 'tirl-pop-to-buffer)))
     (?k  . ( (call-interactively 'tirl-kill-tmp)))
     (?m  . ( (tirl-mark timu-:prefix-arg 'verb)))
     (?r  . ( (call-interactively 'tirl-reload)))
     (?s  . ( (call-interactively 'tirl-status)))
     (?c  . ( (call-interactively 'tirl-chmod-toggle)))
     ))
  "
Menu for rcs rlog buffer:
u      unlock, cancel co
U      Unlock first lock. Prefix arg unlocks all.

o      co version at current point
l      co with lock the version at current point
h      co the HEAD version

f      find-file this version to temporary buffer.
p      pop to the file where rlog belongs

k      Kill temporary version buffer files.
m      mark temporary version buffer files.. Prefix arg unmarks.

r      to reload buffer.
s      show status of file
c      toggle -r--r--r-- and -rw-r--r-- chmod.
")


;;; ----------------------------------------------------------------------
;;;
(defun tirl-:mode-menu  (&optional arg)
  "Call Echo area mode menu with ARG."
  (interactive "P")
  (timu-menu 'tirl-:mode-menu arg))

;;}}}
;;{{{ Macros

;;; .......................................................... &macros ...

;;; ----------------------------------------------------------------------
;;;
(put 'tirl-do-macro 'lisp-indent-function 0)
(defmacro tirl-do-macro (&rest body)
  "Store info to variables 'ver' and 'file'. Variable VERB must e also bound.
If 'ver' of 'file' cannot be set, print message and do nothing with BODY."
  (`
   (when (and (or (setq ver (tirl-get-version))
		  (error "No version found on the line."))
	      (or (setq file (tirl-get-filename))
		  (error "Can't find rcs file name from buffer.")))

     ;;  We must find absolute path; this isn't enough
     ;;
     ;;  Working file: tm-view.el
     ;;
     (unless (string-match "/" file)
       (let (buffer
	     )
	 (cond
	  ((setq buffer (get-buffer file))
	   (save-excursion
	     (set-buffer buffer)
	     (setq file buffer-file-name)
	     ))
	  (t
	   (error "Can't find absolute filename %s" file)
	   ))))

     (,@ body)
     )))



;;; ----------------------------------------------------------------------
;;;
(put 'tirl-file-confirm-macro 'lisp-indent-function 2)
(defmacro tirl-file-confirm-macro (file verb &rest body)
  "Make sure FILE is read-only before continuing.
If VERB is nil, don't do any checkings or ask from user when
executing BODY."
  (`
   (when (or (null (, verb))
	     (and (, verb)
		  (or (file-read-only-p (, file))
		      (y-or-n-p
		       (format "Writable %s exist, are you sure "
			       (file-name-nondirectory (, file))))
		      )))
     (,@ body)
     )))



;;; ----------------------------------------------------------------------
;;;
(defsubst tirl-lock-listed-p  ()
  "See if there is locks in listing."
  (save-excursion
    (PMIN)
    ;;    locks: strict
    ;;	jaalto: 1.1
    (re-search-forward "^locks:")
    (forward-line 1)
    (looking-at "^[ \t]+")
    ))


;;; ----------------------------------------------------------------------
;;;
(defsubst tirl-get-tmp-list (file)
  "Return temp buffers."
  (when file
    (setq file (file-name-nondirectory file))
    (ti::b-buffer-list (format "\\*%s.*[0-9]+\\*" file) 'temp-too)
    ))


;;; ----------------------------------------------------------------------
;;;
(put 'tirl-do-over-locks-macro 'lisp-indent-function 2)
(defmacro tirl-do-over-locks-macro (user ver &rest body)
  "USER and VER is currently unused. Execute BODY over 'locks:' keyword.

While the macro loops each line; the varaibles 'user' and 'ver'
are updated. If you want to terminate macro, move point away from the
lock lines: eg by (goto-char (point-min)))."
  (`
   (save-excursion
     (PMIN) (re-search-forward "^locks:") (forward-line 1)

     (while (looking-at "^[ \t]+\\([^:]+\\):[ \t]\\([.0-9]+\\)")
       (setq user (read-match 1) ver (read-match 2))
       (,@ body)
       (forward-line 1)
       )
     )))


;;}}}
;;{{{ Rcs interface

;;; ............................................................. &rcs ...

;;; ----------------------------------------------------------------------
;;;
(defsubst tirl-cmd-cd-template  (file &optional options)
  "Create 'cd' command template: \"cd DIR; %s FILE\""
  (interactive)
  (concat  "cd " (file-name-directory file) "; "
	   "%s " (or options "") " " (file-name-nondirectory file)))


;;; ----------------------------------------------------------------------
;;;
(defun tirl-cmd-get-cvs  (sym)
  "Return RCS executable according to SYM."
  ;;
  ;; This is default function. User may return "my-co" for 'co
  ;; command in certain situations etc in his own function.
  ;;
  (cond
   ;;  "cvs co" is for
   ;;  the initial checkout of a file only, after that,
   ;;  "cvs update" is used.
   ((eq 'co sym)	"cvs update")


   ((eq 'ci sym)	"cvs commit")
   ((eq 'rcs sym)	"cvs rcs")
   ((eq 'rlog sym)      "cvs log")
   ((eq 'rcsdiff sym)   "cvs diff")  ;; -j<old-rev> -j<new-rev>"
   (t
    (error "No cmd %s " sym))
   ))


;;; ----------------------------------------------------------------------
;;;
(defun tirl-cmd-get-rcs  (sym)
  "Return RCS executable according to SYM."
  ;;
  ;; This is default function. User may return "my-co" for 'co
  ;; command in certain situations etc in his own function.
  ;;
  (cond
   ((eq 'co sym)	(symbol-name sym))
   ((eq 'ci sym)	(symbol-name sym))
   ((eq 'rcs sym)	(symbol-name sym))
   ((eq 'rlog sym)      (symbol-name sym))
   ((eq 'rcsdiff sym)   (symbol-name sym))
   (t
    (error "No cmd %s " sym))
   ))



;;; ----------------------------------------------------------------------
;;;
(defun tirl-cmd-exec (sym shell-cmd &optional buffer noerr)
  "Execute shell command. If error, show result buffer.

Input:

  SYM		Command symbol like 'co
  SHELL-CMD	Full shell command. If this command has %s in
                a string, then RCS-SH-EXE is sprintf'd into that position.
                Normally the rcs exe is prepended to the command.
  BUFFER	where to put shell command results
  NOERR		ignore errors

References:
  `tirl-:cmd-function'"
  (let* ((exe    (funcall tirl-:cmd-function sym))
	 (send   (if (string-match "%s" shell-cmd)
		     (format shell-cmd exe sym)
		   (format "%s %s" exe shell-cmd)))
	  (out   (or buffer (use-tmp-buffer tirl-:shell-buffer 'clear)))
	  ret
	  )
    (setq ret (shell-execute  send out))
    (when (and (null noerr)
	       (shell-exec-nok-p ret))
      (pop-to-buffer out)
      (error "Shell error %d " ret)
      )
    ret
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tirl-cmd-diff-p  (file &optional options)
  "Return t if there is diff for FILE with diff OPTIONS."
  (if (zerop
       (tirl-cmd-exec
	'rcsdiff
	(tirl-cmd-cd-template file (or options "-q"))
	nil
	'noerr
	))
      nil t))

;;; ----------------------------------------------------------------------
;;;
(defun tirl-cmd-rcs  (file &optional options)
  "Run rcs command on FILE with OPTIONS."
  (tirl-cmd-exec 'rcs (tirl-cmd-cd-template file options)))




;;}}}
;;{{{ Misc

;;; ............................................................ &misc ...


;;; ----------------------------------------------------------------------
;;;
(defun tirl-lock-list ()
  "Return lock list. '((USER . LOCK-VER) ..)"
  (let (list
	user
	ver
	)
    (tirl-do-over-locks-macro user ver
      (push (cons user ver) list))
    (nreverse list)
    ))



;;; ----------------------------------------------------------------------
;;;
(defun tirl-reload  (&optional verb)
  "Replace buffer with current log. VERB."
  (interactive)
  (let ((file (tirl-get-filename))
	)
    (VERB)
    (tirl-cmd-exec 'rlog file nil 'noerr)
    (erase-buffer)
    (tirl-load-to-buffer (current-buffer))
    (run-hooks-until 'tirl-:vc-print-log-hook)
    (if verb
	(message "Updated."))
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tirl-status  ()
  "Show file status."
  (interactive)
  (let* ((file   (tirl-get-filename))
	 (buffer (get-file-buffer file))
	 str
	 fn
	 ver
	 )

    (if buffer
	(setq ver (ti::m-rcs-buffer-version buffer)))

    (setq fn  (file-name-nondirectory file)
	  str (ti::c-int-to-access-mode-string
	       (file-modes file)))

    (message "%s%s has modes %s " (if ver (concat ver " "))  fn str)
    ))


;;; ----------------------------------------------------------------------
;;;
(defun tirl-chmod-toggle (&optional verb)
  "Toggle between =r and +w. VERB."
  (interactive)
  (let* ((file  (tirl-get-filename))
	 (modes  (file-modes file))
         )
    (VERB)
    (if (file-read-only-p file)
	(set-file-modes file (file-mode-make-writable modes))
      (set-file-modes file (file-mode-make-read-only-all modes)))
    (if verb
	(tirl-status))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tirl-highlight  ()
  "Do basic highlight."
  (when (ti::xe-window-system)
    (save-excursion
      (PMIN)
      (tima-looking-at ".*file:[ \t]+\\([^\n]+\\)" 1 'region)
      (tima-re-search-forward "^locks:[ \t]+\\([^\n]+\\)" 1 'region)
      (tima-re-search-forward "revision[ \t]+\\([^\n]+\\)" 1)
      )))


;;; ----------------------------------------------------------------------
;;;
(defun tirl-select-backend  ()
  "Select RCS or CVS command for the log buffer: set `tirl-:cmd-function'."
  (interactive)
  (let* ((ibuffer  tirl-:invoked-buffer)
	 file
	 type
	 )
    (when (and
	   ibuffer
	   (get-buffer ibuffer)
	   (setq file
		 (job-buffer ibuffer (buffer-file-name)))
	   )
      (if (fboundp 'vc-buffer-backend)	;19.30+
	  (setq type (sfuncall 'vc-buffer-backend))
	;;  nope; that function does not exist. (19.28)
	;;
	(setq type (vc-file-getprop file 'vc-backend)))

      (make-local-variable 'tirl-:cmd-function)
      (cond
       ((equal type 'RCS) (setq tirl-:cmd-function 'tirl-cmd-get-rcs))
       ((equal type 'CVS) (setq tirl-:cmd-function 'tirl-cmd-get-rcs))
       ))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tirl-rename-buffer ()
  "Rename buffer to *vcl* if the the previous buffer name was *vc*.
Other vc commands normally destroy the log buffer, so renaming
it keeps it alive until next rlog command."
  (interactive)
  (let* ((buffer (get-buffer "*vcl*"))
	 )
    (cond
     ((string= "*vc*" (buffer-name))
      ;;  Remove old log buffer if it exists.
      ;;
      (if buffer (kill-buffer buffer))
      (rename-buffer "*vcl*")
      ))
    ))



;;; ----------------------------------------------------------------------
;;;
(defsubst tirl-char-mark-p  (&optional remove)
  "Check if there is marker char at the beginning of line. Moves point."
  (beginning-of-line)
  (char= (following-char) ?>))


;;; ----------------------------------------------------------------------
;;;
(defun tirl-char-mark  (&optional unmark)
  "Mark the line, or UNMARK."
  (interactive)
  (cond
   ((and unmark (tirl-char-mark-p))
    (delete-char 1))
   ((and (null unmark) (tirl-char-mark-p))
    ;;  there is already mark
    )
   ((null unmark)
    (insert ">")
    )))


;;; ----------------------------------------------------------------------
;;;
(defun tirl-buffer-version  (file)
  "If FILE is in emacs, return version number."
  (if (get-file-buffer file)			;Loaded into emacs already
      (ti::m-rcs-buffer-version (get-file-buffer file))))


;;; ----------------------------------------------------------------------
;;;
(defun tirl-get-filename  ()
  "Return filename or nil"
  (save-excursion
    (PMIN)
    (cond
     ((re-search-forward "RCS file:[ \t]+\\([^ \n\t]+\\)")
      (remove-properties (ti::m-rcs-normal-file (read-match 1))))

     ((re-search-forward "Working file:[ \t]+\\([^ \n\t]+\\)")
      (remove-properties (read-match 1)))
     )))

;;; ----------------------------------------------------------------------
;;;
(defun tirl-get-version  ()
  "Return version on current line."
  (let* ((line  (read-full-line))
	 (ver   (strmatget "[0-9]+\\.[0-9.]+" 0 line))
	 (dots  (and ver
		      (count-char-in-string ?. ver)))
         )
    (if (and ver
	     ;; Must be odd count
	     (not (eq 1 (% dots 2))))	;odd, 1.1  or 1.1.1.1
	(setq ver nil))			;cancel

    ver
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tirl-load-to-buffer  (dest &optional noerr)
  "Examine `tirl-:shell-buffer' and copy the output to DEST buffer.
If DEST does not exist, it is created. NOERR ignores errors."
  (interactive)
  (let ((shell tirl-:shell-buffer)
	point
	)
    (job-buffer shell
      (PMIN)

      (cond				; -p switch
       ((looking-at ".*-->[ \t]+stdout")
	(forward-line 2) (setq point (point))
	)
       ((save-excursion			; rlog
	  (forward-line 1)
	  (looking-at "RCS file:")
	  (setq point (point))
	  ))
       )


      (when point			;Only if start point set
	(if (not (get-buffer dest))
	    (setq dest (use-tmp-buffer dest 'clear))
	  (clear-buffer dest))
	(append-to-buffer dest point (point-max))
	))

    (when (and (null noerr)
	       (null point))
      (pop-to-buffer shell)
      (error "Nothing to load  from shell buffer."))

    point				;return success status
    ))




;;; ----------------------------------------------------------------------
;;;
(defun tirl-file-untouched-p (file)
  "Check if buffer is in emacs and that FILE is not modified.
If the file is not in emacs, run rcsdiff.

Return:
 str	buffer's RCS version if untouched.
 t      if file was not in emacs and there was no rcsdiff."
  (let* (buffer
	 untouched
	 ret
	 )
    (setq buffer (get-file-buffer file))
    (cond
     ((null buffer)
      (if (file-read-only-p file)	;If it's readonly, suppose no diffs
	  (setq ret t)
	(if (tirl-cmd-diff-p file)
	    (setq ret t))
	))

     (buffer			;Loaded into emacs already
      (save-excursion
	(set-buffer buffer)

	(cond
	 ((not (buffer-modified-p))
	  (setq untouched t)
	  ))

	(if untouched
	    (setq ret (ti::m-rcs-buffer-version)))
	)))
    ret
    ))


;;}}}
;;{{{ interactive

;;; ..................................................... &interactive ...

;;; ----------------------------------------------------------------------
;;;
(defun tirl-mark (&optional unmark verb)
  "Mark revisions that were loaded by \\[tirl-find-file-tmp].
Optionally UNMARK. VERB."
  (interactive "P")
  (let* ((list (tirl-get-tmp-list (tirl-get-filename)))
	 (len  (if list (length list)))
	 beg end
	 re
	 ver
	 elt
	 )
    (VERB)
    (save-excursion
      (list-loop list elt
	(setq ver (strmatget "[0-9]+[.0-9]+" 0 elt))
	(setq re  (format "\\(revision\\)[ \t]+%s[^ \t]*$" ver))

	(PMIN)
	(when (re-search-forward re nil t)

	  (cond
	   ((ti::xe-window-system)	;Windowed -- use colors
	    (setq beg (match-beginning 1)  end (match-end 1))

	    (if unmark
		(put-text-property beg end 'face 'default)
	      (put-text-property beg end 'face 'region))
	    )
	   (t				;Non-Windowed
	    (tirl-char-mark unmark)
	    )))
	))

    (if verb
	(if (null len)
	    (message "No temporary files.")
	  (message "%smarked %d items" (if unmark "un" "") len)))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tirl-kill-tmp  ()
  "Kill all tmp buffers that were loaded from call \\[tirl-find-file-tmp]."
  (interactive)
  (let ((file  (tirl-get-filename))
	list
	)
    (if (null file)
	(message "No RCS filename found.")
      (setq list (tirl-get-tmp-list file))
      (list-loop list file
	(kill-buffer file)
	))))

;;; ----------------------------------------------------------------------
;;;
(defun tirl-pop-to-buffer  ()
  "Pop to buffer accordig to this rlog."
  (interactive)
  (let* ((file    (tirl-get-filename))
	 (buffer  (get-file-buffer file))
         )
    (if buffer
	(pop-to-buffer buffer)
      (if (y-or-n-p (format "No %s buffer, load "
			    (file-name-nondirectory file)))
	  (find-file file)))
    ))



;;; ----------------------------------------------------------------------
;;;
(defun tirl-find-file-tmp  (&optional no-pop verb)
  "Find the current version into Emacs.
The file will be Checked Out by using pipes and the created
temporary buffer will not have any filename association.

You can use this function to eg. get version 1.1.1.1 and 1.1.1.2 into
emacs while your workfile stays somewhere else. Nice for pasting
text from other versions.

Input:
 NO-POP     do not `pop-to-buffer' after rcs call.
 VERB	    Verbose messages."
  (interactive "P")
  (let* (file
	 ver
	 buffer
         )
    (VERB)

    (tirl-do-macro
      (setq buffer (format "*%s %s*" (file-name-nondirectory file) ver))
      (if (get-buffer buffer)
	  (if (null no-pop)
	      (pop-to-buffer buffer))
	(tirl-cmd-exec 'co (format "-p -r%s %s" ver file))
	(tirl-load-to-buffer buffer)
	(if (null no-pop)
	    (pop-to-buffer buffer))
	))
    (if (and verb no-pop)
	(message "Loaded %s" ver))

    ))


;;; ----------------------------------------------------------------------
;;;
(defun tirl-unlock-unsafely (&optional all verb)
  "Read 'locks:' keyword and unlock first locked version in the list.
If there is no locks, then do nothing. ALL unlocks all locks.

This is unsafe function, because no attempt is made to
check if the file has changes. You may loose data is you call
this fnction without checking the diffs.

Input:

  ALL	    Unlock all version locked.
  VERB	    Verbose messags.

Notes:
 `tirl-:locker-name'	other locks are not touched ever.
 No buffer reverting is attempted."
  (interactive)
  (let* ((name  tirl-:locker-name)
	 user
	 ver
	 file
	 done
         )
    (VERB)
    (if (and verb
	     (null (y-or-n-p "unlock: Are you absolutely sure ")))
	(error "Aborted."))

    (setq file (tirl-get-filename))
    (if (and verb
	     (not (file-read-only-p file)))
	(if (null (y-or-n-p (format "%s is writable, proceed "
				    (file-name-nondirectory file))))
	    (error "Aborted.")))

    (set-file-modes file (file-mode-make-read-only-all (file-modes file)))

    (tirl-do-over-locks-macro user ver
      (when (string= user name)
	(if verb
	    (message "Unlocking %s" ver))
	(tirl-cmd-exec 'co (format "-u%s %s" ver file))
	(setq done t)
	(if (null all)
	    (PMIN))			;Terminate lock macro loop
	)
      )
    (when done
      (tirl-reload)
      (if verb (message "done."))
      )
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tirl-cancel-co  (&optional verb)
  "Cancel Checkout for current revision, so that file is no more locked.
Notice that the lock status is based on the buffer content. Do
\\[tirl-reload] to update the log if needed. VERB.

Chmod undelying file to read-only."
  (interactive)
  (let* (buffer
	 ver
	 file
	 llist
         )
    (VERB)
    (tirl-do-macro
      (setq llist  (tirl-lock-list))

      (if (null llist)
	  (if verb (message "Lock list seems to be empty."))
	(if (not (rassoc ver llist))
	    (if verb
		(message "%s is not locked." ver))
	  (set-file-modes file 292)	;444oct, rrr
	  (tirl-cmd-exec 'co (format "-u%s %s" ver file))

	  (tirl-reload)			;Update
	  (when (setq buffer (get-file-buffer file))
	    (pop-to-buffer buffer)
	    (call-interactively 'revert-buffer)
	    (message "")
	    )


	  (if verb (message "Revision %s unlocked." ver))
	  )))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tirl-do-co-l  ()
  "Do co and lock the version number on the line."
  (interactive)
  (let* (old-buffer
	 ver
	 file
         )
    (tirl-do-macro
      (setq old-buffer (get-file-buffer file))

      (if (not (tirl-file-untouched-p file))
	  (error "'%s' modified or buffer not read-only." file))

      (if (file-writable-p file)
	  (error "Can't lock: Underlying file is writable."))

      (when (or  (null (tirl-lock-listed-p))
		 (y-or-n-p "There is already lock, proceed? "))
	(tirl-cmd-exec 'co (format "-l%s %s" ver file))
	(tirl-reload)
	(pop-to-buffer (find-file-noselect file))
	(when old-buffer
	  (call-interactively 'revert-buffer)
	  (message "")
	  ))

      )))



;;; ----------------------------------------------------------------------
;;;
(defun tirl-do-co-head  ()
  "CheckOut the HEAD revision."
  (interactive)
  (let* (
         )
    (PMIN)
    (if (re-search-forward "^head: ")
	(call-interactively 'tirl-do-co)
      (message "Hm, Can't find the 'head:' tag anywhere? "))
    ))

;;; ----------------------------------------------------------------------
;;;
(defun tirl-do-co  (&optional replace verb)
  "Checkout specific revision around current point.
REPLACE current emacs buffer with this version if the existing file in emacs
is read-only. VERB."
  (interactive "P")
  (let* (verb
	 ver file
	 untouched
	 buffer
	 buffer-ver
	 ret
         )
    (VERB)
    (tirl-do-macro
      (setq buffer     (find-buffer-visiting  file)
	    untouched  (tirl-file-untouched-p file)
	    buffer-ver (or (tirl-buffer-version file) "")
	    )

      (tirl-file-confirm-macro file verb
	(cond
	 ((string= ver buffer-ver)
	  (if verb
	      (message (format "%s v%s already in emacs." buffer ver)))
	  (setq ret buffer)
	  )

	 ((or (and
	       (file-writable-p file)
	       (y-or-n-p "Writable file, needs chmod, ok? ")
	       (progn
		 (set-file-modes
		  file (file-mode-make-read-only (file-modes file)))
		 t
		 ))
	      (null buffer)
	      untouched
	      )

	  (when (or (null verb)
		    (null buffer)
		    (and verb
			 (y-or-n-p
			  (format "Untouched %s, replace %s with version %s ?"
				  (file-name-nondirectory file)
				  buffer-ver ver)))
		    )
	    (if buffer (kill-buffer buffer))
	    (tirl-cmd-exec 'co (format "-r%s %s " ver file))
	    (setq ret (find-file-noselect file))
	    (if verb  (pop-to-buffer ret))
	    ))
	 (t
	  (if verb
	      (message (format "Changed buffer exist, cancelled.")))
	  ))
	))
    ret
    ))


;;; ----------------------------------------------------------------------
;;;
(defadvice vc-print-log (around tirl act)
  "Run hook tirl-:vc-print-log-hook."
  (let* ((BuffeR (current-buffer))
	 )

    ad-do-it

    (make-local-variable 'tirl-:invoked-buffer)
    (setq tirl-:invoked-buffer BuffeR)
;;;    (d! tirl-:invoked-buffer)
    (run-hooks 'tirl-:vc-print-log-hook)
    ))

;;}}}

(add-hook 'tirl-:mode-define-keys-hook 'tirl-mode-define-keys)

(provide   'tinyrlog)
(run-hooks 'tirl-:load-hook)

;;; tinyrlog.el ends here
